/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_audio.h"

#define MIN_AUDIO_BUFFER_SIZE     (1 << 12)
#define DEFAULT_AUDIO_BUFFER_SIZE (1 << 14)
#define MAX_AUDIO_BUFFER_SIZE     (1 << 16)
#define CX88_AUDIO_SPEED 48000

/* Maximum line size, in bytes (dictated by capacity of line length register) */
#define MAX_AUDIO_LINE_SIZE 0xFFF

static uint32_t cx88_audio_formats[] = 
{
#if __FreeBSD_version >= 800097
   SND_FORMAT(AFMT_S16_LE, 1, 0),
   SND_FORMAT(AFMT_S16_LE, 2, 0),
#else
   AFMT_S16_LE,
   AFMT_STEREO | AFMT_S16_LE,
#endif
   0
};

static struct pcmchan_caps cx88_audio_caps = 
{
   .minspeed = CX88_AUDIO_SPEED,
   .maxspeed = CX88_AUDIO_SPEED,
   .fmtlist = cx88_audio_formats, 
   .caps = 0
};

static void cx88_audio_reconfigure(cx88_audio_softc *sc);
static int cx88_audio_dmaprog(cx88_buffer *buffer, uint8_t bufindex);
static int cx88_audio_configure(
   cx88_audio_softc *sc,
   cx88_audio_config *conf,
   uint32_t blocksize,
   struct thread *context);
static void cx88_audio_close_session(cx88_audio_softc *sc);
static int cx88_audio_start_capture(cx88_audio_softc *sc);
static void cx88_audio_bufsync(cx88_audio_softc *sc);
static int cx88_audio_transfer(cx88_audio_softc *sc, cx88_audio_wait *wait, boolean_t nonblocking);
static void *cx88_pcmchan_init(
   kobj_t obj,
   void *devinfo,
   struct snd_dbuf *b,
   struct pcm_channel *c,
   int dir);
static int cx88_pcmchan_free(kobj_t obj, void *data);
static int cx88_pcmchan_setformat(kobj_t obj, void *data, uint32_t format);
static uint32_t cx88_pcmchan_setspeed (kobj_t obj, void *data, uint32_t speed);
static uint32_t cx88_pcmchan_setblocksize(kobj_t obj, void *data, uint32_t blocksize);
static int cx88_pcmchan_trigger (kobj_t obj, void *data, int go);
static uint32_t cx88_pcmchan_getptr(kobj_t obj, void *data);
static struct pcmchan_caps * cx88_pcmchan_getcaps(kobj_t obj, void *devinfo);
static int cx88_mixer_init(struct snd_mixer *m);
static int cx88_mixer_uninit(struct snd_mixer *m);
static int cx88_mixer_set(
   struct snd_mixer *m,
   unsigned int dev,
   unsigned int left,
   unsigned int right);
static uint32_t cx88_mixer_setrecsrc(struct snd_mixer *m, uint32_t src);
static int cx88_pcm_probe(device_t dev);
static int cx88_pcm_attach(device_t dev);
static int cx88_pcm_detach(device_t dev);
static int cx88audio_mod_handler(module_t mod, int what, void *arg);

static d_open_t cx88_audio_open;
static d_close_t cx88_audio_close;
static d_mmap_t cx88_audio_mmap;
static d_ioctl_t cx88_audio_ioctl;
static d_poll_t cx88_audio_poll;

static struct cdevsw cx88_audio_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88audio",
   .d_open = cx88_audio_open,   /* open */
   .d_close = cx88_audio_close, /* close */
   .d_mmap = cx88_audio_mmap,   /* mmap */
   .d_ioctl = cx88_audio_ioctl, /* ioctl */
   .d_poll = cx88_audio_poll,   /* poll */
};

static struct unrhdr *cx88_audio_units = NULL;

static const cx88_audio_stream_desc cx88_audio_streams[] = 
{
   {
      .stream = CX88_AUDIO_STREAM_CX23880,
      .dma_channel = CX88_CH_AUDIO_IN,
      .reg_gp_count = CX23880_AUDIO_REG_IN_GP_COUNT,
      .reg_dma_control = CX23880_AUDIO_REG_DMA_CONTROL,
      .dma_control_mask = CX23880_AUDIO_IN_RISC_ENABLE | CX23880_AUDIO_IN_FIFO_ENABLE,
      .reg_line_length = CX23880_AUDIO_REG_IN_LINE_LENGTH,
      .reg_pci_intr_mask = CX23880_REG_PCI_INTMASK,
      .pci_intr_mask = CX23880_PCI_INT_AUDIO,
      .reg_intr_mask = CX23880_AUDIO_REG_INT_MASK,
      .reg_intr_stat = CX23880_AUDIO_REG_INT_STAT,
      .reg_intr_mstat = CX23880_AUDIO_REG_INT_MSTAT,
      .intr_mask = CX23880_AUDIO_INT_IN_SYNC | CX23880_AUDIO_INT_OPCODE_ERR | 
         CX23880_AUDIO_INT_PCI_ABORT | CX23880_AUDIO_INT_PAR_ERR | 
         CX23880_AUDIO_INT_RIP_ERR | CX23880_AUDIO_INT_IN_RISCI1,
      .risci1_mask = CX23880_AUDIO_INT_IN_RISCI1,
      .opcode_err_mask = CX23880_AUDIO_INT_OPCODE_ERR,
      .sync_err_mask = CX23880_AUDIO_INT_IN_SYNC,
      .bus_err_mask = CX23880_AUDIO_INT_PCI_ABORT | CX23880_AUDIO_INT_PAR_ERR |
         CX23880_AUDIO_INT_RIP_ERR,
      .overflow_mask = CX23880_AUDIO_INT_IN_OVERFLOW,
      .ber_mask = CX23880_AUDIO_INT_BER,
      .reg_devcntl2 = CX23880_REG_DEVCNTL2
   },
   {
      .stream = CX88_AUDIO_STREAM_CX23885_INTA,
      .dma_channel = CX88_CH_AUDIO_IN,
      .reg_gp_count = CX23885_AUDIO_REG_INTA_GP_COUNT,
      .reg_dma_control = CX23885_AUDIO_REG_INT_DMA_CONTROL,
      .dma_control_mask = CX23885_AUDIO_INTA_RISC_ENABLE | CX23885_AUDIO_INTA_FIFO_ENABLE,
      .reg_line_length = CX23885_AUDIO_REG_INTA_LINE_LENGTH,
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_AUDIO_INT,
      .reg_intr_mask = CX23885_AUDIO_REG_INT_INT_MASK,
      .reg_intr_stat = CX23885_AUDIO_REG_INT_INT_STAT,
      .reg_intr_mstat = CX23885_AUDIO_REG_INT_INT_MSTAT,
      .intr_mask = CX23885_AUDIO_INT_INTA_SYNC |
         CX23885_AUDIO_INT_INTA_OPCODE_ERR | CX23885_AUDIO_INT_INTA_RISCI1,
      .risci1_mask = CX23885_AUDIO_INT_INTA_RISCI1,
      .opcode_err_mask = CX23885_AUDIO_INT_INTA_OPCODE_ERR,
      .sync_err_mask = CX23885_AUDIO_INT_INTA_SYNC,
      .overflow_mask = CX23885_AUDIO_INT_INTA_OVERFLOW,
      .reg_devcntl2 = CX23885_REG_DEVCNTL2,
      .reg_mode = CX23885_AUDIO_REG_INTA_MODE
   },
   {
      .stream = CX88_AUDIO_STREAM_CX23885_INTB,
      .dma_channel = CX88_CH_AUDIO_IN_2,
      .reg_gp_count = CX23885_AUDIO_REG_INTB_GP_COUNT,
      .reg_dma_control = CX23885_AUDIO_REG_INT_DMA_CONTROL,
      .dma_control_mask = CX23885_AUDIO_INTB_RISC_ENABLE | CX23885_AUDIO_INTB_FIFO_ENABLE,
      .reg_line_length = CX23885_AUDIO_REG_INTB_LINE_LENGTH,
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_AUDIO_INT,
      .reg_intr_mask = CX23885_AUDIO_REG_INT_INT_MASK,
      .reg_intr_stat = CX23885_AUDIO_REG_INT_INT_STAT,
      .reg_intr_mstat = CX23885_AUDIO_REG_INT_INT_MSTAT,
      .intr_mask = CX23885_AUDIO_INT_INTB_SYNC |
         CX23885_AUDIO_INT_INTB_OPCODE_ERR | CX23885_AUDIO_INT_INTB_RISCI1,
      .risci1_mask = CX23885_AUDIO_INT_INTB_RISCI1,
      .opcode_err_mask = CX23885_AUDIO_INT_INTB_OPCODE_ERR,
      .sync_err_mask = CX23885_AUDIO_INT_INTB_SYNC,
      .overflow_mask = CX23885_AUDIO_INT_INTB_OVERFLOW,
      .reg_devcntl2 = CX23885_REG_DEVCNTL2,
      .reg_mode = CX23885_AUDIO_REG_INTB_MODE
   },
   {
      .stream = CX88_AUDIO_STREAM_CX23885_EXT,
      .dma_channel = CX88_CH_AUDIO_IN_3,
      .reg_gp_count = CX23885_AUDIO_REG_EXT_GP_COUNT,
      .reg_dma_control = CX23885_AUDIO_REG_EXT_DMA_CONTROL,
      .dma_control_mask = CX23885_AUDIO_EXT_RISC_ENABLE | CX23885_AUDIO_EXT_FIFO_ENABLE,
      .reg_line_length = CX23885_AUDIO_REG_EXT_LINE_LENGTH,
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_AUDIO_EXT,
      .reg_intr_mask = CX23885_AUDIO_REG_EXT_INT_MASK,
      .reg_intr_stat = CX23885_AUDIO_REG_EXT_INT_STAT,
      .reg_intr_mstat = CX23885_AUDIO_REG_EXT_INT_MSTAT,
      .intr_mask = CX23885_AUDIO_INT_EXT_SYNC |
         CX23885_AUDIO_INT_EXT_OPCODE_ERR | CX23885_AUDIO_INT_EXT_RISCI1,
      .risci1_mask = CX23885_AUDIO_INT_EXT_RISCI1,
      .opcode_err_mask = CX23885_AUDIO_INT_EXT_OPCODE_ERR,
      .sync_err_mask = CX23885_AUDIO_INT_EXT_SYNC,
      .overflow_mask = CX23885_AUDIO_INT_EXT_OVERFLOW,
      .reg_devcntl2 = CX23885_REG_DEVCNTL2,
      .reg_mode = CX23885_AUDIO_REG_EXT_MODE
   }
};

int cx88_audio_setup(
   cx88_audio_softc *sc,
   cx88_audio_stream_t stream,
   cx88_common_softc *common,
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t),
   device_t audiocore_dev,
   uint8_t audiocore_path)
{
   int error = 0;
   if (stream >= CX88_AUDIO_NUM_STREAMS)
   {
      return EINVAL;
   }
   sc->stream = &(cx88_audio_streams[stream]);
   sc->common = common;
   sc->i2c_accessor = i2c_accessor;
   sc->audiocore_dev = audiocore_dev;
   sc->audiocore_path = audiocore_path;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   sc->user_configured = FALSE;
   sc->pcm_dev = NULL;
   sc->pcm_chan = NULL;
   sc->pcm_bufsize = 0;
	sc->last_sync_buffer = 0;
   sc->buffer_available = 0;
   mtx_init(&(sc->isrlock), "CX88 Audio ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Audio SX lock");
   sc->initialized = TRUE;
   if ((error = cx88_init_channel(&(sc->channel), common,
      sc->stream->dma_channel, MAX_AUDIO_BUFFER_SIZE)))
   {
      goto fail_init_channel;
   }
   cx88_lock_device_tree(common);
   sc->pcm_dev = device_add_child(common->dev, "pcm", -1);
   if (sc->pcm_dev == NULL)
   {
      error = ENXIO;
      goto fail_add_pcm;
   }
   device_set_ivars(sc->pcm_dev, sc);
   if ((error = device_probe_and_attach(sc->pcm_dev)))
   {
      goto fail_attach_pcm;  
   }
   cx88_unlock_device_tree(common);
   if ((sc->unit = alloc_unr(cx88_audio_units)) == -1)
   {
      goto fail_alloc_unit;
   }
   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_audio_cdevsw, sc->unit, UID_ROOT, GID_WHEEL, 0644, "cx88audio%d", sc->unit);
   if (sc->devnode == NULL)
   {
      goto fail_make_dev;
   }
	sc->devnode->si_drv1 = sc;
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED);
   goto success;
   fail_make_dev:
      free_unr(cx88_audio_units, sc->unit);
   fail_alloc_unit:
      cx88_lock_device_tree(common);
      device_detach(sc->pcm_dev);
   fail_attach_pcm:
      device_delete_child(common->dev, sc->pcm_dev);
   fail_add_pcm:
      cx88_unlock_device_tree(common);
      cx88_destroy_channel(&(sc->channel));  
   fail_init_channel:
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   success:
      return error;
}

int cx88_audio_teardown(cx88_audio_softc *sc)
{
   if (!sc->initialized)
   {
      return 0;
   }
   int error = 0;
   sx_slock(&(sc->sxlock));
   cx88_audio_reset(sc, CX88_AUDIO_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel(&(sc->channel));
   sc->pcm_blocksize = 0;
   sx_xunlock(&(sc->sxlock));
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sc->devnode = NULL;  
   }
   if (sc->unit >= 0)
   {
      free_unr(cx88_audio_units, sc->unit);
      sc->unit = -1;
   }
   if (sc->pcm_dev != NULL)
   {
      cx88_lock_device_tree(sc->common);
      error = device_detach(sc->pcm_dev);
      if (!error)
      {
         device_delete_child(sc->common->dev, sc->pcm_dev);
         sc->pcm_dev = NULL;
      }
      cx88_unlock_device_tree(sc->common);
   }
   if (!error)
   {
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      sc->initialized = FALSE;
   }
   return error;
}

void cx88_audio_reset(cx88_audio_softc *sc, cx88_audio_state_t state)
{
   mtx_lock(&(sc->isrlock));
   /* Protect access to shared registers. */
   cx88_lock_board(sc->common);
   /* Stop any data transfer */
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, 
      cx88_read_reg(sc->common, sc->stream->reg_dma_control) &
         ~(sc->stream->dma_control_mask));
   /* Disable interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask,
      cx88_read_reg(sc->common, sc->stream->reg_intr_mask) & 
         ~(sc->stream->intr_mask));
   cx88_unlock_board(sc->common);

   CX88_AUDIO_STOP_CAPTURE(sc->audiocore_dev, sc->audiocore_path);

   /* Reset capture status values */
   sc->isr_status = CX88_AUDIO_WAIT_ABORTED;
   
   /* Our state value should never be higher than CX88_AUDIO_CONFIGURED, as we won't be capturing once reset is done */
   sc->buffer_available = 0;
   sc->state = CX88_AUDIO_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   
   wakeup(sc);
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
}

static void cx88_audio_bufsync(cx88_audio_softc *sc)
{
   if (sc->buffer_available > 0)
   {
      /* Perform any necessary cache/bounce buffer synchronization.
       * Do this in a loop to account for the unlikely case where
       * we've missed an interrupt and multiple buffers need to be 
       * synchronized */
      uint8_t sync_bufindex = sc->last_sync_buffer;
      do
      {
         if (sync_bufindex >= sc->channel.num_buffers)
         {
            sync_bufindex = 0;
         }
         cx88_buffer *buffer = &(sc->channel.buffers[sync_bufindex]);
         /*Synchronize the buffer */
         cx88_mark_transfer_complete(buffer);
         cx88_prepare_for_transfer(buffer);
      }
      while ((++sync_bufindex) != sc->buffer_available);
      sc->last_sync_buffer = sc->buffer_available;
   }
}

static int cx88_audio_transfer(cx88_audio_softc *sc, cx88_audio_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_AUDIO_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   while (!error && !(sc->isr_status & CX88_AUDIO_CAP_FATAL) && (sc->buffer_available == 0))
   {
      if (!nonblocking)
      {
         error = msleep(sc, &(sc->isrlock), PZERO | PCATCH, "cx88au", wait->in_timeout_ms);
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   cx88_audio_bufsync(sc);
   wait->out_status = sc->isr_status;
   wait->out_buffer_available = sc->buffer_available;
   wait->out_stereo_format = 
      (CX88_AUDIO_GET_STEREO(sc->audiocore_dev, sc->audiocore_path) ?
         CX88_AUDIO_STEREO_ON : CX88_AUDIO_STEREO_OFF);
   sc->isr_status = CX88_AUDIO_CAP_NOERR;
   uint32_t stat = cx88_read_reg(sc->common, sc->stream->reg_intr_stat);
   if (stat & sc->stream->ber_mask)
   {
      wait->out_status |= CX88_AUDIO_BERTHRESH;
   }
   if (stat & sc->stream->overflow_mask)
   {
      wait->out_status |= CX88_AUDIO_OVERFLOW;
   }
   sc->buffer_available = 0;
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_ms)
   {
      wait->out_status |= CX88_AUDIO_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_AUDIO_WAIT_ABORTED;
   }
   return error;
}  

static int cx88_audio_start_capture(cx88_audio_softc *sc)
{
   cx88_audio_reset(sc, sc->state);
   if (sc->state != CX88_AUDIO_CONFIGURED)
   {
      return ENXIO;
   }
   int error = 0;
   if ((error = cx88_start_channel(&(sc->channel))))
   {
      return error;
   }
   if (sc->stream->stream != CX88_AUDIO_STREAM_CX23880)
   {
      cx88_write_reg(sc->common, sc->stream->reg_mode, 1);
   }
   cx88_write_reg(sc->common, sc->stream->reg_line_length, sc->channel.line_size);
	sc->last_sync_buffer = 0;
   sc->buffer_available = 0;
   sc->isr_status = CX88_AUDIO_CAP_NOERR;
   sc->state = CX88_AUDIO_CAPTURING;
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_stat,
      cx88_read_reg(sc->common, sc->stream->reg_intr_stat));
   if ((error = CX88_AUDIO_START_CAPTURE(sc->audiocore_dev, sc->audiocore_path)))
   {
      return error;
   }
   /* Protect access to shared registers. */
   cx88_lock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_pci_intr_mask,
      (cx88_read_reg(sc->common, sc->stream->reg_pci_intr_mask) | 
         sc->stream->pci_intr_mask));
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 
      cx88_read_reg(sc->common, sc->stream->reg_intr_mask) | sc->stream->intr_mask);
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, 
      cx88_read_reg(sc->common, sc->stream->reg_dma_control) |
         sc->stream->dma_control_mask);
   cx88_unlock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_devcntl2, CX88_RUN_RISC);
   return 0;
}

/* Configures channel, buffers, and I2C devices */
static int cx88_audio_configure(
   cx88_audio_softc *sc,
   cx88_audio_config *config,
   uint32_t blocksize,
   struct thread *context)
{
   int error = 0;
   if (sc->state == CX88_AUDIO_PCM)
   {
      return EBUSY;
   }
   if (sc->state < CX88_AUDIO_UNCONFIGURED)
   {
      return ENXIO;
   }
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED); /* Ensure there is no outstanding capture */
   uint32_t last_buffer_size = 0;
   uint8_t num_buffers = 0;
   /* Validate ioctl parameters */
   if ((config->in_num_samples < (MIN_AUDIO_BUFFER_SIZE / CX88_AUDIO_SAMPLE_SIZE)) ||
      ((config->in_buftype == CX88_AUDIO_BUF_USER) && (config->bufconfig.in_userbuf == NULL)) ||
      ((config->in_buftype != CX88_AUDIO_BUF_USER) && (config->in_buftype != CX88_AUDIO_BUF_KERNEL)))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Configuring audio input channel\n"))
   if (!error)
   {
      /* Allocate the channel */
      error = cx88_configure_channel(
         &(sc->channel),
         0,
         CX88_AUDIO_SAMPLE_SIZE,
         MAX_AUDIO_LINE_SIZE,
         config->in_num_samples);
   }
   if (!error)
   {
      uint32_t total_lines = (config->in_num_samples * CX88_AUDIO_SAMPLE_SIZE) / sc->channel.line_size;
      uint32_t lines_per_buffer = blocksize / sc->channel.line_size;
      if (lines_per_buffer == 0)
      {
         lines_per_buffer = total_lines / 2;
         if (lines_per_buffer > (sc->pcm_bufsize / sc->channel.line_size / 2))
         {
            lines_per_buffer = sc->pcm_bufsize / sc->channel.line_size / 2;
         }
      }
      num_buffers = total_lines / lines_per_buffer;
      sc->pcm_blocksize = lines_per_buffer * sc->channel.line_size;
      /* Number of samples in last buffer will be less than samples_per_buffer
       * if num_samples is not a multiple of num_buffers */
      last_buffer_size = (total_lines * sc->channel.line_size) - ((num_buffers - 1) * sc->pcm_blocksize);
      config->out_num_buffers = num_buffers;
      config->out_samples_per_buffer = sc->pcm_blocksize / CX88_AUDIO_SAMPLE_SIZE;
      if (config->in_buftype == CX88_AUDIO_BUF_KERNEL)
      {
         /* Calculate the total amount of memory the application will need to mmap(), along
          * with the buffer size the application will need to use in calculating the offsets
          * of individual buffers within that mmap'ed region */
         config->bufconfig.out_kernelbuf.out_buffer_size = round_page(sc->pcm_blocksize);
         config->bufconfig.out_kernelbuf.out_mmap_size = 
            ((num_buffers - 1) * config->bufconfig.out_kernelbuf.out_buffer_size) + last_buffer_size;
      }
      /* Request the buffer structures */
      error = cx88_request_buffers (&(sc->channel), num_buffers, 0);
   }
   if (!error)
   {
      uint32_t i;
      for (i = 0; i < num_buffers; ++i)
      {
         void *bufptr = NULL;
         struct thread *bufcontext = NULL;
         size_t buffer_size = ((i == (num_buffers - 1)) ? last_buffer_size : sc->pcm_blocksize);
         uint32_t byte_offset = i * sc->pcm_blocksize;
         if (config->in_buftype == CX88_AUDIO_BUF_USER)
         {
            /* Each cx88 buffer is a segment of the user-allocated buffer */
            bufptr = (uint8_t *)(config->bufconfig.in_userbuf) + byte_offset;
            bufcontext = context;
         }
         else 
         {
            if (blocksize == 0)
            {
               byte_offset = round_page(byte_offset);
            }
            if ((byte_offset + buffer_size) <= sc->pcm_bufsize)
            {
               DIAGNOSTIC(device_printf(sc->common.dev, "Using PCM data buffer for %d\n", i))
               bufptr = (uint8_t *)(sc->pcm_databuf) + byte_offset;
            }
         }
         
         /* Configure each buffer */
         error = cx88_configure_buffer(
            &(sc->channel.buffers[i]),
            bufptr,
            bufcontext,
            buffer_size,
            ((i == (num_buffers - 1)) ? 0 : (i + 1)),
            cx88_audio_dmaprog,
            CX88_RISC_REPROGRAM_IF_BUFFER_CHANGED);
         if (error)
         {
            break;
         }
      }
   }
   if (!error)
   {
      /* Now commit all changes we've made to the channel configuration */
      error = cx88_commit_channel(&(sc->channel));
   }
   if (!error)
   {
      sc->state = CX88_AUDIO_CONFIGURED;
   }
   else
   {
      cx88_release_channel(&(sc->channel));
   }
   return error;
}

/* Open file operation */
static int cx88_audio_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_audio_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }
   sc->in_use = TRUE;
   sx_xunlock(&sc->sxlock);
   return 0;
}

static void cx88_audio_close_session(cx88_audio_softc *sc)
{
   if (sc->reconfig_requested)
   {
      cx88_audio_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
}

/* Close file operation */
static int cx88_audio_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{   
   cx88_audio_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->state != CX88_AUDIO_PCM)
   {
      cx88_audio_reset(sc, sc->state);
      /* If we're using user-allocated buffers, release them as
      * they can't be persisted across open() and close() in
      * different processes */
      cx88_free_userbufs(&(sc->channel), NULL);
      cx88_audio_close_session(sc);
   }
   sc->in_use = FALSE;
   sc->user_configured = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Mmap file operation */
static int cx88_audio_mmap(
   struct cdev *dev,
#if __FreeBSD_version >= 900006
   vm_ooffset_t offset,
   vm_paddr_t *paddr,
   int nprot,
   vm_memattr_t *memattr)
#else
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
#endif
{
   cx88_audio_softc *sc = dev->si_drv1;
   sx_slock(&(sc->sxlock));
   DIAGNOSTIC(device_printf(sc->common.dev, "Attempting to mmap device memory\n"))
   int error = cx88_common_mmap(&(sc->channel), (vm_ooffset_t)offset, paddr, nprot);
   sx_sunlock(&(sc->sxlock));
   return error;
}

/* Ioctl file operation */
static int cx88_audio_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_audio_softc *sc = dev->si_drv1;
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common->dev);
         break;
      case CX88_AUDIO_FORMAT: /* Establish broadcast/stereo format */
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_AUDIO_UNCONFIGURED) ?
            ENXIO : CX88_AUDIO_SET_FORMAT(sc->audiocore_dev, sc->audiocore_path, (cx88_audio_format*)data));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_CONFIG: /* Configure buffers */
         sx_xlock(&(sc->sxlock));
         error = cx88_audio_configure(sc, (cx88_audio_config*)data, 0, td);
         if (!error)
         {
            sc->user_configured = TRUE;  
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG: /* Configure I2C devices */
         {
            cx88_common_i2c_config *conf = (cx88_common_i2c_config*)data;
            cx88_i2c_softc *i2c_sc = sc->i2c_accessor(sc->common, conf->in_bus_number);
            if (i2c_sc == NULL)
            {
               error = EINVAL;
            }
            else
            {
               error = cx88_i2c_update_devices(i2c_sc, conf);
            }
         }
         break;
      case CX88_AUDIO_CAPTURE: /* Start capturing data */
         sx_xlock(&(sc->sxlock));
         error = cx88_audio_start_capture(sc);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_STOP: /* Stop capturing data */
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         cx88_audio_reset(sc, sc->state);
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_WAIT: /* Wait for data to become available */
         sx_slock(&(sc->sxlock));
         error = cx88_audio_transfer(sc, (cx88_audio_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO:
         cx88_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX88_COMMON_PERIPH_RESET:
         cx88_peripheral_reset(sc->common, *((uint32_t*)data));
         break;
      case CX23885_COMMON_MC_GPIO:
         error = cx23885_mc_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX23885_COMMON_BRIDGE:
         snprintf((char*)data, CX88_COMMON_MAX_DRIVER_NAME, "%s", device_get_nameunit(sc->common->dev));
         break;
      case CX23885_COMMON_GET_STREAMS:
      {
         uint32_t *stream = (uint32_t*)data;
         switch(sc->stream->stream)
         {
            case CX88_AUDIO_STREAM_CX23885_INTA:
               *stream = CX23885_STREAM_AUDIO1;
               break;
            case CX88_AUDIO_STREAM_CX23885_INTB:
               *stream = CX23885_STREAM_AUDIO2;
               break;
            default:
               *stream = CX23885_STREAM_NONE;
               break;
         }
         break;
      }
      case CX88_AUDIO_VOLUME:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_AUDIO_UNCONFIGURED) ?
            ENXIO : CX88_AUDIO_SET_VOLUME(sc->audiocore_dev, sc->audiocore_path, *((cx88_audio_volume*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_MUTE:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_AUDIO_UNCONFIGURED) ?
            ENXIO : CX88_AUDIO_SET_MUTE(sc->audiocore_dev, sc->audiocore_path,
               !CX88_AUDIO_GET_MUTE(sc->audiocore_dev, sc->audiocore_path)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_GET_STATUS:
         sx_slock(&(sc->sxlock));
         error = ((sc->state < CX88_AUDIO_UNCONFIGURED) ?
            ENXIO : CX88_AUDIO_GET_INFO(sc->audiocore_dev, sc->audiocore_path, (cx88_audio_status*)data));
         sx_sunlock(&(sc->sxlock));
         break;
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Select/poll file operation */
static int cx88_audio_poll(struct cdev *dev, int events, struct thread *td)
{   
   cx88_audio_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_AUDIO_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->buffer_available > 0)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

int cx88_audio_filter(cx88_audio_softc *sc)
{
	if (cx88_read_reg(sc->common, sc->stream->reg_intr_mstat) != 0)
	{
		/* Mask out the hardware interrupt until the ISR can handle it */
		cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 0);
		return FILTER_SCHEDULE_THREAD;
	}
	return FILTER_STRAY;
}

/* Interrupt service routine */
void cx88_audio_intr(cx88_audio_softc *sc)
{
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt */
   uint32_t intrs = cx88_read_reg(sc->common, sc->stream->reg_intr_stat);
   boolean_t pcm_intr = FALSE;
   if (intrs & sc->stream->intr_mask)
   {
      if (sc->state < CX88_AUDIO_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      if (intrs & sc->stream->sync_err_mask)
      {
         sc->isr_status |= CX88_AUDIO_SYNC_ERROR;
      }
      /* Check for fatal errors */
      if (intrs & sc->stream->opcode_err_mask)
      {
         sc->isr_status |= CX88_AUDIO_DMA_PROGRAM;
      }
      if (intrs & sc->stream->bus_err_mask)
      {
         sc->isr_status |= CX88_AUDIO_BUS_ERROR;
      }
      if (sc->isr_status & CX88_AUDIO_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_lock_board(sc->common);
         cx88_write_reg(sc->common, sc->stream->reg_dma_control, 
            cx88_read_reg(sc->common, sc->stream->reg_dma_control) & (~(sc->stream->dma_control_mask)));
         cx88_unlock_board(sc->common);
      }
      if (intrs & sc->stream->risci1_mask)
      {
         /* The counter indicates the most recently completed buffer */
         uint32_t bufindex = cx88_read_reg(sc->common, sc->stream->reg_gp_count);
         if (bufindex < sc->channel.num_buffers)
         {
            /* Available index is 1-based */
            sc->buffer_available = bufindex + 1;
            if (sc->state == CX88_AUDIO_PCM)
            {
					cx88_audio_bufsync(sc);
               pcm_intr = TRUE;  
            }
         }
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(sc->common, sc->stream->reg_intr_stat, intrs);
      if (!pcm_intr)
      {
         wakeup(sc);
         selwakeup(&(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   cx88_lock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 
      cx88_read_reg(sc->common, sc->stream->reg_intr_mask) | sc->stream->intr_mask);
   cx88_unlock_board(sc->common);
   if (pcm_intr)
   {
      chn_intr(sc->pcm_chan);
   }
}

static int cx88_audio_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (bufindex == 0)
   {
      /* Initial SYNC at beginning of capture */
      cx88_risc_sync(buffer, 0, CX88_RISC_NO_SYNC_INTR, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   }
   /* Transfer the entire buffer */
   error = cx88_risc_transfer_dmabuf(buffer, 0, (uint32_t)(buffer->data_buffer_size), CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   if (!error)
   {
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
       * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         (bufindex == (buffer->channel->num_buffers - 1) ? CX88_RISC_SYNC_SIZE : 0),
         CX88_TRIGGER_RISCI1,
         ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT));
   }
   return error;
}

static void cx88_audio_reconfigure(cx88_audio_softc *sc)
{
   cx88_free_sram(&(sc->channel));
   if (sc->state > CX88_AUDIO_UNCONFIGURED)
   {
      sc->state = CX88_AUDIO_UNCONFIGURED;  
   }
}

int cx88_audio_request_reconfig(cx88_audio_softc *sc)
{   
   int error = 0;
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if ((sc->in_use) || (sc->state == CX88_AUDIO_PCM))
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_audio_reconfigure(sc);  
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error; 
}

/* Suspend DEVMETHOD */
int cx88_audio_suspend(cx88_audio_softc *sc)
{
   sx_slock(&(sc->sxlock));
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel(&(sc->channel));
   sx_xunlock(&(sc->sxlock));
   return 0;
}

static void *cx88_pcmchan_init(
   kobj_t obj,
   void *devinfo,
   struct snd_dbuf *b,
   struct pcm_channel *c,
   int dir)
{
   cx88_audio_softc *sc = devinfo;
   sc->pcm_buf = b;
   sc->pcm_chan = c;
   sc->pcm_bufsize = round_page(pcm_getbuffersize(sc->pcm_dev,
      MIN_AUDIO_BUFFER_SIZE, DEFAULT_AUDIO_BUFFER_SIZE, MAX_AUDIO_BUFFER_SIZE));
   sc->pcm_databuf = malloc(sc->pcm_bufsize, M_CX88, M_NOWAIT);
   if (sc->pcm_databuf == NULL)
   {
      sc->pcm_bufsize = 0;
      return NULL;
   }
   if (sndbuf_setup(b, sc->pcm_databuf, sc->pcm_bufsize))
   {
      sc->pcm_bufsize = 0;
      free(sc->pcm_databuf, M_CX88);
      sc->pcm_databuf = NULL;
      return NULL;
   }
   return sc;
}

static int cx88_pcmchan_free(kobj_t obj, void *data)
{
   cx88_audio_softc *sc = data;
   sc->pcm_bufsize = 0;
   free(sc->pcm_databuf, M_CX88);  
   sc->pcm_databuf = NULL;
   return 0;
}

static int cx88_pcmchan_setformat(kobj_t obj, void *data, uint32_t format)
{   
   int error = 0;
   cx88_audio_softc *sc = data;
   if (!(sx_try_xlock(&(sc->sxlock))))
   {
      return EBUSY;  
   }
   if (sc->user_configured)
   {
      error = EBUSY;  
   }
#if __FreeBSD_version >= 800097
   boolean_t stereo = ((AFMT_CHANNEL(format) > 1) ? TRUE : FALSE);
#else
   boolean_t stereo = ((format & AFMT_STEREO) ? TRUE : FALSE);
#endif
   if (!error)
   {
      error = CX88_AUDIO_SET_STEREO(sc->audiocore_dev, sc->audiocore_path, stereo);
   }
   sx_xunlock(&(sc->sxlock));
   return error;
}

static uint32_t cx88_pcmchan_setspeed (kobj_t obj, void *data, uint32_t speed)
{
   return CX88_AUDIO_SPEED;  
}
      
static uint32_t cx88_pcmchan_setblocksize(kobj_t obj, void *data, uint32_t blocksize)
{
   int error = 0;
   cx88_audio_softc *sc = data;
   sx_xlock(&(sc->sxlock));
   if (sc->user_configured)
   {
      error = EBUSY;
   }
   else if (sc->pcm_buf == NULL)
   {
      error = ENXIO;
   }
   else
   {
      cx88_audio_config config;
      config.in_num_samples = sc->pcm_bufsize / CX88_AUDIO_SAMPLE_SIZE;
      config.in_buftype = CX88_AUDIO_BUF_KERNEL;
      error = cx88_audio_configure(sc, &config, blocksize, NULL);
   }
   if (error)
   {
      sc->pcm_blocksize = 0;
   }
   sx_xunlock(&(sc->sxlock));
   return sc->pcm_blocksize;
}

static int cx88_pcmchan_trigger (kobj_t obj, void *data, int go)
{
   int error = 0;
   cx88_audio_softc *sc = data;
   if (!(sx_try_xlock(&(sc->sxlock))))
   {
      return EBUSY;  
   }
   switch (go)
   {
      case PCMTRIG_START:
         if (sc->user_configured)
         {
            error = EBUSY;
         }
         else
         {
            error = cx88_audio_start_capture(sc);
            if (!error)
            {
               sc->state = CX88_AUDIO_PCM;
            }
         }
         break;
      case PCMTRIG_ABORT:
      case PCMTRIG_STOP:
         if (sc->state == CX88_AUDIO_PCM)
         {
            cx88_audio_reset(sc, sc->state);
            if (!(sc->in_use))
            {
               cx88_audio_close_session(sc);
            }
         }
         break;
      default:
         break;
   }
   sx_xunlock(&(sc->sxlock));
   return error;
}

static uint32_t cx88_pcmchan_getptr(kobj_t obj, void *data)
{
   cx88_audio_softc *sc = data;
   if (sc->buffer_available <= sc->channel.num_buffers)
   {
      if (sc->buffer_available == sc->channel.num_buffers)
      {
         return sc->pcm_bufsize;  
      }
      else
      {
         return (sc->buffer_available * sc->pcm_blocksize);  
      }
   }
   return 0;
}

static struct pcmchan_caps * cx88_pcmchan_getcaps(kobj_t obj, void *devinfo)
{
   return (&cx88_audio_caps);
}

static kobj_method_t cx88_pcmchan_methods[] = 
{
   KOBJMETHOD(channel_init,            cx88_pcmchan_init),
   KOBJMETHOD(channel_free,            cx88_pcmchan_free),
   KOBJMETHOD(channel_setformat,       cx88_pcmchan_setformat),
   KOBJMETHOD(channel_setspeed,        cx88_pcmchan_setspeed),
   KOBJMETHOD(channel_setblocksize,    cx88_pcmchan_setblocksize),
   KOBJMETHOD(channel_trigger,         cx88_pcmchan_trigger),
   KOBJMETHOD(channel_getptr,          cx88_pcmchan_getptr),
   KOBJMETHOD(channel_getcaps,         cx88_pcmchan_getcaps),
   {NULL, NULL}
};
CHANNEL_DECLARE(cx88_pcmchan);

static int cx88_mixer_init(struct snd_mixer *m)
{
   mix_setdevs(m, SOUND_MASK_LINE);
   mix_setrecdevs(m, SOUND_MASK_LINE);
   return 0;  
}

static int cx88_mixer_uninit(struct snd_mixer *m)
{
   cx88_audio_softc *sc = mix_getdevinfo(m);
   CX88_AUDIO_SET_MUTE(sc->audiocore_dev, sc->audiocore_path, TRUE);
   return 0;  
}

static int cx88_mixer_set(
   struct snd_mixer *m,
   unsigned int dev,
   unsigned int left,
   unsigned int right)
{
   cx88_audio_softc *sc = mix_getdevinfo(m);
   cx88_audio_volume vol;
   vol.in_left = (uint8_t)left;
   vol.in_right = (uint8_t)right;
   int retval = (left | (right << 8));
   if (!(sx_try_xlock(&(sc->sxlock))))
   {
      return -1;  
   }
   if (CX88_AUDIO_SET_VOLUME(sc->audiocore_dev, sc->audiocore_path, vol) != 0)
   {
      retval = -1;
   }
   sx_xunlock(&(sc->sxlock));
   return retval;
}

static uint32_t cx88_mixer_setrecsrc(struct snd_mixer *m, uint32_t src)
{
   return SOUND_MASK_LINE;  
}

static kobj_method_t cx88_mixer_methods[] = 
{
   KOBJMETHOD(mixer_init,          cx88_mixer_init),
   KOBJMETHOD(mixer_uninit,        cx88_mixer_uninit),
   KOBJMETHOD(mixer_set,           cx88_mixer_set),
   KOBJMETHOD(mixer_setrecsrc,     cx88_mixer_setrecsrc),
   {NULL, NULL}
};
MIXER_DECLARE(cx88_mixer);

static int cx88_pcm_probe(device_t dev)
{
   device_set_desc(dev, "CX2388x PCM interface");
   return BUS_PROBE_SPECIFIC;
}

static int cx88_pcm_attach(device_t dev)
{
   int error = 0;
   char status[SND_STATUSLEN];
   cx88_audio_softc *sc = device_get_ivars(dev);
   error = mixer_init(dev, &cx88_mixer_class, sc);
   if (!error)
   {
      pcm_setflags(dev, pcm_getflags(dev) | SD_F_MPSAFE); 
      error = pcm_register(dev, sc, 0, 1);
   }
   if (!error)
   {
      error = pcm_addchan(dev, PCMDIR_REC, &cx88_pcmchan_class, sc);
   }
   if (!error)
   {
      snprintf(status, SND_STATUSLEN, "on %s", device_get_nameunit(device_get_parent(dev)));
      pcm_setstatus(dev, status);
   }
   return error;
}

static int cx88_pcm_detach(device_t dev)
{
   return pcm_unregister(dev);
}

static device_method_t cx88_pcm_methods[] = 
{
   DEVMETHOD(device_probe,             cx88_pcm_probe),
   DEVMETHOD(device_attach,            cx88_pcm_attach),
   DEVMETHOD(device_detach,            cx88_pcm_detach),
   {NULL, NULL}
};

static int cx88audio_mod_handler(module_t mod, int what, void *arg)
{
   switch(what)
   {
      case MOD_LOAD:
         cx88_audio_units = new_unrhdr(0, INT_MAX, NULL);
         if (cx88_audio_units == NULL)
         {
            return ENOMEM;
         }
         return 0;
      case MOD_UNLOAD:
         if (cx88_audio_units != NULL)
         {
            delete_unrhdr(cx88_audio_units);
            cx88_audio_units = NULL;
         }
         return 0;
      default:
         return 0;
   }
   return 0;
}

/* Module declarations */
static moduledata_t cx88audio_mod_data =
{
   "cx88audiocore",
   cx88audio_mod_handler,
   NULL
};

DECLARE_MODULE(cx88audiocore, cx88audio_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_DEPEND(cx88audiocore, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);
MODULE_DEPEND(cx88audiocore, cx88_i2c, CX88_I2C_MINVER, CX88_I2C_MODVER, CX88_I2C_MAXVER);
MODULE_VERSION(cx88audiocore, CX88_AUDIO_MODVER);

driver_t cx88_pcm_driver = 
{
   "pcm",
   cx88_pcm_methods,
   PCM_SOFTC_SIZE
};

MODULE_DEPEND(cx88pcm, sound, SOUND_MINVER, SOUND_PREFVER, SOUND_MAXVER);
MODULE_VERSION(cx88pcm, CX88_AUDIO_MODVER);
