/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#ifndef __CX88_AUDIO_H__
#define __CX88_AUDIO_H__

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include <dev/sound/pcm/sound.h>
#include "mixer_if.h"
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_audio_regs.h"
#include "cx88_audio_ioctl.h"
#include "cx88_audio_if.h"

#define CX88_AUDIO_MINVER 2
#define CX88_AUDIO_MODVER 2
#define CX88_AUDIO_MAXVER 2

typedef enum
{
   CX88_AUDIO_STREAM_CX23880 = 0,
   CX88_AUDIO_STREAM_CX23885_INTA,
   CX88_AUDIO_STREAM_CX23885_INTB,
   CX88_AUDIO_STREAM_CX23885_EXT,
   CX88_AUDIO_NUM_STREAMS
} cx88_audio_stream_t;

typedef struct
{
   cx88_audio_stream_t stream;
   cx88_channel_t dma_channel;
   uint32_t reg_gp_count;
   uint32_t reg_dma_control;
   uint32_t dma_control_mask;
   uint32_t reg_line_length;
   uint32_t reg_pci_intr_mask;
   uint32_t pci_intr_mask;
   uint32_t reg_intr_mask;
   uint32_t reg_intr_stat;
   uint32_t reg_intr_mstat;
   uint32_t intr_mask;
   uint32_t risci1_mask;
   uint32_t opcode_err_mask;
   uint32_t sync_err_mask;
   uint32_t bus_err_mask;
   uint32_t overflow_mask;
   uint32_t ber_mask;
   uint32_t reg_devcntl2;
   uint32_t reg_mode;
} cx88_audio_stream_desc;

/* Driver states */
typedef enum
{
   CX88_AUDIO_INVALID = 0, /* Uninitialized */
   CX88_AUDIO_UNCONFIGURED, /* Not configured or formatted */
   CX88_AUDIO_CONFIGURED, /* Configured (I2C devices, buffers, etc.) */
   CX88_AUDIO_CAPTURING, /* Configured and capturing */
   CX88_AUDIO_PCM
} cx88_audio_state_t;

typedef struct
{
   const cx88_audio_stream_desc *stream;
   cx88_common_softc *common;
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t);
   cx88_channel channel;
   device_t audiocore_dev;
   uint8_t audiocore_path;
   device_t pcm_dev;
   struct pcm_channel *pcm_chan;
   struct snd_dbuf *pcm_buf;
   struct cdev *devnode;
   int unit; /* Unit number for devnode */
   struct mtx isrlock;
   struct sx sxlock;
   struct selinfo selectors;
   boolean_t initialized;
   boolean_t in_use;
   boolean_t reconfig_requested;
   boolean_t user_configured;
   cx88_audio_state_t state;
   volatile uint8_t isr_status; /* Error status from interrupt service routine */
	uint8_t last_sync_buffer;
   volatile uint8_t buffer_available;
   unsigned int pcm_bufsize;
   uint32_t pcm_blocksize;
   void *pcm_databuf;
} cx88_audio_softc;

extern driver_t cx88_pcm_driver;

#define CX88_PCM_DRIVER_MODULE(host) \
   DRIVER_MODULE(cx88pcm, host, cx88_pcm_driver, pcm_devclass, NULL, NULL); \
   MODULE_DEPEND(host, sound, SOUND_MINVER, SOUND_PREFVER, SOUND_MAXVER);

int cx88_audio_setup(
   cx88_audio_softc *sc,
   cx88_audio_stream_t stream,
   cx88_common_softc *common,
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t),
   device_t audiocore_dev,
   uint8_t audiocore_path);
int cx88_audio_teardown(cx88_audio_softc *sc);
int cx88_audio_filter(cx88_audio_softc *sc);
void cx88_audio_intr(cx88_audio_softc *sc);
int cx88_audio_suspend(cx88_audio_softc *sc);
void cx88_audio_reset(cx88_audio_softc *sc, cx88_audio_state_t state);
int cx88_audio_request_reconfig(cx88_audio_softc *sc);

#endif
