/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <sys/errno.h>
#include "cx88_video_capture_config.h"
#include "cx88_ffmpeg_capture.h"

cx88_ffmpeg_capture::cx88_ffmpeg_capture(const char *filename, int &error)
   : cx88_file_capture(filename, error, 0),
     m_context(NULL)
{
   for (uint8_t i = 0; i < 3; ++i)
   {
      m_buffers[i] = NULL;
      m_src_linesize[i] = 0;
      m_src_lines[i] = 0;
      m_dst_linesize[i] = 0;
      m_dst_lines[i] = 0;
   }
}

cx88_ffmpeg_capture::~cx88_ffmpeg_capture(void)
{   
   if (m_context != NULL)
   {
      sws_freeContext(m_context);
      m_context = NULL;
   }
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      free(m_buffers[i]);
      m_buffers[i] = NULL;
      m_src_linesize[i] = 0;
      m_src_lines[i] = 0;
      m_dst_linesize[i] = 0;
      m_dst_lines[i] = 0;
   }
}

int cx88_ffmpeg_capture::initialize(cx88_capture_config *conf)
{
   cx88_video_capture_config *videoconfig = dynamic_cast<cx88_video_capture_config*>(conf);
   if (videoconfig == NULL)
   {
      return EINVAL;
   }
   cx88_video_format *format = videoconfig->get_format();
   cx88_video_config *config = videoconfig->get_config();
   for (uint8_t i = 0; i < 3; ++i)
   {
      m_src_linesize[i] = (int)(format->out_geoms[i].line_size + config->bufconfig[i].in_x_header + 
         config->bufconfig[i].in_x_trailer);
      m_src_lines[i] = (int)(format->out_geoms[i].num_lines + config->bufconfig[i].in_y_header +
         config->bufconfig[i].in_y_trailer);
   }
   int pixfmt = PIX_FMT_NONE;
   uint8_t swap = format->in_even_field.swap | format->in_odd_field.swap;
   uint8_t pixel_size = 1;
   switch (format->in_pixel_format)
   {
      case CX88_VIDEO_RGB32:
         if (swap == (CX88_VIDEO_SWAP_WORDS | CX88_VIDEO_SWAP_BYTES))
         {
            pixfmt = PIX_FMT_RGB32;
         }
         else
         {
            pixfmt = PIX_FMT_BGR32;
         }
         pixel_size = 4;
         break;
      case CX88_VIDEO_BGR24:
         pixfmt = PIX_FMT_BGR24;
         pixel_size = 3;
         break;
      case CX88_VIDEO_RGB16:
         if (swap & CX88_VIDEO_SWAP_BYTES)  
         {
            pixfmt = PIX_FMT_RGB565;
         }
         else
         {
            pixfmt = PIX_FMT_BGR565;
         }
         pixel_size = 2;
         break;
      case CX88_VIDEO_RGB15:
         if (swap & CX88_VIDEO_SWAP_BYTES)  
         {
            pixfmt = PIX_FMT_RGB555;
         }
         else
         {
            pixfmt = PIX_FMT_BGR555;
         }
         pixel_size = 2;
         break;
      case CX88_VIDEO_YUV422:
         if (swap & CX88_VIDEO_SWAP_BYTES)  
         {
            pixfmt = PIX_FMT_UYVY422;
         }
         else
         {
            pixfmt = PIX_FMT_YUYV422;
         }
         break;
      case CX88_VIDEO_Y8:
         pixfmt = PIX_FMT_GRAY8;
         break;
      case CX88_VIDEO_YCC422:
         pixfmt = PIX_FMT_YUV422P;
         break;
      case CX88_VIDEO_YCC411:
         pixfmt = PIX_FMT_YUV411P;
         break;
      default:
         CX88ERR << "No conversion available for pixel format " << format->in_pixel_format << endl;
         return EINVAL;
   }
   m_dst_linesize[0] = (m_src_linesize[0] + (pixel_size - 1)) / pixel_size;
   m_dst_linesize[1] = (m_dst_linesize[0] + 1) / 2;
   m_dst_linesize[2] = m_dst_linesize[1];
   m_dst_lines[0] = m_src_lines[0];
   m_dst_lines[1] = (m_dst_lines[0] + 1) / 2;
   m_dst_lines[2] = m_dst_lines[1];
   for (uint8_t i = 0; i < 3; ++i)
   {
      size_t bufsize = (size_t)(m_dst_linesize[i] * m_dst_lines[i]);
      void *buffer = realloc(m_buffers[i], bufsize);
      if (buffer == NULL)
      {
         return ENOMEM;
      }
      m_buffers[i] = (uint8_t*)buffer;
   }
   if (m_context != NULL)
   {
      sws_freeContext(m_context);
      m_context = NULL;
   }
   m_context = sws_getContext(
      m_dst_linesize[0],
      m_dst_lines[0],
      pixfmt,
      m_dst_linesize[0],
      m_dst_lines[0],
      PIX_FMT_YUV420P,
      SWS_BICUBIC,
      NULL, NULL, NULL);
   if (m_context == NULL)
   {
      CX88ERR << "Unable to generate swscaler context" << endl;
      return ENOMEM;
   }
   return 0;
}

void cx88_ffmpeg_capture::process_buffer(void *data, size_t num_bytes, size_t packet_size)
{
   uint8_t *buffers[] = {(uint8_t*)data, NULL, NULL};
   size_t bufsize = (size_t)(m_src_linesize[0] * m_src_lines[0]);
   for (uint8_t i = 1; (i < 3) && (bufsize < num_bytes); ++i)
   {
      buffers[i] = (uint8_t*)data + bufsize;
      bufsize += (size_t)(m_src_linesize[i] * m_src_lines[i]);  
   }
   if (bufsize != num_bytes)
   {
      CX88ERR << "cx88_ffmpeg_capture: Ignoring unexpected frame size " << num_bytes << endl;
      return;
   }
   int error = sws_scale(m_context, buffers, m_src_linesize, 0,
      m_src_lines[0], m_buffers, m_dst_linesize);
   if (error)
   {
      CX88ERR << "cx88_ffmpeg_capture: sws_scale() returned error " << error << endl;
      return;
   }
   for (uint8_t i = 0; i < 3; ++i)
   {
      cx88_file_capture::process_buffer(m_buffers[i],
         (size_t)(m_dst_linesize[i] * m_dst_lines[i]), packet_size);
   }
}
