/* cx88_url_capture.c
 *
 * open connection to another machine
 *
 * URLs currently supported:
 *
 *   tcp://hostname:port
 *   udp://hostname:port
 *
 * URLs not currently supported, but might be in the future:
 *
 *   file:///filename   (This would free up the -f option.)
 *
 * Notes and Caveats:
 *
 *   The protocol (tcp or udp) is case insensitive.
 *
 *   The hostname can be a numeric IP address (e.g. 10.0.200.66)
 *   or a hostname (e.g. blue).
 *   Full domain names have not been tested. (e.g. foo.bar.com)
 *   Only hosts on a LAN have been tested.
 *   (Maybe someday we will all have 20 Mbps IP connections.)
 *   Hostnames longer than 1000 characters are not supported.
 *
 *   The port is required.
 *   Port numbers greater than 65535 are not supported.
 *   Trailing garbage after the port is ignored.  E.g. tcp://blue:1234garbage
 *
 *   The MTU defaults to 1500.  
 */

/*-
 * Copyright 2007,2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <ctype.h>
#include <netinet/in.h>
#include "cx88_debug.h"
#include "cx88_url_capture.h"

using namespace std;

cx88_url_capture::cx88_url_capture(
   const char *url_arg,  /* URL argument from the command line */
   int &error,
   size_t mtu /* = CX88_DEFAULT_MTU*/)
   : cx88_capture(error),
     m_socket(-1),
     m_mtu(mtu)
{
   if (error)
   {
      return;
   }
   const char *protocol = url_arg;      /* tcp or udp */
   const char *hostname = strstr(protocol, "://");   /* hostname or IP address of target machine */
  
   struct addrinfo hints, *res, *res0;
  
   /* Parse the URL argument. */
  
   /* parse protocol */
  
   if (hostname == NULL)
   {
      CX88ERR << "Missing protocol in URL; must be protocol://hostname:port" << endl;
      error = EINVAL;
      return;
   }
   hostname += 3;  // Account for the "://"
  
   /* parse port */
   char *port = strchr(hostname, ':');
   if (port == NULL)
   {
      CX88ERR << "Missing port in URL; must be protocol://hostname:port" << endl;
      error = EINVAL;
      return;
   }
   ++port;
   char *endptr = NULL;
   unsigned long lport = strtoul(port, &endptr, 10);
   if (endptr == port)
   {
      CX88ERR << "Empty or invalid port string after ':'" << endl;
      error = EINVAL;
      return;
   }
   if (lport > 65535)
   {
      CX88ERR << "Invalid port number (" << lport << "), must be <= 65535" << endl;
      error = EINVAL;
      return;
   }
   
   /* Extract hostname */
   
   size_t hostname_len = port - hostname;
   char *hostname_copy = (char*)(malloc(hostname_len));
   if (hostname_copy == NULL)
   {
      CX88ERR << "Unable to allocate memory for hostname parsing" << endl;
      error = ENOMEM;
      return;  
   }
   strncpy(hostname_copy, hostname, hostname_len - 1);
   hostname_copy[hostname_len - 1] = '\0';
  
   /* Set up connection (code based on example from FreeBSD's getaddrinfo(3) man page) */
  
   bzero(&hints, sizeof(hints));
   hints.ai_family = PF_UNSPEC;
   if ((hostname - protocol) != 6)
   {
      CX88ERR << "Protocol must be tcp or udp" << endl;
      free(hostname_copy);
      error = EINVAL;
      return;
   }
   if (strncmp(protocol, "tcp", 3) == 0)
   {
      hints.ai_socktype = SOCK_STREAM;
      hints.ai_protocol = IPPROTO_TCP;
   }
   else if (strncmp(protocol, "udp", 3) == 0)
   {
      hints.ai_socktype = SOCK_DGRAM;
      hints.ai_protocol = IPPROTO_UDP;
   }
   else
   {
      CX88ERR << "Protocol must be tcp or udp" << endl;
      free(hostname_copy);
      error = EINVAL;
      return;
   }
   DIAGNOSTIC(CX88ERR << "Attempting to connect to host " << hostname_copy << " on port " << port << endl);
   int net_error = getaddrinfo(hostname_copy, port, &hints, &res0);
   free(hostname_copy);
   if (net_error)
   {
      CX88ERR << "Unable to getaddrinfo() " << gai_strerror(net_error) << endl;
      error = EIO;
      return;
   }
   for (res = res0; res; res = res->ai_next)
   {
      m_socket = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (m_socket < 0)
	   {
	      error = errno;
	      continue;
	   }
      if ((error = connect(m_socket, res->ai_addr, res->ai_addrlen)) < 0)
	   {
	      close(m_socket);
	      m_socket = -1;
	      continue;
	   }
      error = 0;
      break;  /* okay we got one */
   }
   if (error)
   {
      CX88ERR << "Unable to connect to host: " << strerror(error) << endl;
   }
   freeaddrinfo(res0);
}


cx88_url_capture::~cx88_url_capture(void)
{
   stop();
   if (m_socket >= 0)
   {
      close(m_socket);
   }
}

void cx88_url_capture::process_buffer(void *data, size_t num_bytes, size_t packet_size)
{
   char *bufptr = reinterpret_cast<char*>(data);
   size_t count = 0;
   ssize_t writecount;
   ssize_t msglen = m_mtu;
   if (packet_size != 0)
   {
      msglen = (m_mtu / packet_size) * packet_size;
   }
   while (count < num_bytes)
   {
      writecount = num_bytes - count;
      if (writecount > msglen)
      {
         writecount = msglen;
      }
      writecount = write(m_socket, bufptr, writecount);
      if (writecount > 0)
      {
         bufptr += writecount;
         count += writecount;
      }
      if (writecount < 0)
      {
         DIAGNOSTIC(CX88ERR << "Unable to write to socket: " << strerror(errno) << endl);
         break;
      }
   }
}
