/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/errno.h>
#include <sstream>
#include "cx88_debug.h"
#include "cx88_xml_config.h"

using namespace std;

cx88_xml_config::cx88_xml_config(const char *filename, int &error)
   : m_file(NULL),
     m_default_channel(CX88_XML_CHANNEL_FIRST)
{
   if (error)
   {
      return;
   }
   m_file = ezxml_parse_file(filename);
   if (m_file == NULL)
   {
      CX88ERR << "Error parsing XML file" << endl;
      error = EINVAL;
      return;
   }
   ezxml_t config = ezxml_child(m_file, CX88_XML_CONFIG_NAME);
   if (config == NULL)
   {
      return;
   }
   ezxml_t alias = ezxml_child(config, CX88_XML_ALIAS_NAME);
   try
   {
      while (alias != NULL)
      {  
         string name(ezxml_attr(alias, CX88_XML_NAME_NAME));
         string channel(ezxml_attr(alias, CX88_XML_CHANNEL_NAME));
         m_aliases.insert(pair<string, string> (name, channel));
         alias = alias->next;     
      }
      ezxml_t default_channel = ezxml_child(config, CX88_XML_CHANNEL_DEFAULT);
      if (default_channel != NULL)
      {
         m_default_channel = default_channel->txt;  
      }
   }
   catch(...)
   {
      error = ENOMEM;
   }
}

cx88_xml_config::~cx88_xml_config(void)
{
   if (m_file != NULL)
   {
      ezxml_free(m_file);
      m_file = NULL;  
   }
}

cx88_xml_channel::cx88_xml_channel(cx88_xml_config &config, tuner_config &tunerconf, int &error)
   : m_config(config),
     m_profile(NULL),
     m_range(m_ranges.end()),
     m_alias(config.m_aliases.end()),
     m_channel(0),
     m_subchannel(0),
     m_has_subchannel(false),
     m_tunerconf(tunerconf)
{
   if (error)
   {
      return;
   }
   ezxml_t config_section = ezxml_child(config.m_file, CX88_XML_CONFIG_NAME);
   if (config_section == NULL)
   {
      return;
   }
   if ((error = parse_libtuner(config_section, m_tunerconf)))
   {
      return;
   }
}

int cx88_xml_channel::parse_libtuner(ezxml_t parent, tuner_config &conf)
{
   int error = 0;
   ezxml_t libtuner = ezxml_child(parent, CX88_XML_LIBTUNER_NAME);
   while (libtuner != NULL)
   {
      if ((error = conf.load_string(libtuner->txt)))
      {
         return error;
      }
      libtuner = libtuner->next;
   }
   return 0;
}

int cx88_xml_channel::select_profile(const char *name)
{
   ezxml_t curprofile = ezxml_child(m_config.m_file, CX88_XML_PROFILE_NAME);
   if ((m_profile != NULL) && ((name == NULL) || (strcmp(name, ezxml_attr(m_profile, CX88_XML_NAME_NAME)) == 0)))
   {
      return 0;
   }
   while (curprofile != NULL)
   {
      const char *profile_name = ezxml_attr(curprofile, CX88_XML_NAME_NAME);
      if ((name == NULL) || (strcmp(name, profile_name) == 0))
      {
         if ((curprofile != m_profile) && (m_range != m_ranges.end()))
         {
            m_tunerconf.remove_config(m_range->tunerconf);
         }
         m_profile = curprofile;
         try
         {
            m_ranges.clear();
            ezxml_t cur_range = ezxml_child(m_profile, CX88_XML_RANGE_NAME);
            while (cur_range != NULL)
            {
               unsigned long start_chan = 0, end_chan = 0;
               ezxml_t start = ezxml_child(cur_range, CX88_XML_START_NAME);
               ezxml_t end = ezxml_child(cur_range, CX88_XML_END_NAME);
               if (start != NULL)
               {
                  start_chan = strtoul(start->txt, NULL, 10);
               }
               if (end != NULL)
               {
                  end_chan = strtoul(end->txt, NULL, 10);
               }
               cx88_range_t range;
               range.start = start_chan;
               range.end = end_chan;
               range.elem = cur_range;
               int error = parse_libtuner(range.elem, range.tunerconf);
               if (error != 0)
               {
                  return error;
               }
               m_ranges.push_back(range);
               cur_range = cur_range->next;
            }
         }
         catch (...)
         {
            return ENOMEM;
         }
         break;
      }
      curprofile = curprofile->next;  
   }
   m_range = m_ranges.begin();
   if (m_range != m_ranges.end())
   {
      m_channel = m_range->start;
   }
   return 0;
}

const char *cx88_xml_channel::get_profile_name(void)
{
   if (m_profile == NULL)
   {
      return NULL;
   }
   return ezxml_attr(m_profile, CX88_XML_NAME_NAME);   
}

int cx88_xml_channel::select_channel(unsigned long channel)
{
   if (m_profile == NULL)
   {
      return EINVAL;
   }
   if ((channel >= m_range->start) && (channel <= m_range->end))
   {
      m_channel = channel;
      return 0;  
   }
   for (rangelist::iterator iter = m_ranges.begin(); iter != m_ranges.end(); ++iter)
   {
      if ((iter->start <= channel) && (iter->end >= channel))
      {
         m_tunerconf.remove_config(m_range->tunerconf);
         m_range = iter;
         m_channel = channel;
         return m_tunerconf.add_config(m_range->tunerconf);
      }
   }
   return ENOENT;
}

int cx88_xml_channel::select_channel(const string& name)
{
   int error = 0;
   string channel = name;
   if (name == CX88_XML_CHANNEL_UP)
   {
      return channel_up();
   }
   else if (name == CX88_XML_CHANNEL_DOWN)
   {
      return channel_down();
   }
   else if (name == CX88_XML_CHANNEL_DEFAULT)
   {
      channel = m_config.m_default_channel;  
   }
   m_alias = m_config.m_aliases.find(channel);
   if (m_alias != m_config.m_aliases.end())
   {
      channel = m_alias->second;
   }
   m_subchannel = 0;
   m_has_subchannel = false;
   string::size_type chan_begin = channel.find_first_of(':');
   string::size_type chan_end = channel.find_last_of('-');
   if (chan_begin == string::npos)
   {
      chan_begin = 0;
   }
   else
   {
      chan_begin += 1;
   }
   if (chan_end == string::npos)
   {
      chan_end = channel.length();
   }
   string chanstr = channel.substr(chan_begin, chan_end - chan_begin);;
   string profile;
   const char *profile_name = NULL;
   if (chan_begin != 0)
   {
      profile = channel.substr(0, chan_begin - 1);
      profile_name = profile.c_str();
   }
   error = select_profile(profile_name);
   if (!error)
   {
      if (chanstr == CX88_XML_CHANNEL_FIRST)
      {
         return select_first_channel();
      }
      else if (chanstr == CX88_XML_CHANNEL_LAST)
      {
         return select_last_channel();
      }
      if (chan_end != channel.length())
      {
         string subchanstr = channel.substr(chan_end + 1, channel.length() - chan_end - 1);
         m_subchannel = strtoul(subchanstr.c_str(), NULL, 10);
         m_has_subchannel = true;
      }
      error = select_channel(strtoul(chanstr.c_str(), NULL, 10));
   }
   return error;
}

const char *cx88_xml_channel::get_property(const char *property)
{
   ezxml_t child = ezxml_child(m_range->elem, property);
   ezxml_t cur_level = m_range->elem;
   while (child == NULL)
   {
      cur_level = cur_level->parent;
      if (cur_level == NULL)
      {
         break;
      }
      child = ezxml_child(cur_level, property);
   }
   if (child != NULL)
   {
      return child->txt;
   }
   return NULL;
}

template <typename payload>
size_t cx88_xml_channel::lookup(struct table_entry<payload> *table, size_t table_size, const char *value)
{
   DIAGNOSTIC(
      CX88LOG << "Checking table for correct sequence" << endl;
      for (size_t i = 1; i < table_size; ++i)
      {
         if (strcasecmp(table[i].name, table[i - 1].name) < 0)
         {
            CX88ERR << "Table entry " << table[i].name << " is not lexicographically greater than table entry " << table[i - 1].name << endl;
            return table_size;
         }
      }
   )
   size_t min_index = 0, max_index = table_size;
   while (min_index != max_index)
   {
      size_t middle_index = (min_index + max_index) / 2;
      int cmp = strcasecmp(table[middle_index].name, value);
      if (cmp == 0)
      {
         return middle_index;
      }
      else if (cmp < 0)
      {
         min_index = middle_index + 1;  
      }
      else
      {
         max_index = middle_index;  
      }
   }
   return table_size;
}

int cx88_xml_channel::parse_modulation(const char *mod_string, dvb_modulation_t &modulation)
{
   int error = 0;
   static struct table_entry<dvb_modulation_t> dvb_mod_table[] =
   {
      {"16PSK", DVB_MOD_16PSK},
      {"8PSK", DVB_MOD_8PSK},
      {"OFDM", DVB_MOD_OFDM},  
      {"QAM", DVB_MOD_QAM_AUTO},
      {"QAM128", DVB_MOD_QAM_128},
      {"QAM16", DVB_MOD_QAM_16},
      {"QAM256", DVB_MOD_QAM_256},
      {"QAM32", DVB_MOD_QAM_32},
      {"QAM64", DVB_MOD_QAM_64},
      {"QPSK", DVB_MOD_QPSK},
      {"QPSK-NBC", DVB_MOD_QPSK_NBC},
      {"VSB16", DVB_MOD_VSB_16},
      {"VSB8", DVB_MOD_VSB_8}
   };
   size_t table_size = sizeof(dvb_mod_table) / sizeof(dvb_mod_table[0]);
   size_t index = lookup(dvb_mod_table, table_size, mod_string);
   if (index >= table_size)
   {
      CX88ERR << "Unrecognized DVB modulation type: " << modulation << endl;
      error = EINVAL;
   }
   else
   {
      modulation = dvb_mod_table[index].value;  
   }
   return error;  
}

int cx88_xml_channel::parse_fec(const char *fec_string, dvb_fec_t &fec)
{
   int error = 0;
   static struct table_entry<dvb_fec_t> dvb_fec_table[] =
   {
      {"1/2", DVB_FEC_1_2},  
      {"2/3", DVB_FEC_2_3},
      {"3/4", DVB_FEC_3_4},
      {"3/5", DVB_FEC_3_5},
      {"4/5", DVB_FEC_4_5},
      {"5/6", DVB_FEC_5_6},
      {"6/7", DVB_FEC_6_7},
      {"7/8", DVB_FEC_7_8},
      {"8/9", DVB_FEC_8_9},
      {"9/10", DVB_FEC_9_10},
      {"auto", DVB_FEC_AUTO},
      {"none", DVB_FEC_NONE}
   };
   size_t table_size = sizeof(dvb_fec_table) / sizeof(dvb_fec_table[0]);
   size_t index = lookup(dvb_fec_table, table_size, fec_string);
   if (index >= table_size)
   {
      CX88ERR << "Unrecognized code rate type: " << fec_string << endl;
      error = EINVAL;
   }
   else
   {
      fec = dvb_fec_table[index].value;
   }
   return error;
}

int cx88_xml_channel::parse_qam(dvb_qam_settings &qam)
{
   int error = 0;
   qam.symbol_rate = DVB_SYMBOL_RATE_AUTO;
   qam.code_rate = DVB_FEC_AUTO;
   const char *symbol_rate = get_property(CX88_XML_SYMBOL_RATE);
   if (symbol_rate != NULL)
   {
      qam.symbol_rate = (uint32_t)(strtoul(symbol_rate, NULL, 10));
   }
   const char *fec = get_property(CX88_XML_CODE_RATE);
   if (fec != NULL)
   {
      error = parse_fec(fec, qam.code_rate);
   }
   return error;
}

int cx88_xml_channel::parse_ofdm(dvb_ofdm_settings &ofdm)
{
   int error = 0;
   ofdm.subcarrier_modulation = DVB_MOD_UNKNOWN;
   ofdm.high_prio_code_rate = DVB_FEC_AUTO;
   ofdm.low_prio_code_rate = DVB_FEC_AUTO;
   ofdm.subcarrier_mode = DVB_OFDM_AUTO_SUBCARRIERS;
   ofdm.guard_interval = DVB_OFDM_GUARD_AUTO;
   ofdm.hierarchy = DVB_OFDM_HIERARCHY_AUTO;
   size_t table_size, index;
   const char *submodulation = get_property(CX88_XML_SUBMODULATION);
   if (submodulation != NULL)
   {
      error = parse_modulation(submodulation, ofdm.subcarrier_modulation);  
   }
   const char *hp_code_rate = get_property(CX88_XML_CODE_RATE_HP);
   if (!error && (hp_code_rate != NULL))
   {
      error = parse_fec(hp_code_rate, ofdm.high_prio_code_rate);
   }
   const char *lp_code_rate = get_property(CX88_XML_CODE_RATE_LP);
   if (!error && (lp_code_rate != NULL))
   {
      error = parse_fec(lp_code_rate, ofdm.low_prio_code_rate);
   }
   if (error)
   {
      return error;
   }
   const char *subcarriers = get_property(CX88_XML_SUBCARRIERS);
   static struct table_entry<dvb_subcarrier_mode_t> subcarriers_table[] =
   {
      {"2k", DVB_OFDM_2K_SUBCARRIERS},  
      {"4k", DVB_OFDM_4K_SUBCARRIERS},
      {"8k", DVB_OFDM_8K_SUBCARRIERS},
      {"auto", DVB_OFDM_AUTO_SUBCARRIERS}
   };
   if (subcarriers != NULL)
   {
      table_size = sizeof(subcarriers_table) / sizeof(subcarriers_table[0]);
      index = lookup(subcarriers_table, table_size, subcarriers);
      if (index >= table_size)
      {
         CX88ERR << "Unrecognized subcarrier mode: " << subcarriers << endl;
         return EINVAL;
      }
      else
      {
         ofdm.subcarrier_mode = subcarriers_table[index].value;
      }
   }
   const char *guard_interval = get_property(CX88_XML_GUARD_INTERVAL);
   static struct table_entry<dvb_guard_interval_t> ofdm_guard_table[] =
   {
      {"1/16", DVB_OFDM_GUARD_1_16},  
      {"1/32", DVB_OFDM_GUARD_1_32},
      {"1/4", DVB_OFDM_GUARD_1_4},
      {"1/8", DVB_OFDM_GUARD_1_8},
      {"auto", DVB_OFDM_GUARD_AUTO}
   };
   if (guard_interval != NULL)
   {
      table_size = sizeof(ofdm_guard_table) / sizeof(ofdm_guard_table[0]);
      index = lookup(ofdm_guard_table, table_size, guard_interval);
      if (index >= table_size)
      {
         CX88ERR << "Unrecognized guard interval: " << guard_interval << endl;
         return EINVAL;
      }
      else
      {
         ofdm.guard_interval = ofdm_guard_table[index].value;
      }
   }
   const char *hierarchy = get_property(CX88_XML_HIERARCHY);
   if (hierarchy != NULL)
   {
      if (strcasecmp(hierarchy, "none") == 0)
      {
         ofdm.hierarchy = DVB_OFDM_HIERARCHY_NONE;
      }
      else if (strcasecmp(hierarchy, "auto") != 0)
      {
         switch (strtoul(hierarchy, NULL, 10))
         {
            case 1:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_1;
               break;
            case 2:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_2;
               break;
            case 4:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_4;
               break;
            default:  
               CX88ERR << "Unrecognized hierarchy value: " << hierarchy << endl;
               return EINVAL;
         }
      }
   }
   return error;
}

int cx88_xml_channel::parse_psk(dvb_psk_settings &psk)
{
   int error = 0;
   psk.symbol_rate = DVB_SYMBOL_RATE_AUTO;
   psk.code_rate = DVB_FEC_AUTO;
   psk.pilot = DVB_PILOT_AUTO;
   psk.rolloff_factor = DVB_ROLLOFF_AUTO;
   size_t table_size, index;
   const char *symbol_rate = get_property(CX88_XML_SYMBOL_RATE);
   if (symbol_rate != NULL)
   {
      psk.symbol_rate = (uint32_t)(strtoul(symbol_rate, NULL, 10));
   }
   const char *fec = get_property(CX88_XML_CODE_RATE);
   if (fec != NULL)
   {
      error = parse_fec(fec, psk.code_rate);
   }
   if (error)
   {
      return error;
   }
   const char *pilot = get_property(CX88_XML_PILOT);
   static struct table_entry<dvb_pilot_t> psk_pilot_table[] =
   {
      {"auto", DVB_PILOT_AUTO},  
      {"off", DVB_PILOT_OFF},
      {"on", DVB_PILOT_ON}
   };
   if (pilot != NULL)
   {
      table_size = sizeof(psk_pilot_table) / sizeof(psk_pilot_table[0]);
      index = lookup(psk_pilot_table, table_size, pilot);
      if (index >= table_size)
      {
         CX88ERR << "Unrecognized PSK pilot setting: " << pilot << endl;
         return EINVAL;
      }
      else
      {
         psk.pilot = psk_pilot_table[index].value;
      }
   }
   const char *rolloff = get_property(CX88_XML_ROLLOFF);
   static struct table_entry<dvb_rolloff_t> psk_rolloff_table[] =
   {  
      {"0.20", DVB_ROLLOFF_0_20},
      {"0.25", DVB_ROLLOFF_0_25},
      {"0.35", DVB_ROLLOFF_0_35},
      {"auto", DVB_ROLLOFF_AUTO}
   };
   if (rolloff != NULL)
   {
      table_size = sizeof(psk_rolloff_table) / sizeof(psk_rolloff_table[0]);
      index = lookup(psk_rolloff_table, table_size, rolloff);
      if (index >= table_size)
      {
         CX88ERR << "Unrecognized PSK rolloff factor: " << rolloff << endl;
         return EINVAL;
      }
      else
      {
         psk.rolloff_factor = psk_rolloff_table[index].value;
      }
   }
   return error;
   
}

int cx88_xml_channel::select_dvb_channel(const string &channel, dvb_channel &dvbchan)
{
   int error = select_channel(channel);
   if (!error)
   {
      const char *bandwidth = get_property(CX88_XML_BANDWIDTH_NAME);
      const char *base_freq = get_property(CX88_XML_FREQ_NAME);
      const char *modulation = get_property(CX88_XML_MODULATION_NAME);
      if ((bandwidth != NULL) && (base_freq != NULL) && (modulation != NULL))
      {
         dvbchan.bandwidth_hz = (uint32_t) (strtoul(bandwidth, NULL, 10));
         dvbchan.frequency_hz = (uint64_t)
            (strtoull(base_freq, NULL, 10) + ((m_channel - m_range->start) * dvbchan.bandwidth_hz));
         error = parse_modulation(modulation, dvbchan.modulation);
         if (!error)
         {
            switch (dvbchan.modulation)
            {
               case DVB_MOD_QAM_16:
               case DVB_MOD_QAM_32:
               case DVB_MOD_QAM_64:
               case DVB_MOD_QAM_128:
               case DVB_MOD_QAM_256:
               case DVB_MOD_QAM_AUTO:
                  error = parse_qam(dvbchan.modulation_settings.qam);
                  break;
               case DVB_MOD_OFDM:
                  error = parse_ofdm(dvbchan.modulation_settings.ofdm);
                  break;
               case DVB_MOD_QPSK:
               case DVB_MOD_QPSK_NBC:
               case DVB_MOD_8PSK:
               case DVB_MOD_16PSK:
                  error = parse_psk(dvbchan.modulation_settings.psk);
                  break;
               default:
                  break;
            }
         }
      }
      else
      {
         CX88ERR << "Bandwidth, base frequency, or modulation not specified" << endl;
         return EINVAL;
      }
      dvbchan.inversion = DVB_INVERSION_AUTO;
      const char *inversion = get_property(CX88_XML_INVERSION);
      if (!error && (inversion != NULL))
      {
         if (strcasecmp(inversion, "off") == 0)
         {
            dvbchan.inversion = DVB_INVERSION_OFF;    
         }
         else if (strcasecmp(inversion, "on") == 0)
         {
            dvbchan.inversion = DVB_INVERSION_ON;
         }
         else if (strcasecmp(inversion, "auto") != 0)
         {
            CX88ERR << "Invalid spectral inversion setting: " << inversion << endl;
            return EINVAL;  
         }
      }
   }
   return error;
}

int cx88_xml_channel::parse_analog_video(const char *video_string, avb_video_fmt_t &fmt)
{
   int error = 0;  
   static struct table_entry<avb_video_fmt_t> analog_video_table[] =
   {
      {"none", AVB_VIDEO_FMT_NONE},
      {"NTSC-443", AVB_VIDEO_FMT_NTSC_443},
      {"NTSC-M", AVB_VIDEO_FMT_NTSC_M},
      {"NTSC-M-Jp", AVB_VIDEO_FMT_NTSC_J},
      {"NTSC-N", AVB_VIDEO_FMT_NTSC_N},
      {"PAL-60", AVB_VIDEO_FMT_PAL_60},
      {"PAL-B", AVB_VIDEO_FMT_PAL_B},
      {"PAL-D", AVB_VIDEO_FMT_PAL_D},
      {"PAL-D1", AVB_VIDEO_FMT_PAL_D1},
      {"PAL-G", AVB_VIDEO_FMT_PAL_G},
      {"PAL-H", AVB_VIDEO_FMT_PAL_H},
      {"PAL-I", AVB_VIDEO_FMT_PAL_I},
      {"PAL-K", AVB_VIDEO_FMT_PAL_K},
      {"PAL-M", AVB_VIDEO_FMT_PAL_M},
      {"PAL-N", AVB_VIDEO_FMT_PAL_N},
      {"PAL-Nc", AVB_VIDEO_FMT_PAL_NC},
      {"SECAM-B", AVB_VIDEO_FMT_SECAM_B},
      {"SECAM-D", AVB_VIDEO_FMT_SECAM_D},
      {"SECAM-G", AVB_VIDEO_FMT_SECAM_G},
      {"SECAM-H", AVB_VIDEO_FMT_SECAM_H},
      {"SECAM-K", AVB_VIDEO_FMT_SECAM_K},
      {"SECAM-K1", AVB_VIDEO_FMT_SECAM_K1},
      {"SECAM-L", AVB_VIDEO_FMT_SECAM_L},
      {"SECAM-Lc", AVB_VIDEO_FMT_SECAM_LC}
   }; 
   size_t table_size = sizeof(analog_video_table) / sizeof(analog_video_table[0]);
   size_t index = lookup(analog_video_table, table_size, video_string);
   if (index >= table_size)
   {
      CX88ERR << "Unrecognized analog video format: " << video_string << endl;
      error = EINVAL;
   }
   else
   {
      fmt = analog_video_table[index].value;  
   }
   return error;  
}

int cx88_xml_channel::parse_analog_audio(const char *audio_string, avb_audio_fmt_t &fmt)
{
   int error = 0;  
   static struct table_entry<avb_audio_fmt_t> analog_audio_table[] =
   {
      {"A2", AVB_AUDIO_FMT_A2},
      {"A2-SAP", AVB_AUDIO_FMT_A2_SAP},
      {"AM-Mono", AVB_AUDIO_FMT_AM_MONO},
      {"AM-Stereo", AVB_AUDIO_FMT_AM_STEREO},
      {"BTSC", AVB_AUDIO_FMT_BTSC},
      {"BTSC-SAP", AVB_AUDIO_FMT_BTSC_SAP},
      {"EIAJ", AVB_AUDIO_FMT_EIAJ},
      {"EIAJ-SAP", AVB_AUDIO_FMT_EIAJ_SAP},
      {"FM-Mono", AVB_AUDIO_FMT_FM_MONO},
      {"FM-Mono-50", AVB_AUDIO_FMT_FM_MONO_NON_USA},
      {"FM-Mono-75", AVB_AUDIO_FMT_FM_MONO_USA},
      {"FM-Stereo", AVB_AUDIO_FMT_FM_STEREO},
      {"FM-Stereo-50", AVB_AUDIO_FMT_FM_STEREO_NON_USA},
      {"FM-Stereo-75", AVB_AUDIO_FMT_FM_STEREO_USA},
      {"NICAM", AVB_AUDIO_FMT_NICAM},
      {"NICAM-SAP", AVB_AUDIO_FMT_NICAM_SAP},
      {"none", AVB_AUDIO_FMT_NONE}
   }; 
   size_t table_size = sizeof(analog_audio_table) / sizeof(analog_audio_table[0]);
   size_t index = lookup(analog_audio_table, table_size, audio_string);
   if (index >= table_size)
   {
      CX88ERR << "Unrecognized analog audio format: " << audio_string << endl;
      error = EINVAL;
   }
   else
   {
      fmt = analog_audio_table[index].value;  
   }
   return error;  
}

int cx88_xml_channel::select_avb_channel(const string &channel, avb_channel &avbchan)
{
   int error = select_channel(channel);
   if (error)
   {
      return error;
   }
   avbchan.bandwidth_hz = 0;
   avbchan.frequency_hz = 0;
   const char *bandwidth = get_property(CX88_XML_BANDWIDTH_NAME);
   const char *base_freq = get_property(CX88_XML_FREQ_NAME);
   const char *video = get_property(CX88_XML_ANALOG_VIDEO_NAME);
   const char *audio = get_property(CX88_XML_ANALOG_AUDIO_NAME);
   if (bandwidth != NULL)
   {
      avbchan.bandwidth_hz = (uint32_t) (strtoul(bandwidth, NULL, 10));
   }
   if (base_freq != NULL)
   {
      avbchan.frequency_hz = (uint64_t)
         (strtoull(base_freq, NULL, 10) + ((m_channel - m_range->start) * avbchan.bandwidth_hz));
   }
   if (video == NULL)
   {
      avbchan.video_format = AVB_VIDEO_FMT_NONE;
   }
   else
   {
      error = parse_analog_video(video, avbchan.video_format);
   }
   if (!error)
   {
      if (audio == NULL)
      {
         avbchan.audio_format = AVB_AUDIO_FMT_NONE;
      }
      else
      {  
         error = parse_analog_audio(audio, avbchan.audio_format);
      }
   }
   if (!error)
   {
      if ((avbchan.video_format == AVB_VIDEO_FMT_NONE) && 
         ((avbchan.audio_format == AVB_AUDIO_FMT_NONE)))
      {
         CX88ERR << "At least an analog audio or video format must be specified" << endl;
         error = EINVAL;  
      }
   }
   return error;
}

int cx88_xml_channel::channel_up(void)
{
   if (m_alias != m_config.m_aliases.end())
   {
      ++m_alias;
      if (m_alias == m_config.m_aliases.end())
      {
         m_alias = m_config.m_aliases.begin();    
      }
      return select_channel(m_alias->first);
   }
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   else if (m_channel < m_range->end)
   {
      ++m_channel;
      return 0;
   }
   else
   {
      m_tunerconf.remove_config(m_range->tunerconf);
      ++m_range;
      if (m_range == m_ranges.end())
      {
         m_range = m_ranges.begin();
      }
      m_channel = m_range->start;
      return m_tunerconf.add_config(m_range->tunerconf);
   }
}

int cx88_xml_channel::channel_down(void)
{
   rangelist::iterator range = m_range;
   if (m_alias != m_config.m_aliases.end())
   {
      if (m_alias == m_config.m_aliases.begin())
      {
         m_alias = m_config.m_aliases.end();
      }
      --m_alias;
      return select_channel(m_alias->first);
   }
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   else if (m_channel > m_range->start)
   {
      --m_channel;
      return 0;
   }
   else if (m_range == m_ranges.begin())
   {
      m_range = m_ranges.end();
   }
   m_tunerconf.remove_config(range->tunerconf);
   --m_range;
   m_channel = m_range->end;
   return m_tunerconf.add_config(m_range->tunerconf);
}

int cx88_xml_channel::select_first_channel(void)
{
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   rangelist::iterator range = m_range;
   m_range = m_ranges.begin();
   m_channel = m_range->start;
   if (range != m_range)
   {
      m_tunerconf.remove_config(range->tunerconf);
      return m_tunerconf.add_config(m_range->tunerconf);  
   }
   return 0;
}
      
int cx88_xml_channel::select_last_channel(void)
{
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   rangelist::iterator range = m_range;
   m_range = m_ranges.end();
   --m_range;
   m_channel = m_range->end;
   if (range != m_range)
   {
      m_tunerconf.remove_config(range->tunerconf);
      return m_tunerconf.add_config(m_range->tunerconf);  
   }
   return 0;
}

void cx88_xml_channel::get_channel_name(string &name)
{
   if (m_alias != m_config.m_aliases.end())
   {
      name = m_alias->first;
   }
   else
   {      
      ostringstream stream;
      stream << get_profile_name() << ':' << get_channel();
      if (m_has_subchannel)
      {
         stream << '-' << get_subchannel();
      }
      name = stream.str();
   }
}

