/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __PCHDTV_HD3000_H__
#define __PCHDTV_HD3000_H__

#include <sys/errno.h>
#include "dtt7612.h"
#include "or51132.h"
#include "tda9887.h"
#include "cx88_debug.h"
#include "cx88_mpeg_driver.h"
#include "cx88_analog_driver.h"

namespace cx88_mpeg_driver_factory
{

   class pchdtv_hd3000
      : public cx88_mpeg_driver
   {
      public:
         
         pchdtv_hd3000(
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            int &error)
            : cx88_mpeg_driver(cx88conf, tunerconf, devnode, captures, error)
         {}
         
         virtual ~pchdtv_hd3000(void) {}
         
      protected:
         
         virtual int create_i2c_drivers(void)
         {
            m_i2c_devs[0].driver = new dtt7612(m_tunerconf, *(m_i2c_devs[0].dev));
            m_i2c_devs[1].driver = new or51132(m_tunerconf, *(m_i2c_devs[1].dev));
            if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
            {
               CX88ERR << "Error initializing tuner/demodulator driver for pcHDTV HD-3000" << std::endl;
               return ENOMEM;
            }
            return 0;
         }
         
         virtual void destroy_i2c_drivers(void)
         {
            delete m_i2c_devs[1].driver;
            delete m_i2c_devs[0].driver;  
         }
         
         virtual int create_i2c_devices(void)
         {
            m_i2c_devs[0].conf.in_addr = 0x61;
            m_i2c_devs[1].conf.in_addr = 0x15;
            return 0;
         }
   
   };
   
};

namespace cx88_analog_driver_factory
{
   class pchdtv_hd3000
      : public cx88_analog_driver
   {
      public:
         
         pchdtv_hd3000(
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            cx88_analog_mode mode,
            cx88_analog_driver *slave,
            cx88_analog_driver *pip_driver,
            int &error)
            : cx88_analog_driver(cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error)
         {}
         
         virtual ~pchdtv_hd3000(void) {}
         
      protected:
         
         uint8_t m_input_mux;
         uint8_t m_input_type;
                  
         virtual int create_i2c_drivers(void)
         {
            m_i2c_devs[0].driver = new dtt7612(m_tunerconf, *(m_i2c_devs[0].dev));
            m_i2c_devs[1].driver = new tda9887(m_tunerconf, *(m_i2c_devs[1].dev), TDA9887_OPTION_RADIO_IF_41_3 | TDA9887_OPTION_RADIO_GAIN_NORM);
            if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
            {
               CX88ERR << "Error initializing analog tuner/demodulator drivers for pcHDTV HD-3000" << std::endl;
               return ENOMEM;
            }
            return 0;
         }
         
         virtual void destroy_i2c_drivers(void)
         {
            delete m_i2c_devs[0].driver;
            delete m_i2c_devs[1].driver;
         }
         
         virtual int create_i2c_devices(void)
         {
            m_i2c_devs[0].conf.in_addr = 0x61;
            m_i2c_devs[0].source_mask = CX88_SOURCE_DEFAULT;
            m_i2c_devs[1].conf.in_addr = 0x43;
            m_i2c_devs[1].source_mask = CX88_SOURCE_DEFAULT;
            return 0;
         }
      
         virtual int select_source(const char *name)
         {
            if (name != NULL)
            {
               if (strcasecmp(name, "Composite") == 0)
               {
                  m_source_mask = 0;
                  m_input_mux = CX88_VIDEO_INPUT_MUX2;
                  m_input_type = CX88_VIDEO_INPUT_COMPOSITE;
                  return 0;
               }
               else if (strcasecmp(name, "S-Video") == 0)
               {
                  m_source_mask = 0;
                  m_input_mux = CX88_VIDEO_INPUT_MUX3;
                  m_input_type = CX88_VIDEO_INPUT_SVIDEO;
                  return 0;
               }
            }
            m_source_mask = CX88_SOURCE_DEFAULT;
            m_input_mux = CX88_VIDEO_INPUT_MUX1;
            m_input_type = CX88_VIDEO_INPUT_COMPOSITE;
            return 0;
         }
         
         virtual int set_capture_options(cx88_video_capture &capture)
         {
            capture.in_input_source = m_input_mux;
            capture.in_input_type = m_input_type;
            return 0;
         }
   
   };
   
};

#endif

