/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __WINTV_HVR1850_H__
#define __WINTV_HVR1850_H__

#include <sys/errno.h>
#include "tda18271.h"
#include "s5h1411.h"
#include "cx88_debug.h"
#include "cx88_mpeg_driver.h"
#include "cx88_analog_driver.h"
#include "cx88_cards.h"

namespace cx88_mpeg_driver_factory
{

   class wintv_hvr1850
      : public cx88_mpeg_driver
   {
      public:
         
         wintv_hvr1850(
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            int &error)
            : cx88_mpeg_driver(cx88conf, tunerconf, devnode, captures, error)
         {}
         
         virtual ~wintv_hvr1850(void) {}
         
      protected:

         static int tda18271_interface_config(tda18271 &driver, const dvb_channel &channel, tda18271_interface &ifc)
         {
            ifc.std = 0x0;
            ifc.if_level = 0x06;
            switch (channel.modulation)
            {
               case DVB_MOD_VSB_8:
               case DVB_MOD_VSB_16:
                  ifc.std = 0x03;
                  ifc.ifreq_hz = 5380000;
                  break;
               default:
                  break;
            }
            return 0;
         }

         virtual int create_i2c_drivers(void)
         {
            int error = 0;
            cx88_common_gpio gpio;
            gpio.in_write_mask = 0x6000;
            gpio.in_read_mask = 0x0;
            gpio.inout_gpio = 0x0;
            if (ioctl(m_fd, CX23885_COMMON_MC_GPIO, &gpio) != 0)
            {
               perror("Error during microcontroller GPIO ioctl");
               return errno;
            }
            usleep(100000);
            gpio.inout_gpio = 0x4000;
            if (ioctl(m_fd, CX23885_COMMON_MC_GPIO, &gpio) != 0)
            {
               perror("Error during microcontroller GPIO ioctl");
               return errno;
            }
            usleep(100000);
            m_i2c_devs[1].driver = new s5h1411(m_tunerconf, *(m_i2c_devs[1].dev), *(m_i2c_devs[2].dev), DVB_INPUT_SERIAL, 
               S5H1411_IFREQ_5_38MHZ, S5H1411_IFREQ_4MHZ, S5H1411_GPIO_DISABLE, S5H1411_CLK_CONT_NINV, error);
            m_i2c_devs[0].driver = new tda18271(m_tunerconf, *(m_i2c_devs[0].dev), TDA18271_MODE_SINGLE,
                  NULL, tda18271_interface_config, error);
            if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
            {
               error = ENOMEM;
            }
            if (error)
            {
               CX88ERR << "Error initializing tuner/demodulator driver for WinTV HVR-1850" << std::endl;
               return error;
            }
            return 0;
         }
         
         virtual void destroy_i2c_drivers(void)
         {
            delete m_i2c_devs[0].driver;
            delete m_i2c_devs[1].driver;  
         }
         
         virtual int create_i2c_devices(void)
         {
            m_i2c_devs[0].conf.in_addr = 0x60;
            // The TDA18271 on the 1850 will NAK 400KHz transactions.
            // This is odd, since the datasheet indicates it supports
            // 400KHz, and the earlier-revision TDA18271 on
            // the 1800 seems fine w/ 400KHz.
            //m_i2c_devs[0].conf.in_speed = CX88_I2C_SPEED_400KHZ;
            m_i2c_devs[1].conf.in_addr = 0x19;
            m_i2c_devs[1].conf.in_speed = CX88_I2C_SPEED_400KHZ;
            m_i2c_devs[2].conf.in_addr = 0x1A;
            m_i2c_devs[2].conf.in_speed = CX88_I2C_SPEED_400KHZ;
            return 0;
         }
         
         virtual int set_capture_options(cx88_mpeg_capture &capture)
         {
            capture.in_start_byte_count = 0;
            return 0;
         }
         
         virtual int set_channel_options(dvb_channel &channel) 
         {
            if (channel.inversion == DVB_INVERSION_AUTO)
            {
               channel.inversion = DVB_INVERSION_ON;
            }
            return 0;
         }

   };
};

namespace cx88_analog_driver_factory
{
   class wintv_hvr1850
      : public cx88_analog_driver
   {
      public:
         
         wintv_hvr1850(
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            cx88_analog_mode mode,
            cx88_analog_driver *slave,
            cx88_analog_driver *pip_driver,
            int &error)
            : cx88_analog_driver(cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error),
              m_bridge_fd(-1)
         {}
         
         virtual ~wintv_hvr1850(void) {}
         
      protected:
      
         uint8_t m_audio_input;
         
         virtual int select_source(const char *name)
         {
            if (name != NULL)
            {
               if (strcasecmp(name, "line-in") == 0)
               {
                  m_source_mask = 0;
                  m_audio_input = CX23885_AUDIO_INPUT_AC97;
                  return 0;
               }
            }
            m_source_mask = CX88_SOURCE_DEFAULT;
            m_audio_input = CX23885_AUDIO_INPUT_BCAST2;
            return 0;
         }

         virtual int set_format_options(cx88_audio_format &format)
         {
            int error = 0;
            format.in_input = m_audio_input;
            if (m_source_mask == CX88_SOURCE_DEFAULT)
            {
               cx23888_dif_control dif_control;
               static const uint8_t fm_dif_config[] = {
                  0x3d, 0xa,  0xd7, 0x9d, 0xc,  0xbc, 0x3,  0x65,
                  0x85, 0x8c, 0x3,  0xbd, 0xa,  0x64, 0xb4, 0x1d,
                  0x0,  0x88, 0x0,  0x0,  0x0,  0x18, 0x40, 0x22,
                  0x0,  0x26, 0x0,  0x40, 0x0,  0x26, 0x26, 0xc2,
                  0x0,  0x26, 0x26, 0xc2, 0x0,  0x26, 0x0,  0x40,
                  0x0,  0xc2, 0x0,  0x0,  0x20, 0xe6, 0x50, 0x2,
                  0x0,  0x1,  0x0,  0x27, 0x13, 0x2b, 0x31, 0x0,
                  0x0,  0x0,  0x0,  0x0,  0x0,  0xc8, 0x0,  0x68,
                  0x6,  0xbf, 0xef, 0x1b, 0xe8, 0x35, 0x0,  0x0,
                  0xa,  0x0,  0x1,  0x0,  0x61, 0x0,  0xb,  0x0,
                  0xad, 0x0,  0x57, 0x0,  0xf4, 0x1,  0xe1, 0x0,
                  0xca, 0x1,  0xeb, 0x2,  0x64, 0x7,  0x9a, 0x0,
                  0x1b, 0x5,  0x35, 0x6,  0x60, 0x3,  0x24, 0x4,
                  0xac, 0x12, 0xdd, 0x12, 0x69, 0x13, 0xd7, 0x12,
                  0xdb, 0x35, 0x6a, 0x14, 0x6,  0x3a, 0xbc, 0x37,
                  0xa8, 0x3f, 0xaf, 0x3c, 0x38, 0x6,  0xdd, 0x2,
                  0xf7, 0xc,  0x9f, 0x9,  0xb,  0x13, 0x24, 0x10,
                  0xa4, 0x17, 0x93, 0x15, 0x1d, 0x1a, 0x2c, 0x19,
                  0x0,  0x0,  0x6f, 0x1a, 0x0,  0x0,  0x0,  0x0
               };
               switch (format.in_audio_format)
               {
                  case CX88_AUDIO_FMT_FM_RADIO:
                  case CX88_AUDIO_FMT_FM_RADIO_50:
                  case CX88_AUDIO_FMT_FM_RADIO_75:
                     dif_control.regs = fm_dif_config;
                     dif_control.offset = 0;
                     dif_control.count = sizeof(fm_dif_config);
                     break;
                  default:
                     CX88ERR << "Audio format " << format.in_audio_format <<
                        "not currently supported for HVR-1850" << std::endl;
                     error = EINVAL;
                     break;
               }
               if (!error && (ioctl(m_bridge_fd, CX23888_COMMON_DIF_ENABLE, &dif_control) == -1))
               {
                  perror("Error configuring CX23888 demodulator");
                  error = errno;
               }
            }
            else if (ioctl(m_bridge_fd, CX23888_COMMON_DIF_DISABLE) == -1)
            {
               perror("Error disabling CX23888 demodulator");
               error = errno;
            }
            return error;
         }

         virtual int create_i2c_devices(void)
         {
            m_i2c_devs[0].bus = 1;
            m_i2c_devs[0].conf.in_addr = 0x60;
            return 0;
         }

         virtual void destroy_i2c_drivers(void)
         {
            delete m_i2c_devs[0].driver;
            close(m_bridge_fd);
         }

         static int tda18271_interface_config(tda18271 &driver, const avb_channel &channel, tda18271_interface &ifc)
         {
            ifc.fm_rfn = 0;
            return 0;
         }

         virtual int create_i2c_drivers(void)
         {
            int error = 0;
            m_i2c_devs[0].driver = new tda18271(m_tunerconf, *(m_i2c_devs[0].dev), TDA18271_MODE_SINGLE,
                  tda18271_interface_config, NULL, error);
            if (m_i2c_devs[0].driver == NULL)
            {
               error = ENOMEM;
            }
            if (error)
            {
               CX88ERR << "Error initializing analog tuner driver for WinTV HVR-1850" << std::endl;
               return error;
            }
            m_bridge_fd = cx88_open_bridge(m_fd);
            if (m_bridge_fd == -1)
            {
               error = ENOENT;
            }
            return error;
         }
         
         int m_bridge_fd;
   };
};

#endif
