/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sstream>
#include "cx88_debug.h"
#include "cx88_mpeg_driver.h"

using namespace std;
               
#define PACKET_DELAY_OPTION   "packet_delay"
#define DEFAULT_PACKET_DELAY  100

cx88_mpeg_driver::cx88_mpeg_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : cx88_driver(cx88conf, tunerconf, devnode, captures, error),
     m_new_pat_version(0),
     m_packet_offset(0)
{
   if (error)
   {
      return;
   }
   try
   {
      m_pids.reserve(10);
      m_programs.reserve(10);
   }
   catch (...)
   {
      CX88ERR << "Unable to reserve space for stream PIDs" << endl;
      error = ENOMEM;
   }
   init_psfilter();
}

cx88_mpeg_driver::~cx88_mpeg_driver(void)
{}

int cx88_mpeg_driver::capture(const string &channel_string, cx88_capture_params &params)
{
   int error = 0;
   cx88_mpeg_config config;
   if ((m_fd = open(m_devnode, O_RDWR)) < 0)
   {
      CX88ERR << "Unable to open device node " << m_devnode << ": " << strerror(errno) << endl;
      error = errno; 
   }
   if (!error)
   {
      config.in_buftype = CX88_MPEG_BUF_KERNEL;
      config.in_packet_size = CX88_MPEG_DEFAULT_PACKET_SIZE;
      config.inout_num_packets = 10000;
      error = set_config_options(config);
   }
   if (!error && (error = ioctl(m_fd, CX88_MPEG_CONFIG, &config)))
   {
      CX88ERR << "Error issuing CONFIG ioctl: " << strerror(errno) << endl;
   }
   void *buf = MAP_FAILED;
   if (!error)
   {
      buf = mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ | PROT_WRITE, MAP_SHARED, m_fd, 0);
      if (buf == MAP_FAILED)
      {
         CX88ERR << "Unable to map buffer memory from device: " << strerror(errno) << endl;
         error = errno;
      }
   }
   error = (error ? error : start_captures());
   error = (error ? error : i2c_setup());
   error = (error ? error : capture(buf, config, params));
   i2c_teardown();
   stop_captures();
   if (buf != MAP_FAILED)
   {
      munmap(buf, config.bufconfig.out_kernelbuf.out_mmap_size);
   }
   if (m_fd >= 0)
   {
      close(m_fd);
   }
   if (params.completion != NULL)
   {
      params.completion(params.completion_arg, error);  
   }
   return error;
}

void cx88_mpeg_driver::init_psfilter(void)
{
   m_pmt_pid = 0;
   m_pat_ready = false;
   m_pmt_ready = false;
   m_pat_version = 0;
   m_pmt_version = 0;
   m_new_pat_version = (m_new_pat_version + 1) % 32;
   m_pids.clear();
   m_programs.clear();
}

int cx88_mpeg_driver::submit_buffer(void *data, size_t num_bytes, size_t packet_size)
{
   if (!m_channel.has_subchannel())
   {
      return cx88_driver::submit_buffer(data, num_bytes, packet_size);
   }
   size_t i = 0;
   uint8_t *bytebuffer = reinterpret_cast <uint8_t*> (data);
   if ((data != m_last_packet) && (m_packet_offset != 0))
   {
      memcpy(&m_last_packet[packet_size - m_packet_offset], data, m_packet_offset);
      int error = submit_buffer(m_last_packet, packet_size, packet_size);
      if (error)
      {
         return error;
      }
      bytebuffer += m_packet_offset;
      i = m_packet_offset;
   }
   uint8_t *buffer = bytebuffer;
   uint8_t *valid_packet_start = NULL;
   while ((i + packet_size) <= num_bytes)
   {
      bool valid_packet = true;
      if (bytebuffer[0] != CX88_MPEG_DEFAULT_START_BYTE)
      {
         DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Invalid packet at offset " << i << endl)
         valid_packet = false;
         for (size_t j = i; j < (num_bytes - packet_size); ++j)
         {
            if ((buffer[j] == CX88_MPEG_DEFAULT_START_BYTE) &&
                (buffer[j + packet_size] == CX88_MPEG_DEFAULT_START_BYTE))
            {
               i = j;
               bytebuffer = reinterpret_cast <uint8_t*> (data) + j;
               m_packet_offset = (m_packet_offset + j) % packet_size;
               DIAGNOSTIC(CX88LOG << "Found new packet offset " << m_packet_offset << endl)
               valid_packet = true;
               break;
            }
         }
      }
      uint16_t pid = get_pid(bytebuffer);
      if (valid_packet)
      {
         if (pid == 0)
         {
            valid_packet = process_pat(bytebuffer, packet_size);
         }
         else if ((m_pmt_pid != 0) && (pid == m_pmt_pid))
         {
            process_pmt(bytebuffer, packet_size);
         }
         else if (m_pat_ready && (m_pmt_pid == 0))
         {
            CX88ERR << "<MPEG PS Filter> Invalid subchannel: " << m_channel.get_channel() << "-" << m_channel.get_subchannel() << endl;
            CX88ERR << "Available subchannels:" << endl;
            for (program_vec::size_type i = 0; i < m_programs.size(); ++i)
            {
               CX88ERR << "\t" << m_channel.get_channel() << "-" << m_programs[i] << endl;
            }
            return EINVAL;
         }
         else
         {
            program_vec::size_type j;
            for (j = 0; (j < m_pids.size()) && (m_pids[j] != pid); ++j);
            if (j == m_pids.size())
            {
               valid_packet = false;
            }
         }
      }
      if (valid_packet && (valid_packet_start == NULL))
      {
         valid_packet_start = bytebuffer;
      }
      i += packet_size;
      bytebuffer += packet_size;
      if (valid_packet_start != NULL)
      {
         size_t valid_len = 0;
         if (!valid_packet)
         {
            valid_len = bytebuffer - packet_size - valid_packet_start;
         }
         else if ((num_bytes - i) < packet_size)
         {
            valid_len = bytebuffer - valid_packet_start;
         }
         if (valid_len != 0)
         {
            for (capvec::size_type j = 0; j < m_captures.size(); ++j)
            {
               int error = m_captures[j]->submit_buffer(valid_packet_start, valid_len, packet_size);
               if (error)
               {
                  CX88ERR << "<MPEG PS Filter> Error submitting buffer to captures: " << strerror(error) << endl;
                  return error;
               }
            }
            valid_packet_start = NULL;
         }
      }
      if ((num_bytes - i) < packet_size)
      {
         memcpy(m_last_packet, bytebuffer, num_bytes - i);
      }
   }
   return 0;
}

bool cx88_mpeg_driver::process_pat(uint8_t *packet, size_t size)
{
   uint16_t len = ((packet[6] & 0x0F) << 8) | packet[7] + 4;
   uint8_t version = (packet[10] & 0x3E) >> 1;
   uint16_t program_index = 0;
   if ((len > (size - 4)) || (len < 5))
   {
      DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Invalid PAT length; ignoring" << endl)
      return false;
   }
   uint32_t packet_crc = ((packet[len] << 24) | (packet[len + 1] << 16) | (packet[len + 2] << 8) | packet[len + 3]);
   if (packet_crc != crc32(&(packet[5]), len - 5))
   {
      DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Invalid PAT CRC; ignoring" << endl)
      return false;
   }
   if (version != m_pat_version)
   {
      if (m_pat_ready && (version != ((m_pat_version + 1) % 32)))
      {
         DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Unexpected PAT version; ignoring" << endl)
         return false;
      }
      m_pids.clear();
      m_programs.clear();
      m_pmt_pid = 0;
      m_pat_ready = false;
      m_pmt_ready = false;
      m_new_pat_version = (m_new_pat_version + 1) % 32;
   }
   for (uint16_t i = 13; i <= (len - 4); i += 4)
   {
      uint16_t program = (packet[i] << 8) | packet[i + 1];
      if (!m_pat_ready)
      {
         m_programs.push_back(program);
      }
      if (program == m_channel.get_subchannel()) 
      {
         program_index = i;
      }
   }
   if (packet[11] == packet[12])
   {
      m_pat_ready = true;
   }
   if (program_index != 0)
   {
      m_pmt_pid = ((packet[program_index + 2] & 0x1) << 8) | packet[program_index + 3];
      m_pat_version = version;
      packet[6] &= 0xF0;
      packet[7] = 13;
      packet[10] = (packet[10] & 0xC0) | (m_new_pat_version << 1) | 1;
      packet[11] = 0;
      packet[12] = 0;
      packet[13] = packet[program_index];
      packet[14] = packet[program_index + 1];
      packet[15] = packet[program_index + 2];
      packet[16] = packet[program_index + 3];
      uint32_t crc = crc32(&(packet[5]), 12);
      packet[17] = (crc & 0xFF000000) >> 24;
      packet[18] = (crc & 0x00FF0000) >> 16;
      packet[19] = (crc & 0x0000FF00) >> 8;
      packet[20] = crc & 0x000000FF;
      memset(&(packet[21]), 0xFF, size - 21);      
      return true;
   }
   return false;
}

void cx88_mpeg_driver::process_pmt(uint8_t *packet, size_t size)
{
   uint16_t len = ((packet[6] & 0x0F) << 8) | packet[7] + 4;
   uint8_t version = (packet[10] & 0x3E) >> 1;
   if ((len > (size - 4)) || (len < 5))
   {
      DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Invalid PMT length; ignoring" << endl)
      return;
   }
   uint32_t packet_crc = ((packet[len] << 24) | (packet[len + 1] << 16) | (packet[len + 2] << 8) | packet[len + 3]);
   if (packet_crc != crc32(&(packet[5]), len - 5))
   {
      DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Invalid PMT CRC; ignoring" << endl)
      return;
   }
   if (version != m_pmt_version)
   {
      if (m_pmt_ready && (version != ((m_pmt_version + 1) % 32)))
      {
         DIAGNOSTIC(CX88ERR << "<MPEG PS Filter> Warning: Unexpected PMT version; ignoring" << endl)
         return;
      }
      m_pids.clear();
      m_pmt_ready = false;
      m_pmt_version = version;
   }
   else if (m_pmt_ready)
   {
      return;
   }
   if (packet[11] == packet[12])
   {
      m_pmt_ready = true;
   }
   uint16_t i = 17 + (((packet[15] & 0x0F) << 8) | packet[16]);
   while (i <= (len - 5))
   {
      m_pids.push_back(get_pid(&(packet[i])));
      i += (5 + packet[i + 4]);
   }
}

// MPEG CRC algorithm adapted from NIST ATSC reference parser (NDRI)
uint32_t cx88_mpeg_driver::crc32(uint8_t *packet, size_t size)
{
   static const uint32_t polynomial = 0x04C11DB6;
   uint8_t data_bit;
   uint32_t shift_reg = 0xFFFFFFFF;

   for (size_t i = 0; i < size; ++i)
   {
      for (int j = 7; j >= 0; --j)
      {
         data_bit = (packet[i] >> j) & 0x01;
         data_bit ^= ((shift_reg >> 31) & 0x01);
         shift_reg = (shift_reg << 1);
         if (data_bit == 1)
         {
            shift_reg = (shift_reg ^ polynomial) | 1 ;
         }
      }
   }
   return shift_reg;
}

int cx88_mpeg_driver::change_channel(dvb_interface &ifc, cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   dvb_channel channel;
   m_channel_changed = false; 
   int error = m_channel.select_dvb_channel(m_channel_name, channel);
   init_psfilter();
   i2c_stop();
   error = (error ? error : select_source(m_channel.get_property(CX88_XML_MPEG_SOURCE)));
   error = (error ? error : set_channel_options(channel));
   for (uint8_t i = 0; !error && (i < MAX_I2C_DEVS); ++i)
   {
      if (m_i2c_devs[i].source_mask & m_source_mask)
      {
         dvb_driver *driver = dynamic_cast <dvb_driver*> (m_i2c_devs[i].driver);
         if (driver != NULL)
         {   
            if ((error = driver->set_channel(channel, ifc)))
            {
               CX88ERR << "Error setting channel on peripheral device" << endl;
            }
         }
      }
   }
   error = (error ? error : i2c_start(params));
   return error;
}

int cx88_mpeg_driver::capture(void *buffer, cx88_mpeg_config &config, cx88_capture_params &params)
{
   cx88_mpeg_capture capture;
   capture.in_start_byte = CX88_MPEG_DEFAULT_START_BYTE;
   capture.in_start_byte_count = CX88_MPEG_DEFAULT_START_BYTE_COUNT;
   capture.in_options = 0;
   dvb_interface ifc;
   cx88_mpeg_wait wait;
   wait.in_timeout_s = 10;
   int error = 0;
   time_t start_time = time(NULL);;
   uint64_t packet_count;
   uint32_t packet_delay = m_tunerconf.get_number<uint32_t>(PACKET_DELAY_OPTION, DEFAULT_PACKET_DELAY);
   while (!m_stop)
   {
      error = change_channel(ifc, params);
      if (params.channel_event != NULL)
      {
         params.channel_event(params.channel_arg, error);
      }
      if (error)
      {
         error = wait_for_event(params.duration_s);
         if (error)
         {
            return error;
         }
         continue;
      }
      if (ifc.input_width_bits == DVB_INPUT_SERIAL)
      {
         capture.in_options |= CX88_MPEG_SERIAL_INPUT;
      }
      if (ifc.clock == DVB_IFC_PUNC_CLCK)
      {
         capture.in_options |= CX88_MPEG_PUNCTURED_CLOCK;
      }
      if (ifc.polarity == DVB_IFC_POS_POL)
      {
         capture.in_options |= CX88_MPEG_POS_POLARITY;
      }
      if (ifc.bit_endianness == DVB_IFC_BIT_LE)
      {
         capture.in_options |= CX88_MPEG_LITTLE_ENDIAN;
      }
      if ((error = set_capture_options(capture)))
      {
         return error;
      }
      packet_count = 0;
      if ((error = ioctl(m_fd, CX88_MPEG_CAPTURE, &capture)))
      {
         CX88ERR << "Error issuing CAPTURE ioctl on device: " << strerror(errno) << endl;
         return error;
      }
      uint8_t last_buffer = 0;
      while (!m_stop && !m_channel_changed)
      {
         if ((error = ioctl(m_fd, CX88_MPEG_WAIT, &wait)))
         {
            CX88ERR << "Error issuing WAIT ioctl on device: " << strerror(errno) << endl;
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;
         }
         if ((wait.out_status & CX88_MPEG_BUS_ERROR) == CX88_MPEG_BUS_ERROR)
         {
            CX88ERR << "Fatal error: PCI bus error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_DMA_PROGRAM) == CX88_MPEG_DMA_PROGRAM)
         {
            CX88ERR << "Fatal error: DMA program error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_WAIT_ABORTED) == CX88_MPEG_WAIT_ABORTED)
         {
            CX88ERR << "Fatal error: Wait aborted" << endl;
         }
         if ((wait.out_status & CX88_MPEG_CAP_TIMEOUT) == CX88_MPEG_CAP_TIMEOUT)
         {
            CX88ERR << "Nonfatal error: Capture timed out" << endl;
         }
         if ((wait.out_status & CX88_MPEG_BAD_PACKET) == CX88_MPEG_BAD_PACKET)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_bad_packets << " bad packet(s)" << endl;
         }
         if ((wait.out_status & CX88_MPEG_HW_OVERFLOW) == CX88_MPEG_HW_OVERFLOW)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_overflows << " hardware buffer overflow(s)" << endl;
         }
         if ((wait.out_packet_count - packet_count) >= config.inout_num_packets)
         {
            CX88ERR << "Nonfatal error: Software buffer overflow" << endl;
         }
         if (wait.out_status & CX88_MPEG_CAP_FATAL)
         {            
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;
         }
         if ((wait.out_buffer_available == 0) || (wait.out_buffer_available > config.out_num_buffers))
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;
         }
         while (last_buffer != wait.out_buffer_available)
         {
            ++last_buffer;
            if (last_buffer > config.out_num_buffers)
            {
               last_buffer = 1;
            }
            size_t bufoffset = (last_buffer - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
            size_t bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
            void *bufptr = (char *)buffer + bufoffset;
            if (bufsize > (config.out_packets_per_buffer * config.in_packet_size))
            {
               bufsize = config.out_packets_per_buffer * config.in_packet_size;
            }
            if (packet_delay > packet_count)
            {
               size_t byte_delay = (packet_delay - packet_count) * config.in_packet_size;
               if (byte_delay >= bufsize)
               {
                  continue;
               }
               else
               {
                  bufsize -= byte_delay;
                  bufptr = (char *)bufptr + byte_delay;
               }
            }
            if ((error = submit_buffer(bufptr, bufsize, config.in_packet_size)))
            {
               ioctl(m_fd, CX88_MPEG_STOP);
               error = wait_for_event(params.duration_s);
               if (error)
               {
                  return error;
               }
               break;
            }
         }
         if ((params.duration_s > 0) && ((unsigned int)(time(NULL) - start_time) >= params.duration_s))
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;  
         }
         packet_count = wait.out_packet_count;
      }
      ioctl(m_fd, CX88_MPEG_STOP);
   }
   return error;
}

int cx88_mpeg_driver::get_signal(dvb_signal &signal)
{
   int error = 0;
   signal.locked = false;
   for (uint8_t i = 0; i < CX88_COMMON_MAX_I2C_DEVICES; ++i)
   {
      dvb_driver *driver = dynamic_cast<dvb_driver *>(m_i2c_devs[i].driver);
      if (driver != NULL)
      {   
         if ((error = driver->get_signal(signal)))
         {
            return error;
         }
      }
   }
   return error;
}

double cx88_mpeg_driver::get_signal_strength(void)
{
   cx88_driver_lock lock(this);
   double strength = 0.0;  
   dvb_signal signal;
   get_signal(signal);
   if (signal.locked)
   {
      if (signal.strength == 0.0)
      {
         strength = 100.0;
      }
      else
      {
         strength = signal.strength;
      }  
   }
   return strength;
}

int cx88_mpeg_driver::get_info(string &info)
{
   cx88_driver_lock lock(this);
   int error = 0;
   try
   {
      dvb_signal signal;
      ostringstream stream;
      string channel;
      m_channel.get_channel_name(channel);
      stream << "Tuned to channel " << channel << endl;
      if ((error = get_signal(signal)))
      {
         return error;
      }
      if (!(signal.locked))
      {
         stream << "\tNo signal" << endl;
      }
      else
      {
         if (signal.strength > 0.0)
         {
            stream << "\tSignal strength: " << signal.strength << "%" << endl;
         }
         if (signal.snr > 0.0)
         {
            stream << "\tSignal-to-noise ratio: " << signal.snr << " dB" << endl;
         }
         if (signal.ber > 0)
         {
            stream << "\tBit-error rate: " << signal.ber << endl;
         }
         if (signal.uncorrected_blocks > 0)
         {
            stream << "\tUncorrected blocks: " << signal.uncorrected_blocks << endl;
         }
         if (m_programs.size() > 0)
         {
            stream << "\tCurrent subchannel: " << m_channel.get_channel() << "-" << m_channel.get_subchannel() << endl;
            stream << "\tAvailable subchannels: ";
            for (program_vec::size_type i = 0; i < m_programs.size(); ++i)
            {
               stream << m_channel.get_channel() << "-" << m_programs[i] << " ";
            }
            stream << endl;
         }
      }
      info = stream.str();
   }
   catch(...)
   {
      CX88ERR << "Insufficient memory to generate info string" << endl;
      error = ENOMEM;
   }
   return 0;
}
