/*-
 * Copyright 2007 Jason Harmening
 * 
 * Copyright 2009 Konstantin Dimitrov <kosio.dimitrov@gmail.com>
 * Discovered SRAM layout for CX23885/7/8
 * 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>
#include <sys/mman.h>
#include <sys/endian.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>
#include <vm/vm_map.h>
#include <sys/module.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_regs.h"

#define CX88_DEFAULT_LATENCY       0x40
#define CX88_DMA_ALIGNMENT         4
#define CX88_DMA_BOUNDARY          0
#define CX88_CHANNEL_MASK(channel) (1 << channel)

/* RISC opcodes */
typedef enum
{
   CX88_RISC_WRITE = 0x1,
   CX88_RISC_SKIP = 0x2,
   CX88_RISC_WRITEC = 0x5,
   CX88_RISC_JUMP = 0x7,
   CX88_RISC_SYNC = 0x8,
   CX88_RISC_READ = 0x9,
   CX88_RISC_READC = 0xA,
   CX88_RISC_WRITERM = 0xB,
   CX88_RISC_WRITECM = 0xC,
   CX88_RISC_WRITECR = 0xD
} cx88_risc_opcode;

/* Line position indicator */
enum
{
   CX88_RISC_EOL = 0x1, /* End of line */
   CX88_RISC_SOL /* Start of line */
};

typedef enum 
{
   CX88_SRAM_FIFO,
   CX88_SRAM_CONTROL
} cx88_sram_t;

static int cx88_release_softc(cx88_common_softc *sc);
static int cx88_release_client(cx88_common_softc *sc);
static cx88_size_t cx88_fit_sram(struct rman *sram, cx88_size_t chunk_size, cx88_size_t used_sram);
static int cx88_alloc_sram(cx88_channel *channel, cx88_sram_t sram_type, cx88_size_t min_sram_size);
static void cx88_dmaprogmap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error);
static void cx88_dmamap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error);
static void cx88_dmamap_load_uio(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   bus_size_t mapsize,
   int error);
static int cx88_alloc_dmaprog_buffer(cx88_buffer *buffer, uint32_t size);
static int cx88_alloc_dmaprogram(cx88_channel *channel);
static void cx88_release_dmaprogram(cx88_buffer *buffer);
static void cx88_free_data_buffer(cx88_buffer *buffer);
static int cx88_configure_fifos(cx88_channel *channel);
static void *cx88_get_sibling_board_data(device_t dev);
static int cx88_init_sram(
   struct rman *sram,
   uint32_t base,
   uint32_t size
   );
static void cx88_init_board_data(cx88_board_data *data);
static int cx23880_init_board_data(void *arg);
static void cx23885_common_init(cx88_board_data *data);
static int cx23885_init_board_data(void *arg);
static int cx23887_init_board_data(void *arg);
static int cx88_destroy_board_data(void *arg);
static void cx88_init_dma_counters(cx88_buffer *buffer);
static uint8_t cx88_get_linepos(
   cx88_buffer *buffer,
   uint32_t *count,
   bus_addr_t *physaddr);
static void cx88_risc_seek(cx88_buffer *buffer, bus_size_t offset);
static void cx88_sram_instr(cx88_buffer *buffer, uint32_t instr);
static void cx88_ram_instr(cx88_buffer *buffer, uint32_t instr);
static void cx88_write_addr64(cx88_buffer *buffer, bus_addr_t address);
static void cx88_write_addr32(cx88_buffer *buffer, bus_addr_t address);
static void cx23880_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value);
static uint32_t cx23880_read_gpio(cx88_common_softc *sc, uint32_t mask);
static void cx23885_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value);
static uint32_t cx23885_read_gpio(cx88_common_softc *sc, uint32_t mask);
static int cx23885_mc_wait(cx88_common_softc *sc);
static void cx23885_mc_write_byte(cx88_common_softc *sc, uint8_t addr, uint8_t byte);
static uint8_t cx23885_mc_read_byte(cx88_common_softc *sc, uint8_t addr);

MALLOC_DEFINE(M_CX88, "cx88", "cx88 buffers");

/* Channel descriptor table. min_size should sum to
 * CX88_USABLE_SRAM */
static const cx88_channel_desc cx23880_channels[] =
{
   /*channel                  dma_number  direction         min_fifo_size  min_ctrl_size */
   /*Video*/
   {CX88_CH_VIDEO_PACKED,     0,          CX88_DOWNSTREAM,  0x2400,        0},
   {CX88_CH_VIDEO_U,          1,          CX88_DOWNSTREAM,  0x0800,        0},
   {CX88_CH_VIDEO_V,          2,          CX88_DOWNSTREAM,  0x0800,        0},
   {CX88_CH_VIDEO_VBI,        3,          CX88_DOWNSTREAM,  0x1200,        0},
   /*Audio*/
   {CX88_CH_AUDIO_IN,         4,          CX88_DOWNSTREAM,  0x0260,        0},
   {CX88_CH_AUDIO_OUT,        5,          CX88_UPSTREAM,    0x0260,        0},
   /*MPEG*/
   {CX88_CH_MPEG_TS,          7,          CX88_DOWNSTREAM,  0x1200,        0},
   /*Video Interface Port*/
   {CX88_CH_VIP_IN,           8,          CX88_DOWNSTREAM,  0x0800,        0},
   {CX88_CH_VIP_OUT,          9,          CX88_UPSTREAM,    0x0800,        0},
   /*General Purpose I/O*/
   {CX88_CH_GPHOST_IN,        10,         CX88_DOWNSTREAM,  0x0800,        0},
   {CX88_CH_GPHOST_OUT,       11,         CX88_UPSTREAM,    0x0800,        0}
};

static const cx88_channel_desc cx23885_channels[] = 
{
   /*channel                  dma_number  direction         min_fifo_size  min_ctrl_size */
   {CX88_CH_VIDEO_PACKED,     0,          CX88_DOWNSTREAM,  0x2000,        0xA0},
   {CX88_CH_VIDEO_VBI,        1,          CX88_DOWNSTREAM,  0x1200,        0x90},
   {CX88_CH_VIDEO_PACKED_2,   2,          CX88_DOWNSTREAM,  0x2000,        0xA0},
   {CX88_CH_MPEG_TS,          2,          CX88_DOWNSTREAM,  0x1200,        0x90},
   {CX88_CH_VIDEO_VBI_2,      3,          CX88_DOWNSTREAM,  0x1200,        0x90},
   {CX88_CH_MPEG_TS_2,        4,          CX88_DOWNSTREAM,  0x1200,        0x90},
   {CX88_CH_AUDIO_IN,         5,          CX88_DOWNSTREAM,  0x0340,        0x80},
   {CX88_CH_AUDIO_IN_2,       6,          CX88_DOWNSTREAM,  0x0340,        0x80},
   {CX88_CH_AUDIO_IN_3,       7,          CX88_DOWNSTREAM,  0x0340,        0x80}
};

/* Static mutex for shared data initialization */
static struct mtx interdev_lock;

/* Module callback */
static int cx88_mod_handler(module_t mod, int what, void *arg)
{
   switch(what)
   {
      case MOD_LOAD:
         mtx_init(&interdev_lock, "CX88 Inter-device lock", NULL, MTX_DEF);
         return 0;
      case MOD_UNLOAD:
      case MOD_SHUTDOWN:
         mtx_lock(&interdev_lock);
         mtx_destroy(&interdev_lock);
         return 0;
      default:
         return 0;
   }
}

/* Module declarations */
static moduledata_t cx88_mod_data =
{
   "cx88",
   cx88_mod_handler,
   NULL
};

DECLARE_MODULE(cx88, cx88_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_VERSION(cx88, CX88_MODVER);

static void cx88_release_dmaprogram(cx88_buffer *buffer)
{
   if (buffer->dma_program_map_state == CX88_DMAMAP_LOADED)
   {
      bus_dmamap_unload(buffer->dma_program_tag, buffer->dma_program_map);
      buffer->dma_program_map_state = CX88_DMAMAP_CREATED;
   }
   if (buffer->dma_program_map_state == CX88_DMAMAP_CREATED)
   {
      bus_dmamem_free(buffer->dma_program_tag, buffer->dma_program_buffer, buffer->dma_program_map);
      buffer->dma_program_buffer = NULL;
      buffer->dma_program_map_state = CX88_DMAMAP_TAGGED;
   }
   if (buffer->dma_program_map_state == CX88_DMAMAP_TAGGED)
   {
      bus_dma_tag_destroy(buffer->dma_program_tag);
      buffer->dma_program_map_state = CX88_DMAMAP_NULL;  
   }
   buffer->dma_program_state = CX88_DMAPROG_SIZE;
   if (buffer->dma_program_size > 0)
   {
      buffer->dma_program_size = 0;
      buffer->channel->is_dirty = TRUE;  
   }
}

static void cx88_free_data_buffer(cx88_buffer *buffer)
{
   if (buffer->data_buffer_map_state == CX88_DMAMAP_LOADED)
   {
      bus_dmamap_unload(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
      buffer->data_buffer_map_state = CX88_DMAMAP_CREATED;
   }
   switch (buffer->data_buffer_type)
   {
      case CX88_DATABUF_MALLOC: /* Kernel-allocated buffer */
         DIAG(DIAG_BUFFER, device_printf(buffer->channel->sc->dev, "Releasing kernel-allocated buffer\n"))
         free(buffer->data_buffer, M_CX88);
         break;
      case CX88_DATABUF_USER: /* User-allocated buffer */
         DIAG(DIAG_BUFFER, device_printf(buffer->channel->sc->dev, "Releasing user-allocated buffer\n"))
         /* Call vm_map_unwire directly instead of going through
            * vslock(), as we may not be executing in the context of
            * the allocating process */
         vm_map_unwire(&(buffer->owner->p_vmspace->vm_map),
            trunc_page((vm_offset_t)(buffer->data_buffer)),
            round_page((vm_offset_t)(buffer->data_buffer) + buffer->data_buffer_size),
            VM_MAP_WIRE_SYSTEM | VM_MAP_WIRE_NOHOLES);
         buffer->owner = NULL;
         break;
      default:
         break;
   }
   buffer->data_buffer = NULL;
   buffer->data_buffer_type = CX88_DATABUF_NONE;
   buffer->dma_program_state = CX88_DMAPROG_SIZE;
   if (buffer->data_buffer_size > 0)
   {
      buffer->data_buffer_size = 0;
      buffer->channel->is_dirty = TRUE;
   }
}

/* Free buffer resources */
void cx88_release_buffer(cx88_buffer *buffer)
{
   DIAG(DIAG_BUFFER, device_printf(buffer->channel->sc->dev, "Releasing buffer for channel ID %d\n", buffer->channel->descriptor->channel))
   cx88_release_dmaprogram(buffer); /* Free the RISC program */
   free(buffer->sgl, M_CX88);
   buffer->sgl = NULL;
   buffer->sgl_size = 0;
   cx88_free_data_buffer(buffer);
   buffer->next_buffer_index = 0;
   if (buffer->data_buffer_map_state == CX88_DMAMAP_CREATED)
   {
      bus_dmamap_destroy(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
      buffer->data_buffer_map_state = CX88_DMAMAP_TAGGED;
   }
}

/* Free user-allocated buffers for a channel, if they belong to the
 * specified process.  Pass NULL for owner to free all user-allocated
 * buffers. */
void cx88_free_userbufs(cx88_channel *channel, struct proc *owner)
{
   if (channel->buffers != NULL)
   {
      uint8_t i;
      for (i = 0; i < channel->num_buffers; ++i)
      {
         if ((channel->buffers[i].data_buffer_type == CX88_DATABUF_USER) &&
            ((owner == NULL) || (owner == channel->buffers[i].owner)))
         {
            cx88_free_data_buffer(&(channel->buffers[i]));
         }
      }
   }  
}

void cx88_free_sram(cx88_channel *channel)
{
   if (channel->fifo_sram != NULL)
   {
      rman_release_resource(channel->fifo_sram);
      channel->fifo_sram = NULL;
   }
   if (channel->control_sram != NULL)
   {
      rman_release_resource(channel->control_sram);
      channel->control_sram = NULL;
   }
   channel->is_dirty = TRUE;
}

/* Free channel resources */
void cx88_release_channel(cx88_channel *channel)
{
   DIAG(DIAG_CHANNEL, device_printf(channel->sc->dev, "Releasing channel ID %d\n", channel->descriptor->channel))
   if (channel->buffers != NULL)
   {
      uint8_t i;
      for (i = 0; i < channel->num_buffers; ++i)
      {
         cx88_release_buffer(&(channel->buffers[i]));
      }
      free(channel->buffers, M_CX88);
      channel->buffers = NULL;
   }
   channel->num_buffers = 0;
   cx88_free_sram(channel);
}

/* Initialize basic channel data */
int cx88_init_channel(
   cx88_channel *channel,
   cx88_common_softc *sc,
   cx88_channel_t channel_id,
   uint32_t max_buffer_size)
{
   cx88_size_t i;
   int error = 0;
   DIAG(DIAG_CHANNEL, printf("Initializing channel %d\n", channel_id))
   channel->descriptor = NULL;
   for (i = 0; i < sc->board_data->table_size; ++i)
   {
      if (sc->board_data->channel_table[i].channel == channel_id)
      {
         channel->descriptor = &(sc->board_data->channel_table[i]);
         break;
      }
   }
   if (channel->descriptor == NULL)
   {
      return EINVAL;
   }
   channel->sc = sc;
   channel->buffers = NULL;
   channel->num_buffers = 0;
   channel->packet_size = 0;
   channel->is_dirty = TRUE;
   channel->max_buffer_size = 0;
   channel->start_buffer = 0;
   channel->fifo_sram = NULL;
   channel->control_sram = NULL;
   channel->max_line_size = 0;
   channel->line_boundary = 0;
   channel->line_size = 0;
   if (max_buffer_size != 0)
   {
      int nsegments = (max_buffer_size + PAGE_SIZE - 1) / PAGE_SIZE;
      error = bus_dma_tag_create(
         bus_get_dma_tag(sc->dev),
         CX88_DMA_ALIGNMENT,
         CX88_DMA_BOUNDARY,
         ((sc->board_data->addr_dwords > 1) ? BUS_SPACE_MAXADDR : BUS_SPACE_MAXADDR_32BIT),
         BUS_SPACE_MAXADDR,
         NULL,
         NULL,
         max_buffer_size,
         nsegments,
         max_buffer_size,
         0,
         NULL,
         NULL,
         &(channel->data_buffer_tag));
      if (error)
      {
         device_printf(sc->dev, "Unable to create DMA tag for data buffer\n");
         return error;
      }
      channel->max_buffer_size = max_buffer_size;
   }
   /* Modify the channel bitmask in the board data structure */
   cx88_lock_board(sc);
   if (sc->board_data->present_dma_channels & CX88_CHANNEL_MASK(channel->descriptor->dma_number))
   {
      device_printf(sc->dev, "DMA channel %d already reserved\n", channel->descriptor->dma_number);
      error = EINVAL;
   }
   else if (sc->board_data->present_channels & CX88_CHANNEL_MASK(channel->descriptor->channel))
   {
      device_printf(sc->dev, "Channel ID %d already reserved\n", channel_id);
      error = EINVAL;
   }
   else
   {
      sc->board_data->present_channels |= CX88_CHANNEL_MASK(channel_id);
      sc->board_data->present_dma_channels |= CX88_CHANNEL_MASK(channel->descriptor->dma_number);
   }
   for (i = 0; i < channel->sc->num_siblings; ++i)
   {
      CX88_REQUEST_RECONFIG(channel->sc->siblings[i]);
   }
   cx88_unlock_board(sc);
   if (error)
   {
      bus_dma_tag_destroy(channel->data_buffer_tag);
   }
   return error;
}

/* Final destruction of channel */
void cx88_destroy_channel(cx88_channel *channel)
{
   uint8_t i;
   if (channel->max_buffer_size != 0)
   {
      cx88_release_channel(channel); /* Free any allocated resources */
      bus_dma_tag_destroy(channel->data_buffer_tag);
      channel->max_buffer_size = 0;
   }
   cx88_lock_board(channel->sc);
   /* Remove this channel from the bitmask of allocated channels */
   channel->sc->board_data->present_channels &= (~(CX88_CHANNEL_MASK(channel->descriptor->channel)));
   channel->sc->board_data->present_dma_channels &= (~(CX88_CHANNEL_MASK(channel->descriptor->dma_number)));
   /* If this channel was accounted for in any SRAM allocation, 
    * attempt to rebalance resources */
   DIAG(DIAG_CHANNEL, device_printf(channel->sc->dev, "Rebalancing resources\n"))
   for (i = 0; i < channel->sc->num_siblings; ++i)
   {
      CX88_REQUEST_RECONFIG(channel->sc->siblings[i]);
   }
   cx88_unlock_board(channel->sc);
}

/* Use 16.16 fixed-point math to determine the size ratio of a chunk of SRAM
 * to a larger region, and expand the chunk to occupy the same ratio of the entire
 * SRAM.  Used below to optimize channel allocation sizes when not all DMA
 * channels are used */
static cx88_size_t cx88_fit_sram(struct rman *sram, cx88_size_t chunk_size, cx88_size_t used_sram)
{
   uint32_t fixedpoint = (((uint32_t) chunk_size) << 16) / used_sram;
   fixedpoint *= (sram->rm_end - sram->rm_start + 1);
   cx88_size_t new_chunk_size = ((cx88_size_t) (fixedpoint >> 16));
   new_chunk_size &= (~7UL);
   return new_chunk_size;
}

/* Allocates a chunk of SRAM for a given DMA channel */
static int cx88_alloc_sram(cx88_channel *channel, cx88_sram_t sram_type, cx88_size_t min_sram_size)
{
   DIAG(DIAG_CHANNEL | DIAG_SRAM, device_printf(channel->sc->dev, "Allocating channel ID %d\n", channel->descriptor->channel))
   uint16_t present_channels;
   cx88_size_t total_size = 0, channel_size = 0, channel_offset = 0, next_channel_offset = 0;
   cx88_size_t min_size = 0;
   size_t i;
   size_t num_channels = channel->sc->board_data->table_size;
   const cx88_channel_desc *channels = channel->sc->board_data->channel_table;
   boolean_t realloc_attempted = FALSE;
   present_channels = channel->sc->board_data->present_channels;
   struct resource **sram_region = NULL;
   struct rman *sram_block = NULL;
   switch (sram_type)
   {
      case CX88_SRAM_FIFO:
         min_size = channel->descriptor->min_fifo_size;
         sram_block = &(channel->sc->board_data->fifo_sram);
         sram_region = &(channel->fifo_sram);
         break;
      case CX88_SRAM_CONTROL:
         min_size = channel->descriptor->min_ctrl_size;
         sram_block = &(channel->sc->board_data->control_sram);
         sram_region = &(channel->control_sram);
         break;
      default:
         return EINVAL;
   }
   if (!(CX88_CHANNEL_MASK(channel->descriptor->channel) & present_channels))
   {
      device_printf(channel->sc->dev, "Channel ID %d not reserved by client\n", channel->descriptor->channel);
      return EACCES;
   }
   /* Sum the min_size fields of all used channels for the board and
   * calculate the current channel's offset for the minimum allocation */
   for (i = 0; i < num_channels; ++i)
   {
      if (channel->descriptor->channel == channels[i].channel)
      {
         channel_offset = total_size;
         next_channel_offset = total_size + min_size;
      }
      if (present_channels & CX88_CHANNEL_MASK(channels[i].channel))
      {
         total_size += ((sram_type == CX88_SRAM_FIFO) ? channels[i].min_fifo_size : channels[i].min_ctrl_size);
      }
   }
   /* Now expand the minimum offset calculations to fit the full SRAM size */
   channel_offset = cx88_fit_sram(sram_block, channel_offset, total_size);
   next_channel_offset = cx88_fit_sram(sram_block, next_channel_offset, total_size);
   channel_size = next_channel_offset - channel_offset;
   if (min_sram_size > channel_size)
   {
      channel_size = min_sram_size;
      channel_offset = 0;
   }
   if (min_sram_size > min_size)
   {
      min_size = min_sram_size;
   }
   while (*sram_region == NULL)
   {
      /* Step 1: try to allocate at channel_offset to help prevent fragging SRAM
         if present_channels remains constant (the most likely case) */
      DIAG(DIAG_SRAM,
         device_printf(channel->sc->dev, "Attempting to allocate SRAM region of size 0x%x at offset 0x%x\n",
            channel_size, channel_offset))
      *sram_region = rman_reserve_resource(sram_block,
         sram_block->rm_start + channel_offset, sram_block->rm_start + channel_offset + channel_size - 1, channel_size, 
         RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      if (*sram_region == NULL)
      {
         DIAG(DIAG_SRAM, 
            device_printf(channel->sc->dev, "Attempting to allocate full SRAM region at any location\n"))
         /* Step 2: Now try to allocate anywhere */
         *sram_region = rman_reserve_resource(sram_block,
            sram_block->rm_start, sram_block->rm_end, channel_size, 
            RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      }
      if ((*sram_region == NULL) && !realloc_attempted)
      {
         /* Step 3: See if siblings can reallocate */
         device_printf(channel->sc->dev, 
            "Unable to allocate SRAM region for channel ID %d; reallocation request posted to siblings\n",
            channel->descriptor->channel);
         realloc_attempted = TRUE;
         int error = 0;
         cx88_lock_board(channel->sc);
         boolean_t retry = FALSE;
         for (i = 0; i < channel->sc->num_siblings; ++i)
         {
            error = CX88_REQUEST_RECONFIG(channel->sc->siblings[i]);
            if (error == 0)
            {
               device_printf(channel->sc->siblings[i], "Reallocating\n");
               /* If a sibling has reallocated, we should loop back and try
                * to allocate our SRAM chunk again */
               retry = TRUE;
            }
            else if (error == EINPROGRESS)
            {
               device_printf(channel->sc->siblings[i], "Busy; will reallocate later\n");
            }
            else
            {
               device_printf(channel->sc->siblings[i], "Unable to reallocate\n");
            }
         }
         cx88_unlock_board(channel->sc);
         if (retry)
         {
            /* Loop back and retry maximum allowable allocation */
            continue;
         }
      }
      if ((*sram_region == NULL) && (channel_size > min_size))
      {
         /* Step 4: Try to allocate smallest allowable region */
         *sram_region = rman_reserve_resource(sram_block,
            sram_block->rm_start, sram_block->rm_end, min_size, 
            RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      }
      if (*sram_region == NULL)
      {
         /* We've exhausted all allocation possibilities */
         return ENOMEM;  
      }
   }
   DIAG(DIAG_CHANNEL | DIAG_SRAM,
      device_printf(
         channel->sc->dev,
         "Allocated channel ID %d of size 0x%lx at offset 0x%lx\n",
         channel->descriptor->channel,
         rman_get_size(*sram_region),
         rman_get_start(*sram_region)
         )
      )
   return 0;
}

/* Reset DMA program generation data */
static void cx88_init_dma_counters(cx88_buffer *buffer)
{
   buffer->dma_program_counter = 0;
   buffer->sgl_index = buffer->sgl;
   buffer->sgl_offset = 0;
   buffer->buffer_offset = 0;
   buffer->line_offset = 0;
   buffer->needs_relink = FALSE;
   buffer->target_addr_valid = TRUE;
}

/* Allocate FIFO space within a channel's SRAM chunk */
static int cx88_configure_fifos(cx88_channel *channel)
{
   DIAG(DIAG_CHANNEL | DIAG_SRAM, device_printf(channel->sc->dev, "Configuring SRAM FIFOs for channel ID %d\n", channel->descriptor->channel))
   uint32_t fifo_space = rman_get_size(channel->fifo_sram);
   uint32_t used_fifo_space = fifo_space;
   uint32_t control_space = 0;
   /* Set aside 1/32 of the chunk for the instruction queue */
   cx88_size_t instrq_size = rman_get_size(channel->fifo_sram) >> 5;
   cx88_size_t i;
   cx88_size_t min_num_fifos;
   cx88_size_t cdt_entry_size = channel->sc->board_data->cdt_entry_size;
   channel->num_fifos = 0;
   instrq_size &= (~7UL);
   if ((instrq_size < channel->sc->board_data->min_instrq_size) || (channel->control_sram != NULL))
   {
      instrq_size = channel->sc->board_data->min_instrq_size;
   }
   /* Initialize the variable to track the amount of SRAM we're trying to use for FIFOs,
    * make sure it's a multiple of the packet size */
   if (channel->control_sram == NULL)
   {
      fifo_space -= instrq_size;
      used_fifo_space = fifo_space - (2 * cdt_entry_size);
   }
   else
   {
      control_space = rman_get_size(channel->control_sram) - instrq_size;
   }
   used_fifo_space -= (used_fifo_space % channel->packet_size);
   /* We need at least 2 FIFOs, each with space for at least one packet */
   if (fifo_space < (2 * channel->packet_size))
   {
      device_printf(channel->sc->dev, "SRAM allocation too small for channel ID %d\n", channel->descriptor->channel);
      return ENOMEM;
   }
   /* Loop until we've either calculated an optimum FIFO size & number of FIFOs, or we
    * can't possibly use any less of the SRAM chunk */
   while ((used_fifo_space >= (2 * channel->packet_size)) && (channel->num_fifos == 0))
   {
      /* Step 1: Calculate the minimum number of FIFOs that we can possibly get away with.
       * It must be at least 2, and may be greater given maximum line size constraints */
      min_num_fifos = (used_fifo_space + channel->max_line_size - 1) / channel->max_line_size;
      if (min_num_fifos < 2)
      {
         min_num_fifos = 2;  
      }
      
      cx88_size_t num_packets = used_fifo_space / channel->packet_size;      
      /* Step 2: Loop until we've found a way to divide the usable FIFO space */
      for (i = min_num_fifos; 
          ((i <= num_packets) && (i <= channel->sc->board_data->max_num_fifos)); 
          ++i)
      {
         if (((control_space == 0) && ((used_fifo_space + (i * cdt_entry_size)) > fifo_space)) ||
             ((control_space != 0) && ((i * cdt_entry_size) > control_space)))
         {
            /* The required CDT space will only grow as i increases, so if we
             * can't fit them in now, just bail out of the loop */
            break;
         }
         /* The number of packets that can be stored in our FIFO space must
          * be divisible by the number of FIFOs, and the number of packets in each line
          * must not cross the specified boundary */
         if (((num_packets % i) == 0) && ((channel->line_boundary % (num_packets / i)) == 0))
         {
            cx88_size_t line_size = used_fifo_space / i;
            /* The cx88 takes FIFO sizes in qwords, so make sure we align to an 8-byte boundary */
            cx88_size_t fifo_size = (line_size + 7) & (~7UL);
            /* Now make sure the aligned FIFOs and CDTs will fit in our SRAM chunk */
            if (((control_space == 0) && ((i * (cdt_entry_size + fifo_size)) <= fifo_space)) ||
                ((control_space != 0) && ((i * cdt_entry_size) <= control_space) && ((i * fifo_size) <= fifo_space)))
            {
               channel->fifo_size = fifo_size;
               channel->line_size = line_size;
               channel->num_fifos = i;
               break;
            }
         }
      }
      /* If we haven't yet found a suitable FIFO arrangement, subtract 
       * a packet and go back around */
      used_fifo_space -= channel->packet_size;
   }
   if (channel->num_fifos == 0)
   {
      device_printf(channel->sc->dev, "Unable to allocate SRAM FIFOs for channel ID %d\n", channel->descriptor->channel);
      return ENOMEM;
   }
   channel->fifo_offset = rman_get_start(channel->fifo_sram);
   if (control_space == 0)
   {
      channel->cdt_offset = channel->fifo_offset + (channel->num_fifos * channel->fifo_size);
   }
   else
   {
      channel->cdt_offset = rman_get_start(channel->control_sram);
   }
   DIAG(DIAG_CHANNEL | DIAG_SRAM,
      device_printf(
         channel->sc->dev,
         "Channel ID %d:\n"
         "\tCDT offset = 0x%x\n"
         "\tNumber of FIFOs = %d\n"
         "\tFIFO offset = 0x%x\n"
         "\tFIFO size = 0x%x\n"
         "\tLine size = 0x%x\n",
         channel->descriptor->channel,
         channel->cdt_offset,
         channel->num_fifos,
         channel->fifo_offset,
         channel->fifo_size,
         channel->line_size
         )
      )
   for (i = 0; i < channel->num_fifos; ++i)
   {
      /* Write the FIFO base addresses to the CDTs */
      cx88_write_reg(channel->sc, 
         (channel->cdt_offset + (i * cdt_entry_size)),
         (channel->fifo_offset + (i * channel->fifo_size))); 
      DIAG(DIAG_CHANNEL | DIAG_SRAM,
         device_printf(
            channel->sc->dev,
            "Writing FIFO base address 0x%x to SRAM offset 0x%x\n",
            (channel->fifo_offset + (i * channel->fifo_size)),
            (channel->cdt_offset + (i * cdt_entry_size))))
   }
   return 0;
}

/* bus_dma map load callback for RISC program buffers */
static void cx88_dmaprogmap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error)
{
   cx88_buffer *buffer = arg;
   /* We only allow physically contiguous RISC programs */
   buffer->dma_program_physaddr = sgl->ds_addr;
}

/* Allocates a DMA program buffer */
static int cx88_alloc_dmaprog_buffer(cx88_buffer *buffer, uint32_t size)
{
   int error = 0;
   DIAG(DIAG_BUFFER, printf("Allocating DMA program buffer\n"))
   if ((buffer->dma_program_buffer != NULL) && (buffer->dma_program_size == size))
   {
      return 0;  
   }
   cx88_release_dmaprogram(buffer);
   error = bus_dma_tag_create(
      bus_get_dma_tag(buffer->channel->sc->dev),
      CX88_DMA_ALIGNMENT,
      CX88_DMA_BOUNDARY,
      ((buffer->channel->sc->board_data->addr_dwords > 1) ? BUS_SPACE_MAXADDR : BUS_SPACE_MAXADDR_32BIT),
      BUS_SPACE_MAXADDR,
      NULL,
      NULL,
      size,
      1,
      size,
      0,
      NULL,
      NULL,
      &(buffer->dma_program_tag));
   if (error)
   {
      device_printf(buffer->channel->sc->dev, "Unable to create DMA program tag\n");
      cx88_release_dmaprogram(buffer);
      return error;
   }
   buffer->dma_program_map_state = CX88_DMAMAP_TAGGED;
   error = bus_dmamem_alloc(buffer->dma_program_tag, 
      (void**)(&(buffer->dma_program_buffer)), BUS_DMA_WAITOK | BUS_DMA_COHERENT, 
      &(buffer->dma_program_map));
   if (error)
   {
      device_printf(buffer->channel->sc->dev, "Unable to allocate DMA program buffer\n");
      cx88_release_dmaprogram(buffer);
      return error;
   }
   buffer->dma_program_map_state = CX88_DMAMAP_CREATED;
   buffer->dma_program_size = size;
   error = bus_dmamap_load(
      buffer->dma_program_tag,
      buffer->dma_program_map,
      buffer->dma_program_buffer,
      size,
      cx88_dmaprogmap_load,
      buffer,
      BUS_DMA_NOWAIT);
   if (error)
   {
      DIAG(DIAG_BUFFER, printf("Unable to load DMA program map\n"))
      cx88_release_dmaprogram(buffer);
      return error;
   }
   buffer->dma_program_map_state = CX88_DMAMAP_LOADED;
   buffer->dma_program_state = CX88_DMAPROG_LOAD;
   buffer->generate_instr = cx88_ram_instr;
   return 0;
}

/* Allocate DMA and generate DMA programs for all buffers in the channel */
static int cx88_alloc_dmaprogram(cx88_channel *channel)
{
   int error = 0;
   uint8_t i;
   /* Whatever isn't used by the FIFOs goes to the instruction queue by default */
   channel->instrq_offset = channel->cdt_offset +
      (channel->sc->board_data->cdt_entry_size * channel->num_fifos);
   if (channel->control_sram == NULL)
   {
      channel->instrq_size = rman_get_start(channel->fifo_sram) +    
         rman_get_size(channel->fifo_sram) - channel->instrq_offset;
   }
   else
   {
      channel->instrq_size = rman_get_start(channel->control_sram) +    
         rman_get_size(channel->control_sram) - channel->instrq_offset;
   }
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      /* Make sure the data buffer has actually been allocated */
      if (buffer->data_buffer_type != CX88_DATABUF_NONE)
      {
         uint32_t new_program_size = buffer->dma_program_size;
         /* If the data buffer has been allocated but the DMA program hasn't
          * yet been generated, the size for the DMA program will need to be calculated.
          * The client-supplied generate_dma_program() callback needs to take this
          * state into account and only increase the DMA program counter without
          * actually writing the DMA program.  In reality, client callbacks shouldn't
          * need to worry about dma_program_state as they will build DMA programs through
          * the cx88_risc* functions, which correctly account for it. */
         if (buffer->dma_program_state == CX88_DMAPROG_SIZE)
         {
            cx88_init_dma_counters(buffer);
            DIAG(DIAG_BUFFER, device_printf(channel->sc->dev, "Sizing buffer %d\n", i))
            if ((error = buffer->generate_dma_program(buffer, i)))
            {
               return error;
            }
            new_program_size = cx88_risc_offset(buffer);
            buffer->dma_program_state = CX88_DMAPROG_LOAD;
         }
         if (new_program_size <= (channel->instrq_size - channel->sc->board_data->min_instrq_size))
         {
            /* The instruction queue is actually large enough to house this entire DMA program
             * with enough space left to maintain the minimum safe size, so go ahead and store
             * the program in SRAM */
            DIAG(DIAG_BUFFER | DIAG_SRAM, device_printf(channel->sc->dev, "Allocating DMA program of size %d in SRAM\n", new_program_size))
            if (buffer->dma_program_buffer != NULL)
            {
               cx88_release_dmaprogram(buffer);
               buffer->dma_program_state = CX88_DMAPROG_LOAD;
            }
            else if (buffer->dma_program_physaddr != channel->instrq_offset)
            {
               buffer->dma_program_state = CX88_DMAPROG_LOAD;
            }
            buffer->dma_program_physaddr = channel->instrq_offset;
            buffer->generate_instr = cx88_sram_instr;
            channel->instrq_size -= new_program_size;
            channel->instrq_offset += new_program_size;
         }
         else
         {
            /* We can't fit the DMA program in SRAM, so just allocate it in RAM */
            DIAG(DIAG_BUFFER | DIAG_SRAM, device_printf(channel->sc->dev, "Allocating DMA program of size %d in RAM\n", new_program_size))
            error = cx88_alloc_dmaprog_buffer(buffer, new_program_size);
            if (error) 
            {
               device_printf(channel->sc->dev, "Unable to allocate DMA program buffer\n");
               return error;
            }
         }
         buffer->dma_program_size = new_program_size;
      }
   }
   if (channel->instrq_size > channel->sc->board_data->max_instrq_size)
   {
      channel->instrq_size = channel->sc->board_data->max_instrq_size;
   }
   DIAG(DIAG_CHANNEL | DIAG_SRAM,
      device_printf(
         channel->sc->dev,
         "Channel ID %d:\n"
         "\tInstrq offset = 0x%x\n"
         "\tInstrq size = 0x%x\n",
         channel->descriptor->channel,
         channel->instrq_offset,
         channel->instrq_size
         )
      )
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      if ((buffer->dma_program_state == CX88_DMAPROG_DONE) &&
          (channel->buffers[buffer->next_buffer_index].dma_program_state ==
           CX88_DMAPROG_LOAD))
      {
         buffer->dma_program_state = CX88_DMAPROG_LINK;  
      }
   }
   /* Now that we've sized and allocated the DMA programs, it's time
    * to actually generate them.*/
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);      
      if ((buffer->dma_program_state == CX88_DMAPROG_LOAD) ||
          (buffer->dma_program_state == CX88_DMAPROG_LINK))
      {
         DIAG(DIAG_BUFFER, 
            if (buffer->dma_program_state == CX88_DMAPROG_LINK)
            {
               device_printf(channel->sc->dev, "Linking buffer %d\n", i);
            }
            else
            {
               device_printf(channel->sc->dev, "Loading buffer %d\n", i);
            })
         cx88_init_dma_counters(buffer);
         if ((error = buffer->generate_dma_program(buffer, i)))
         {
            return error;
         }
         if (buffer->needs_relink)
         {
            buffer->dma_program_state = CX88_DMAPROG_LINK;  
         }
         else
         {
            buffer->dma_program_state = CX88_DMAPROG_DONE;
         }
      }
   }
   return 0;
}

/* Commit the FIFO, instruction queue, and DMA program info to the cx88 registers.
 * After this, we're ready to transfer data */
int cx88_commit_channel(cx88_channel *channel)
{
   DIAG(DIAG_CHANNEL, device_printf(channel->sc->dev, "Committing channel ID %d configuration to registers\n", channel->descriptor->channel))
   if (channel->is_dirty)
   {
      if ((channel->fifo_sram == NULL) ||
          (channel->start_buffer >= channel->num_buffers) ||
          (channel->buffers[channel->start_buffer].data_buffer_type == CX88_DATABUF_NONE))
      {
         return EINVAL;
      }
      int error = cx88_alloc_dmaprogram(channel);
      if (error)
      {
         cx88_release_channel(channel);
         return error;
      }
      uint32_t cdt_size = (channel->num_fifos * channel->sc->board_data->cdt_entry_size) >> 3;
      uint32_t instrq_size = channel->instrq_size >> 2;
      uint32_t cmds_offset = channel->sc->board_data->cmds_offset + 
         (channel->descriptor->dma_number * channel->sc->board_data->cmds_entry_size);
      if (channel->buffers[channel->start_buffer].dma_program_buffer == NULL)
      {
         /* The cx88 stores the flag indicating an SRAM resident DMA program
          * in the same dword as the instruction queue size */
         instrq_size |= (1 << 31);
      }
      /* Generate the CMDS */
      cx88_write_reg(channel->sc, cmds_offset, 
          (uint32_t)(channel->buffers[channel->start_buffer].dma_program_physaddr & 0xFFFFFFFF));
      if (channel->sc->board_data->addr_dwords > 1)
      {
         cmds_offset += 0x4;
         cx88_write_reg(channel->sc, cmds_offset, 
            (uint32_t)((uint64_t)(channel->buffers[channel->start_buffer].dma_program_physaddr) >> 32));
      }
      cx88_write_reg(channel->sc, cmds_offset + 0x4, channel->cdt_offset);
      cx88_write_reg(channel->sc, cmds_offset + 0x8, cdt_size);
      cx88_write_reg(channel->sc, cmds_offset + 0xC, channel->instrq_offset);
      cx88_write_reg(channel->sc, cmds_offset + 0x10, instrq_size);
      channel->is_dirty = FALSE;
   }
   return 0;
}

/* Prepare a channel for starting/restarting a DMA transfer */
int cx88_start_channel(cx88_channel *channel)
{
   if (channel->is_dirty)
   {
      return ENXIO;
   }
   int error = 0;
   cx88_size_t cdt_entry_size = channel->sc->board_data->cdt_entry_size;
   uint32_t cdt_size = (channel->num_fifos * cdt_entry_size) >> 3;   
   uint32_t reg_PTR2 = channel->sc->board_data->ptr2_base + 
      (4 * channel->descriptor->dma_number);
   uint32_t reg_CNT1 = channel->sc->board_data->cnt1_base + 
      (4 * channel->descriptor->dma_number);
   uint32_t reg_CNT2 = channel->sc->board_data->cnt2_base + 
      (4 * channel->descriptor->dma_number);
   /* Write CDT and FIFO info */
   cx88_write_reg(channel->sc, reg_PTR2, channel->cdt_offset);
   cx88_write_reg(channel->sc, reg_CNT1, ((channel->fifo_size >> 3) - 1));
   cx88_write_reg(channel->sc, reg_CNT2, cdt_size);
   for (uint8_t i = 0; i < channel->num_buffers; ++i)
   {
      channel->buffers[i].ready_for_transfer = FALSE;
      if ((channel->descriptor->direction == CX88_DOWNSTREAM) &&
          (channel->buffers[i].data_buffer_type != CX88_DATABUF_NONE) &&
          !error)
      {
         error = cx88_prepare_for_transfer(&(channel->buffers[i]));
      }
   }
   return error;
}

/* Configure channel properties */
int cx88_configure_channel(
   cx88_channel *channel,
   cx88_size_t min_sram_size,
   cx88_size_t packet_size,
   cx88_size_t max_line_size,
   uint32_t line_boundary)
{
   DIAG(DIAG_CHANNEL, device_printf(channel->sc->dev, "Configuring channel ID %d\n", channel->descriptor->channel))
   int error = 0;
   uint8_t i;
   if ((channel->fifo_sram != NULL) && (min_sram_size > rman_get_size(channel->fifo_sram)))
   {
      cx88_free_sram(channel);
   }
   if (!error &&
      ((channel->fifo_sram == NULL) || (channel->packet_size != packet_size) || 
       (channel->max_line_size != max_line_size) || (channel->line_boundary != line_boundary)))
   {
      cx88_size_t old_packet_size = channel->packet_size;
      cx88_size_t old_line_size = channel->line_size;
      /* Allocate the channel if it doesn't already have an SRAM chunk */
      if (channel->fifo_sram == NULL)
      {
         error = cx88_alloc_sram(channel, CX88_SRAM_FIFO, min_sram_size);
      }
      if (!error && (channel->sc->board_data->bridge != CX88_BRIDGE_CX23880) &&
         (channel->control_sram == NULL))
      {
         error = cx88_alloc_sram(channel, CX88_SRAM_CONTROL, 0);
      }
      if (!error)
      {
         /* If the packet size, the maximum line size, or the line boundary are changing,
         * we must re-calculate the FIFO arrangement */
         channel->packet_size = packet_size;
         channel->max_line_size = max_line_size;
         channel->line_boundary = line_boundary;
         error = cx88_configure_fifos(channel);
         channel->is_dirty = TRUE;
         if ((packet_size != old_packet_size) || (channel->line_size != old_line_size))
         {
            for (i = 0; i < channel->num_buffers; ++i)
            {
               channel->buffers[i].dma_program_state = CX88_DMAPROG_SIZE;  
            }
         }
      }
   }
   if (error)
   {
      cx88_free_sram(channel);
   }
   return error;
}

/* Request a given number of buffers for a channel */
int cx88_request_buffers(
   cx88_channel *channel, 
   uint8_t num_buffers,
   uint8_t start_buffer)
{
   uint8_t i;
   uint8_t prev_buffers = channel->num_buffers;
   cx88_buffer *requested_bufs;
   if (start_buffer >= num_buffers)
   {
      return EINVAL;
   }
   if (prev_buffers != num_buffers)
   {
      /* Release resources for any existing buffers
       * beyond the number we're requesting */
      for (i = num_buffers; i < prev_buffers; ++i)
      {
         cx88_release_buffer(&(channel->buffers[i]));
      }
      requested_bufs = realloc(channel->buffers, sizeof(*(channel->buffers)) * num_buffers, M_CX88, M_WAITOK);
      if (requested_bufs == NULL)
      {
         return ENOMEM;
      }
      channel->is_dirty = TRUE;
      channel->buffers = requested_bufs;
      channel->num_buffers = num_buffers;
      /* Initialize any buffers that have been added to the buffer array */
      for (i = prev_buffers; i < num_buffers; ++i)
      {
         cx88_buffer * buffer = &(channel->buffers[i]);
         buffer->channel = channel;
         buffer->data_buffer_size = 0;
         buffer->data_buffer_type = CX88_DATABUF_NONE;
         buffer->data_buffer = NULL;
         buffer->data_buffer_map_state = CX88_DMAMAP_TAGGED;
         buffer->dma_program_size = 0;
         buffer->dma_program_buffer = NULL;
         buffer->dma_program_state = CX88_DMAPROG_SIZE;
         buffer->dma_program_map_state = CX88_DMAMAP_NULL;
         buffer->generate_dma_program = NULL;
         buffer->generate_instr = NULL;
         if (channel->sc->board_data->addr_dwords > 1)
         {
            DIAG(DIAG_BUFFER, device_printf(channel->sc->dev, "Using 64-bit DMA engine\n"))
            buffer->generate_addr = cx88_write_addr64;
         }
         else
         {
            DIAG(DIAG_BUFFER, device_printf(channel->sc->dev, "Using 32-bit DMA engine\n"))
            buffer->generate_addr = cx88_write_addr32;
         }
         buffer->sgl = NULL;
         buffer->sgl_size = 0;
         buffer->owner = NULL;
         buffer->next_buffer_index = 0;
      }
   }
   if (start_buffer != channel->start_buffer)
   {
      channel->start_buffer = start_buffer;
      channel->is_dirty = TRUE;
   }
   return 0;
}

/* bus_dma map load callback for data buffers */
static void cx88_dmamap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error)
{
   cx88_buffer *buffer = arg;
   DIAG(DIAG_BUFFER, printf("Allocating SGL of size %d\n", sgl_size))
   if (error)
   {
      return;
   }
   if (buffer->sgl_size != sgl_size)
   {
      free(buffer->sgl, M_CX88);
      /* Store the scatter/gather list in the buffer */
      if ((buffer->sgl = malloc(sizeof(*sgl) * sgl_size, M_CX88, M_WAITOK)) == NULL)
      {
         printf("CX88: Unable to allocate scatter/gather list\n");
         buffer->sgl_size = 0;
         return;
      }
      buffer->sgl_size = sgl_size;
   }
   memcpy(buffer->sgl, sgl, sizeof(*sgl) * sgl_size);
}

/* bus_dma map load callback for user-allocated buffers */
static void cx88_dmamap_load_uio(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   bus_size_t mapsize,
   int error)
{
   cx88_dmamap_load(arg, sgl, sgl_size, error);
}

/* Configure a buffer structure and allocate its data buffer */
int cx88_configure_buffer(
   cx88_buffer *buffer,
   void *data_buffer,
   struct thread *context,
   size_t buffer_size,
   uint8_t next_buffer_index,
   cx88_risc_generator dma_program_generator,
   cx88_risc_reprogram_t reprogram)
{
   DIAG(DIAG_BUFFER, device_printf(buffer->channel->sc->dev, "Configuring DMA data buffer\n"))
   int error = 0;
   if (next_buffer_index >= buffer->channel->num_buffers)
   {
      return EINVAL;
   }
   if (next_buffer_index != buffer->next_buffer_index)
   {
      if (buffer->dma_program_state >= CX88_DMAPROG_LINK)
      {
         buffer->dma_program_state = CX88_DMAPROG_LINK;  
      }
      buffer->next_buffer_index = next_buffer_index;
      buffer->channel->is_dirty = TRUE;
   }
   if ((reprogram == CX88_RISC_REPROGRAM_FORCE) || 
       (buffer->generate_dma_program != dma_program_generator))
   {
      /* If we need to re-generate the DMA program, assign the new
       * generator and indicate that the DMA program needs to be resized
       * and reloaded */
      buffer->generate_dma_program = dma_program_generator;
      buffer->dma_program_state = CX88_DMAPROG_SIZE;
      buffer->channel->is_dirty = TRUE;
   }
   /* The remainder of this function deals with the actual buffer
    * allocation and DMA mapping.  If the physical buffer is not
    * actually changing, we can just return here */
   if (context == NULL)
   {
      if ((buffer->data_buffer_size == buffer_size) &&
         (((buffer->data_buffer_type == CX88_DATABUF_MALLOC) && (data_buffer == NULL)) ||
            ((buffer->data_buffer_type == CX88_DATABUF_KERNEL) && (data_buffer == buffer->data_buffer))))
      {
         return 0;
      }
   }
   else
   {
      if ((buffer->data_buffer_size == buffer_size) &&
            (buffer->data_buffer_type == CX88_DATABUF_USER) && 
            (data_buffer == buffer->data_buffer) && (buffer->owner == context->td_proc))
      {
         return 0;
      }
   }
   buffer->channel->is_dirty = TRUE;
   if (buffer->data_buffer_map_state == CX88_DMAMAP_TAGGED)
   {
      error = bus_dmamap_create(buffer->channel->data_buffer_tag, 0, &(buffer->data_buffer_map));
      if (error)
      {
         device_printf(buffer->channel->sc->dev, "Unable to create DMA map for buffer\n");
         cx88_release_buffer(buffer);
         return error;
      }
      buffer->data_buffer_map_state = CX88_DMAMAP_CREATED;
   }
   cx88_free_data_buffer(buffer);  
   buffer->data_buffer_size = buffer_size;
   buffer->dma_program_state = CX88_DMAPROG_SIZE;
   if (context == NULL)
   {
      if (data_buffer == NULL)
      {
         /* Don't use bus_dmamem_alloc as it will allocate the maximum allowable
         * buffer size, which we probably don't need.  And it uses contigmalloc() underneath.  Here we
         * use malloc() to ensure greater flexibility in allocating larger data buffers, at the possible
         * expense of bounce buffering on non-IOMMU machines with >4GB of RAM (CX23880 only) */
         if ((buffer->data_buffer = malloc(round_page(buffer_size), M_CX88, M_WAITOK | M_ZERO)) == NULL)
         {
            device_printf(buffer->channel->sc->dev, "Unable to allocate DMA buffer for buffer\n");
            cx88_release_buffer(buffer);
            return ENOMEM;
         }
         buffer->data_buffer_type = CX88_DATABUF_MALLOC;
      }
      else
      {
         buffer->data_buffer = data_buffer;  
         buffer->data_buffer_type = CX88_DATABUF_KERNEL;
      }
      error = bus_dmamap_load(
         buffer->channel->data_buffer_tag,
         buffer->data_buffer_map,
         buffer->data_buffer,
         buffer_size,
         cx88_dmamap_load,
         buffer,
         BUS_DMA_NOWAIT);
   }
   else
   {
      if ((data_buffer == NULL) || (context->td_proc != curproc))
      {
         device_printf(buffer->channel->sc->dev, 
            "User-mode data buffer must be non-NULL and allocated by current process\n");
         error = EINVAL;  
      }
      else
      {
         error = vslock(data_buffer, buffer_size);
      }
      if (error)
      {
         device_printf(buffer->channel->sc->dev, "Unable to wire buffer\n");
         cx88_release_buffer(buffer);
         return error;
      }
      buffer->data_buffer = data_buffer;
      buffer->data_buffer_type = CX88_DATABUF_USER;
      buffer->owner = context->td_proc;
      struct iovec uvec;
      struct uio ubuf;
      uvec.iov_base = buffer->data_buffer;
      uvec.iov_len = buffer_size;
      ubuf.uio_iov = &uvec;
      ubuf.uio_iovcnt = 1;
      ubuf.uio_segflg = UIO_USERSPACE;
      ubuf.uio_td = context;
      ubuf.uio_resid = buffer_size;
      ubuf.uio_offset = 0;
      ubuf.uio_rw = ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ? UIO_WRITE : UIO_READ);
      error = bus_dmamap_load_uio(
         buffer->channel->data_buffer_tag,
         buffer->data_buffer_map,
         &ubuf,
         cx88_dmamap_load_uio,
         buffer,
         BUS_DMA_NOWAIT);
   }
   if (!error && (buffer->sgl == NULL))
   {
      error = ENOMEM;
   }
   if (error)
   {
      cx88_release_buffer(buffer);
      device_printf(buffer->channel->sc->dev, "Unable to load DMA map for buffer\n");
      return error;
   }
   buffer->data_buffer_map_state = CX88_DMAMAP_LOADED;
   return 0;
}

/* Synchronize RISC and data buffers before DMA */
int cx88_prepare_for_transfer(cx88_buffer *buffer)
{
   if ((buffer->channel->is_dirty) ||
       (buffer->data_buffer_type == CX88_DATABUF_NONE) ||
       ((buffer->data_buffer_type == CX88_DATABUF_USER) && 
        (buffer->owner != curproc)))
   {
      return EINVAL;
   }
   if (buffer->dma_program_buffer != NULL)
   {
      bus_dmamap_sync(buffer->dma_program_tag, buffer->dma_program_map, BUS_DMASYNC_PREWRITE);
   }
   bus_dmamap_sync(
      buffer->channel->data_buffer_tag,
      buffer->data_buffer_map,
      ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ? 
         BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE));
   buffer->ready_for_transfer = TRUE;
   return 0;
}

/* Synchronize RISC and data buffers after DMA */
int cx88_mark_transfer_complete(cx88_buffer *buffer)
{
   buffer->ready_for_transfer = FALSE;
   if ((buffer->channel->is_dirty) ||
       (buffer->data_buffer_type == CX88_DATABUF_NONE) ||
       ((buffer->data_buffer_type == CX88_DATABUF_USER) && 
        (buffer->owner != curproc)))
   {
      return EINVAL;
   }
   if (buffer->dma_program_buffer != NULL)
   {
      bus_dmamap_sync(buffer->dma_program_tag, buffer->dma_program_map, BUS_DMASYNC_POSTWRITE);
   }
   bus_dmamap_sync(
      buffer->channel->data_buffer_tag,
      buffer->data_buffer_map,
      ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ?
         BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE));
   return 0;
}

/* Probe routine */
int cx88_common_probe(device_t dev, const cx88_client_driver * client) 
{
   uint16_t pid = pci_get_device(dev);
   uint16_t vid = pci_get_vendor(dev);
   /* Match against Conexant's vendor ID and a product ID generated
    * from the client's function index */
   if (vid != CX88_VID)
   {
      return ENXIO;
   }
   switch (client->bridge)
   {
      case CX88_BRIDGE_CX23880:
         if ((pid == (client->function + CX23880_BASE_PID)) &&
             (client->function < CX88_NUM_FUNCTIONS))
         {
            device_set_desc(dev, client->description);
            return BUS_PROBE_DEFAULT;
         }
         break;
      case CX88_BRIDGE_CX23885:
         if (pid == CX23885_PID)
         {
            device_set_desc(dev, client->description);
            return BUS_PROBE_DEFAULT;
         }     
         break;
      case CX88_BRIDGE_CX23887:
         if (pid == CX23887_PID)
         {
            device_set_desc(dev, client->description);
            return BUS_PROBE_DEFAULT;
         }
         break;
   }
   return ENXIO;
}

/* Helper function for freeing client-specific data in the softc
 * structure.  This has been separated from cx88_release_softc() to
 * allow clients access to the softc via their devices/cdevs during
 * detachment (see cx88_common_detach()) */
static int cx88_release_client(cx88_common_softc *sc)
{
   int error = 0;
   if (sc->irq_cookie != NULL)
   {
      if ((error = bus_teardown_intr(sc->dev, sc->irq_res, sc->irq_cookie)) != 0)
      {
         return error;
      }
      sc->irq_cookie = NULL;
   }
   if (sc->is_configured && (sc->client->teardown != NULL))
   {
      /* Lock the board to ensure all devmethods using
       * cx88_get_softc() will get a NULL softc after we
       * set is_configured to FALSE */
      cx88_lock_board(sc);
      sc->is_configured = FALSE;
      cx88_unlock_board(sc);
      if ((error = sc->client->teardown(sc)) != 0)
      {
         sc->is_configured = TRUE;
         return error;
      }
   }
   return error;
}

/* Free a common softc structure.  Used both for device detachment
 * and error cases during device attachment, so must check
 * each softc item before freeing it */
static int cx88_release_softc(cx88_common_softc *sc)
{
   int error = cx88_release_client(sc);
   if (error != 0)
   {
      return error;
   }
   if (sc->irq_res != NULL)
   {
      bus_release_resource(sc->dev, SYS_RES_IRQ, sc->irq_rid, sc->irq_res);
      sc->irq_res = NULL;
   }
   if (sc->msi)
   {
      pci_release_msi(sc->dev);
      sc->msi = FALSE;
   }
   if (sc->mem_res != NULL)
   {
      if ((sc->client != NULL) && (sc->client->bridge != CX88_BRIDGE_CX23880))
      {
         cx88_write_reg(sc, CX23885_REG_SOFT_RESET, 1);
      }
      bus_release_resource(sc->dev, SYS_RES_MEMORY, sc->mem_rid, sc->mem_res);
      sc->mem_res = NULL;
   }
   if ((error = cx88_remove_board_data(sc, (void **)(&(sc->board_data)), cx88_destroy_board_data)) != 0)
   {
      return error;
   }
   /* Softc is externally allocated (see below) */
   device_set_softc(sc->dev, NULL);
   free(sc, M_CX88);
   return error;
}

/* Attach routine */
int cx88_common_attach(device_t dev, const cx88_client_driver *client) 
{
   /* Externally allocate softc so that we can set it to NULL during detachment
    * without freeing it--since our DEVMETHODs will still be valid, DEVMETHODs
    * that may be called from other threads can use cx88_get_softc() to
    * determine that the device is no longer in a valid state */
   if (client->softc_size < sizeof(cx88_common_softc))
   {
      return EINVAL;
   }
   int (*board_setup)(void*) = NULL;
   switch (client->bridge)
   {
      case CX88_BRIDGE_CX23880:
         board_setup = cx23880_init_board_data;
         break;
      case CX88_BRIDGE_CX23885:
         board_setup = cx23885_init_board_data;
         break;
      case CX88_BRIDGE_CX23887:
         board_setup = cx23887_init_board_data;
         break;
      default:
         return EINVAL;
   }
   cx88_common_softc *sc = malloc(client->softc_size, M_CX88, M_NOWAIT | M_ZERO);
   uint8_t i;
   int error = 0;
   if (sc == NULL)
   {
      return ENOMEM;
   }
   device_set_softc(dev, sc);
   sc->board_data = NULL;
   sc->mem_res = NULL;
   sc->irq_res = NULL;
   sc->irq_cookie = NULL;
   sc->client = client;
   sc->msi = FALSE;
   sc->is_configured = FALSE;
   sc->dev = dev;
   sc->num_siblings = 0;
   if (resource_disabled(device_get_name(dev), device_get_unit(dev)))
   {
      device_printf(dev, "Disabled by device hint\n");
      return ENXIO;
   }
   if (client->bridge == CX88_BRIDGE_CX23880)
   {
      /* Generate the array of sibling devices */
      for (i = 0; i < CX88_NUM_FUNCTIONS; ++i)
      {
         device_t sibling;
         /* Skip our own function */
         if (i == client->function)
         {
            continue;
         }
         sibling = pci_find_bsf(pci_get_bus(dev), pci_get_slot(dev), i);
         if (sibling != NULL)
         {
            DIAG(DIAG_DEVICE, device_printf(dev, "Found sibling at function %d\n", i))
            sc->siblings[(sc->num_siblings)++] = sibling;
         }
      }
   }
   if ((error = cx88_common_resume(dev)))
   {
      device_printf(dev, "Unable to initialize PCI configuration\n");
      cx88_release_softc(sc);
      return error;
   }
   sc->irq_rid = 0;
   int msix_count = pci_msix_count(dev);
   int msi_count = pci_msi_count(dev);
   int intr_count = 1;
   int msi_enabled = 0;
   if ((resource_int_value(device_get_name(dev), device_get_unit(dev), "msi", &msi_enabled) == 0) && msi_enabled &&
      (((msix_count == 1) && (pci_alloc_msix(dev, &intr_count) == 0)) ||
       ((msi_count == 1) && (pci_alloc_msi(dev, &intr_count) == 0))))
   {
      if (intr_count == 1)
      {
         DIAG(DIAG_DEVICE, device_printf(dev, "Using MSI(X), msix_count = %d, msi_count = %d\n", msix_count, msi_count))
         sc->irq_rid = 1;
         sc->msi = TRUE;
      }
      else
      {
         pci_release_msi(dev);
      }
   }
   sc->irq_res = bus_alloc_resource_any(dev, SYS_RES_IRQ, &sc->irq_rid, RF_ACTIVE | RF_SHAREABLE);
   if (sc->irq_res == NULL)
   {
      device_printf(dev, "Unable to allocate interrupt resource\n");
      cx88_release_softc(sc);
      return ENXIO;
   }
   /* Load the common board data structure */
   if ((error = cx88_add_board_data(sc, (void **)(&(sc->board_data)),
      sizeof(*(sc->board_data)), cx88_get_sibling_board_data, board_setup)))
   {
      cx88_release_softc(sc);
      return error;
   }
   /* Call client-specific softc setup routine */
   if ((client->setup != NULL) && ((error = client->setup(sc))))
   {
      cx88_release_softc(sc);
      return error;
   }
   sc->is_configured = TRUE;
   if (((client->interrupt_filter != NULL) || (client->interrupt_handler != NULL)) &&
      (error = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_AV | INTR_MPSAFE,
      client->interrupt_filter, client->interrupt_handler, sc, &(sc->irq_cookie))))
   {
      device_printf(dev, "Unable to establish interrupt handler\n");
      cx88_release_softc(sc);
      return error;
   }
   return 0;
}

/* Resume routine: Initialize global register settings */
int cx88_common_resume(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   uint32_t cmdr;
   cmdr = pci_read_config(dev, PCIR_COMMAND, 2);
   /* Enable memory window, bus mastering, and write caching */
   cmdr |= (PCIM_CMD_MEMEN | PCIM_CMD_BUSMASTEREN | PCIM_CMD_MWRICEN);
   pci_write_config(dev, PCIR_COMMAND, cmdr, 2);
   if (pci_read_config(dev, PCIR_LATTIMER, 1) == 0)
   {
      pci_write_config(dev, PCIR_LATTIMER, CX88_DEFAULT_LATENCY, 1);
   }
   if (sc->mem_res == NULL)
   {
      sc->mem_rid = PCIR_BAR(0);
      sc->mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &sc->mem_rid, RF_ACTIVE);
      if (sc->mem_res == NULL)
      {
         device_printf(dev, "Unable to map memory\n");
         cx88_release_softc(sc);
         return ENXIO;
      }
      sc->mem_tag = rman_get_bustag(sc->mem_res);
      sc->mem_handle = rman_get_bushandle(sc->mem_res);
   }
   /* Work around CX23885 bug which raises a non-maskable interrupt if bit 4 of TLCTL0 is set */
   if (sc->client->bridge == CX88_BRIDGE_CX23885)
   {
      cx88_write_reg(sc, CX23885_REG_TLCTL0, cx88_read_reg(sc, CX23885_REG_TLCTL0) & (~(1UL << 4)));
   }
   return 0;
}

/* Detach routine */
int cx88_common_detach(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   int error = cx88_release_client(sc);
   if (error != 0)
   {
      return error;
   }
   /* Acquire global lock and board lock to ensure a sibling is not calling
    * us.  After we release the locks, subsequent sibling calls
    * will get a NULL softc. */
   mtx_lock(&interdev_lock);
   cx88_lock_board(sc);
   device_set_softc(dev, NULL);
   cx88_unlock_board(sc);
   mtx_unlock(&interdev_lock);
   if ((error = cx88_release_softc(sc)))
   {
      /* Reset the softc so that we can try releasing it later */
      device_set_softc(dev, sc);  
   }
   if (!error)
   {
      error = bus_generic_detach(dev);
   }
   return error;
}

static int cx88_init_sram(
   struct rman *sram,
   uint32_t base,
   uint32_t size
   )
{
   int error = 0;
   uint32_t sram_end = base + size - 1;
   sram->rm_start = base;
   sram->rm_end = sram_end;
   sram->rm_type = RMAN_ARRAY;
   sram->rm_descr = "CX88 SRAM";
   error = rman_init(sram);
   if (!error)
   {
      if ((error = rman_manage_region(sram, base, sram_end)) != 0)
      {
         rman_fini(sram);
      }
   }
   return error;
}

static void cx88_init_board_data(cx88_board_data *data)
{
   mtx_init(&(data->lock), "CX88 board lock", NULL, MTX_DEF | MTX_RECURSE);
   data->present_channels = 0;
   data->present_dma_channels = 0;
}

/* Initializes a newly-allocated cx88_board_data structure.
 * Called when the current client is the first driver to reference
 * the board data */
static int cx23880_init_board_data(void *arg)
{
   DIAG(DIAG_DEVICE, printf("Initializing CX23880 board data\n"))
   cx88_board_data *data = arg;
   cx88_init_board_data(data);
   data->addr_dwords = 1;
   data->cmds_entry_size = 0x40;
   data->cmds_offset = 0x180040;
   data->cdt_entry_size = 0x10;
   data->min_instrq_size = 0x40;
   data->max_instrq_size = 0x3FC;
   data->max_num_fifos = 0x3FF;
   data->bridge = CX88_BRIDGE_CX23880;
   data->channel_table = cx23880_channels;
   data->table_size = sizeof(cx23880_channels) / sizeof(cx88_channel_desc);
   data->ptr1_base = 0x300080;
   data->ptr2_base = 0x3000C0;
   data->cnt1_base = 0x300100;
   data->cnt2_base = 0x300140;
   return cx88_init_sram(&(data->fifo_sram), 0x180340, 0x7CC0);
}

static void cx23885_common_init(cx88_board_data *data)
{
   cx88_init_board_data(data);
   data->addr_dwords = 2;
   data->cmds_entry_size = 0x50;
   data->cmds_offset = 0x10000;
   data->cdt_entry_size = 0x10;
   data->min_instrq_size = 0x60;
   data->max_instrq_size = 0x3FC;
   data->max_num_fifos = 0x3FF;
   data->ptr1_base = 0x100000;
   data->ptr2_base = 0x100040;
   data->cnt1_base = 0x100080;
   data->cnt2_base = 0x1000C0;
   data->channel_table = cx23885_channels;
   data->table_size = sizeof(cx23885_channels) / sizeof(cx88_channel_desc);
}

static int cx23885_init_board_data(void *arg)
{
   DIAG(DIAG_DEVICE, printf("Initializing CX23885 board data\n"))
   cx88_board_data *data = arg;
   int error = 0;
   cx23885_common_init(data);
   data->bridge = CX88_BRIDGE_CX23885;
   error = cx88_init_sram(&(data->fifo_sram), 0x40, 0x7FC0);
   if (!error)
   {
      error = cx88_init_sram(&(data->control_sram), 0x10380, 0x480);
      if (error)
      {
         rman_fini(&(data->fifo_sram));
      }
   }
   return error;
}

static int cx23887_init_board_data(void *arg)
{
   DIAG(DIAG_DEVICE, printf("Initializing CX23887 board data\n"))
   cx88_board_data *data = arg;
   int error = 0;
   cx23885_common_init(data);
   data->bridge = CX88_BRIDGE_CX23887;
   error = cx88_init_sram(&(data->fifo_sram), 0x40, 0x8FC0);
   if (!error)
   {
      error = cx88_init_sram(&(data->control_sram), 0x105B0, 0x3650);
      if (error)
      {
         rman_fini(&(data->fifo_sram));
      }
   }
   return error;
}

/* Cleans up the board data structure when the current client is the
 * last to reference it and it is about to be freed */
static int cx88_destroy_board_data(void *arg)
{
   cx88_board_data *data = arg;
   mtx_destroy(&(data->lock));
   rman_fini(&(data->fifo_sram));
   if (data->bridge != CX88_BRIDGE_CX23880)
   {
      rman_fini(&(data->control_sram));
   }
   return 0;
}

/* Accessor function for board data on siblings */
static void *cx88_get_sibling_board_data(device_t dev)
{
   return CX88_GET_BOARD_DATA(dev);
}

/* Generic function for adding data to be shared by multiple
 * drivers for the same board */
int cx88_add_board_data(
   cx88_common_softc *sc,
   void **data,
   size_t data_size,
   get_board_data_func accessor,
   int (*setup)(void*)
   )
{
   /* Ensure we can add or update the board data structure atomically
    * w/ respect to siblings */
   mtx_lock(&interdev_lock);
   *data = NULL;
   int error = 0;
   uint8_t i;
   /* First check if a siblings has allocated the structure */
   for (i = 0; i < sc->num_siblings; ++i)
   {
      if ((*data = accessor(sc->siblings[i])) != NULL)
      {
         DIAG(DIAG_DEVICE, device_printf(sc->dev, "Found data on sibling board\n"))
         break;
      }
   }
   if (*data == NULL)
   {
      /* No sibling has allocated the structure, so we need to */
      size_t *base_addr;
      if ((base_addr = malloc(data_size + sizeof(size_t), M_CX88, M_NOWAIT | M_ZERO)) == NULL)
      {
         device_printf(sc->dev, "Unable to allocate board data\n");
         mtx_unlock(&interdev_lock);
         return ENOMEM;
      }
      DIAG(DIAG_DEVICE, device_printf(sc->dev, "Allocated board data\n"))
      /* Place the refcount for the structure in the word immediately preceding the 
       * the address that will be returned to the client */
      *base_addr = 1;
      *data = base_addr + 1;
      /* If the client supplied an initialization routine, call it now */
      if (setup != NULL && (error = setup(*data)))
      {
         free(base_addr, M_CX88);
         *data = NULL;
      }
   }
   else
   {
      /* A sibling has already allocated the structure, so just increment its
       * refcount */
      size_t *base_addr = ((size_t *) (*data)) - 1;
      ++(*base_addr);
   }
   mtx_unlock(&interdev_lock);
   return error;
}

/* Function for removing a reference to a shared data structure */
int cx88_remove_board_data(
   cx88_common_softc *sc,
   void **data,
   int (*teardown)(void*)
   )
{
   int error = 0;
   mtx_lock(&interdev_lock);
   if ((*data) != NULL)
   {
      size_t *base_addr = ((size_t *)(*data)) - 1;
      if (*base_addr == 1)
      {
         /* The refcount is 1, so we're the last driver to use the structure.
          * Go ahead and free it */
         DIAG(DIAG_DEVICE, device_printf(sc->dev, "Freeing board data\n"))
         if (teardown != NULL)
         {
            /* If the client supplied a cleanup routine, call it now */
            error = teardown(*data);
         }
         if (!error)
         {
            free(base_addr, M_CX88);
         }
      }
      else
      {
         /* We're not the last to use it, so just decrement the refcount */
         --(*base_addr);
      }
      *data = NULL;
   }
   mtx_unlock(&interdev_lock);
   return error;
}

/* Retrieve a driver's softc, only if it has been
 * allocated and initialized */
void *cx88_get_softc(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   if ((sc == NULL) || !(sc->is_configured))
   {
      return NULL;
   }
   return sc;
}

/* DEVMETHOD for client drivers to implement so that siblings
 * can access their common board data structure */
cx88_board_data *cx88_common_get_board_data(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   if (sc != NULL)
   {
      return sc->board_data;
   }
   return NULL;
}

/* Generic mmap function that clients can call from their d_mmap routines */
int cx88_common_mmap(
   cx88_channel *channel,
   vm_ooffset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   int i;
   vm_offset_t phys_offset = 0;
   int error = EINVAL;
   if (nprot & PROT_EXEC)
   {
      return error;
   }
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      if ((buffer->data_buffer_type == CX88_DATABUF_NONE) ||
         (buffer->data_buffer_type == CX88_DATABUF_USER))
      {
         /* Only map kernel-allocated buffers */
         continue;
      }
      vm_offset_t data_buffer = (vm_offset_t)(buffer->data_buffer);
      if ((data_buffer - trunc_page(data_buffer)) != 0)
      {
         device_printf(channel->sc->dev, "Unable to mmap: buffer %d not page-aligned\n", i);
         return EINVAL;
      }
      if ((phys_offset + buffer->data_buffer_size) > offset)
      {
         //Use vtophys instead of the stored sgl--we want the CPU-visible physical address
         //of the selected offset in the original buffer, while the sgl stores device-visible
         //physical addresses (they may differ on machines with IOMMUs or in cases where 
         //bounce buffers are used).
         *paddr = vtophys ((char*)data_buffer + (offset - phys_offset));
         DIAG(DIAG_MMAP, device_printf(channel->sc->dev, "MMAP: Using buffer %d at offset 0x%lx\n", i, (unsigned long)(offset - phys_offset)))
         return 0;
      }
      else
      {
         phys_offset += round_page(buffer->data_buffer_size);
      }
   }
   return error;
}

/* Write/read GPIO pins */
void cx88_gpioctl(cx88_common_softc *sc, cx88_common_gpio *gpio)
{
   cx88_lock_board(sc);
   switch (sc->board_data->bridge)
   {
      case CX88_BRIDGE_CX23880:
         if (gpio->in_write_mask != 0)
         {
            /* First write all required GPIO pins */
            cx23880_write_gpio(sc, gpio->in_write_mask, gpio->inout_gpio);
         }
         if (gpio->in_read_mask != 0)
         {
            /* Next read all required GPIO pins */
            gpio->inout_gpio = cx23880_read_gpio(sc, gpio->in_read_mask);
         }
         break;
      case CX88_BRIDGE_CX23885:
      case CX88_BRIDGE_CX23887:
         if (gpio->in_write_mask != 0)
         {
            /* First write all required GPIO pins */
            cx23885_write_gpio(sc, gpio->in_write_mask, gpio->inout_gpio);
         }
         if (gpio->in_read_mask != 0)
         {
            /* Next read all required GPIO pins */
            gpio->inout_gpio = cx23885_read_gpio(sc, gpio->in_read_mask);
         }
         break;
   }
   cx88_unlock_board(sc);
}

/* Write GPIO pins */
static void cx23880_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value)
{
   static uint32_t regs[] = {CX23880_REG_GPIO0, CX23880_REG_GPIO1, CX23880_REG_GPIO2, CX23880_REG_GPIO3};
   int i;
   uint8_t shift;
   uint32_t write_data = mask & value;
   uint32_t write_mask;
   for (i = 0; i < (sizeof(regs) / sizeof(uint32_t)); ++i)
   {
      shift = i * 8;
      write_mask = (mask >> shift) & 0xFF;
      if (write_mask != 0)
      {
         uint32_t field = 0;
         /* Enable writing for only the bits we want to write */
         field |= cx88_write_field(CX23880_GPIO_BIT_WRITE, write_mask);
         field |= cx88_write_field(CX23880_GPIO_OUT_ENABLE, write_mask);
         field |= cx88_write_field(CX23880_GPIO_IO, (write_data >> shift));
         cx88_write_reg(sc, regs[i], field);
      }
   }
}

/* Write GPIO pins */
static void cx23885_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value)
{
   uint32_t field = 0;
   field |= cx88_write_field(CX23885_GPIO_LINE_MASK, mask);
   field |= cx88_write_field(CX23885_GPIO_IO, value);
   cx88_write_reg(sc, CX23885_REG_GPIO, field);
}

/* Read GPIO pins */
static uint32_t cx23880_read_gpio(cx88_common_softc *sc, uint32_t mask)
{
   uint32_t value = 0;
   static uint32_t regs[] = {CX23880_REG_GPIO0, CX23880_REG_GPIO1, CX23880_REG_GPIO2, CX23880_REG_GPIO3};
   int i;
   uint8_t shift;
   uint32_t read_mask;
   for (i = 0; i < (sizeof(regs) / sizeof(uint32_t)); ++i)
   {
      shift = i * 8;
      read_mask = (mask >> shift) & 0xFF;
      if (read_mask != 0)
      {
         value |= ((cx88_read_field(cx88_read_reg(sc, regs[i]), CX23880_GPIO_IO) & read_mask) << shift);
      }
   }
   return value;
}

/* Read GPIO pins */
static uint32_t cx23885_read_gpio(cx88_common_softc *sc, uint32_t mask)
{
   return (cx88_read_field(cx88_read_reg(sc, CX23885_REG_GPIO), CX23885_GPIO_IO) & mask);
}

void cx88_peripheral_reset(cx88_common_softc *sc, uint32_t reset_val)
{
   if (sc->board_data->bridge == CX88_BRIDGE_CX23880)
   {
      cx88_write_reg(sc, CX23880_REG_SRST, reset_val);   
   }
}

static int cx23885_mc_wait(cx88_common_softc *sc)
{
   struct timeval start;
   struct timeval current;
   microtime(&start);
   while (!(cx88_read_reg(sc, CX23885_REG_MC_RWD) & CX23885_MC_READY))
   {
      microtime(&current);
      timevalsub(&current, &start);
      if (current.tv_usec >= 1000)
      {
         return EIO;
      }
   }
   return 0;
}

static void cx23885_mc_write_byte(cx88_common_softc *sc, uint8_t addr, uint8_t byte)
{
   uint32_t reg = cx88_write_field(CX23885_MC_ADDR, addr) | CX23885_MC_READ | CX23885_MC_READY | byte;
   cx88_write_reg(sc, CX23885_REG_MC_RWD, reg);
   cx88_write_reg(sc, CX23885_REG_MC_RWD, reg | CX23885_MC_CHIP_SEL | CX23885_MC_WRITE);
}

static uint8_t cx23885_mc_read_byte(cx88_common_softc *sc, uint8_t addr)
{
   uint32_t reg = cx88_write_field(CX23885_MC_ADDR, addr) | CX23885_MC_READ | CX23885_MC_READY;
   cx88_write_reg(sc, CX23885_REG_MC_RWD, reg);
   reg = (reg & ~(CX23885_MC_READ)) | CX23885_MC_WRITE;
   cx88_write_reg(sc, CX23885_REG_MC_RWD, reg);
   uint8_t byte = (uint8_t)(cx88_read_field(cx88_read_reg(sc, CX23885_REG_MC_RWD), CX23885_MC_DATA));
   cx88_write_reg(sc, CX23885_REG_MC_RWD, CX23885_MC_WRITE | CX23885_MC_READ | CX23885_MC_CHIP_SEL | CX23885_MC_READY);
   return byte;
}

int cx23885_mc_write_reg(cx88_common_softc *sc, uint16_t address, uint32_t value)
{
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY);
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_DATA0, (uint8_t)(value & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_DATA1, (uint8_t)((value >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_DATA2, (uint8_t)((value >> 16) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_DATA3, (uint8_t)((value >> 24) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_ADDR0, (uint8_t)(address & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_ADDR1, (uint8_t)((address >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_MODE, CX23885_MC_DATA_REG_WRITE);
   return cx23885_mc_wait(sc);
}

int cx23885_mc_read_reg(cx88_common_softc *sc, uint16_t address, uint32_t *value)
{
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY);
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_ADDR0, (uint8_t)(address & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_ADDR1, (uint8_t)((address >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_REG_MODE, CX23885_MC_DATA_REG_READ);
   int error = cx23885_mc_wait(sc);
   if (error)
   {
      return error;
   }
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY | cx88_write_field(CX23885_MC_DATA, 0xFF));
   uint8_t byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_REG_DATA0);
   *value = byte;
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_REG_DATA1);
   *value |= ((uint32_t)byte << 8);
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_REG_DATA2);
   *value |= ((uint32_t)byte << 16);
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_REG_DATA3);
   *value |= ((uint32_t)byte << 24);
   return 0;
}

int cx23885_mc_write_mem(cx88_common_softc *sc, uint32_t address, uint32_t value)
{
   if (address > 0x3FFFFF)
   {
      return EINVAL;
   }
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY);
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_DATA0, (uint8_t)(value & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_DATA1, (uint8_t)((value >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_DATA2, (uint8_t)((value >> 16) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_DATA3, (uint8_t)((value >> 24) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR2, CX23885_MC_DATA_MEM_WRITE | ((address >> 16) & 0x3F));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR1, (uint8_t)((address >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR0, (uint8_t)(address & 0xFF));
   return cx23885_mc_wait(sc);
}

int cx23885_mc_read_mem(cx88_common_softc *sc, uint32_t address, uint32_t *value)
{
   if (address > 0x3FFFFF)
   {
      return EINVAL;
   }
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY);
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR2, CX23885_MC_DATA_MEM_READ | ((address >> 16) & 0x3F));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR1, (uint8_t)((address >> 8) & 0xFF));
   cx23885_mc_write_byte(sc, CX23885_MC_ADDR_MEM_ADDR0, (uint8_t)(address & 0xFF));
   int error = cx23885_mc_wait(sc);
   if (error)
   {
      return error;
   }
   cx88_write_reg(sc, CX23885_REG_MC_OEN, CX23885_MC_READY | cx88_write_field(CX23885_MC_DATA, 0xFF));
   uint8_t byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_MEM_DATA3);
   *value = (uint32_t)byte << 24;
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_MEM_DATA2);
   *value |= ((uint32_t)byte << 16);
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_MEM_DATA1);
   *value |= ((uint32_t)byte << 8);
   byte = cx23885_mc_read_byte(sc, CX23885_MC_ADDR_MEM_DATA0);
   *value |= byte;
   return 0;
}

int cx23885_mc_gpioctl(cx88_common_softc *sc, cx88_common_gpio *gpio)
{
   int error = 0;
   if (sc->board_data->bridge == CX88_BRIDGE_CX23880)
   {
      return ENODEV;
   }
   cx88_lock_board(sc);
   cx88_write_reg(sc, CX23885_REG_MC_CTL,
      cx88_write_field(CX23885_MC_CTL_SPEED, CX23885_MC_SPEED_FAST) |
      cx88_write_field(CX23885_MC_CTL_GPIO_SEL, CX23885_MC_GPIO_SEL_GPIO3) |
      CX23885_MC_CTL_GPIO_ENABLE);
   uint32_t write_mask = gpio->in_write_mask & 0xFFFF;
   uint32_t read_mask = gpio->in_read_mask & 0xFFFF;
   if (write_mask != 0)
   {
      error = (error ? error : cx23885_mc_write_reg(sc, 0x9020, write_mask));
      error = (error ? error : cx23885_mc_write_reg(sc, 0x900C, gpio->inout_gpio));
   }
   if (!error && (read_mask != 0))
   {
      error = cx23885_mc_read_reg(sc, 0x900C, &(gpio->inout_gpio));
      gpio->inout_gpio &= read_mask;
   }
   cx88_unlock_board(sc);
   return error;
}

/* Generate a RISC instruction */
static inline uint32_t cx88_format_instr(
   cx88_risc_opcode opcode,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   return ((opcode << 28) | (intrs << 24) | (countctl << 16));
}

/* Write a RISC instruction to SRAM */
static void cx88_sram_instr(cx88_buffer *buffer, uint32_t instr)
{
   cx88_write_reg(
      buffer->channel->sc,
      (uint32_t)(buffer->dma_program_physaddr + (buffer->dma_program_counter * sizeof(uint32_t))), 
      instr);
   ++(buffer->dma_program_counter);
}

/* Write a RISC instruction to host memory */
static void cx88_ram_instr(cx88_buffer *buffer, uint32_t instr)
{
   /* Place the instruction in RAM, in PCI little-endian format */
   buffer->dma_program_buffer[buffer->dma_program_counter++] = htole32(instr); 
}

/* Write a 64-bit DMA target address */
static void cx88_write_addr64(cx88_buffer *buffer, bus_addr_t address)
{
   DIAG(DIAG_RISC, printf("CX88-RISC: Generating 64-bit addr 0x%lx\n", address))
   buffer->generate_instr(buffer, (uint32_t)(address & 0xFFFFFFFF));
   buffer->generate_instr(buffer, (uint32_t)((uint64_t)address >> 32));
}

/* Write a 32-bit DMA target address */
static void cx88_write_addr32(cx88_buffer *buffer, bus_addr_t address)
{
   DIAG(DIAG_RISC, printf("CX88-RISC: Generating 32-bit addr 0x%lx\n", address))
   buffer->generate_instr(buffer, (uint32_t)address);
}

/* Determine the maximum number of bytes that can be skipped/transferred with a single
 * instruction, return that number of bytes (via count) and return the line position
 * (SOL/EOL) flags required by the instruction. */
static uint8_t cx88_get_linepos(
   cx88_buffer *buffer,
   uint32_t *count,
   bus_addr_t *physaddr)
{
   uint8_t linepos = 0;
   uint32_t actualcount = (*count);
   DIAG(DIAG_RISC, printf("CX88-RISC: Getting line position for 0x%x-byte transfer\n", actualcount))
   if (buffer->sgl_index > &(buffer->sgl[buffer->sgl_size - 1]))
   {
      *count = 0;
      return (CX88_RISC_SOL | CX88_RISC_EOL);
   }
   if (physaddr != NULL)
   {
      *physaddr = buffer->sgl_index->ds_addr + buffer->sgl_offset;
   }
   if (buffer->sgl_offset == 0)
   {
      buffer->target_addr_valid = FALSE;
   }
   if ((buffer->sgl_index->ds_len - buffer->sgl_offset) < actualcount)
   {
      /* A single instruction can only skip/transfer until the end of the buffer segment */
      actualcount = buffer->sgl_index->ds_len - buffer->sgl_offset;
   }
   if (buffer->line_offset == 0)
   {
      DIAG(DIAG_RISC, printf("CX88-RISC: Start of line\n"))
      linepos |= CX88_RISC_SOL;
   }
   if ((buffer->line_offset + actualcount) >= buffer->channel->line_size)
   {
      DIAG(DIAG_RISC, printf("CX88-RISC: End of line\n"))
      linepos |= CX88_RISC_EOL;
      /* A single instruction can also only skip/transfer until the end of the current line */
      actualcount = buffer->channel->line_size - buffer->line_offset;
   }
   buffer->line_offset += actualcount;
   if (buffer->line_offset == buffer->channel->line_size)
   {
      /* If we've used up the current line, reset the line offset to indicate
       * beginning of the next line for the next skip/transfer */
      buffer->line_offset = 0;
   }
   buffer->sgl_offset += actualcount;
   if (buffer->sgl_offset == buffer->sgl_index->ds_len)
   {
      /* If we've used up the current buffer segment, reset the segment offset
       * and move to the next segment for the next skip/transfer */
      buffer->sgl_offset = 0;
      ++(buffer->sgl_index);
   }
   buffer->buffer_offset += actualcount;
   *count = actualcount;
   DIAG(DIAG_RISC, printf("CX88-RISC: Able to transfer 0x%x bytes\n", actualcount));
   return linepos;
}

static void cx88_risc_seek(cx88_buffer *buffer, bus_size_t offset)
{
   if (offset == buffer->buffer_offset)
   {
      return;
   }
   buffer->target_addr_valid = FALSE;
   bus_size_t current_offset = buffer->buffer_offset - buffer->sgl_offset;
   bus_dma_segment_t *current_seg = buffer->sgl_index;
   if (offset < current_offset)
   {
      current_offset = 0;
      current_seg = buffer->sgl;  
   }
   while (current_seg <= &(buffer->sgl[buffer->sgl_size - 1]))
   {
      if ((current_offset <= offset) && ((current_offset + current_seg->ds_len) > offset))
      {
         buffer->sgl_index = current_seg;
         buffer->sgl_offset = offset - current_offset;
         buffer->buffer_offset = offset;
         break;
      }
      current_offset += (current_seg->ds_len);
      ++current_seg;
   }
}

/* Generic single-instruction transfer */
uint32_t cx88_risc_transfer(
   cx88_buffer *buffer,
   bus_size_t offset,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   /* Determine whether to use READ(C) or WRITE(C) */
   boolean_t is_read = ((buffer->channel->descriptor->direction == CX88_UPSTREAM) ? TRUE : FALSE);
   uint32_t actualcount = count;
   cx88_risc_seek(buffer, offset);
   /* Calculate target physical address and maximum transfer count */
   bus_addr_t physaddr;
   uint8_t linepos = cx88_get_linepos(buffer, &actualcount, &physaddr);
   boolean_t is_continued = FALSE;
   if ((!(linepos & CX88_RISC_SOL)) && buffer->target_addr_valid)
   {
      /* If we're picking up at the current offset (i.e. we don't need to
       * specify a new physical address), and we're not at the start of the 
       * line or the DMA segment, we don't need to supply the physical
       * address and can use WRITEC or READC */
      is_continued = TRUE;
   }
   else
   {
      buffer->target_addr_valid = TRUE;
   }
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      /* We're actually writing the DMA program, so generate the instruction */
      cx88_risc_opcode operation;
      if (is_continued)
      {
         operation = (is_read ? CX88_RISC_READC : CX88_RISC_WRITEC);
         DIAG(DIAG_RISC, is_read ? printf("READC:") : printf("WRITEC:"))
      }
      else
      {
         operation = (is_read ? CX88_RISC_READ : CX88_RISC_WRITE);
         DIAG(DIAG_RISC, is_read ? printf("READ:") : printf("WRITE:"))
      }
      DIAG(DIAG_RISC, printf("\t0x%x bytes at offset 0x%lx\n", actualcount, (unsigned long)offset))
      uint32_t instr = cx88_format_instr(operation, intrs, countctl); 
      instr |= ((linepos & 0x3) << 26);
      instr |= (actualcount & 0xFFF);
      buffer->generate_instr(buffer, instr);
      if (!is_continued)
      {
         buffer->generate_addr(buffer, physaddr);
      }
   }
   else
   {
      /* We're sizing the program, so just bump the counter */
      buffer->dma_program_counter += 
         (is_continued ? CX88_RISC_READC_WRITEC_SIZE : CX88_RISC_READ_WRITE_SIZE(buffer->channel->sc));
   }
   return actualcount;
}

/* Sync to a specified line number */
void cx88_risc_sync(
   cx88_buffer *buffer,
   uint16_t lineno,
   cx88_risc_sync_t type,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   buffer->line_offset = 0;
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      DIAG(DIAG_RISC, printf("SYNC to line %d\n", lineno))
      uint32_t sync_instr = cx88_format_instr(CX88_RISC_SYNC, intrs, countctl);
      if (type == CX88_RISC_NO_SYNC_INTR)
      {
         sync_instr |= (1 << 15);
      }
      sync_instr |= (lineno & 0x3FF);
      buffer->generate_instr(buffer, sync_instr);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_SYNC_SIZE;
   }
}

/* Jump to arbitrary address in destination buffer's RISC program */
int cx88_risc_jump(
   cx88_buffer *src,
   uint8_t dest_index,
   uint32_t offset,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (dest_index >= src->channel->num_buffers)
   {
      return EINVAL;
   }
   cx88_buffer *dest = &(src->channel->buffers[dest_index]);
   if ((src->dma_program_state == CX88_DMAPROG_LOAD) ||
       ((dest != src) && (src->dma_program_state == CX88_DMAPROG_LINK)))
   {
      bus_addr_t target;
      if (dest->data_buffer_type == CX88_DATABUF_NONE)
      {
         return EINVAL;
      }
      if ((dest != src) && (dest_index != src->next_buffer_index))
      {
         src->needs_relink = TRUE;
      }
      /* Calculate physical target address */
      target = dest->dma_program_physaddr + (offset * sizeof(uint32_t));
      DIAG(DIAG_RISC, printf("JUMP to 0x%lx\n", (unsigned long)target))
      uint32_t jump_instr = cx88_format_instr(CX88_RISC_JUMP, intrs, countctl);
      if (dest->dma_program_buffer == NULL)
      {
         /* Specify a jump into an SRAM-resident program */
         jump_instr |= 1;
      }
      src->generate_instr(src, jump_instr);
      src->generate_addr(src, target);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      src->dma_program_counter += CX88_RISC_JUMP_SIZE(src->channel->sc);
   }
   return 0;
}

/* Skip as many bytes as possible (up to count) with single instruction */
uint32_t cx88_risc_skip(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t skipcount = 0;
   uint8_t linepos = 0; 
   while (!(linepos & CX88_RISC_EOL) && (skipcount < count))
   {
      uint32_t actualcount = count - skipcount;
      linepos = cx88_get_linepos(buffer, &actualcount, NULL);
      if (actualcount == 0)
      {
         break;
      }
      skipcount += actualcount;
   }
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      DIAG(DIAG_RISC, printf("SKIP 0x%x bytes\n", skipcount))
      uint32_t skip_instr = cx88_format_instr(CX88_RISC_SKIP, intrs, countctl);
      skip_instr |= ((linepos & 0x3) << 26);
      skip_instr |= (skipcount & 0xFFF);
      buffer->generate_instr(buffer, skip_instr);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_SKIP_SIZE;
   }
   return skipcount;
}

/* Write cx88 register to host memory */
void cx88_risc_write_reg_to_mem(
   cx88_buffer *buffer,
   uint32_t reg,
   bus_addr_t target_addr,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writerm_instr = cx88_format_instr(CX88_RISC_WRITERM, intrs, countctl);
      writerm_instr |= 1;
      buffer->generate_instr(buffer, writerm_instr);
      buffer->generate_addr(buffer, target_addr);
      buffer->generate_instr(buffer, (reg & 0xFFFFFF));
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITERM_SIZE(buffer->channel->sc);
   }
}

/* Write constant to host memory */
void cx88_risc_write_const_to_mem(
   cx88_buffer *buffer,
   uint32_t constant,
   bus_addr_t target_addr,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writecm_instr = cx88_format_instr(CX88_RISC_WRITECM, intrs, countctl);
      buffer->generate_instr(buffer, writecm_instr);
      buffer->generate_addr(buffer, target_addr);
      buffer->generate_instr(buffer, constant);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITECM_SIZE(buffer->channel->sc);
   }
}

/* Write constant to cx88 register */
void cx88_risc_write_const_to_reg(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t mask,
   uint32_t reg,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writecr_instr = cx88_format_instr(CX88_RISC_WRITECR, intrs, countctl);
      writecr_instr |= 1;
      buffer->generate_instr(buffer, writecr_instr);
      buffer->generate_instr(buffer, (reg & 0xFFFFFF));
      buffer->generate_instr(buffer, constant);
      buffer->generate_instr(buffer, mask);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITECR_SIZE;
   }
}

/* Transfer arbitrary number of bytes */
int cx88_risc_transfer_dmabuf(
   cx88_buffer *buffer,
   bus_size_t offset,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t bytes_transferred = 0;
   uint32_t current_transfer;
   bus_size_t current_offset = offset;
   if ((offset + count) > (buffer->data_buffer_size))
   {
      return EINVAL;
   }
   while (bytes_transferred < count)
   {
      uint32_t bytes_remaining = count - bytes_transferred;
      current_transfer = cx88_risc_transfer (buffer, current_offset, bytes_remaining, intrs, countctl);
      bytes_transferred += current_transfer;
      current_offset += current_transfer;
   }
   return 0;
}

/* Skip arbitrary number of bytes */
int cx88_risc_skip_dmabuf(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t bytes_skipped = 0;
   if ((buffer->buffer_offset + count) > (buffer->data_buffer_size))
   {
      return EINVAL;
   }
   while (bytes_skipped < count)
   {
      uint32_t bytes_remaining = count - bytes_skipped;
      bytes_skipped += cx88_risc_skip (buffer, bytes_remaining, intrs, countctl);
   }
   return 0;
}
