/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
 
 /* File: cx88_common.h
  * Purpose: Header file for core cx88 channel, buffer, and driver management
  * functions
  */

#ifndef __CX88_COMMON_H__
#define __CX88_COMMON_H__

#include <sys/types.h>
#include <sys/param.h>
#include <sys/bus.h>
#include <machine/bus.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/rman.h>
#include <dev/pci/pcivar.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include "cx88_common_ioctl.h"

#define DIAG_GENERAL     (1 << 0)
#define DIAG_DEVICE      (1 << 1)
#define DIAG_CHANNEL     (1 << 2)
#define DIAG_BUFFER      (1 << 3)
#define DIAG_SRAM        (1 << 4)
#define DIAG_RISC        (1 << 5)
#define DIAG_MMAP        (1 << 6)
#if (defined(_DIAGNOSTIC) && (_DIAGNOSTIC != 0))
#define DIAG(level, stmt)     \
   if (_DIAGNOSTIC & (level)) \
   {                          \
      stmt;                   \
   }
#define DIAGNOSTIC(stmt) DIAG(DIAG_GENERAL, stmt)
#else
#define DIAG(level, stmt)
#define DIAGNOSTIC(stmt)
#endif

#define CX88_MINVER 12
#define CX88_MODVER 12
#define CX88_MAXVER 12

#define CX88_VID                   0x14f1
#define CX23880_BASE_PID           0x8800
#define CX23885_PID                0x8852
#define CX23887_PID                0x8880

MALLOC_DECLARE(M_CX88);

typedef uint16_t cx88_size_t; /* cx88 SRAM size */

/* Forward declarations */
typedef struct _cx88_common_softc cx88_common_softc;

typedef struct _cx88_buffer cx88_buffer;

typedef struct _cx88_channel cx88_channel;

typedef enum
{
   CX88_BRIDGE_CX23880,
   CX88_BRIDGE_CX23885,
   CX88_BRIDGE_CX23887
} cx88_bridge_t;

/* cx88 PCI functions */
typedef enum
{
   CX88_FUNC_VIDEO = 0,
   CX88_FUNC_AUDIO,
   CX88_FUNC_MPEGTS,
   CX88_FUNC_VIP,
   CX88_FUNC_GPHOST,
   CX88_NUM_FUNCTIONS
} cx88_function_t;

#define CX88_MAX_SIBLINGS (CX88_NUM_FUNCTIONS - 1)

/* Defines vital statistics of a client driver */
typedef struct
{
   size_t softc_size; /* total size of the client's softc */
   cx88_bridge_t bridge;
   cx88_function_t function; /* PCI function */
   int (*setup)(cx88_common_softc*); /* initialization callback for client's softc */
   int (*teardown)(cx88_common_softc*); /* cleanup callback for client's softc */
   driver_filter_t *interrupt_filter; /* client's interrupt filter */
   driver_intr_t *interrupt_handler; /* client's interrupt handler */
   const char *description; /* description of client */
} cx88_client_driver;

/* cx88 DMA channel numbers.  The cx88 datasheet numbers these
 * in the same order, but starting at 21 instead of 0 */
typedef enum
{
   CX88_CH_VIDEO_PACKED = 0,
   CX88_CH_VIDEO_U,
   CX88_CH_VIDEO_V,
   CX88_CH_VIDEO_VBI,
   CX88_CH_AUDIO_IN,
   CX88_CH_AUDIO_OUT,
   CX88_CH_MPEG_TS,
   CX88_CH_VIP_IN,
   CX88_CH_VIP_OUT,
   CX88_CH_GPHOST_IN,
   CX88_CH_GPHOST_OUT,
   CX88_CH_MPEG_TS_2,
   CX88_CH_AUDIO_IN_2,
   CX88_CH_AUDIO_IN_3,
   CX88_CH_VIDEO_PACKED_2,
   CX88_CH_VIDEO_VBI_2
} cx88_channel_t;

/* Vital data about a DMA channel */
typedef struct
{
   cx88_channel_t             channel;
   uint8_t                    dma_number;
   enum
   {
      CX88_UPSTREAM,
      CX88_DOWNSTREAM
   }                          direction;
   cx88_size_t                min_fifo_size;
   cx88_size_t                min_ctrl_size;
} cx88_channel_desc;

/* Data structure to be shared by all cx88 client drivers for
 * a particular card.  Allocated by first client driver to attach
 * and freed by last client to detach */
typedef struct
{
   cx88_bridge_t             bridge;
   struct mtx                lock;
   uint32_t                  cmds_offset;
   cx88_size_t               cmds_entry_size;
   cx88_size_t               cdt_entry_size;
   cx88_size_t               min_instrq_size;
   cx88_size_t               max_instrq_size;
   cx88_size_t               max_num_fifos;
   struct rman               fifo_sram;
   struct rman               control_sram;
   const cx88_channel_desc  *channel_table;
   cx88_size_t               table_size;
   uint32_t                  ptr1_base;
   uint32_t                  ptr2_base;
   uint32_t                  cnt1_base;
   uint32_t                  cnt2_base;
   uint32_t                  present_channels;
   uint16_t                  present_dma_channels;
   uint8_t                   addr_dwords;
} cx88_board_data;

/* Common softc structure. This MUST be the first member of each client's softc */
struct _cx88_common_softc
{
   device_t                  dev;
   device_t                  siblings[CX88_MAX_SIBLINGS];
   uint8_t                   num_siblings;
   bus_space_tag_t           mem_tag; 
   bus_space_handle_t        mem_handle;
   struct resource          *mem_res;
   int                       mem_rid;
   struct resource          *irq_res;
   int                       irq_rid;
   void                     *irq_cookie;
   const cx88_client_driver *client;
   cx88_board_data          *board_data;
   boolean_t                 msi;
   boolean_t                 is_configured;
};

/* Client drivers should provide functions for generating the 
 * DMA RISC program */
typedef int (*cx88_risc_generator) (cx88_buffer*, uint8_t);

typedef void (*cx88_instr_generator) (cx88_buffer*, uint32_t);
typedef void (*cx88_addr_generator) (cx88_buffer*, bus_addr_t);

/* Determines which (if any) RISC interrupts are triggered when
 * a RISC instruction is executed */
typedef enum
{
   CX88_NO_TRIGGER = 0x0,
   CX88_TRIGGER_RISCI1, /* RISC interrupt 1 */
   CX88_TRIGGER_RISCI2, /* RISC interrupt 2 */
   CX88_TRIGGER_RISCI1_RISCI2 /* both RISC interrupts */
} cx88_risci_trigger;

/* Determines how the internal counter should be dealt with
 * when a RISC instruction is executed */
typedef enum
{
   CX88_IGNORE_COUNT = 0x0, /* No change */
   CX88_INCREMENT_COUNT = 0x1, /* Increment */
   CX88_RESET_COUNT = 0x3 /* Reset to 0 */
} cx88_count_control;

/* Size of each instruction (in 4-byte words) */
#define CX88_RISC_SYNC_SIZE           1
#define CX88_RISC_READC_WRITEC_SIZE   1
#define CX88_RISC_READ_WRITE_SIZE(sc) (1 + sc->board_data->addr_dwords)
#define CX88_RISC_JUMP_SIZE(sc)       (1 + sc->board_data->addr_dwords)
#define CX88_RISC_SKIP_SIZE           1
#define CX88_RISC_WRITERM_SIZE(sc)    (2 + sc->board_data->addr_dwords)
#define CX88_RISC_WRITECM_SIZE(sc)    (2 + sc->board_data->addr_dwords)
#define CX88_RISC_WRITECR_SIZE        4

typedef enum
{
   CX88_DMAMAP_NULL,
   CX88_DMAMAP_TAGGED,
   CX88_DMAMAP_CREATED,
   CX88_DMAMAP_LOADED
} cx88_dmamap_state_t;

/* Describes a data buffer and its DMA program. Each channel
 * may contain multiple buffers */
struct _cx88_buffer
{
   cx88_channel              *channel;
   enum
   {
      CX88_DMAPROG_SIZE,
      CX88_DMAPROG_LOAD,
      CX88_DMAPROG_LINK,
      CX88_DMAPROG_DONE
   }                          dma_program_state;
   uint32_t                  *dma_program_buffer;
   bus_dma_tag_t              dma_program_tag;
   bus_dmamap_t               dma_program_map;
   cx88_dmamap_state_t        dma_program_map_state;
   bus_addr_t                 dma_program_physaddr;
   uint32_t                   dma_program_size;
   uint32_t                   dma_program_counter;
   enum
   {
      CX88_DATABUF_NONE,
      CX88_DATABUF_MALLOC,
      CX88_DATABUF_KERNEL,
      CX88_DATABUF_USER
   }                          data_buffer_type;
   struct proc               *owner;
   void                      *data_buffer;
   bus_dmamap_t               data_buffer_map;
   cx88_dmamap_state_t        data_buffer_map_state;
   bus_dma_segment_t         *sgl;
   int                        sgl_size;
   bus_dma_segment_t         *sgl_index;
   bus_size_t                 sgl_offset;
   bus_size_t                 buffer_offset;
   bus_size_t                 data_buffer_size;
   cx88_risc_generator        generate_dma_program;
   cx88_instr_generator       generate_instr;
   cx88_addr_generator        generate_addr;
   cx88_size_t                line_offset;
   uint8_t                    next_buffer_index;
   boolean_t                  needs_relink;
   boolean_t                  target_addr_valid;
   boolean_t                  ready_for_transfer;
};

/* Describes a DMA channel, its SRAM layout, and data buffers.
 * Client drivers are responsible for storing as many of these
 * structures as necessary in their softc structures.
 * All SRAM offsets are based on the beginning of the cx88 memory
 * space, not the beginning of the SRAM in that memory space */
struct _cx88_channel
{
   const cx88_channel_desc   *descriptor;
   cx88_common_softc         *sc;
   struct resource           *fifo_sram;
   struct resource           *control_sram;
   uint32_t                   instrq_offset;
   uint32_t                   cdt_offset;
   uint32_t                   fifo_offset;
   cx88_size_t                packet_size;
   cx88_size_t                line_size;
   cx88_size_t                instrq_size;
   cx88_size_t                fifo_size;
   cx88_size_t                num_fifos;
   cx88_buffer               *buffers;
   uint8_t                    num_buffers;
   uint8_t                    start_buffer;
   boolean_t                  is_dirty;
   bus_dma_tag_t              data_buffer_tag;
   uint32_t                   max_buffer_size;
   uint32_t                   line_boundary; /* Number of packets, not bytes */
   cx88_size_t                max_line_size;
};

/* Probe function to be called by clients' probe DEVMETHODs.
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_probe(device_t dev, const cx88_client_driver *client);

/* Attach function to be called by clients' attach DEVMETHODs 
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_attach(device_t dev, const cx88_client_driver *client);

/* Resume function for power management */
int cx88_common_resume(device_t dev);

/* Detach function to be called by clients' detach DEVMETHODs
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_detach(device_t dev);

/* mmap function to be called by clients' d_mmap
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_mmap(
   cx88_channel *channel, /* Channel whose buffers should be mapped */
   vm_ooffset_t offset,
   vm_paddr_t *paddr,
   int nprot);

typedef enum
{
   CX88_RISC_REPROGRAM_IF_BUFFER_CHANGED, /* Re-generate DMA program only if buffer attributes change */
   CX88_RISC_REPROGRAM_FORCE /* Force DMA program to be re-generated (program has changed independently of buffer) */
} cx88_risc_reprogram_t;

/* Clients should call this to configure a buffer in a channel's 
 * buffer array.  Returns 0 on success, appropriate error otherwise. */
int cx88_configure_buffer(
   cx88_buffer *buffer,
   void *data_buffer, /* Buffer to store the data. If NULL, buffer will be allocated */
   struct thread *context, /* VA space context in which buffer is allocated (NULL=>kernel) */
   size_t buffer_size,
   uint8_t next_buffer_index,
   cx88_risc_generator dma_program_generator, /* Generates this buffer's RISC program */
   cx88_risc_reprogram_t reprogram);

/* Frees all resources associated with a buffer */
void cx88_release_buffer(cx88_buffer *buffer);

/* Allocates an initializes buffers for a channel.  Each buffer should
 * then be configured with cx88_configure_buffer(). Returns 0 on success, appropriate
 * error otherwise. */
int cx88_request_buffers(
   cx88_channel *channel,
   uint8_t num_buffers,
   uint8_t start_buffer); /* Index of buffer whose RISC program will run first when DMA is started */
   
/* Free user-allocated buffers for a channel, if they belong to the
 * specified process.  Pass NULL for owner to free all user-allocated
 * buffers. */
void cx88_free_userbufs(cx88_channel *channel, struct proc *owner);

/* Frees the channel's SRAM allocation */
void cx88_free_sram(cx88_channel *channel);

/* Configures a DMA channel. Returns 0 on success, appropriate error otherwise. */
int cx88_configure_channel(
   cx88_channel *channel,
   cx88_size_t min_sram_size, /* Minimum SRAM chunk size to be allocated (0 -> automatic allocation) */
   cx88_size_t packet_size, /* Packet size (or line size) for channel */
   cx88_size_t max_line_size, /* Maximum allowable line size (in bytes).  Must be equal to packet_size for line-based channels */
   uint32_t line_boundary); /* Lines in a buffer must not cross multiple of this no. of packets (0 -> no boundary) */
   
/* Frees all resources associated with a chanel, including buffers */
void cx88_release_channel(cx88_channel *channel);

/* Once a channel and its buffers have been configured, generate the DMA programs and
 * write the channel data to the cx88 registers and CMDS. Returns 0 on success, appropriate
 * error otherwise. */
int cx88_commit_channel(cx88_channel *channel);

/* Prepare a channel for starting/restarting a DMA transfer */
int cx88_start_channel(cx88_channel *channel);

/* Initialize a channel structure in client's softc.  Most likely called
 * from the client's softc setup function during driver attachment.  Returns
 * 0 on success, appropriate error otherwise. */
int cx88_init_channel(
   cx88_channel *channel,
   cx88_common_softc *sc,
   cx88_channel_t channel_number,
   uint32_t max_buffer_size); /* Maximum allowable data buffer size */
   
/* De-initializes a channel structure.  Most likely called from client's softc
 * teardown function during driver detachment.  Will free any channel and buffer
 * resources */
void cx88_destroy_channel(cx88_channel *channel);

/* Synchronizes data buffer memory prior to transfer.
 * Returns 0 on success, appropriate error otherwise. */
int cx88_prepare_for_transfer(cx88_buffer *buffer);

/* Synchronizes data buffer memory following transfer */
int cx88_mark_transfer_complete(cx88_buffer *buffer);

/* Accessor for this device's board data structure.  Returns NULL
 * if device has not allocated board data or retrieved it from a sibling. */
cx88_board_data *cx88_common_get_board_data(device_t dev);

/* Defines an accessor function for generic shared data allocated by sibing */
typedef void *(*get_board_data_func) (device_t);

/* Adds a data structure to be shared by sibling devices.  Will attempt to
 * retrieve the structure from siblings using the provided accessor function,
 * and will allocate the structure if all siblings return NULL.
 * In spite of the name, this (along w/ cx88_remove_board_data()) is
 * designed to support not just the cx88_board_data structure used by the
 * common layer, but any arbitrary data a driver may want to share with its
 * siblings.
 * Returns 0 on success, appropriate error otherwise. */
int cx88_add_board_data(
   cx88_common_softc *sc,
   void **data,
   size_t data_size,
   get_board_data_func accessor,
   int (*setup)(void*)); /* Initialization function to be called if the shared data needs to be allocated */
   
/* Removes this driver's reference to a shared data structure.  If this driver is the last
 * driver to reference the structure, it will be freed.  Returns 0 on success, appropriate
 * error otherwise. */
int cx88_remove_board_data(
   cx88_common_softc *sc,
   void **data,
   int (*teardown)(void*)); /* Will be called if the structure is to be freed */
   
/* Accessor for client softc.  If the softc is NULL or uninitialized (client's softc
 * setup callback has not been called), will return NULL */
void *cx88_get_softc(device_t dev);

/* Write/read GPIO pins */
void cx88_gpioctl(cx88_common_softc *sc, cx88_common_gpio *gpio);

/* Write microcontroller register */
int cx23885_mc_write_reg(cx88_common_softc *sc, uint16_t address, uint32_t value);

/* Read microcontroller register */
int cx23885_mc_read_reg(cx88_common_softc *sc, uint16_t address, uint32_t *value);

/* Write microcontroller memory */
int cx23885_mc_write_mem(cx88_common_softc *sc, uint32_t address, uint32_t value);

/* Read microcontroller memory */
int cx23885_mc_read_mem(cx88_common_softc *sc, uint32_t address, uint32_t *value);

/* Write/read microcontroller GPIO lines */
int cx23885_mc_gpioctl(cx88_common_softc *sc, cx88_common_gpio *gpio);

/* Control SRSTO pin which resets peripheral devices */
void cx88_peripheral_reset(cx88_common_softc *sc, uint32_t reset_val);

/* Will transfer as many bytes as possible with a single RISC
 * instruction, up to a masximum of count. Uses buffer's channel 
 * direction and line/DMA segment offsets to determine whether to
 * use READ/READC/WRITE/WRITEC, and returns the actual number of 
 * bytes that can be transferred. */
uint32_t cx88_risc_transfer(
   cx88_buffer *buffer,
   bus_size_t offset, /* Byte offset in buffer at which to begin transfer */
   uint32_t count, /* Maximum number of bytes */
   cx88_risci_trigger intrs, /* Applies to each generated instruction */
   cx88_count_control countctl); /* Applies to each generated instruction */

typedef enum
{
   CX88_RISC_NO_SYNC_INTR,
   CX88_RISC_SYNC_INTR
} cx88_risc_sync_t;

/* Generates a SYNC instruction */
void cx88_risc_sync(
   cx88_buffer *buffer,
   uint16_t lineno,
   cx88_risc_sync_t type,
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Ignores as many bytes as possible with a a single SKIP instruction,
 * up to a maximum of count bytes.  Returns the actual number of bytes that
 * can be skipped. */
uint32_t cx88_risc_skip(
   cx88_buffer *buffer,
   uint32_t count, /* Maximum number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);
   
/* Jump to arbitrary address in destination buffer's RISC program */  
int cx88_risc_jump(
   cx88_buffer *src, /* source buffer */
   uint8_t dest_index, /* destination buffer index (may point to same buffer as src) */
   uint32_t offset, /* offset into destination buffer's DMA program */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);
   
/* Jump to next buffer */
static inline int cx88_risc_next_buffer(
   cx88_buffer *buffer,
   uint32_t offset,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   return cx88_risc_jump(buffer, buffer->next_buffer_index, offset, intrs, countctl);
}

/* Write cx88 register contents to memory */
void cx88_risc_write_reg_to_mem(
   cx88_buffer *buffer,
   uint32_t reg, /* Offset into cx88 memory window */
   bus_addr_t target_addr, /* Target physical address */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Write constant to memory */
void cx88_risc_write_const_to_mem(
   cx88_buffer *buffer,
   uint32_t constant,
   bus_addr_t target_addr, /* Target physical address */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Write constant to cx88 register */
void cx88_risc_write_const_to_reg(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t mask, /* Bitmask to apply to constant */
   uint32_t reg, /* Offset into cx88 memory window */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Generic function for transferring an arbitrary number of bytes.
 * Calls cx88_risc_transfer() as many times as necessary to generate the
 * instructions to transfer count bytes */
int cx88_risc_transfer_dmabuf (
   cx88_buffer *buffer,
   bus_size_t offset, /* Byte offset in buffer at which to begin transfer */
   uint32_t count, /* Number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);
   
/* Generic function for ignoring an arbitrary number of bytes.
 * Calls cx88_risc_skip() as many times as necessary to generate the 
 * required SKIP instructions. */
int cx88_risc_skip_dmabuf(
   cx88_buffer *buffer,
   uint32_t count, /* Number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

static inline uint32_t cx88_risc_offset(cx88_buffer *buffer)
{
   return (buffer->dma_program_counter * sizeof (*(buffer->dma_program_buffer)));
}

/* Accesses card vendor data in PCI subsystem ID */
static inline uint32_t cx88_get_card(device_t dev)
{
   return (((uint32_t)(pci_get_subvendor(dev)) << 16) | pci_get_subdevice(dev));
}

/* Writes a cx88 register */
static inline void cx88_write_reg(
   cx88_common_softc *sc,
   uint32_t offset, /* Offset in cx88 memory window */
   uint32_t value)
{
   bus_space_write_4(sc->mem_tag, sc->mem_handle, offset, value);
}

/* Writes a contiguous series of registers */
static inline void cx88_write_region(
   cx88_common_softc *sc,
   uint32_t offset,
   const uint32_t *values,
   uint32_t count)
{
   bus_space_write_region_4(sc->mem_tag, sc->mem_handle, offset, values, count);
}

/* Reads a contiguous series of registers */
static inline void cx88_read_region(
   cx88_common_softc *sc,
   uint32_t offset,
   uint32_t *values,
   uint32_t count)
{
   bus_space_read_region_4(sc->mem_tag, sc->mem_handle, offset, values, count);  
}

/* Reads a cx88 register */
static inline uint32_t cx88_read_reg(
   cx88_common_softc *sc,
   uint32_t offset) /* Offset in cx88 memory window */
{
   return bus_space_read_4(sc->mem_tag, sc->mem_handle, offset);    
}

/* Generates a bitfield */
static inline uint32_t cx88_write_field(
   uint32_t mask,
   uint8_t shift,
   uint32_t value)
{
   return ((value & mask) << shift);
}

/* Extracts a bitfield */
static inline uint32_t cx88_read_field(
   uint32_t value,
   uint32_t mask,
   uint8_t shift)
{
   return ((value >> shift) & mask);
}

static inline uint32_t cx88_update_field(
   uint32_t reg,
   uint32_t mask,
   uint32_t shift,
   uint32_t value)
{
   return ((reg & (~(mask << shift))) | cx88_write_field(mask, shift, value));
}

typedef struct
{
   uint32_t offset;
   uint32_t value;
}
cx88_reg;

static inline void cx88_write_regs(
   cx88_common_softc *sc,
   const cx88_reg *regs,
   uint32_t num_regs)
{
   for (uint32_t i = 0; i < num_regs; ++i)
   {
      cx88_write_reg(sc, regs[i].offset, regs[i].value);
   }
}

static inline void cx88_read_regs(
   cx88_common_softc *sc,
   cx88_reg *regs,
   uint32_t num_regs)
{
   for (uint32_t i = 0; i < num_regs; ++i)
   {
      regs[i].value = cx88_read_reg(sc, regs[i].offset);
   }
}

/* Acquires exclusive access to data that may be shared by sibling devices */
static inline void cx88_lock_board(cx88_common_softc *sc)
{
   mtx_lock(&(sc->board_data->lock));  
}

/* Releases exclusive access to data that may be shared by sibling devices */
static inline void cx88_unlock_board(cx88_common_softc *sc)
{
   mtx_unlock(&(sc->board_data->lock));
}

static inline void cx88_lock_device_tree(cx88_common_softc *sc)
{
   if (sc->is_configured)
   {
   //#if __FreeBSD_version >= 800500
   //   newbus_xlock();
   //#else
      mtx_lock(&Giant);
   //#endif
   }
}

static inline void cx88_unlock_device_tree(cx88_common_softc *sc)
{
   if (sc->is_configured)
   {
   //#if __FreeBSD_version >= 800500
   //   newbus_xunlock();
   //#else
      mtx_unlock(&Giant);
   //#endif
   }
}

/* DEVMETHODs all client drivers should implement */
#define CX88_METHODS(reconfig_request) \
   DEVMETHOD(cx88_get_board_data,      cx88_common_get_board_data), \
   DEVMETHOD(cx88_request_reconfig,    reconfig_request)

/* All client drivers should declare dependency on this module */
#define CX88_DRIVER_MODULE(host) \
   MODULE_DEPEND(host, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);

#endif
