/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_audio.h"

typedef struct
{
   cx88_common_softc common;
   cx88_i2c_softc i2c;
   cx88_audio_softc audio;
   cx88_audio_format format;
   cx88_audio_volume volume;
   uint8_t volume_reg;
   int8_t balance_reg;
} cx23880_audio_softc;

static int cx23880_audio_probe(device_t dev);
static int cx23880_audio_attach(device_t dev);
static cx88_i2c_softc *cx23880_audio_i2c(cx88_common_softc *sc, uint8_t bus);
static int cx23880_audio_setup(cx88_common_softc *arg);
static int cx23880_audio_teardown(cx88_common_softc *arg);
static void cx23880_audio_stop_capture(device_t dev, uint8_t path);
static int cx23880_audio_fm_radio(cx23880_audio_softc *sc, uint32_t *ctl_reg);
static int cx23880_audio_btsc(cx23880_audio_softc *sc, uint32_t *ctl_reg);
static int cx23880_audio_eiaj(cx23880_audio_softc *sc, uint32_t *ctl_reg);
static int cx23880_audio_a2(cx23880_audio_softc *sc, uint32_t *ctl_reg);
static int cx23880_audio_nicam(cx23880_audio_softc *sc, uint32_t *ctl_reg);
static int cx23880_audio_load_format(cx23880_audio_softc *sc);
static int cx23880_audio_start_capture(device_t dev, uint8_t path);
static int cx23880_audio_set_format(device_t dev, uint8_t path, cx88_audio_format *fmt);
static int cx23880_audio_set_volume(device_t dev, uint8_t path, cx88_audio_volume vol);
static cx88_audio_volume cx23880_audio_get_volume(device_t dev, uint8_t path);
static int cx23880_audio_set_stereo(device_t dev, uint8_t path, boolean_t stereo);
static boolean_t cx23880_audio_get_stereo(device_t dev, uint8_t path);
static int cx23880_audio_set_mute(device_t dev, uint8_t path, boolean_t mute);
static boolean_t cx23880_audio_get_mute(device_t dev, uint8_t path);
static int cx23880_audio_get_info(device_t dev, uint8_t path, cx88_audio_status *status);
static int cx23880_audio_filter(void *arg);
static void cx23880_audio_intr(void *arg);
static cx88_i2c_bus *cx23880_audio_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx23880_audio_get_i2c_dev(device_t dev, uint8_t addr);
static int cx23880_audio_request_reconfig(device_t dev);
static int cx23880_audio_suspend(device_t dev);
static int cx23880_audio_shutdown(device_t dev);

static const cx88_client_driver cx23880_audio_client = 
{
   .softc_size               = sizeof(cx23880_audio_softc),
   .bridge                   = CX88_BRIDGE_CX23880,
   .function                 = CX88_FUNC_AUDIO,
   .setup                    = cx23880_audio_setup,
   .teardown                 = cx23880_audio_teardown,
	.interrupt_filter         = cx23880_audio_filter,
   .interrupt_handler        = cx23880_audio_intr,
   .description              = "Conexant CX23880/1/2/3 Analog Audio"
};

typedef int (*broadcast_format_func) (cx23880_audio_softc*, uint32_t*);

typedef struct
{
   uint32_t init_reg;
   broadcast_format_func set_format;
} cx23880_audio_broadcast_format;

static cx23880_audio_broadcast_format broadcast_formats [] = 
{
   /* init_reg             set_format */
   {AUDIO_INIT_FM_RADIO,   cx23880_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO */
   {AUDIO_INIT_FM_RADIO,   cx23880_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO_50 */
   {AUDIO_INIT_FM_RADIO,   cx23880_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO_75 */
   {AUDIO_INIT_BTSC,       cx23880_audio_fm_radio}, /* CX88_AUDIO_FMT_BTSC_FM */
   {AUDIO_INIT_BTSC,       cx23880_audio_btsc},     /* CX88_AUDIO_FMT_BTSC */
   {AUDIO_INIT_EIAJ,       cx23880_audio_eiaj},     /* CX88_AUDIO_FMT_EIAJ */
   {AUDIO_INIT_A2,         cx23880_audio_a2},       /* CX88_AUDIO_FMT_AM_L */
   {AUDIO_INIT_A2,         cx23880_audio_a2},       /* CX88_AUDIO_FMT_A2_BG */
   {AUDIO_INIT_A2,         cx23880_audio_a2},       /* CX88_AUDIO_FMT_A2_DK */
   {AUDIO_INIT_A2,         cx23880_audio_a2},       /* CX88_AUDIO_FMT_A1_I */
   {AUDIO_INIT_NICAM,      cx23880_audio_nicam},    /* CX88_AUDIO_FMT_NICAM_BGDK */
   {AUDIO_INIT_NICAM,      cx23880_audio_nicam},    /* CX88_AUDIO_FMT_NICAM_I */
   {AUDIO_INIT_NICAM,      cx23880_audio_nicam}     /* CX88_AUDIO_FMT_NICAM_L */
};

static cx88_i2c_softc *cx23880_audio_i2c(cx88_common_softc *sc, uint8_t bus)
{
   if (bus != 0)
   {
      return NULL;
   }
   return &(((cx23880_audio_softc*)sc)->i2c);
}

static int cx23880_audio_setup(cx88_common_softc *arg)
{
   cx23880_audio_softc *sc = (cx23880_audio_softc*)arg;
   int error = cx88_i2c_attach(&(sc->i2c), 0, &(sc->common), sc->common.dev);
   if (!error)
   {
      error = cx88_audio_setup(&(sc->audio), CX88_AUDIO_STREAM_CX23880, &(sc->common), 
         cx23880_audio_i2c, sc->common.dev, 0);
      if (error)
      {
         cx88_i2c_detach(&(sc->i2c));
      }
   }
   return error;
}

static int cx23880_audio_teardown(cx88_common_softc *arg)
{
   cx23880_audio_softc *sc = (cx23880_audio_softc*)arg;
   int error = cx88_audio_teardown(&(sc->audio));
   if (!error)
   {
      error = cx88_i2c_detach(&(sc->i2c));
   }
   return error;
}

static void cx23880_audio_stop_capture(device_t dev, uint8_t path)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   /* Mute all audio */
   cx88_write_reg(&(sc->common), 
      AUDIO_REG_VOL_CTL, VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   /* Disable chroma ADC gain */
   cx88_write_reg(&(sc->common), AUDIO_REG_AFE_12DB, 0);
}

static int cx23880_audio_fm_radio(cx23880_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   switch (sc->format.in_stereo_format)
   {
      case CX88_AUDIO_STEREO_OFF:
         src_sel = AUD_SRC_FM_RADIO_FORCE_MONO;
         break;
      case CX88_AUDIO_STEREO_ON:
         src_sel = AUD_SRC_FM_RADIO_FORCE_STEREO;
         break;
      case CX88_AUDIO_STEREO_AUTO:
         src_sel = AUD_SRC_FM_RADIO_AUTO_STEREO;
         break;
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   static const cx88_reg regs_fm_radio_75[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x0000091B},
      {AUDIO_REG_DEEMPH0_A0, 0x00006B68},
      {AUDIO_REG_DEEMPH0_B0, 0x000011EC},
      {AUDIO_REG_DEEMPH1_G0, 0x00000AA0},
      {AUDIO_REG_DEEMPH1_A0, 0x00006B68},
      {AUDIO_REG_DEEMPH1_B0, 0x000011EC},
   };
   static const cx88_reg regs_fm_radio_50[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x00000C45},
      {AUDIO_REG_DEEMPH0_A0, 0x00006262},
      {AUDIO_REG_DEEMPH0_B0, 0x00001C29},
      {AUDIO_REG_DEEMPH1_G0, 0x00000D80},
      {AUDIO_REG_DEEMPH1_A0, 0x00006262},
      {AUDIO_REG_DEEMPH1_B0, 0x00001C29}
   };
   if ((sc->format.in_audio_format == CX88_AUDIO_FMT_BTSC_FM) ||
       (sc->format.in_audio_format == CX88_AUDIO_FMT_FM_RADIO_75))
   {
      cx88_write_regs(&(sc->common), regs_fm_radio_75, 
         sizeof(regs_fm_radio_75) / sizeof(cx88_reg));
   }
   else if (sc->format.in_audio_format == CX88_AUDIO_FMT_FM_RADIO_50)
   {
      cx88_write_regs(&(sc->common), regs_fm_radio_50,
         sizeof(regs_fm_radio_50) / sizeof(cx88_reg));  
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_DEEMPH1_A1, 0x0003FC66);
   cx88_write_reg(&(sc->common), AUDIO_REG_DEEMPH1_B1, 0x0000399A);
   cx88_write_reg(&(sc->common), AUDIO_REG_POLYPH80SCALEFAC, 3);
   return 0;
}

static int cx23880_audio_btsc(cx23880_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   if (sc->format.in_second_lang)
   {
      src_sel = AUD_SRC_BTSC_FORCE_SAP;
   }
   else
   {
      switch (sc->format.in_stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_BTSC_FORCE_MONO;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_BTSC_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_BTSC_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg |= AUD_CTL_FMRADIO_RDS;
   static const cx88_reg regs_btsc[] =
   {
      {AUDIO_REG_OUT1_SEL, 0X00000013},
      {AUDIO_REG_OUT1_SHIFT, 0X00000000},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0X0012010C},
      {AUDIO_REG_DMD_RA_DDS, 0X00C3E7AA},
      {AUDIO_REG_DBX_IN_GAIN, 0X00004734},
      {AUDIO_REG_DBX_WBE_GAIN, 0X00004640},
      {AUDIO_REG_DBX_SE_GAIN, 0X00008D31},
      {AUDIO_REG_DCOC_0_SRC, 0X0000001A},
      {AUDIO_REG_IIR1_4_SEL, 0x00000021},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DN0_FREQ, 0x0000283B},
      {AUDIO_REG_DN2_SRC_SEL, 0x00000008},
      {AUDIO_REG_DN2_FREQ, 0x00003000},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DN2_SHFT, 0x00000000},
      {AUDIO_REG_IIR2_2_SEL, 0x00000020},
      {AUDIO_REG_IIR2_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR2_3_SEL, 0x0000001F},
      {AUDIO_REG_IIR2_3_SHIFT, 0x00000000},
      {AUDIO_REG_CRDC1_SRC_SEL, 0x000003CE},
      {AUDIO_REG_CRDC1_SHIFT, 0x00000000},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_RDSI_SEL, 0X00000008},
      {AUDIO_REG_RDSQ_SEL, 0X00000008},
      {AUDIO_REG_RDSI_SHIFT, 0X00000000},
      {AUDIO_REG_RDSQ_SHIFT, 0X00000000},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003}
   };
   static const cx88_reg regs_btsc_sap[] = 
   {
      {AUDIO_REG_DBX_IN_GAIN, 0x00007200},
      {AUDIO_REG_DBX_WBE_GAIN, 0x00006200},
      {AUDIO_REG_DBX_SE_GAIN, 0x00006200},
      {AUDIO_REG_IIR1_1_SEL, 0x00000000},
      {AUDIO_REG_IIR1_3_SEL, 0x00000001},
      {AUDIO_REG_DN1_SRC_SEL, 0x00000007},
      {AUDIO_REG_IIR1_4_SHIFT, 0x00000006},
      {AUDIO_REG_IIR2_1_SHIFT, 0x00000000},
      {AUDIO_REG_IIR2_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_1_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SEL, 0x0000000D},
      {AUDIO_REG_IIR3_1_SEL, 0x0000000E},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x00000014},
      {AUDIO_REG_DEEMPH1_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH1_G0, 0x00004000},
      {AUDIO_REG_DEEMPH1_A0, 0x00000000},
      {AUDIO_REG_DEEMPH1_B0, 0x00000000},
      {AUDIO_REG_DEEMPH1_A1, 0x00000000},
      {AUDIO_REG_DEEMPH1_B1, 0x00000000},
      {AUDIO_REG_OUT0_SEL, 0x0000003F},
      {AUDIO_REG_OUT1_SEL, 0x0000003F},
      {AUDIO_REG_DN1_AFC, 0x00000002},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR1_0_SEL, 0x0000001D},
      {AUDIO_REG_IIR1_2_SEL, 0x0000001E},
      {AUDIO_REG_IIR2_1_SEL, 0x00000002},
      {AUDIO_REG_IIR2_2_SEL, 0x00000004},
      {AUDIO_REG_IIR3_2_SEL, 0x0000000F},
      {AUDIO_REG_DCOC2_SHIFT, 0x00000001},
      {AUDIO_REG_IIR3_2_SHIFT, 0x00000001},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x00000014},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000006},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0x000E4DB2},
      {AUDIO_REG_DMD_RA_DDS, 0x00F696E6},
      {AUDIO_REG_IIR2_3_SEL, 0x00000025},
      {AUDIO_REG_IIR1_4_SEL, 0x00000021},
      {AUDIO_REG_DN1_FREQ, 0x0000C965},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_RDSI_SEL, 0x00000009},
      {AUDIO_REG_RDSQ_SEL, 0x00000009},
      {AUDIO_REG_RDSI_SHIFT, 0x00000000},
      {AUDIO_REG_RDSQ_SHIFT, 0x00000000},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003}
   };
   if (sc->format.in_second_lang)
   {
      cx88_write_regs(&(sc->common), regs_btsc_sap,
         sizeof(regs_btsc_sap) / sizeof(cx88_reg));  
   }
   else
   {
      cx88_write_regs(&(sc->common), regs_btsc,
         sizeof(regs_btsc) / sizeof(cx88_reg));  
   }
   return 0;
}

static int cx23880_audio_eiaj(cx23880_audio_softc *sc, uint32_t *ctl_reg)
{
   /* TODO: Linux driver doesn't have EIAJ register values, so these
    * are taken from EIAJ example in CX2388x datasheet.  For BTSC at
    * least, the datasheet example values will work, but the cribbed
    * Linux driver values give better quality.  So hopefully these will
    * work at least as a stopgap. */
   uint32_t src_sel = 0;
   if (sc->format.in_second_lang)
   {
      src_sel = AUD_SRC_EIAJ_FORCE_MONO2;
   }
   else
   {
      switch (sc->format.in_stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_EIAJ_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_EIAJ_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_EIAJ_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   static const cx88_reg regs_eiaj[] = 
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x00008000},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00C00000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_C2_UP_THR, 0x00001000},
      {AUDIO_REG_C2_LO_THR, 0x00000400},
      {AUDIO_REG_C1_UP_THR, 0x00001800},
      {AUDIO_REG_C1_LO_THR, 0x00001000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000080},
      {AUDIO_REG_START_TIMER, 0x00000200},
      {AUDIO_REG_DEEMPH0_G0, 0x000023F8},
      {AUDIO_REG_PHASE_FIX_CTL, 0x00000009},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000006},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000006}
   };
   cx88_write_regs(&(sc->common), regs_eiaj, sizeof(regs_eiaj) / sizeof(cx88_reg));
   return 0;  
}

static int cx23880_audio_a2(cx23880_audio_softc *sc, uint32_t *ctl_reg)
{   
   uint32_t src_sel = 0;
   if (sc->format.in_second_lang)
   {
      src_sel = AUD_SRC_A2_FORCE_MONO2;
   }
   else
   {
      switch (sc->format.in_stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_A2_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_A2_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_A2_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg |= AUD_CTL_FMRADIO_RDS;
   static const cx88_reg regs_a2_common[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_AGC_HYST, 0x0000001A},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x0000755B},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00551340},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x006D30BE},
      {AUDIO_REG_PILOT_BQD_1_K3, 0xFFD394AF},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00040000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x002A4841},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000040},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000007},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_DEEMPH0_G0, 0x00000380},
      {AUDIO_REG_DEEMPH1_G0, 0x00000380},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC0_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_IIR3_0_SEL, 0x00000021},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR3_1_SEL, 0x00000023},
      {AUDIO_REG_RDSI_SEL, 0x00000017},
      {AUDIO_REG_RDSI_SHIFT, 0x00000000},
      {AUDIO_REG_RDSQ_SEL, 0x00000017},
      {AUDIO_REG_RDSQ_SHIFT, 0x00000000},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000001},
      {AUDIO_REG_START_TIMER, 0x00000000}
   };
   static const cx88_reg regs_a2_bg[] =
   {
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000}
   };
   static const cx88_reg regs_a2_dk[] =
   {
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000},
      {AUDIO_REG_DN0_FREQ, 0x00003A1C},
      {AUDIO_REG_DN2_FREQ, 0x0000D2E0}
   };
   static const cx88_reg regs_deemph_50[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x00000380},
      {AUDIO_REG_DEEMPH1_G0, 0x00000380},
      {AUDIO_REG_DEEMPHGAIN_R, 0x000011E1},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x0002A7BC},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x0003023C}
   };
   static const cx88_reg regs_a1_i[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x933A0000},
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000004},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x0000755B},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00551340},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x006D30BE},
      {AUDIO_REG_PILOT_BQD_1_K3, 0xFFD394AF},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00040000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x002A4841},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000060},
      {AUDIO_REG_AGC_HYST, 0x0000000A},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000007},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC0_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_IIR3_0_SEL, 0x00000021},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR3_1_SEL, 0x00000023},
      {AUDIO_REG_DN0_FREQ, 0x000035A3},
      {AUDIO_REG_DN2_FREQ, 0x000029C7},
      {AUDIO_REG_CRDC0_SRC_SEL, 0x00000511},
      {AUDIO_REG_IIR1_0_SEL, 0x00000001},
      {AUDIO_REG_IIR1_1_SEL, 0x00000000},
      {AUDIO_REG_IIR3_2_SEL, 0x00000003},
      {AUDIO_REG_IIR3_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SEL, 0x00000002},
      {AUDIO_REG_IIR2_0_SEL, 0x00000021},
      {AUDIO_REG_IIR2_0_SHIFT, 0x00000002},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x0000000B},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x0000000B},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000001},
      {AUDIO_REG_START_TIMER, 0x00000000}
   };
   static const cx88_reg regs_am_l[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0xF53D4800},
      {AUDIO_REG_QAM_MODE, 0x00000000},
      {AUDIO_REG_PHACC_FREQ, 0x4A3A0000},
      {AUDIO_REG_DEEMPHGAIN_R, 0x00006680},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x000353DE},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x000001B1},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_FM_MODE_ENABLE, 0x00000007},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_AGC_GAIN, 0x00000000},
      {AUDIO_REG_AGC_HYST, 0x00000018},
      {AUDIO_REG_AGC_DEF, 0x00000020},
      {AUDIO_REG_DN0_FREQ, 0x00000000},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0x000E4DB2},
      {AUDIO_REG_DCOC_0_SRC, 0x00000021},
      {AUDIO_REG_IIR1_0_SEL, 0x00000000},
      {AUDIO_REG_IIR1_0_SHIFT, 0x00000007},
      {AUDIO_REG_IIR1_1_SEL, 0x00000002},
      {AUDIO_REG_IIR1_1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SRC, 0x00000003},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000000},
      {AUDIO_REG_IIR1_2_SEL, 0x00000023},
      {AUDIO_REG_IIR1_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR1_3_SEL, 0x00000004},
      {AUDIO_REG_IIR1_3_SHIFT, 0x00000007},
      {AUDIO_REG_IIR1_4_SEL, 0x00000005},
      {AUDIO_REG_IIR1_4_SHIFT, 0x00000007},
      {AUDIO_REG_IIR3_0_SEL, 0x00000007},
      {AUDIO_REG_IIR3_0_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x00000011},
      {AUDIO_REG_DEEMPH0_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH0_G0, 0x00007000},
      {AUDIO_REG_DEEMPH0_A0, 0x00000000},
      {AUDIO_REG_DEEMPH0_B0, 0x00000000},
      {AUDIO_REG_DEEMPH0_A1, 0x00000000},
      {AUDIO_REG_DEEMPH0_B1, 0x00000000},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x00000011},
      {AUDIO_REG_DEEMPH1_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH1_G0, 0x00007000},
      {AUDIO_REG_DEEMPH1_A0, 0x00000000},
      {AUDIO_REG_DEEMPH1_B0, 0x00000000},
      {AUDIO_REG_DEEMPH1_A1, 0x00000000},
      {AUDIO_REG_DEEMPH1_B1, 0x00000000},
      {AUDIO_REG_OUT0_SEL, 0x0000003F},
      {AUDIO_REG_OUT1_SEL, 0x0000003F},
      {AUDIO_REG_DMD_RA_DDS, 0x00F5C285},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_RATE_THRES_DMD, 0x000000C0}
   };
   switch (sc->format.in_audio_format)
   {
      case CX88_AUDIO_FMT_AM_L:
         cx88_write_regs(&(sc->common), regs_am_l, sizeof(regs_am_l) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A2_BG:
         cx88_write_regs(&(sc->common), regs_a2_common, sizeof(regs_a2_common) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_a2_bg, sizeof(regs_a2_bg) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A2_DK:
         cx88_write_regs(&(sc->common), regs_a2_common, sizeof(regs_a2_common) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_a2_dk, sizeof(regs_a2_dk) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A1_I:
         cx88_write_regs(&(sc->common), regs_a1_i, sizeof(regs_a1_i) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
   }
   return 0;
}

static int cx23880_audio_nicam(cx23880_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   if (sc->format.in_second_lang)
   {
      src_sel = AUD_SRC_NICAM_FORCE_MONO2;
   }
   else
   {
      switch (sc->format.in_stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_NICAM_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_NICAM_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_NICAM_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_DEMATRIX, AUD_DEMATRIX_LR);
   *ctl_reg |= AUD_CTL_DEMATRIX_BYPASS;
   static const cx88_reg regs_nicam_l[] =
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000060},
      {AUDIO_REG_RATE_ADJ2, 0x000000F9},
      {AUDIO_REG_RATE_ADJ3, 0x000001CC},
      {AUDIO_REG_RATE_ADJ4, 0x000002B3},
      {AUDIO_REG_RATE_ADJ5, 0x00000726},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_DMD_RA_DDS, 0x00C00000},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_START_TIMER, 0x00000000},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x000353DE},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x000001B1},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000},
      {AUDIO_REG_DEEMPHGAIN_R, 0x00006680},
      {AUDIO_REG_RATE_THRES_DMD, 0x000000C0}   
   };
   static const cx88_reg regs_nicam_bgdki[] =
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000010},
      {AUDIO_REG_RATE_ADJ2, 0x00000040},
      {AUDIO_REG_RATE_ADJ3, 0x00000100},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00001000},
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x000003FF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x000000FF},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000003F},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_DEEMPHGAIN_R, 0x000023C2},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x0002A7BC},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x0003023E},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_QAM_MODE, 0x00000005}
   };
   static const cx88_reg regs_nicam_i[] =
   { 
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_PHACC_FREQ, 0x933A0000}
   };
   static const cx88_reg regs_nicam_bgdk[] =
   {
      {AUDIO_REG_PDF_DDS_CNST, 0x16820600},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000}
   };
   switch (sc->format.in_audio_format)
   {
      case CX88_AUDIO_FMT_NICAM_BGDK:
         cx88_write_regs(&(sc->common), regs_nicam_bgdki, sizeof(regs_nicam_bgdki) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_nicam_bgdk, sizeof(regs_nicam_bgdk) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_NICAM_I:
         cx88_write_regs(&(sc->common), regs_nicam_bgdki, sizeof(regs_nicam_bgdki) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_nicam_i, sizeof(regs_nicam_i) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_NICAM_L:
         cx88_write_regs(&(sc->common), regs_nicam_l, sizeof(regs_nicam_l) / sizeof(cx88_reg));
         break;
   }
   return 0;
}

static int cx23880_audio_load_format(cx23880_audio_softc *sc)
{
   int error = 0;
   uint32_t ctrl_reg = 0;
   uint32_t i2s_ctrl_reg = 0;
   if (sc->format.in_input == CX88_AUDIO_INPUT_I2S)
   {
      ctrl_reg |= AUD_CTL_I2SIN_ENABLE;
   }
   if (sc->format.in_i2s_config & CX88_AUDIO_I2S_TO_DAC)
   {
      ctrl_reg |= (AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2STODAC);
   }
   switch (sc->format.in_i2s_config & 0x70)
   {
      case CX88_AUDIO_I2S_MODE_OSRDAC:  
         ctrl_reg &= (~AUD_CTL_I2SIN_ENABLE);
         ctrl_reg |= AUD_CTL_I2SOUT_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_OS_INSCLK;
         break;
      case CX88_AUDIO_I2S_MODE_OSRADC:
         ctrl_reg &= (~AUD_CTL_I2SOUT_ENABLE);
         ctrl_reg |= AUD_CTL_I2SIN_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_OS_OUTSCLK;
         break;
      case CX88_AUDIO_I2S_MODE_PASS:
         ctrl_reg &= (~AUD_CTL_I2SIN_ENABLE);
         ctrl_reg |= AUD_CTL_I2SOUT_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_PASSTHRU;
         break;
      case CX88_AUDIO_I2S_MODE_FULL:
         ctrl_reg |= (AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2SOUT_ENABLE);
         i2s_ctrl_reg |= (I2S_CTL_SLAVE_INSCLK | I2S_CTL_OS_INSCLK);
         break;
   }
   if (sc->format.in_input == CX88_AUDIO_INPUT_BCAST)
   {
      cx23880_audio_broadcast_format *fmt = &(broadcast_formats[sc->format.in_audio_format]);
      cx88_lock_board(&(sc->common));
      cx88_write_reg(&(sc->common), CX23880_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX23880_REG_ADC_CONFIG) & (~CX23880_CHROMA_VIDEO));
      cx88_unlock_board(&(sc->common));
      cx88_write_reg(&(sc->common), AUDIO_REG_AFE_12DB, 1);
      if ((fmt->init_reg == AUDIO_INIT_BTSC) && sc->format.in_second_lang)
      {
         cx88_write_reg(&(sc->common), AUDIO_REG_INIT, AUDIO_INIT_BTSC_SAP);
      }
      else
      {
         cx88_write_reg(&(sc->common), AUDIO_REG_INIT, fmt->init_reg);
      }
      cx88_write_reg(&(sc->common), AUDIO_REG_INIT_LOAD, 1);
      cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 1);
      error = fmt->set_format(sc, &ctrl_reg);
   }
   else
   { 
      cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 1);
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_RATE_THRES_DMD, 0xC0);
   cx88_write_reg(&(sc->common), AUDIO_REG_I2S_CTL, i2s_ctrl_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_I2SINPUT, 
      sc->format.in_i2s_config & (CX88_AUDIO_I2S_IN_MASTER | CX88_AUDIO_I2S_IN_SONY));
   cx88_write_reg(&(sc->common), AUDIO_REG_I2SOUTPUT,
      (sc->format.in_i2s_config & (CX88_AUDIO_I2S_OUT_SONY)) >> 2);
   cx88_write_reg(&(sc->common), AUDIO_REG_AUD_CTL, ctrl_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 0);
   return error;
}

static int cx23880_audio_start_capture(device_t dev, uint8_t path)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   int error = cx23880_audio_load_format(sc);
   if (error)
   {
      return error;
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_BAL_CTL, sc->balance_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, sc->volume_reg);
   return 0;
}

/* Establish broadcast and stereo format */
static int cx23880_audio_set_format(device_t dev, uint8_t path, cx88_audio_format *fmt)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   if ((fmt->in_audio_format >= (sizeof(broadcast_formats) / sizeof(cx23880_audio_broadcast_format))) || 
       (fmt->in_stereo_format > CX88_AUDIO_STEREO_AUTO) ||
       (fmt->in_input == 0))
   {
      return EINVAL; 
   }
   sc->format = *fmt;
   if (sc->audio.state >= CX88_AUDIO_CAPTURING)
   {
      return cx23880_audio_load_format(sc);  
   }
   return 0;
}

static int cx23880_audio_set_volume(device_t dev, uint8_t path, cx88_audio_volume vol)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   if ((vol.in_left > 100) || (vol.in_right > 100))
   {
      return EINVAL;
   }
   uint16_t temp;
   uint8_t greater = vol.in_right;
   uint8_t lesser = vol.in_left;
   sc->balance_reg = 0;
   if (vol.in_left > vol.in_right)
   {
      greater = vol.in_left;
      lesser = vol.in_right;
      sc->balance_reg = BAL_CTL_RIGHT;
   }
   temp = 0x3F - (((uint16_t)greater * 0x3F) / 100);
   sc->volume_reg = (uint8_t)temp;
   temp = 0x3F - (((uint16_t)lesser * 0x3F) / 100);
   sc->balance_reg |= (uint8_t)(temp - sc->volume_reg);
   if (greater == 0)
   {
      sc->volume_reg |= (VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_BAL_CTL, sc->balance_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, (uint32_t)(sc->volume_reg));
   sc->volume = vol;
   return 0;
}

static cx88_audio_volume cx23880_audio_get_volume(device_t dev, uint8_t path)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   return sc->volume;
}

static int cx23880_audio_set_stereo(device_t dev, uint8_t path, boolean_t stereo)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   sc->format.in_stereo_format = (stereo ? CX88_AUDIO_STEREO_ON : CX88_AUDIO_STEREO_OFF);
   if (sc->audio.state >= CX88_AUDIO_CAPTURING)
   {
      return cx23880_audio_load_format(sc);  
   }
   return 0;
}

static boolean_t cx23880_audio_get_stereo(device_t dev, uint8_t path)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   if (sc->format.in_stereo_format == CX88_AUDIO_STEREO_AUTO)
   {
      uint32_t stereo_mode = cx88_read_field(
         cx88_read_reg(&(sc->common), AUDIO_REG_STATUS), AUDIO_STATUS_STEREO_MODE);
      return ((stereo_mode == AUDIO_STEREO_MODE_STEREO) ? TRUE : FALSE);
   }
   return ((sc->format.in_stereo_format == CX88_AUDIO_STEREO_ON) ? TRUE : FALSE);
}

static int cx23880_audio_set_mute(device_t dev, uint8_t path, boolean_t mute)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   if (mute)
   {
      sc->volume_reg |= (VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   }
   else
   {
      sc->volume_reg &= ~(VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, sc->volume_reg);
   return 0;
}

static boolean_t cx23880_audio_get_mute(device_t dev, uint8_t path)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   return ((sc->volume_reg & VOL_CTL_SRC_MUTE) ? TRUE : FALSE);
}

static int cx23880_audio_get_info(device_t dev, uint8_t path, cx88_audio_status *status)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   status->out_vol_left = sc->volume.in_left;
   status->out_vol_right = sc->volume.in_right;
   status->out_mute = ((sc->volume_reg & VOL_CTL_SRC_MUTE) ? 1 : 0);
   uint32_t status_reg = cx88_read_reg(&(sc->common), AUDIO_REG_STATUS);
   status->out_carrier = cx88_read_field(status_reg, AUDIO_STATUS_PILOT);
   status->out_signal = cx88_read_field(status_reg, AUDIO_STATUS_STEREO_MODE);
   return 0;
}

static int cx23880_audio_filter(void *arg)
{
   cx23880_audio_softc *sc = arg;
   return cx88_audio_filter(&(sc->audio));
}
   
static void cx23880_audio_intr(void *arg)
{
   cx23880_audio_softc *sc = arg;
   cx88_audio_intr(&(sc->audio));
}

static int cx23880_audio_probe(device_t dev) 
{
   if ((pci_get_vendor(dev) == CX88_VID) && (pci_get_device(dev) == 0x8811))
   {
      device_set_desc(dev, cx23880_audio_client.description);
      return BUS_PROBE_DEFAULT;  
   }
   return cx88_common_probe(dev, &cx23880_audio_client);
}

static int cx23880_audio_attach(device_t dev)
{
   return cx88_common_attach(dev, &cx23880_audio_client);
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx23880_audio_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx23880_audio_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

static int cx23880_audio_request_reconfig(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   return cx88_audio_request_reconfig(&(sc->audio));
}

/* Suspend DEVMETHOD */
static int cx23880_audio_suspend(device_t dev)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   return cx88_audio_suspend(&(sc->audio));
}

/* Shutdown DEVMETHOD */
static int cx23880_audio_shutdown(device_t dev)
{
   cx23880_audio_softc *sc = device_get_softc(dev);
   cx88_audio_reset(&(sc->audio), CX88_AUDIO_INVALID);
   return 0;
}

static devclass_t cx23880_audio_devclass;

static device_method_t cx23880_audio_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx23880_audio_probe),
   DEVMETHOD(device_attach,            cx23880_audio_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx23880_audio_shutdown),
   DEVMETHOD(device_suspend,           cx23880_audio_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx23880_audio_request_reconfig),

   /* i2c host methods */
   CX88_I2C_METHODS(cx23880_audio_get_i2c_bus, cx23880_audio_get_i2c_dev),
   
   /* cx88 audio interface */
   DEVMETHOD(cx88_audio_start_capture, cx23880_audio_start_capture),
   DEVMETHOD(cx88_audio_stop_capture,  cx23880_audio_stop_capture),
   DEVMETHOD(cx88_audio_set_mute,      cx23880_audio_set_mute),
   DEVMETHOD(cx88_audio_get_mute,      cx23880_audio_get_mute),
   DEVMETHOD(cx88_audio_set_volume,    cx23880_audio_set_volume),
   DEVMETHOD(cx88_audio_get_volume,    cx23880_audio_get_volume),
   DEVMETHOD(cx88_audio_set_stereo,    cx23880_audio_set_stereo),
   DEVMETHOD(cx88_audio_get_stereo,    cx23880_audio_get_stereo),
   DEVMETHOD(cx88_audio_set_format,    cx23880_audio_set_format),
   DEVMETHOD(cx88_audio_get_info,      cx23880_audio_get_info),

   {NULL, NULL}
};

static driver_t cx23880_audio_driver =
{
   "cx23880audio",
   cx23880_audio_methods,
   0
};

DRIVER_MODULE(cx23880audio, pci, cx23880_audio_driver, cx23880_audio_devclass, NULL, NULL);
DRIVER_MODULE(cx23880audio, cardbus, cx23880_audio_driver, cx23880_audio_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx23880audio);
CX88_I2C_DRIVER_MODULE(cx23880audio);
CX88_PCM_DRIVER_MODULE(cx23880audio);
MODULE_DEPEND(cx23880audio, cx88audiocore, CX88_AUDIO_MINVER, CX88_AUDIO_MODVER, CX88_AUDIO_MAXVER);
MODULE_VERSION(cx23880audio, 1);
