/*-
 * Copyright 2009 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_mpeg.h"

typedef struct
{
   cx88_common_softc common;
   cx88_i2c_softc i2c;
   cx88_mpeg_softc mpeg;
} cx23880_mpeg_softc;

static cx88_i2c_softc *cx23880_mpeg_i2c(cx88_common_softc *sc, uint8_t bus);
static int cx23880_mpeg_setup(cx88_common_softc *arg);
static int cx23880_mpeg_teardown(cx88_common_softc *arg);
static int cx23880_mpeg_filter(void *arg);
static void cx23880_mpeg_intr(void *arg);
static int cx23880_mpeg_probe(device_t dev);
static int cx23880_mpeg_attach(device_t dev);
static int cx23880_mpeg_shutdown(device_t dev);
static int cx23880_mpeg_suspend(device_t dev);
static cx88_i2c_bus *cx23880_mpeg_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx23880_mpeg_get_i2c_dev(device_t dev, uint8_t addr);
static int cx23880_mpeg_request_reconfig(device_t dev); 

/* Client driver descriptor for cx88 common layer */
static const cx88_client_driver cx23880_mpeg_client = 
{
   .softc_size               = sizeof(cx23880_mpeg_softc),
   .bridge                   = CX88_BRIDGE_CX23880,
   .function                 = CX88_FUNC_MPEGTS,
   .setup                    = cx23880_mpeg_setup,
   .teardown                 = cx23880_mpeg_teardown,
	.interrupt_filter         = cx23880_mpeg_filter,
   .interrupt_handler        = cx23880_mpeg_intr,
   .description              = "Conexant CX23880/1/2/3 MPEG Transport Stream"
};

static cx88_i2c_softc *cx23880_mpeg_i2c(cx88_common_softc *sc, uint8_t bus)
{
   if (bus != 0)
   {
      return NULL;
   }
   return &(((cx23880_mpeg_softc*)sc)->i2c);
}

/* Setup callback to be called by cx88 common layer during device attach */
static int cx23880_mpeg_setup(cx88_common_softc *arg)
{
   cx23880_mpeg_softc *sc = (cx23880_mpeg_softc*)arg;
   int error = cx88_i2c_attach(&(sc->i2c), 0, &(sc->common), sc->common.dev);
   if (!error)
   {
      error = cx88_mpeg_setup(&(sc->mpeg), CX88_MPEG_CX23880_TS, &(sc->common), cx23880_mpeg_i2c);
      if (error)
      {
         cx88_i2c_detach(&(sc->i2c));
      }
   }
   return error;
}

/* Cleanup callback to be called by cx88 common layer during device detach */
static int cx23880_mpeg_teardown(cx88_common_softc *arg)
{
   cx23880_mpeg_softc *sc = (cx23880_mpeg_softc*)arg;
   cx88_mpeg_teardown(&(sc->mpeg));
   return cx88_i2c_detach(&(sc->i2c));
}

static int cx23880_mpeg_filter(void *arg)
{
   cx23880_mpeg_softc *sc = arg;
   return cx88_mpeg_filter(&(sc->mpeg));
}
   
static void cx23880_mpeg_intr(void *arg)
{
   cx23880_mpeg_softc *sc = arg;
   cx88_mpeg_intr(&(sc->mpeg));
}

/* Probe DEVMETHOD */
static int cx23880_mpeg_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx23880_mpeg_client);
}

/* Attach DEVMETHOD */
static int cx23880_mpeg_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx23880_mpeg_client);
}

/* Shutdown DEVMETHOD */
static int cx23880_mpeg_shutdown(device_t dev)
{
   cx23880_mpeg_softc *sc = device_get_softc(dev);
   cx88_mpeg_reset(&(sc->mpeg), CX88_MPEG_INVALID);
   return 0;
}

/* Suspend DEVMETHOD */
static int cx23880_mpeg_suspend(device_t dev)
{
   cx23880_mpeg_softc *sc = device_get_softc(dev);
   return cx88_mpeg_suspend(&(sc->mpeg));
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx23880_mpeg_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx23880_mpeg_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

/* Resource reallocation DEVMETHOD */
static int cx23880_mpeg_request_reconfig(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   return cx88_mpeg_request_reconfig(&(sc->mpeg));
}
 
static devclass_t cx23880_mpeg_devclass; 

/* DEVMETHODs */
static device_method_t cx23880_mpeg_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx23880_mpeg_probe),
   DEVMETHOD(device_attach,            cx23880_mpeg_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx23880_mpeg_shutdown),
   DEVMETHOD(device_suspend,           cx23880_mpeg_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx23880_mpeg_request_reconfig),
             
   /* i2c host methods */
   CX88_I2C_METHODS(cx23880_mpeg_get_i2c_bus, cx23880_mpeg_get_i2c_dev),

   {NULL, NULL}
};

/* Module declarations */
static driver_t cx23880_mpeg_driver =
{
   "cx23880mpeg",
   cx23880_mpeg_methods,
   0
};

DRIVER_MODULE(cx23880mpeg, pci, cx23880_mpeg_driver, cx23880_mpeg_devclass, NULL, NULL);
DRIVER_MODULE(cx23880mpeg, cardbus, cx23880_mpeg_driver, cx23880_mpeg_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx23880mpeg);
CX88_I2C_DRIVER_MODULE(cx23880mpeg);
MODULE_DEPEND(cx23880mpeg, cx88mpegcore, CX88_MPEG_MINVER, CX88_MPEG_MODVER, CX88_MPEG_MAXVER);
MODULE_VERSION(cx23880mpeg, 1);
