/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <dev/pci/pcivar.h>
#include <dev/iicbus/iiconf.h>
#include <dev/iicbus/iicbus.h>
#include <sys/module.h>
#include <sys/linker.h>
#include <sys/firmware.h>
#include <sys/endian.h>
#include "cx88_audio.h"
#include "cx23885_avcore_regs.h"
#include "cx23885_avcore.h"

#define CX23885AV_I2C_TIMEOUT      (hz / 16)

static devclass_t cx23885av_devclass;
static devclass_t cx23885adc_devclass;

typedef struct
{
   device_t bus;
   uint8_t addr;
   device_t adc;
   uint8_t adc_addr;
   enum
   {
      CX23885AV_TYPE_CX23885,
      CX23885AV_TYPE_CX23887,
      CX23885AV_TYPE_CX23888
   } type;
   struct
   {
      cx88_audio_format format;
      cx88_audio_volume volume;
      boolean_t mute;
   } audio_config[CX23885AV_NUM_AUDIO_PATHS];
} cx23885av_softc;

typedef struct
{
   cx23885av_softc *sc;
   int error;
   enum
   {
      CX23885AV_IOSTATE_STOPPED = 0,
      CX23885AV_IOSTATE_STARTED,
   } state;
} cx23885av_io_state;

static void cx23885av_io_cleanup(cx23885av_io_state *state);
static void cx23885av_start(uint8_t addr, cx23885av_io_state *state);
static void cx23885av_send_bytes(const void *buffer, int size, cx23885av_io_state *state);
static void cx23885av_recv_bytes(void *buffer, int size, cx23885av_io_state *state);
static void cx23885av_write(
   const uint16_t reg,
   const void *buffer,
   int size,
   cx23885av_io_state *state);
static void cx23885av_read(
   const uint16_t reg,
   void *buffer,
   int size,
   cx23885av_io_state *state);
static void cx23885adc_write(
   const uint8_t reg,
   const void *buffer,
   int size,
   cx23885av_io_state *state);
static void cx23885adc_read(
   const uint8_t reg,
   void *buffer,
   int size,
   cx23885av_io_state *state);
static int cx23885av_probe(device_t dev);
static int cx23885av_attach(device_t dev);
static int cx23885av_detach(device_t dev);
static int cx23885av_suspend(device_t dev);
static int cx23885av_resume(device_t dev);
static int cx23885av_format_audio(device_t dev, uint8_t path, cx88_audio_format *format);
static int cx23885av_start_audio_capture(device_t dev, uint8_t path);
static void cx23885av_stop_audio_capture(device_t dev, uint8_t path);
static int cx23885av_set_mute(device_t dev, uint8_t path, boolean_t mute);
static boolean_t cx23885av_get_mute(device_t dev, uint8_t path);
static int cx23885av_set_volume(device_t dev, uint8_t path, cx88_audio_volume vol);
static cx88_audio_volume cx23885av_get_volume(device_t dev, uint8_t path);
static int cx23885av_set_stereo(device_t dev, uint8_t path, boolean_t stereo);
static boolean_t cx23885av_get_stereo(device_t dev, uint8_t path);
static int cx23885av_get_info(device_t dev, uint8_t path, cx88_audio_status *status);
static int cx23885adc_probe(device_t dev);
static int cx23885adc_attach(device_t dev);
static int cx23885adc_detach(device_t dev);
static int cx23885adc_suspend(device_t dev);
static int cx23885adc_resume(device_t dev);
static void cx23885av_dif_init(cx23885av_io_state *iostate);

static inline void cx23885av_set_error(cx23885av_io_state *iostate, int error)
{
   if (!(iostate->error))
   {
      iostate->error = error; 
   }
}

static inline void cx23885av_io_init(cx23885av_softc *sc, cx23885av_io_state *state)
{
   state->sc = sc;
   state->error = 0;
   state->state = CX23885AV_IOSTATE_STOPPED;
}

static void cx23885av_io_cleanup(cx23885av_io_state *state)
{
   if (state->state == CX23885AV_IOSTATE_STARTED)
   {
      int error = iicbus_stop(state->sc->bus);
      cx23885av_set_error(state, error);
      state->state = CX23885AV_IOSTATE_STOPPED;
   } 
}

static void cx23885av_start(uint8_t addr, cx23885av_io_state *state)
{
   if (!(state->error))
   {
      if (state->state == CX23885AV_IOSTATE_STOPPED)
      {
         state->error = iicbus_start(state->sc->bus, addr, CX23885AV_I2C_TIMEOUT);
         if (!(state->error))
         {
            state->state = CX23885AV_IOSTATE_STARTED;
         }
      }
      else
      {
         state->error = iicbus_repeated_start(state->sc->bus, addr, CX23885AV_I2C_TIMEOUT);
      }
   }
}

static void cx23885av_send_bytes(const void *buffer, int size, cx23885av_io_state *state)
{
   if (!(state->error) && (size > 0))
   {
      int sent_bytes = 0;
      state->error = iicbus_write(state->sc->bus, (const char*)buffer, size, &sent_bytes, CX23885AV_I2C_TIMEOUT);
      if (!(state->error) && (sent_bytes < size))
      {
         state->error = EIO;
      }
   }
}

static void cx23885av_recv_bytes(void *buffer, int size, cx23885av_io_state *state)
{
   if (!(state->error))
   {
      int read_bytes = 0;
      state->error = iicbus_read(state->sc->bus, (char*)buffer, size, &read_bytes, 
         IIC_LAST_READ, CX23885AV_I2C_TIMEOUT);
      if (!(state->error) && (read_bytes < size))
      {
         state->error = EIO;
      }
   }
}

static void cx23885av_write(
   const uint16_t reg,
   const void *buffer,
   int size,
   cx23885av_io_state *state)
{
   cx23885av_start(state->sc->addr << 1, state);
   if (!(state->error))
   {
      int sent_bytes = 0;
      uint16_t bereg = htobe16(reg);
      state->error = iicbus_write(state->sc->bus, (const char*)(&bereg), sizeof(bereg), &sent_bytes, CX23885AV_I2C_TIMEOUT);
      if (!(state->error) && (sent_bytes < sizeof(reg)))
      {
         state->error = EIO;
      }
      cx23885av_send_bytes(buffer, size, state);
   }
}

static void cx23885av_read(
   const uint16_t reg,
   void *buffer,
   int size,
   cx23885av_io_state *state)
{
   cx23885av_write(reg, NULL, 0, state);
   cx23885av_start((state->sc->addr << 1) | 1, state);
   cx23885av_recv_bytes(buffer, size, state);
}

static void cx23885adc_write(
   const uint8_t reg,
   const void *buffer,
   int size,
   cx23885av_io_state *state)
{
   cx23885av_start(state->sc->adc_addr << 1, state);
   if (!(state->error))
   {
      int sent_bytes = 0;
      state->error = iicbus_write(state->sc->bus, (const char*)(&reg), sizeof(reg), &sent_bytes, CX23885AV_I2C_TIMEOUT);
      if (!(state->error) && (sent_bytes < sizeof(reg)))
      {
         state->error = EIO;
      }
      cx23885av_send_bytes(buffer, size, state);
   }
}

static void cx23885adc_read(
   const uint8_t reg,
   void *buffer,
   int size,
   cx23885av_io_state *state)
{
   cx23885adc_write(reg, NULL, 0, state);
   cx23885av_start((state->sc->adc_addr << 1) | 1, state);
   cx23885av_recv_bytes(buffer, size, state);
}

int cx23885av_gpioctl(device_t dev, cx88_common_gpio *gpio)
{
   cx23885av_softc *sc = device_get_softc(dev);
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   uint8_t reg = gpio->in_write_mask & 0xFF;
   if (reg != 0)
   {
      cx23885av_write(CX23885AV_REG_GPIO_OEN, &reg, sizeof(reg), &iostate);
      reg = gpio->inout_gpio & 0xFF;
      cx23885av_write(CX23885AV_REG_GPIO_IO, &reg, sizeof(reg), &iostate);
   }
   if (gpio->in_read_mask & 0xFF)
   {
      cx23885av_read(CX23885AV_REG_GPIO_IO, &reg, sizeof(reg), &iostate);
      gpio->inout_gpio = reg & gpio->in_read_mask;
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

int cx23885av_set_adc(device_t dev, device_t adc)
{
   cx23885av_softc *sc = device_get_softc(dev);
   struct iicbus_ivar *adcaddr = IICBUS_IVAR(adc);
   if (adcaddr == NULL)
   {
      return ENXIO;
   }
   sc->adc = adc;
   sc->adc_addr = (uint8_t)(adcaddr->addr);
   return 0;
}

/* Probe DEVMETHOD */
static int cx23885av_probe(device_t dev)
{
   device_set_desc(dev, "Conexant CX23885/7/8 A/V Core");
   return 0;
}

/* Attach DEVMETHOD */
static int cx23885av_attach(device_t dev)
{
   cx23885av_softc *sc = device_get_softc(dev);
   struct iicbus_ivar *devaddr = IICBUS_IVAR(dev);
   if (devaddr == NULL)
   {
      return ENXIO;  
   }
   sc->bus = device_get_parent(dev);
   sc->addr = (uint8_t)(devaddr->addr);
   sc->adc = NULL;
   device_t bridge = device_get_parent(device_get_parent(sc->bus));
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   DROP_GIANT();
   uint8_t hostreg1 = 0;
   cx23885av_read(CX23885AV_REG_HOSTREG1, &hostreg1, sizeof(hostreg1), &iostate);
   hostreg1 &= (~(HOSTREG1_DIGITAL_POWERDOWN | HOSTREG1_AUTO_INC_DISABLE));
   cx23885av_write(CX23885AV_REG_HOSTREG1, &hostreg1, sizeof(hostreg1), &iostate);
   switch (pci_get_device(bridge))
   {
      case 0x8852:
         sc->type = CX23885AV_TYPE_CX23885;
         device_printf(dev, "Detected CX23885 A/V Core\n");
         break;
      case 0x8880:
         {
            uint16_t ir_tcd = 0;
            cx23885av_read(CX23885AV_REG_IR_TCD, &ir_tcd, sizeof(ir_tcd), &iostate);
            if (ir_tcd == 0)
            {
               device_printf(dev, "Detected CX23888 A/V Core\n");
               sc->type = CX23885AV_TYPE_CX23888;
            }
            else
            {
               device_printf(dev, "Detected CX23887 A/V Core\n");
               sc->type = CX23885AV_TYPE_CX23887;
            }
         }
         break;
      default:
         device_printf(dev, "Unrecognized bridge id 0x%x\n", pci_get_device(bridge));
         cx23885av_set_error(&iostate, ENXIO);
         break;
   }
   uint8_t reg;
   cx23885av_read(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   reg |= MISC_CONTROL_SOFT_RESET;
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   reg &= (~MISC_CONTROL_SOFT_RESET);
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   reg = DOWNLOAD_CONTROL_DL_ENABLE;
   cx23885av_write(CX23885AV_REG_DOWNLOAD_CONTROL, &reg, sizeof(reg), &iostate);
   reg = 0x76;
   cx23885av_write(CX23885AV_REG_HOSTREG3, &reg, sizeof(reg), &iostate);
   reg = 0x40;
   cx23885av_write(CX23885AV_REG_HOSTREG2, &reg, sizeof(reg), &iostate);
   reg = 0x0;
   cx23885av_write(0x130, &reg, sizeof(reg), &iostate);
   reg = 0x1;
   cx23885av_write(0x125, &reg, sizeof(reg), &iostate);
   uint32_t plls[6];
   plls[0] = htole32(0x00000515); // VID_PLL_POSTDIV = 0x5, VID_PLL_COEFFICIENT = 0x15
   plls[1] = htole32(0x13333333); // VID_PLL_FRACTIONAL
   switch (sc->type)
   {
      case CX23885AV_TYPE_CX23885:
         // AUX_PLL_REF_MULT = 0xA, AUX_PLL_POSTDIV = 0x3, AUX_PLL_COEFFICIENT = 0xC
         plls[2] = htole32(0x000A030C);
         plls[3] = htole32(0x01BF0C9E); // AUX_PLL_FRACTIONAL
         plls[4] = htole32(0x00000414); // SYS_PLL_POSTDIV = 0x4, SYS_PLL_COEFFICIENT = 0x14
         plls[5] = htole32(0x0); // SYS_PLL_FRACTIONAL
         break;
      case CX23885AV_TYPE_CX23888:
         reg = 0x71;
         cx23885av_write(CX23885AV_REG_HOSTREG3, &reg, sizeof(reg), &iostate);
      case CX23885AV_TYPE_CX23887:
         // AUX_PLL_REF_MULT = 0xA, AUX_PLL_POSTDIV = 0x3, AUX_PLL_COEFFICIENT = 0x7
         plls[2] = htole32(0x000A030E);
         plls[3] = htole32(0x017DBF48); // AUX_PLL_FRACTIONAL
         plls[4] = htole32(0x00000416); // SYS_PLL_POSTDIV = 0x4, SYS_PLL_COEFFICIENT = 0x16
         plls[5] = htole32(0x01D1744C); // SYS_PLL_FRACTIONAL
         break;
   }
   cx23885av_write(CX23885AV_REG_VID_PLL_COEFFICIENT, plls, sizeof(plls), &iostate);
   cx23885av_dif_init(&iostate);
   uint32_t reg4 = htole32(0x00107D12);
   cx23885av_write(CX23885AV_REG_LUMA_CONTROL, &reg4, sizeof(reg4), &iostate);
   reg4 = htole32(0x3D008282);
   cx23885av_write(CX23885AV_REG_CHROMA_CONTROL, &reg4, sizeof(reg4), &iostate);
   reg = MISC_CONTROL_VID_AUTOCFG_DISABLE;
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   const struct firmware *avcore_fw = firmware_get("cx23885avfw");
   if ((avcore_fw == NULL) || (avcore_fw->data == NULL) || (avcore_fw->datasize > UINT16_MAX))
   {
      device_printf(dev, "Unable to load A/V core firmware\n");
      cx23885av_set_error(&iostate, ENXIO);
   }
   else
   {
      uint16_t reg2 = 0x0000;
      cx23885av_write(CX23885AV_REG_DOWNLOAD_ADDRESS_LOW, &reg2, sizeof(reg2), &iostate);
      reg = DOWNLOAD_CONTROL_DL_ENABLE;
      cx23885av_write(CX23885AV_REG_DOWNLOAD_CONTROL, &reg, sizeof(reg), &iostate);
      hostreg1 |= HOSTREG1_AUTO_INC_DISABLE;
      cx23885av_write(CX23885AV_REG_HOSTREG1, &hostreg1, sizeof(hostreg1), &iostate);
      cx23885av_write(CX23885AV_REG_DOWNLOAD_DATA_CONTROL, avcore_fw->data, (int)(avcore_fw->datasize), &iostate);
      hostreg1 &= (~HOSTREG1_AUTO_INC_DISABLE);
      cx23885av_write(CX23885AV_REG_HOSTREG1, &hostreg1, sizeof(hostreg1), &iostate);
      cx23885av_read(CX23885AV_REG_DOWNLOAD_ADDRESS_LOW, &reg2, sizeof(reg2), &iostate);
      reg2 = le16toh(reg2);
      if (reg2 != avcore_fw->datasize)
      {
         device_printf(dev, "Unexpected firmware transfer count %u; expected %zu\n", reg2, avcore_fw->datasize);
         cx23885av_set_error(&iostate, EIO);
      }
      firmware_put(avcore_fw, FIRMWARE_UNLOAD);
   }
   reg = MUTE_CONTROL_SOFT_MUTE | MUTE_CONTROL_SRC_MUTE;
   cx23885av_write(CX23885AV_REG_PATH1_MUTE_CONTROL, &reg, sizeof(reg), &iostate);
   cx23885av_write(CX23885AV_REG_PATH2_MUTE_CONTROL, &reg, sizeof(reg), &iostate);
   sc->audio_config[CX23885AV_AUDIO_PATH1].mute = TRUE;
   sc->audio_config[CX23885AV_AUDIO_PATH2].mute = TRUE;
   reg = (OUTPUT_SRC3 << OUTPUT_SRC_PAR1_SHIFT)
      | (OUTPUT_SRC4 << OUTPUT_SRC_PAR2_SHIFT);
   cx23885av_write(CX23885AV_REG_OUTPUT_SRC_SEL, &reg, sizeof(reg), &iostate);
   reg = ((INPUT_SRC_PATH1 | INPUT_SRC_CLK_FIFO) << INPUT_SRC3_SHIFT)
      | ((INPUT_SRC_PATH2 | INPUT_SRC_CLK_FIFO) << INPUT_SRC4_SHIFT);
   cx23885av_write(CX23885AV_REG_INPUT_SRC_SEL1, &reg, sizeof(reg), &iostate);
   cx23885av_io_cleanup(&iostate);
   PICKUP_GIANT();
   return iostate.error;
}

/* Detach DEVMETHOD */
static int cx23885av_detach(device_t dev)
{
   return bus_generic_detach(dev);
}

/* Suspend DEVMETHOD */
static int cx23885av_suspend(device_t dev)
{
   return 0;
}

/* Resume DEVMETHOD */
static int cx23885av_resume(device_t dev)
{
   return 0;
}

int cx23885av_dif_enable(device_t dev, cx23888_dif_control *control)
{
   cx23885av_softc *sc = device_get_softc(dev);
   if ((sc->type != CX23885AV_TYPE_CX23888) || 
      ((control->offset + control->count) <= control->offset) ||
      ((control->offset + control->count) > CX23888_DIF_WINDOW_SIZE))
   {
      return EINVAL;
   }
   uint8_t dif_regs[CX23888_DIF_WINDOW_SIZE];
   int error = copyin(control->regs, dif_regs, control->count);
   if (error)
   {
      return error;
   }
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   uint8_t reg = MISC_CONTROL_DIF_ENABLE | MISC_CONTROL_VID_AUTOCFG_DISABLE;
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   cx23885av_write(0x300 + control->offset, dif_regs, control->count, &iostate);
   uint32_t reg4 = 0x0;
   cx23885av_write(CX23885AV_REG_DIF_CONTROL, &reg4, sizeof(reg4), &iostate);
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static void cx23885av_dif_init(cx23885av_io_state *iostate)
{
   uint32_t reg4 = htole32(0x1);
   cx23885av_write(CX23885AV_REG_DIF_BYPASS, &reg4, sizeof(reg4), iostate);
   reg4 = htole32(0x0DF7DF83);
   cx23885av_write(CX23885AV_REG_DIF_SRC_PHASE, &reg4, sizeof(reg4), iostate);
   reg4 = 0x0;
   cx23885av_write(CX23885AV_REG_DIF_CONTROL, &reg4, sizeof(reg4), iostate);
}

int cx23885av_dif_disable(device_t dev)
{
   cx23885av_softc *sc = device_get_softc(dev);
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   uint8_t reg = MISC_CONTROL_VID_AUTOCFG_DISABLE;
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, &reg, sizeof(reg), &iostate);
   cx23885av_dif_init(&iostate);
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static int cx23885av_format_audio(device_t dev, uint8_t path, cx88_audio_format *format)
{
   cx23885av_softc *sc = device_get_softc(dev);
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   uint16_t sel_ctl_reg_addr = 0;
   uint8_t sel_ctl_reg = 0;
   uint8_t input = format->in_input & 0x07;
   uint8_t control_regs[5];
   switch (path)
   {
      case CX23885AV_AUDIO_PATH1:
         sel_ctl_reg_addr = CX23885AV_REG_PATH1_AUDIO_SEL_CTL;
         break;
      case CX23885AV_AUDIO_PATH2:
         sel_ctl_reg_addr = CX23885AV_REG_PATH2_AUDIO_SEL_CTL;
         break;
      default:
         cx23885av_set_error(&iostate, EINVAL);
         break;
   }
   cx23885av_read(CX23885AV_REG_MISC_CONTROL, control_regs, sizeof(control_regs), &iostate);
   switch (input)
   {
      case CX88_AUDIO_INPUT_BCAST:
      case CX23885_AUDIO_INPUT_BCAST2:
      case CX23885_AUDIO_INPUT_BCAST3:
         control_regs[0] &= (~MISC_CONTROL_ADC2_SELECT);
         control_regs[2] |= AFE_CONTROL_VGA_GAIN_CH2_AUDIO;
         control_regs[4] &= (~(AFE_CONTROL_DROOP_COMP_CH2 | AFE_CONTROL_CLAMP_ENABLE_CH2));
         break;
      case CX23885_AUDIO_INPUT_BCAST4:
      case CX23885_AUDIO_INPUT_BCAST5:
         control_regs[0] |= MISC_CONTROL_ADC2_SELECT;
         control_regs[3] |= AFE_CONTROL_VGA_GAIN_CH3_AUDIO;
         control_regs[4] &= (~(AFE_CONTROL_DROOP_COMP_CH3 | AFE_CONTROL_CLAMP_ENABLE_CH3));
         break;
      case CX88_AUDIO_INPUT_I2S:
      case CX23885_AUDIO_INPUT_AC97:
         break;
      default:
         device_printf(dev, "Unsupported audio input type 0x%x\n", input);
         cx23885av_set_error(&iostate, EINVAL);
         break;
   }
   cx23885av_read(sel_ctl_reg_addr, &sel_ctl_reg, sizeof(sel_ctl_reg), &iostate);
   switch (input)
   {
      case CX88_AUDIO_INPUT_I2S:
      {
         uint8_t input_ctl[2];
         uint8_t output_ctl[2];
         input_ctl[0] = 0;
         input_ctl[1] = 0;
         output_ctl[0] = 0;
         output_ctl[1] = 0;
         sel_ctl_reg = (sel_ctl_reg & 0xF0) | AUDIO_SEL_CTL_I2S;
         if (format->in_i2s_config & CX88_AUDIO_I2S_IN_SONY)
         {
            input_ctl[0] |= I2S_INPUT_CTL_SONY;
         }
         if (format->in_i2s_config & CX88_AUDIO_I2S_IN_MASTER)
         {
            input_ctl[1] |= I2S_INPUT_CTL_MASTER;
         }
         if (format->in_i2s_config & CX88_AUDIO_I2S_OUT_SONY)
         {
            output_ctl[0] |= I2S_OUTPUT_CTL_SONY;
         }
         if (format->in_i2s_config & CX23885_AUDIO_I2S_OUT_MASTER)
         {
            output_ctl[1] |= I2S_OUTPUT_CTL_MASTER;
         }
         cx23885av_write(CX23885AV_REG_I2S_INPUT_CTL1, input_ctl, sizeof(input_ctl), &iostate);
         cx23885av_write(CX23885AV_REG_I2S_OUTPUT_CTL1, output_ctl, sizeof(output_ctl), &iostate);
         break;
      }
      case CX23885_AUDIO_INPUT_AC97:
         sel_ctl_reg = (sel_ctl_reg & 0xF0) | AUDIO_SEL_CTL_AC97;
         break;
      default:
      {
         uint8_t audio_control_regs[2];
         audio_control_regs[0] = AUDIO_CONTROL_STD_AUTO;
         audio_control_regs[1] = AUDIO_MODE_PREF_MONO;
         uint8_t format_ctl_reg = 0;
         uint8_t download_ctl_reg = 0;
         if (format->in_second_lang)
         {
            audio_control_regs[1] = AUDIO_MODE_PREF_LANGUAGE_B;
         }
         else if (format->in_stereo_format != CX88_AUDIO_STEREO_OFF)
         {
            audio_control_regs[1] = AUDIO_MODE_PREF_STEREO;
         }
         switch (format->in_audio_format)
         {
            case CX88_AUDIO_FMT_FM_RADIO_50:
               audio_control_regs[1] |= AUDIO_MODE_DEEMPHASIS_50;
            case CX88_AUDIO_FMT_FM_RADIO:
            case CX88_AUDIO_FMT_FM_RADIO_75:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_FM;
               if (format->in_input & CX23885_AUDIO_INPUT_AF)
               {
                  format_ctl_reg |= AUDIO_FORMAT_TUNER_OUTPUT_AF;
               }
               break;
            case CX88_AUDIO_FMT_BTSC:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_BTSC;
               break;
            case CX88_AUDIO_FMT_EIAJ:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_EIAJ;
               break;
            case CX88_AUDIO_FMT_AM_L:
            case CX88_AUDIO_FMT_NICAM_L:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_L;
               break;
            case CX88_AUDIO_FMT_A2_BG:
            case CX88_AUDIO_FMT_NICAM_BGDK:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_BG;
               break;
            case CX88_AUDIO_FMT_A2_DK:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_DK1;
               break;
            case CX88_AUDIO_FMT_A1_I:
            case CX88_AUDIO_FMT_NICAM_I:
               audio_control_regs[0] |= AUDIO_CONTROL_SYSTEM_I;
               break;
            default:
               device_printf(dev, "Unsupported audio format 0x%x\n", format->in_audio_format);
               cx23885av_set_error(&iostate, EINVAL);
               break;
         }
         download_ctl_reg = DOWNLOAD_CONTROL_MC_RUNNING | 0x03;
         cx23885av_write(CX23885AV_REG_DOWNLOAD_CONTROL, &download_ctl_reg,
            sizeof(download_ctl_reg), &iostate);
         cx23885av_write(CX23885AV_REG_AUDIO_CONTROL, audio_control_regs,
            sizeof(audio_control_regs), &iostate);
         cx23885av_write(CX23885AV_REG_AUDIO_FORMAT_CTL, &format_ctl_reg,
            sizeof(format_ctl_reg), &iostate);
         sel_ctl_reg = (sel_ctl_reg & 0xF0) | AUDIO_SEL_CTL_DEMOD_MAIN;
         control_regs[3] |= AFE_CONTROL_AUDIO_ADC2;
         break;
      }
   }
   cx23885av_write(sel_ctl_reg_addr, &sel_ctl_reg, sizeof(sel_ctl_reg), &iostate);
   switch (input)
   {
      case CX88_AUDIO_INPUT_BCAST:
         control_regs[1] &= 0xCF;
         break;
      case CX23885_AUDIO_INPUT_BCAST2:
         control_regs[1] = (control_regs[1] & 0xCF) | 0x10;
         break;
      case CX23885_AUDIO_INPUT_BCAST3:
         control_regs[1] = (control_regs[1] & 0xCF) | 0x20;
         break;
      case CX23885_AUDIO_INPUT_BCAST4:
         control_regs[1] &= 0x3F;
         break;
      case CX23885_AUDIO_INPUT_BCAST5:
         control_regs[1] = (control_regs[1] & 0x3F) | 0x40;
         break;
   }
   cx23885av_write(CX23885AV_REG_MISC_CONTROL, control_regs, sizeof(control_regs), &iostate);
   uint8_t adc_channel = format->in_input & (CX23885_AUDIO_INPUT_ADC1 | CX23885_AUDIO_INPUT_ADC2);
   if (adc_channel)
   {
      if (sc->adc == NULL)
      {
         cx23885av_set_error(&iostate, ENXIO);
      }
      uint8_t channel_regs[2];
      channel_regs[0] = 0;
      channel_regs[1] = 0;
      cx23885adc_read(0xE, channel_regs, 1, &iostate);
      switch (adc_channel)
      {
         case CX23885_AUDIO_INPUT_ADC1:
            channel_regs[0] &= ~(1 << 3);
            break;
         case CX23885_AUDIO_INPUT_ADC2:
            channel_regs[0] |= (1 << 3);
            break;
         default:
            cx23885av_set_error(&iostate, EINVAL);
            break;
      }
      channel_regs[0] |= 0x20;
      cx23885adc_write(0xE, channel_regs, sizeof(channel_regs), &iostate);
   }
   if (!iostate.error)
   {
      sc->audio_config[path].format = *format;
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static int cx23885av_start_audio_capture(device_t dev, uint8_t path)
{
   return cx23885av_set_mute(dev, path, FALSE);
}

static void cx23885av_stop_audio_capture(device_t dev, uint8_t path)
{
   cx23885av_set_mute(dev, path, TRUE);
}

static int cx23885av_set_mute(device_t dev, uint8_t path, boolean_t mute)
{
   if (path >= CX23885AV_NUM_AUDIO_PATHS)
   {
      return EINVAL;
   }
   cx23885av_softc *sc = device_get_softc(dev);
   if (mute == sc->audio_config[path].mute)
   {
      return 0;
   }
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   uint16_t reg_addr = (path == CX23885AV_AUDIO_PATH1 ? 
      CX23885AV_REG_PATH1_MUTE_CONTROL : CX23885AV_REG_PATH2_MUTE_CONTROL);
   uint8_t reg = (mute ? (MUTE_CONTROL_SOFT_MUTE | MUTE_CONTROL_SRC_MUTE) : 0);
   cx23885av_write(reg_addr, &reg, sizeof(reg), &iostate);
   if (!iostate.error)
   {
      sc->audio_config[path].mute = mute;
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static boolean_t cx23885av_get_mute(device_t dev, uint8_t path)
{
   if (path >= CX23885AV_NUM_AUDIO_PATHS)
   {
      return EINVAL;
   }
   cx23885av_softc *sc = device_get_softc(dev);
   return sc->audio_config[path].mute;
}

static int cx23885av_set_volume(device_t dev, uint8_t path, cx88_audio_volume vol)
{
   uint16_t volume_reg_addr;
   switch (path)
   {
      case CX23885AV_AUDIO_PATH1:
         volume_reg_addr = CX23885AV_REG_PATH1_VOLUME;
         break;
      case CX23885AV_AUDIO_PATH2:
         volume_reg_addr = CX23885AV_REG_PATH2_VOLUME;
         break;
      default:
         return EINVAL;
   }
   if ((vol.in_left > 100) || (vol.in_right > 100))
   {
      return EINVAL;
   }
   uint16_t temp;
   uint8_t greater = vol.in_left;
   uint8_t lesser = vol.in_right;
   uint8_t regs[2];
   regs[0] = 0;
   regs[1] = 0;
   if (vol.in_right > vol.in_left)
   {
      greater = vol.in_right;
      lesser = vol.in_left;
      regs[1] = BALANCE_LEFT;
   }
   if (greater == 0)
   {
      return cx23885av_set_mute(dev, path, TRUE);
   }
   temp = 0xE4 - (((uint16_t)greater * 0xE4) / 100);
   regs[0] = (uint8_t)temp;
   temp = ((uint16_t)(greater - lesser) * 0xE4) / 100;
   temp /= 2;
   if (temp > 0x60)
   {
      temp = 0x60;
   }
   regs[1] |= (uint8_t)temp;
   cx23885av_softc *sc = device_get_softc(dev);
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   cx23885av_write(volume_reg_addr, regs, sizeof(regs), &iostate);
   if (!iostate.error)
   {
      sc->audio_config[path].volume = vol;
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static cx88_audio_volume cx23885av_get_volume(device_t dev, uint8_t path)
{
   static cx88_audio_volume vol = {.in_left = 0, .in_right = 0};
   if (path < CX23885AV_NUM_AUDIO_PATHS)
   {
      cx23885av_softc *sc = device_get_softc(dev);
      vol = sc->audio_config[path].volume;
   }
   return vol;
}

static int cx23885av_set_stereo(device_t dev, uint8_t path, boolean_t stereo)
{
   if (path >= CX23885AV_NUM_AUDIO_PATHS)
   {
      return EINVAL;
   }
   cx23885av_softc *sc = device_get_softc(dev);
   if ((stereo && (sc->audio_config[path].format.in_stereo_format == CX88_AUDIO_STEREO_ON)) ||
       (!stereo && (sc->audio_config[path].format.in_stereo_format == CX88_AUDIO_STEREO_OFF)) ||
       sc->audio_config[path].format.in_second_lang)
   {
      return 0;
   }
   uint8_t pref_mode_reg = 0;
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   cx23885av_read(CX23885AV_REG_PREFERRED_MODE, &pref_mode_reg, sizeof(pref_mode_reg), &iostate);
   pref_mode_reg &= 0xF0;
   if (stereo)
   {
      pref_mode_reg |= AUDIO_MODE_PREF_STEREO;
   }
   cx23885av_write(CX23885AV_REG_PREFERRED_MODE, &pref_mode_reg, sizeof(pref_mode_reg), &iostate);
   if (!iostate.error)
   {
      sc->audio_config[path].format.in_stereo_format = 
         (stereo ? CX88_AUDIO_STEREO_ON : CX88_AUDIO_STEREO_OFF);
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static boolean_t cx23885av_get_stereo(device_t dev, uint8_t path)
{
   if (path >= CX23885AV_NUM_AUDIO_PATHS)
   {
      return FALSE;
   }
   cx23885av_softc *sc = device_get_softc(dev);
   if (sc->audio_config[path].format.in_stereo_format == CX88_AUDIO_STEREO_AUTO)
   {   
      cx23885av_io_state iostate;
      cx23885av_io_init(sc, &iostate);
      uint8_t status_reg = 0;
      cx23885av_read(CX23885AV_REG_AUDIO_STATUS_0, &status_reg, sizeof(status_reg), &iostate);
      cx23885av_io_cleanup(&iostate);
      return ((status_reg == AUDIO_STATUS_STEREO) ? TRUE : FALSE);
   }
   return ((sc->audio_config[path].format.in_stereo_format == CX88_AUDIO_STEREO_ON) ?
      TRUE : FALSE);
}

static int cx23885av_get_info(device_t dev, uint8_t path, cx88_audio_status *status)
{
   if (path >= CX23885AV_NUM_AUDIO_PATHS)
   {
      return EINVAL;
   }
   cx23885av_softc *sc = device_get_softc(dev);
   status->out_vol_left = sc->audio_config[path].volume.in_left;
   status->out_vol_right = sc->audio_config[path].volume.in_right;
   status->out_mute = sc->audio_config[path].mute;
   uint8_t status_reg = 0;
   cx23885av_io_state iostate;
   cx23885av_io_init(sc, &iostate);
   cx23885av_read(CX23885AV_REG_AUDIO_STATUS_0, &status_reg, sizeof(status_reg), &iostate);
   status->out_carrier = CX88_AUDIO_CARRIER_1;
   switch (status_reg)
   {
      case AUDIO_STATUS_MONO:
         status->out_signal = CX88_AUDIO_SIGNAL_MONO;
         break;
      case AUDIO_STATUS_STEREO:
         status->out_signal = CX88_AUDIO_SIGNAL_STEREO;
         break;
      case AUDIO_STATUS_DUAL:
         status->out_signal = CX88_AUDIO_SIGNAL_DUAL_MONO;
         break;
      case AUDIO_STATUS_SAP:
         status->out_signal = CX88_AUDIO_SIGNAL_SAP;
         break;
      default:
         status->out_signal = CX88_AUDIO_SIGNAL_MONO;
         status->out_carrier = CX88_AUDIO_NO_CARRIER;
         break;
   }
   cx23885av_io_cleanup(&iostate);
   return iostate.error;
}

static device_method_t cx23885av_methods[] =
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx23885av_probe),
   DEVMETHOD(device_attach,            cx23885av_attach),
   DEVMETHOD(device_detach,            cx23885av_detach),
   DEVMETHOD(device_shutdown,          cx23885av_detach),
   DEVMETHOD(device_suspend,           cx23885av_suspend),
   DEVMETHOD(device_resume,            cx23885av_resume),
   
   /* cx88 audio interface */
   DEVMETHOD(cx88_audio_start_capture, cx23885av_start_audio_capture),
   DEVMETHOD(cx88_audio_stop_capture,  cx23885av_stop_audio_capture),
   DEVMETHOD(cx88_audio_set_mute,      cx23885av_set_mute),
   DEVMETHOD(cx88_audio_get_mute,      cx23885av_get_mute),
   DEVMETHOD(cx88_audio_set_volume,    cx23885av_set_volume),
   DEVMETHOD(cx88_audio_get_volume,    cx23885av_get_volume),
   DEVMETHOD(cx88_audio_set_stereo,    cx23885av_set_stereo),
   DEVMETHOD(cx88_audio_get_stereo,    cx23885av_get_stereo),
   DEVMETHOD(cx88_audio_set_format,    cx23885av_format_audio),
   DEVMETHOD(cx88_audio_get_info,      cx23885av_get_info),
   
   {NULL, NULL}
};

static driver_t cx23885av_driver =
{
   CX23885AV_DRIVER_NAME,
   cx23885av_methods,
   sizeof(cx23885av_softc)
};

DRIVER_MODULE(cx23885av, iicbus, cx23885av_driver, cx23885av_devclass, NULL, NULL);
MODULE_DEPEND(cx23885av, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_DEPEND(cx23885av, firmware, 1, 1, 1);
MODULE_DEPEND(cx23885av, cx88audiocore, CX88_AUDIO_MINVER, CX88_AUDIO_MODVER, CX88_AUDIO_MAXVER);
MODULE_VERSION(cx23885av, CX23885AV_MODVER);

/* Probe DEVMETHOD */
static int cx23885adc_probe(device_t dev)
{
   device_set_desc(dev, "Conexant CX23885/7/8 Audio ADC");
   return 0;
}

/* Attach DEVMETHOD */
static int cx23885adc_attach(device_t dev)
{
   return 0;
}

/* Detach DEVMETHOD */
static int cx23885adc_detach(device_t dev)
{
   return bus_generic_detach(dev);
}

/* Suspend DEVMETHOD */
static int cx23885adc_suspend(device_t dev)
{
   return 0;
}

/* Resume DEVMETHOD */
static int cx23885adc_resume(device_t dev)
{
   return 0;
}

static device_method_t cx23885adc_methods[] =
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx23885adc_probe),
   DEVMETHOD(device_attach,            cx23885adc_attach),
   DEVMETHOD(device_detach,            cx23885adc_detach),
   DEVMETHOD(device_shutdown,          cx23885adc_detach),
   DEVMETHOD(device_suspend,           cx23885adc_suspend),
   DEVMETHOD(device_resume,            cx23885adc_resume),

   {NULL, NULL}
};

static driver_t cx23885adc_driver =
{
   CX23885ADC_DRIVER_NAME,
   cx23885adc_methods,
   0
};

DRIVER_MODULE(cx23885adc, iicbus, cx23885adc_driver, cx23885adc_devclass, NULL, NULL);
MODULE_DEPEND(cx23885adc, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_VERSION(cx23885adc, CX23885ADC_MODVER);
