/*-
 * Copyright 2009 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_mpeg.h"
#include "cx88_audio.h"
#include "cx23885_avcore.h"

#define CX23885_MINVER 2
#define CX23885_MODVER 2
#define CX23885_MAXVER 2

typedef struct
{
   cx88_common_softc common;
   cx88_mpeg_softc ts1;
   cx88_mpeg_softc ts2;
   cx88_audio_softc audio1;
   cx88_audio_softc audio2;
   struct mtx lock;
   struct cdev *devnode;
   uint32_t streams;
   uint32_t stopped_streams;
   boolean_t in_use;
   device_t i2c[CX23885_I2C_BUSES];
   uint8_t num_i2c_buses;
} cx23885_softc;

static cx88_i2c_softc *cx23885_i2c(cx88_common_softc *sc, uint8_t bus);
static int cx23885_setup(cx88_common_softc *arg);
static int cx23885_teardown(cx88_common_softc *arg);
static int cx23885_enable_streams(cx23885_softc *sc, uint32_t streams);
static int cx23885_disable_streams(cx23885_softc *sc, uint32_t streams);
static void cx23885_reconfigure_streams(cx23885_softc *sc, uint32_t streams);
static void cx23885_intr(void *arg);
static int cx23885_lock(cx23885_softc *sc);
static void cx23885_unlock(cx23885_softc *sc);
static int cx23885_probe(device_t dev);
static int cx23885_attach(device_t dev);
static int cx23885_shutdown(device_t dev);
static int cx23885_suspend(device_t dev);
static int cx23885_i2c_probe(device_t dev);
static int cx23885_i2c_attach(device_t dev);
static int cx23885_i2c_detach(device_t dev);
static cx88_i2c_bus *cx23885_i2c_get_bus(device_t dev);
static cx88_i2c_dev *cx23885_i2c_get_dev(device_t dev, uint8_t addr);

static d_open_t cx23885_open;
static d_close_t cx23885_close;
static d_ioctl_t cx23885_ioctl;

/* Devnode descriptor */
static struct cdevsw cx23885_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx23885b",
   .d_open = cx23885_open,   /* open */
   .d_close = cx23885_close, /* close */
   .d_ioctl = cx23885_ioctl, /* ioctl */
};

/* Client driver descriptor for cx88 common layer */
static const cx88_client_driver cx23885_client = 
{
   .softc_size               = sizeof(cx23885_softc),
   .bridge                   = CX88_BRIDGE_CX23885,
   .setup                    = cx23885_setup,
   .teardown                 = cx23885_teardown,
   .interrupt_filter         = NULL,
   .interrupt_handler        = cx23885_intr,
   .description              = "Conexant CX23885 Multimedia Bridge"
};

/* Client driver descriptor for cx88 common layer */
static const cx88_client_driver cx23887_client = 
{
   .softc_size               = sizeof(cx23885_softc),
   .bridge                   = CX88_BRIDGE_CX23887,
   .setup                    = cx23885_setup,
   .teardown                 = cx23885_teardown,
   .interrupt_filter         = NULL,
   .interrupt_handler        = cx23885_intr,
   .description              = "Conexant CX23887/8 Multimedia Bridge"
};

static cx88_i2c_softc *cx23885_i2c(cx88_common_softc *sc, uint8_t bus)
{
   cx23885_softc *full_sc = (cx23885_softc*)sc;
   if ((bus >= full_sc->num_i2c_buses) || (bus == CX23885_INTERNAL_I2C_BUS))
   {
      return NULL;
   }
   return (cx88_i2c_softc*)(device_get_softc(full_sc->i2c[bus]));
}

static int cx23885_setup(cx88_common_softc *arg)
{
   cx23885_softc *sc = (cx23885_softc*)arg;
   int error = 0;
   uint8_t i;
   sc->devnode = NULL;
   sc->stopped_streams = 0;
   sc->streams = CX23885_STREAM_MPEG1 | CX23885_STREAM_MPEG2 | 
      CX23885_STREAM_AUDIO1 | CX23885_STREAM_AUDIO2;
   sc->in_use = FALSE;
   sc->num_i2c_buses = 0;
   mtx_init(&(sc->lock), "CX23885 bridge lock", NULL, MTX_DEF);
   for (i = 0; i < CX23885_I2C_BUSES; ++i, ++(sc->num_i2c_buses))
   {
      sc->i2c[i] = device_add_child(arg->dev, "cx23885i2c", -1);
      if (sc->i2c[i] == NULL)
      {
         error = ENXIO;
      }
      else if ((error = device_probe_and_attach(sc->i2c[i])))
      {
         device_delete_child(arg->dev, sc->i2c[i]);
      }
      if (error)
      {
         break;
      }
   }
   if (!error)
   {
      cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS]);
      i2c_sc->num_i2c_devices = 2;
      error = cx88_i2c_add_device(&(i2c_sc->bus), 0x44, CX23885_I2C_SPEED_2MHZ, CX88_I2C_IO_POLLED,
         CX23885AV_DRIVER_NAME, &(i2c_sc->i2c_devices[0]));
      error = (error ? error : cx88_i2c_add_device(&(i2c_sc->bus), 0x4C, CX23885_I2C_SPEED_2MHZ, CX88_I2C_IO_POLLED,
         CX23885ADC_DRIVER_NAME, &(i2c_sc->i2c_devices[1])));
      error = (error ? error : cx23885av_set_adc(i2c_sc->i2c_devices[0].dev, i2c_sc->i2c_devices[1].dev));
   }
   if (!error)
   {
      error = cx23885_enable_streams(sc, sc->streams);
   }
   if (!error)
   {
      sc->devnode = make_dev(&cx23885_cdevsw, device_get_unit(sc->common.dev), 
         UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(sc->common.dev));
      if (sc->devnode == NULL)
      {
         error = ENOMEM;
      }
      else
      {
         sc->devnode->si_drv1 = sc;
      }
   }
   if (error)
   {
      cx23885_disable_streams(sc, sc->streams);
      sc->streams = 0;
      for (i = 0; i < sc->num_i2c_buses; ++i)
      {
         if (i == CX23885_INTERNAL_I2C_BUS)
         {
            cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[i]);
            cx88_i2c_remove_device(&(i2c_sc->i2c_devices[0]));
            cx88_i2c_remove_device(&(i2c_sc->i2c_devices[1]));
         }
         device_detach(sc->i2c[i]);
         device_delete_child(arg->dev, sc->i2c[i]);
      }
      sc->num_i2c_buses = 0;
      mtx_destroy(&(sc->lock));
   }
   return error;
}

static int cx23885_teardown(cx88_common_softc *arg)
{
   cx23885_softc *sc = (cx23885_softc*)arg;
   int error = 0;
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sc->devnode = NULL;
   }
   error = cx23885_disable_streams(sc, sc->streams);
   if (!error && (sc->num_i2c_buses > CX23885_INTERNAL_I2C_BUS))
   {
      cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS]);
      error = cx88_i2c_remove_device(&(i2c_sc->i2c_devices[0]));
      error = (error ? error : cx88_i2c_remove_device(&(i2c_sc->i2c_devices[1])));
   }
   while (!error && (sc->num_i2c_buses > 0))
   {
      error = device_detach(sc->i2c[sc->num_i2c_buses - 1]);
      if (!error)
      {
         device_delete_child(arg->dev, sc->i2c[sc->num_i2c_buses - 1]);
         --(sc->num_i2c_buses);
      }
   }
   if (!error)
   {
      mtx_destroy(&(sc->lock));
   }  
   return error;
}

static int cx23885_enable_streams(cx23885_softc *sc, uint32_t streams)
{
   int error = 0;
   if (streams & CX23885_STREAM_MPEG1)
   {
      error = cx88_mpeg_setup(&(sc->ts1), CX88_MPEG_CX23885_TS1, &(sc->common), cx23885_i2c);
   }
   if (!error && (streams & CX23885_STREAM_MPEG2))
   {
      error = cx88_mpeg_setup(&(sc->ts2), CX88_MPEG_CX23885_TS2, &(sc->common), cx23885_i2c);
   }
   device_t avcore = ((cx88_i2c_softc*)(device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS])))->i2c_devices[0].dev;
   if (!error && (streams & CX23885_STREAM_AUDIO1))
   {
      error = cx88_audio_setup(&(sc->audio1), CX88_AUDIO_STREAM_CX23885_INTA, &(sc->common),
         cx23885_i2c, avcore, CX23885AV_AUDIO_PATH1);
   }
   if (!error && (streams & CX23885_STREAM_AUDIO2))
   {
      error = cx88_audio_setup(&(sc->audio2), CX88_AUDIO_STREAM_CX23885_INTB, &(sc->common),
         cx23885_i2c, avcore, CX23885AV_AUDIO_PATH2);
   }
   if (error)
   {
      cx23885_disable_streams(sc, streams);
   }
   return error;
}

static int cx23885_disable_streams(cx23885_softc *sc, uint32_t streams)
{
   int error = 0;
   if (streams & CX23885_STREAM_MPEG1)
   {
      cx88_mpeg_teardown(&(sc->ts1));
   }
   if (streams & CX23885_STREAM_MPEG2)
   {
      cx88_mpeg_teardown(&(sc->ts2));
   }
   if (streams & CX23885_STREAM_AUDIO1)
   {
      error = cx88_audio_teardown(&(sc->audio1));
   }
   if (streams & CX23885_STREAM_AUDIO2)
   {
      int error2 = cx88_audio_teardown(&(sc->audio2));
      error = (error ? error : error2);
   }
   return error;
}

static void cx23885_reconfigure_streams(cx23885_softc *sc, uint32_t streams)
{
   if (streams & CX23885_STREAM_MPEG1)
   {
      cx88_mpeg_request_reconfig(&(sc->ts1));
   }
   if (streams & CX23885_STREAM_MPEG2)
   {
      cx88_mpeg_request_reconfig(&(sc->ts2));
   }
   if (streams & CX23885_STREAM_AUDIO1)
   {
      cx88_audio_request_reconfig(&(sc->audio1));
   }
   if (streams & CX23885_STREAM_AUDIO2)
   {
      cx88_audio_request_reconfig(&(sc->audio2));
   }
}

static void cx23885_intr(void *arg)
{
   cx23885_softc *sc = arg;
   mtx_lock(&(sc->lock));
   uint32_t intrs = cx88_read_reg(&(sc->common), CX23885_REG_PCI_INTMSTAT);
   uint32_t streams = sc->streams & (~(sc->stopped_streams));
   if ((intrs & CX23885_PCI_INT_VIDEO_B) && (streams & CX23885_STREAM_MPEG1))
   {
      cx88_mpeg_intr(&(sc->ts1));
   }
   if ((intrs & CX23885_PCI_INT_VIDEO_C) && (streams & CX23885_STREAM_MPEG2))
   {
      cx88_mpeg_intr(&(sc->ts2));
   }
   if (intrs & CX23885_PCI_INT_AUDIO_INT)
   {
      if (streams & CX23885_STREAM_AUDIO1)
      {
         cx88_audio_intr(&(sc->audio1));
      }
      if (streams & CX23885_STREAM_AUDIO2)
      {
         cx88_audio_intr(&(sc->audio2));
      }
   }
   mtx_unlock(&(sc->lock));
}

/* Open file operation */
static int cx23885_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   return 0;
}

/* Close file operation */
static int cx23885_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   return 0;
}

static int cx23885_lock(cx23885_softc *sc)
{
   int error = 0;
   mtx_lock(&(sc->lock));
   if (sc->in_use)
   {
      error = msleep(&(sc->in_use), &(sc->lock), PZERO | PCATCH, "cx885", 0);
      if (!error)
      {
         sc->in_use = TRUE;
      }
   }
   mtx_unlock(&(sc->lock));
   return error;
}

static void cx23885_unlock(cx23885_softc *sc)
{
   mtx_lock(&(sc->lock));
   sc->in_use = FALSE;
   wakeup_one(&(sc->in_use));
   mtx_unlock(&(sc->lock));
}

/* Ioctl file operation */
static int cx23885_ioctl(
   struct cdev *dev,
   u_long cmd,
   caddr_t data, 
   int flags,
   struct thread *td)
{
   cx23885_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   int error = 0;
   switch (cmd)
   {
      case CX23885_COMMON_SET_STREAMS:
      {
         uint32_t streams = *((uint32_t*)data);
         if (streams != sc->streams)
         {
            error = cx23885_lock(sc);
            if (error)
            {
               return error;
            }
            mtx_lock(&(sc->lock));
            sc->stopped_streams = sc->streams & (~streams);
            mtx_unlock(&(sc->lock));
            error = cx23885_disable_streams(sc, sc->stopped_streams);
            if (!error)
            {
               error = cx23885_enable_streams(sc, streams & (~(sc->streams)));
            }
            if (!error)
            {
               cx23885_reconfigure_streams(sc, streams & sc->streams);
               mtx_lock(&(sc->lock));
               sc->streams = streams;
               sc->stopped_streams = 0;
               mtx_unlock(&(sc->lock));
            }
            cx23885_unlock(sc);
         }
         break;
      }
      case CX23885_COMMON_GET_STREAMS:
         *((uint32_t*)data) = sc->streams;
         break;
      case CX23885_COMMON_AV_GPIO:
      {
         cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS]);
         error = cx23885av_gpioctl(i2c_sc->i2c_devices[0].dev, (cx88_common_gpio*)data);
         break;
      }
      case CX23888_COMMON_DIF_ENABLE:
      {
         cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS]);
         error = cx23885av_dif_enable(i2c_sc->i2c_devices[0].dev, (cx23888_dif_control*)data);
         break;
      }
      case CX23888_COMMON_DIF_DISABLE:
      {
         cx88_i2c_softc *i2c_sc = device_get_softc(sc->i2c[CX23885_INTERNAL_I2C_BUS]);
         error = cx23885av_dif_disable(i2c_sc->i2c_devices[0].dev);
         break;
      }
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Probe DEVMETHOD */
static int cx23885_probe(device_t dev) 
{
   int error = cx88_common_probe(dev, &cx23885_client);
   if (error > 0)
   {
      error = cx88_common_probe(dev, &cx23887_client);
   }
   return error;
}

/* Attach DEVMETHOD */
static int cx23885_attach(device_t dev) 
{
   if (cx88_common_probe(dev, &cx23885_client) <= 0)
   {
      return cx88_common_attach(dev, &cx23885_client);
   }
   else
   {
      return cx88_common_attach(dev, &cx23887_client);
   }
}

/* Shutdown DEVMETHOD */
static int cx23885_shutdown(device_t dev)
{
   cx23885_softc *sc = device_get_softc(dev);
   if (sc->streams & CX23885_STREAM_MPEG1)
   {
      cx88_mpeg_reset(&(sc->ts1), CX88_MPEG_INVALID);
   }
   if (sc->streams & CX23885_STREAM_MPEG2)
   {
      cx88_mpeg_reset(&(sc->ts2), CX88_MPEG_INVALID);
   }
   if (sc->streams & CX23885_STREAM_AUDIO1)
   {
      cx88_audio_reset(&(sc->audio1), CX88_AUDIO_INVALID);
   }
   if (sc->streams & CX23885_STREAM_AUDIO2)
   {
      cx88_audio_reset(&(sc->audio2), CX88_AUDIO_INVALID);
   }
   return 0;
}

/* Suspend DEVMETHOD */
static int cx23885_suspend(device_t dev)
{
   cx23885_softc *sc = device_get_softc(dev);
   int error = 0;
   if (sc->streams & CX23885_STREAM_MPEG1)
   {
      error = cx88_mpeg_suspend(&(sc->ts1));
   }
   if (!error && (sc->streams & CX23885_STREAM_MPEG2))
   {
      error = cx88_mpeg_suspend(&(sc->ts2));
   }
   if (sc->streams & CX23885_STREAM_AUDIO1)
   {
      error = cx88_audio_suspend(&(sc->audio1));
   }
   if (sc->streams & CX23885_STREAM_AUDIO2)
   {
      error = cx88_audio_suspend(&(sc->audio2));
   }
   return error;
}

static devclass_t cx23885_devclass; 

/* DEVMETHODs */
static device_method_t cx23885_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx23885_probe),
   DEVMETHOD(device_attach,            cx23885_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx23885_shutdown),
   DEVMETHOD(device_suspend,           cx23885_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   {NULL, NULL}
};

/* Module declarations */
static driver_t cx23885_driver =
{
   "cx23885b",
   cx23885_methods,
   0
};

DRIVER_MODULE(cx23885b, pci, cx23885_driver, cx23885_devclass, NULL, NULL);
DRIVER_MODULE(cx23885b, cardbus, cx23885_driver, cx23885_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx23885b);
CX88_PCM_DRIVER_MODULE(cx23885b);
MODULE_DEPEND(cx23885b, cx88mpegcore, CX88_MPEG_MINVER, CX88_MPEG_MODVER, CX88_MPEG_MAXVER);
MODULE_DEPEND(cx23885b, cx88audiocore, CX88_AUDIO_MINVER, CX88_AUDIO_MODVER, CX88_AUDIO_MAXVER);
MODULE_DEPEND(cx23885b, cx23885av, CX23885AV_MINVER, CX23885AV_MODVER, CX23885AV_MAXVER);
MODULE_DEPEND(cx23885b, cx23885adc, CX23885ADC_MINVER, CX23885ADC_MODVER, CX23885ADC_MAXVER);
MODULE_VERSION(cx23885b, CX23885_MODVER);

static devclass_t cx23885_i2c_devclass;

static int cx23885_i2c_probe(device_t dev)
{
   device_set_desc(dev, "CX23885/7/8 I2C master");
   return BUS_PROBE_SPECIFIC;
}

static int cx23885_i2c_attach(device_t dev)
{
   cx23885_softc *parent_sc = device_get_softc(device_get_parent(dev));
   return cx88_i2c_attach((cx88_i2c_softc*)(device_get_softc(dev)), parent_sc->num_i2c_buses,
      &(parent_sc->common), dev);
}

static int cx23885_i2c_detach(device_t dev)
{
   cx88_i2c_softc *sc = device_get_softc(dev);
   return cx88_i2c_detach(sc);
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx23885_i2c_get_bus(device_t dev)
{
   return cx88_i2c_get_bus((cx88_i2c_softc*)(device_get_softc(dev)));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx23885_i2c_get_dev(device_t dev, uint8_t addr)
{
   return cx88_i2c_get_dev((cx88_i2c_softc*)(device_get_softc(dev)), addr);
}

static device_method_t cx23885_i2c_methods[] = 
{
   DEVMETHOD(device_probe,             cx23885_i2c_probe),
   DEVMETHOD(device_attach,            cx23885_i2c_attach),
   DEVMETHOD(device_detach,            cx23885_i2c_detach),

   /* i2c host methods */
   CX88_I2C_METHODS(cx23885_i2c_get_bus, cx23885_i2c_get_dev),

   {NULL, NULL}
};

static driver_t cx23885_i2c_driver =
{
   "cx23885i2c",
   cx23885_i2c_methods,
   sizeof(cx88_i2c_softc)
};

DRIVER_MODULE(cx23885i2c, cx23885b, cx23885_i2c_driver, cx23885_i2c_devclass, NULL, NULL);
MODULE_DEPEND(cx23885i2c, cx23885b, CX23885_MINVER, CX23885_MODVER, CX23885_MAXVER);
CX88_I2C_DRIVER_MODULE(cx23885i2c);
MODULE_VERSION(cx23885i2c, 1);
