/*-
 * Copyright 2007 Jason Harmening
 *  
 * Copyright 2009 Konstantin Dimitrov <kosio.dimitrov@gmail.com>
 * Add I2C support for CX23885/7/8 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/types.h>
#include <sys/param.h>
#include <sys/bus.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/proc.h>
#include <sys/time.h>
#include <sys/limits.h>
#include "cx88_regs.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"

/* CX23880 register constants */
#define I2C_REG_DATA_CONTROL       0x368000
#define I2C_DATA_CONTROL_SLAVEADDR 0x7F,25
#define I2C_DATA_CONTROL_READ      (1 << 24)
#define I2C_DATA_CONTROL_DATA0     0xFF,24
#define I2C_DATA_CONTROL_DATA1     0xFF,16
#define I2C_DATA_CONTROL_DATA2     0xFF,8
#define I2C_DATA_CONTROL_HWMODE    (1 << 7)
#define I2C_DATA_CONTROL_400KHZ    (1 << 6)
#define I2C_DATA_CONTROL_NOSTOP    (1 << 5)
#define I2C_DATA_CONTROL_NOSTART   (1 << 4)
#define I2C_DATA_CONTROL_SYNC      (1 << 3)
#define I2C_DATA_CONTROL_3BYTEWR   (1 << 2)
#define I2C_DATA_CONTROL_ACKRD     (1 << 2)
#define I2C_DATA_CONTROL_SCL       (1 << 1)
#define I2C_DATA_CONTROL_SDA       (1 << 0)

typedef struct
{
   uint32_t reg_addr;
   uint32_t reg_control;
   uint32_t reg_wdata;
   uint32_t reg_rdata;
   uint32_t ack_bit;
   uint32_t done_bit;
} cx23885_i2c_control;

static cx23885_i2c_control cx23885_i2c_regs[] = 
{
   {
      .reg_addr =    0x180000,
      .reg_control = 0x180008,
      .reg_wdata =   0x180004,
      .reg_rdata =   0x18000C,
      .ack_bit =     CX23885_PCI_INT_I2C0_ACK,
      .done_bit =    CX23885_PCI_INT_I2C0_DONE
   },
   {
      .reg_addr =    0x190000,
      .reg_control = 0x190008,
      .reg_wdata =   0x190004,
      .reg_rdata =   0x19000C,
      .ack_bit =     CX23885_PCI_INT_I2C1_ACK,
      .done_bit =    CX23885_PCI_INT_I2C1_DONE
   },
   {
      .reg_addr =    0x1A0000,
      .reg_control = 0x1A0008,
      .reg_wdata =   0x1A0004,
      .reg_rdata =   0x1A000C,
      .ack_bit =     CX23885_PCI_INT_I2C2_ACK,
      .done_bit =    CX23885_PCI_INT_I2C2_DONE
   }
};

#define CX23885_I2C_CONTROL_2MHZ     (0x07 << 24)
#define CX23885_I2C_CONTROL_400KHZ   (0x27 << 24)
#define CX23885_I2C_CONTROL_100KHZ   (0x9d << 24)
#define CX23885_I2C_CONTROL_NOSTOP   (1 << 4)
#define CX23885_I2C_CONTROL_EXTEND   (1 << 3)
#define CX23885_I2C_CONTROL_READ     (1 << 0)

/* Default interrupt timeout--1s */
#define DEFAULT_TIMEOUT hz

static void *cx88_i2c_get_sibling_data(device_t dev);
static void cx88_i2c_init_board_data(cx88_i2c_board_data *data);
static int cx23880_i2c_init_board_data(void *arg);
static int cx23885_i2c_init_board_data(void *arg);
static int cx88_i2c_destroy_board_data(void *arg);
static int cx88_i2c_filter(void *arg);
static void cx88_i2c_set_intr(cx88_i2c_bus *bus, boolean_t enable);
static void cx88_i2c_terminate(cx88_i2c_bus *bus);
static int cx23880_i2c_commit(cx88_i2c_bus *bus, uint32_t reg, int timeout);
static int cx23885_i2c_commit(
   cx88_i2c_bus *bus, 
   uint32_t wdata,
   uint32_t ctrl,
   int timeout
   );
static int cx23880_i2c_write(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);
static int cx23880_i2c_read(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);
static int cx23885_i2c_write(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);
static int cx23885_i2c_read(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);
static int cx88_i2c_transfer(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);
static int cx88_i2c_set_speed(cx88_i2c_dev *dev, uint8_t speed);
static void cx88_i2c_prepare_io(cx88_i2c_bus *bus);
static void cx88_i2c_complete_io(cx88_i2c_bus *bus, int timeout);
static int cx88_i2c_lock_bus(cx88_i2c_bus *bus, cx88_i2c_io_t iotype, int timeout);
static void cx88_i2c_unlock_bus(cx88_i2c_bus *bus);

/* Accessor for sibling i2c shared data */
static void *cx88_i2c_get_sibling_data(device_t dev)
{
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   if (bus != NULL)
   {
      return bus->board_data;
   }
   return NULL;
}

/* Initializer to be called when i2c shared data is allocated */
static void cx88_i2c_init_board_data(cx88_i2c_board_data *data)
{
   mtx_init(&(data->lock), "CX88 I2C board lock", NULL, MTX_DEF);
   data->bus_owned = FALSE;
   mtx_init(&(data->isr_lock), "CX88 I2C interrupt lock", NULL, MTX_SPIN);
}

static int cx23880_i2c_init_board_data(void *arg)
{
   DIAGNOSTIC(printf("CX88-I2C: Initializing CX23880 board data\n"))
   cx88_i2c_board_data *data = arg;
   cx88_i2c_init_board_data(data);
   data->intr_mask_reg = CX23880_REG_PCI_INTMASK;
   data->intr_stat_reg = CX23880_REG_PCI_INTSTAT;
   data->ack_bit = CX23880_PCI_INT_I2C_ACK;
   data->done_bit = CX23880_PCI_INT_I2C_DONE;
   return 0;
}

static int cx23885_i2c_init_board_data(void *arg)
{
   DIAGNOSTIC(printf("CX88-I2C: Initializing CX23885/7 board data\n"))
   cx88_i2c_board_data *data = arg;
   cx88_i2c_init_board_data(data);
   data->intr_mask_reg = CX23885_REG_PCI_INTMASK;
   data->intr_stat_reg = CX23885_REG_PCI_INTSTAT;
   return 0;
}

/* Cleanup routine to be called when i2c shared data is about to be freed */
static int cx88_i2c_destroy_board_data(void *arg)
{
   DIAGNOSTIC(printf("CX88-I2C: Destroying board data\n"))
   cx88_i2c_board_data *data = arg;
   mtx_destroy(&(data->lock));
   mtx_destroy(&(data->isr_lock));
   return 0;
}

static int cx88_i2c_lock_bus(cx88_i2c_bus *bus, cx88_i2c_io_t iotype, int timeout)
{
   int error = 0;
   mtx_lock(&(bus->board_data->lock));
   if (iotype == CX88_I2C_IO_NORMAL)
   {
      if (bus->board_data->bus_owned)
      {
         error = msleep(bus->board_data, &(bus->board_data->lock), PZERO | PCATCH, "cx88ic", timeout);
      }
   }
   else
   {
      struct timeval start;
      struct timeval current;
      microtime(&start);
      while (!error && bus->board_data->bus_owned)
      {
         mtx_unlock(&(bus->board_data->lock));
         microtime(&current);
         timevalsub(&current, &start);
         if (tvtohz(&current) >= timeout)
         {
            error = IIC_ETIMEOUT;
         }
         mtx_lock(&(bus->board_data->lock));
      }
   }
   if (!error)
   {
      bus->board_data->bus_owned = TRUE;
   }
   mtx_unlock(&(bus->board_data->lock));
   return error;
}

static void cx88_i2c_unlock_bus(cx88_i2c_bus *bus)
{
   mtx_lock(&(bus->board_data->lock));
   bus->board_data->bus_owned = FALSE;
   /* Only one waiter will get to acquire the bus */
   wakeup_one(bus->board_data);
   mtx_unlock(&(bus->board_data->lock));
}

/* Create an i2c bus instance */
int cx88_i2c_create(
   cx88_i2c_bus *bus,
   uint8_t bus_number,
   cx88_common_softc *host_sc,
   device_t parent)
{
   bus->host_sc = host_sc;
   bus->dev = NULL;
   bus->bus_number = bus_number;
   bus->bus_speed = 0;
   bus->board_data = NULL;
   bus->irq_cookie = NULL;
   bus->io_canceled = FALSE;
   bus->isr_error = IIC_NOERR;
   bus->iotype = CX88_I2C_IO_NORMAL;
   bus->transaction_state = CX88_I2C_INACTIVE;
   int error = 0;
   int polled_io = 0;
   uint8_t max_buses = 0;
   switch (host_sc->board_data->bridge)
   {
      case CX88_BRIDGE_CX23880:
         max_buses = CX23880_I2C_BUSES;
         break;
      case CX88_BRIDGE_CX23885:
      case CX88_BRIDGE_CX23887:
         if (bus_number == CX23885_INTERNAL_I2C_BUS)
         {
            bus->bus_speed = CX23885_I2C_CONTROL_2MHZ;
         }
         else
         {
            bus->bus_speed = CX23885_I2C_CONTROL_100KHZ;
         }
         max_buses = CX23885_I2C_BUSES;
         break;
   } 
   if (bus_number >= max_buses)
   {
      return EINVAL;
   }
   int (*init_board_data)(void*) = ((host_sc->board_data->bridge == CX88_BRIDGE_CX23880) ?
      cx23880_i2c_init_board_data : cx23885_i2c_init_board_data);
   /* Allocate shared data structure or retrieve it from sibling driver */
   if ((error = cx88_add_board_data(host_sc, (void**)(&(bus->board_data)),
      sizeof(*(bus->board_data)), cx88_i2c_get_sibling_data, init_board_data)))
   {
      return error;
   }
   if (host_sc->board_data->bridge != CX88_BRIDGE_CX23880)
   {
      bus->board_data->ack_bit = cx23885_i2c_regs[bus_number].ack_bit;
      bus->board_data->done_bit = cx23885_i2c_regs[bus_number].done_bit;
   }
   bus->dev = device_add_child(parent, "iicbus", -1);
   if (bus->dev == NULL)
   {
      cx88_i2c_destroy(bus);
      return ENXIO;
   }
   if ((error = device_probe_and_attach(bus->dev)))
   {
      cx88_i2c_destroy(bus);
      return error;
   }
   if (!polled_io)
   {
      resource_int_value(device_get_name(parent), device_get_unit(parent), "i2c_polling", &polled_io);
   }
   if (!(host_sc->msi) && !polled_io)
   {
      /* Register i2c interrupt service routine */
      if ((error = bus_setup_intr(host_sc->dev, host_sc->irq_res, INTR_TYPE_AV | INTR_MPSAFE,
         cx88_i2c_filter, NULL, bus, &(bus->irq_cookie))))
      {
         cx88_i2c_destroy(bus);
      }
   }
   return error;
}

/* Destroy an i2c bus instance */
int cx88_i2c_destroy(cx88_i2c_bus *bus)
{
   int error = 0;
   if (bus->irq_cookie != NULL)
   {
      error = bus_teardown_intr(bus->host_sc->dev, bus->host_sc->irq_res, bus->irq_cookie);
      if (!error)
      {
         bus->irq_cookie = NULL;
      }
   }
   if (!error && (bus->dev != NULL))
   {
      error = device_detach(bus->dev);
      if (!error)
      {
         /* This is kind of ugly.  Newbus currently provides
          * no way for bus drivers to clean up instance 
          * variables for their children, so I have to do
          * it here.  This is NOT a permanent solution. */
         device_t *children = NULL;
         int num_children = 0, i;
         error = device_get_children(bus->dev, &children, &num_children);
         if (!error)
         {
            for (i = 0; i < num_children; ++i)
            {
               DIAGNOSTIC(device_printf(bus->host_sc->dev, "Freeing iicbus ivar\n"))
               free(IICBUS_IVAR(children[i]), M_DEVBUF);
            }
            free(children, M_TEMP);
            error = device_delete_child(device_get_parent(bus->dev), bus->dev);
            if (!error)
            {
               bus->dev = NULL;
            }
         }
      }
   }
   if (!error && (bus->board_data != NULL))
   {
      /* Release reference to shared data structure */
      error = cx88_remove_board_data(bus->host_sc, (void **)(&(bus->board_data)), cx88_i2c_destroy_board_data);
      if (!error)
      {
         bus->board_data = NULL;
      }
   }
   return error;
}

/* I2C interrupt filter */
static int cx88_i2c_filter(void *arg)
{
   cx88_i2c_bus *bus = arg;
   uint32_t reg;
   /* Ensure the interrupt is actually ours */
   if ((bus->transaction_state != CX88_I2C_INACTIVE) && 
      ((reg = cx88_read_reg(bus->host_sc, bus->board_data->intr_stat_reg)) & bus->board_data->done_bit))
   {
      /* Acknowledge the interrupt */
      cx88_write_reg(bus->host_sc, bus->board_data->intr_stat_reg,
         (bus->board_data->done_bit | bus->board_data->ack_bit));
      if (!(reg & bus->board_data->ack_bit))
      {
         bus->isr_error = IIC_ENOACK;
      }
      if ((bus->irq_cookie != NULL) && (bus->iotype == CX88_I2C_IO_NORMAL))
      {
         mtx_lock_spin(&(bus->board_data->isr_lock));
         wakeup(bus);
         mtx_unlock_spin(&(bus->board_data->isr_lock));
      }
      return FILTER_HANDLED;
   }
   return FILTER_STRAY;
}

static int cx88_i2c_set_speed(cx88_i2c_dev *dev, uint8_t speed)
{
   switch (dev->bus->host_sc->board_data->bridge)
   {
      case CX88_BRIDGE_CX23880:
         switch (speed)
         {
            case CX88_I2C_SPEED_100KHZ:
               dev->speed = 0;
               break;
            case CX88_I2C_SPEED_400KHZ:
               dev->speed = I2C_DATA_CONTROL_400KHZ;
               break;
            default:
               return EINVAL;
         }
         break;
      case CX88_BRIDGE_CX23885:
      case CX88_BRIDGE_CX23887:
         switch (speed)
         {
            case CX88_I2C_SPEED_100KHZ:
               dev->speed = CX23885_I2C_CONTROL_100KHZ;
               break;
            case CX88_I2C_SPEED_400KHZ:
               dev->speed = CX23885_I2C_CONTROL_400KHZ;
               break;
            case CX23885_I2C_SPEED_2MHZ:
               dev->speed = CX23885_I2C_CONTROL_2MHZ;
               break;
            default:
               return EINVAL;
         }
         break;
      default:
         return EINVAL;
   }
   return 0;
}

/* Add a device to a bus instance */
int cx88_i2c_add_device(
   cx88_i2c_bus *bus,
   uint8_t addr,
   uint8_t speed,
   cx88_i2c_io_t iotype,
   const char *name,
   cx88_i2c_dev *dev)
{
   int error = 0;
   if (addr & (1 << 7))
   {
      return EINVAL;
   }
   if (bus->dev == NULL)
   {
      return ENXIO;
   }
   cx88_i2c_init_device(dev);
   dev->bus = bus;
   dev->addr = addr;
   dev->iotype = iotype;
   cx88_i2c_set_speed(dev, speed);
   if (name != NULL)
   {
      dev->dev = BUS_ADD_CHILD(bus->dev, 0, name, -1);
      struct iicbus_ivar *devaddr = IICBUS_IVAR(dev->dev);
      if (devaddr == NULL)
      {
         error = ENXIO;
      }
      else
      {
         devaddr->addr = addr;
         error = device_probe_and_attach(dev->dev);
      }
   }
   if (error)
   {
      cx88_i2c_remove_device(dev);
   }
   return error;
}

int cx88_i2c_remove_device(cx88_i2c_dev *dev)
{
   int error = 0;
   if (dev->dev != NULL)
   {
      struct iicbus_ivar *addr = IICBUS_IVAR(dev->dev);
      /* This kind of sucks--there's no BUS_REMOVE_CHILD method to free ivars for us */
      error = device_detach(dev->dev);
      if (!error)
      {
         error = device_delete_child(dev->bus->dev, dev->dev);
      }
      if (!error)
      {
         free(addr, M_DEVBUF);
      }
   }
   if (!error)
   {
      /* Reset device state to default values */
      cx88_i2c_init_device(dev);
   }
   return error;
}

/* iicbus bus ownership callback--child devices should use this around
 * transactions to prevent other drivers from placing the bus in 
 * an invalid state */
int cx88_i2c_callback(device_t dev, int index, caddr_t data)
{
   /* Just return success in this call.  We previously performed a locked
    * wait on bus->board_data->bus_owned to work around bad locking in
    * the iicbus routine that calls this method in FreeBSD 7 and earlier.
    * FreeBSD 8 fixes the locking but does so by adding a mutex, so our 
    * wait would be unsafe.  But we still need to wait on the available
    * flag in the board data in order to coordinate bus ownership with
    * sibling devices.  Furthermore, some iicbus children don't correctly
    * request the bus anyway.  So we instead acquire the bus in the start()
    * method and release it in the stop() method. */
   return 0;
}

/* Helper function for setting the interrupt mask.  The enable
 * parameter indicates whether interrupts should be enabled or disabled. */
static void cx88_i2c_set_intr(cx88_i2c_bus *bus, boolean_t enable)
{
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(bus->host_sc, bus->board_data->intr_stat_reg,
      (bus->board_data->done_bit | bus->board_data->ack_bit));
   if ((bus->irq_cookie != NULL) && (bus->iotype == CX88_I2C_IO_NORMAL))
   {
      /* Protect access to the shared PCI interrupt register. */
      cx88_lock_board(bus->host_sc);
      /* Enable/disable I2C interrupt */
      cx88_write_reg(
         bus->host_sc,
         bus->board_data->intr_mask_reg,
         (enable ?
            (cx88_read_reg(bus->host_sc, bus->board_data->intr_mask_reg) | bus->board_data->done_bit) :
            (cx88_read_reg(bus->host_sc, bus->board_data->intr_mask_reg) & ~(bus->board_data->done_bit))));
      cx88_unlock_board(bus->host_sc);
   }
}

/* Helper function for terminating an i2c transfer */
static void cx88_i2c_terminate(cx88_i2c_bus *bus)
{
   if (bus->transaction_state != CX88_I2C_INACTIVE)
   {
      cx88_i2c_set_intr(bus, FALSE);
      bus->transaction_state = CX88_I2C_INACTIVE;
      cx88_i2c_unlock_bus(bus);
   }
}

static void cx88_i2c_prepare_io(cx88_i2c_bus *bus)
{
   if ((bus->irq_cookie != NULL) && (bus->iotype == CX88_I2C_IO_NORMAL))
   {
      mtx_lock_spin(&(bus->board_data->isr_lock));
   }
   bus->isr_error = IIC_NOERR;   
}

static void cx88_i2c_complete_io(cx88_i2c_bus *bus, int timeout)
{
   int time_limit = (timeout ? timeout : DEFAULT_TIMEOUT);
   if ((bus->irq_cookie != NULL) && (bus->iotype == CX88_I2C_IO_NORMAL))
   {
      int error = 0;
      if (!(bus->io_canceled))
      {
         error = msleep_spin(bus, &(bus->board_data->isr_lock), 
            "cx88ic", time_limit);
      }
      mtx_unlock_spin(&(bus->board_data->isr_lock));
      if (error)
      {
         bus->isr_error = IIC_ETIMEOUT;
      }
   }
   else
   {
      struct timeval start;
      struct timeval current;
      microtime(&start);
      while ((cx88_i2c_filter(bus) != FILTER_HANDLED) && !(bus->io_canceled))
      {
         microtime(&current);
         timevalsub(&current, &start);
         if (tvtohz(&current) >= time_limit)
         {
            bus->isr_error = IIC_ETIMEOUT;
            break;
         }
      }
   }
   if (bus->io_canceled)
   {
      bus->isr_error = IIC_EBUSBSY;
   }
}

/* Write a byte or sequence of bytes to the data/control register and
 * wait for the interrupt.  The reg parameter should contain the data
 * byte(s), R/W bit, and any required NOSTART/NOSTOP/ACK/NACK bits. */
static int cx23880_i2c_commit(cx88_i2c_bus *bus, uint32_t reg, int timeout)
{
   int error = IIC_NOERR;
   cx88_i2c_prepare_io(bus);
   /* Add required HW I2C flags to the register */
   uint32_t i2c_control = reg | bus->bus_speed | I2C_DATA_CONTROL_HWMODE | 
      I2C_DATA_CONTROL_SYNC | I2C_DATA_CONTROL_SCL | I2C_DATA_CONTROL_SDA;
   cx88_write_reg(bus->host_sc, I2C_REG_DATA_CONTROL, i2c_control);
   cx88_i2c_complete_io(bus, timeout);
   if (bus->isr_error)
   {
      device_printf(bus->host_sc->dev, "[dev 0x%x] I2C I/O Error %d\n", bus->slave_addr, bus->isr_error);
      /* There was an error handling the interrupt, so bail out. */
      cx88_i2c_terminate (bus);
      /* Kind of a hack, but make sure the bus is quiesced by sending out bogus data
       * with NOSTOP and ACK turned off */
      cx88_write_reg(bus->host_sc, I2C_REG_DATA_CONTROL, 
         (i2c_control | I2C_DATA_CONTROL_NOSTART) & (~(I2C_DATA_CONTROL_NOSTOP | I2C_DATA_CONTROL_ACKRD)));
      return bus->isr_error;
   }
   return error;
}

static int cx23885_i2c_commit(
   cx88_i2c_bus *bus, 
   uint32_t wdata,
   uint32_t ctrl,
   int timeout
   )
{
   int error = IIC_NOERR;
   cx88_i2c_prepare_io(bus);
   uint32_t i2c_control = ctrl | bus->bus_speed | (1 << 12) | (1 << 2);
   cx88_write_reg(bus->host_sc, cx23885_i2c_regs[bus->bus_number].reg_addr, bus->slave_addr << 25);
   cx88_write_reg(bus->host_sc, cx23885_i2c_regs[bus->bus_number].reg_wdata, wdata);
   cx88_write_reg(bus->host_sc, cx23885_i2c_regs[bus->bus_number].reg_control, i2c_control);
   cx88_i2c_complete_io(bus, timeout);
   if (bus->isr_error)
   {
      device_printf(bus->host_sc->dev, "[bus %d dev 0x%x] I2C I/O Error 0x%d\n", bus->bus_number, bus->slave_addr, bus->isr_error);
      /* There was an error handling the interrupt, so bail out. */
      cx88_i2c_terminate (bus);
      return bus->isr_error;
   }
   return error;
}

static int cx23880_i2c_write(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   uint32_t reg = 0;
   if (bus->transaction_state == CX88_I2C_START)
   {
      /* This is the first phase of the transaction, so transmit the
       * slave address */
      reg = cx88_write_field(I2C_DATA_CONTROL_SLAVEADDR, bus->slave_addr);
      reg |= cx88_write_field(I2C_DATA_CONTROL_DATA1, bus->last_write.byte);
   }
   else
   {
      reg = cx88_write_field(I2C_DATA_CONTROL_DATA0, bus->last_write.byte);
   }
   if ((transaction != NULL) || (bus->transaction_state == CX88_I2C_RESTART))
   {
      /* If transaction is non-NULL, we have another transaction after this one,
       * so we shouldn't generate the STOP condition.  Transaction will be NULL
       * if we are flushing transactions during a restart operation, but RESTART
       * also indicates there will be subsequent transactions */
      reg |= I2C_DATA_CONTROL_NOSTOP;
   }
   if (bus->transaction_state != CX88_I2C_START)
   {
      reg |= I2C_DATA_CONTROL_NOSTART;
   }
   return cx23880_i2c_commit(bus, reg, bus->last_write.timeout);
}

static int cx23880_i2c_read(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   int error = IIC_NOERR;
   /* Commit the transaction if it is a read */
   uint32_t reg = cx88_write_field(I2C_DATA_CONTROL_SLAVEADDR, bus->slave_addr);
   reg |= I2C_DATA_CONTROL_READ;
   if (!(transaction->last))
   {
      reg |= (I2C_DATA_CONTROL_ACKRD | I2C_DATA_CONTROL_NOSTOP);
   }
   if (bus->transaction_state == CX88_I2C_ACTIVE)
   {
      /* If this read is the first transaction after a start, move
       * the state up to START since reads don't go through an initial
       * caching step */
      bus->transaction_state = CX88_I2C_START;
   }
   if (bus->transaction_state != CX88_I2C_START)
   {
      reg |= I2C_DATA_CONTROL_NOSTART;
   }
   error = cx23880_i2c_commit(bus, reg, transaction->timeout);
   transaction->byte = cx88_read_field(
      cx88_read_reg(bus->host_sc, I2C_REG_DATA_CONTROL),
      I2C_DATA_CONTROL_DATA2);
   return error;
}

static int cx23885_i2c_write(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   uint32_t ctrl = 0;
   if (transaction != NULL)
   {
      /* If we have another write immediately following the one cached 
       * in bus->last_write, then don't send STOP, and use EXTEND to inform
       * the controller that the next transfer won't need the START condition */
      ctrl = (CX23885_I2C_CONTROL_NOSTOP | CX23885_I2C_CONTROL_EXTEND);
   }
   else if (bus->transaction_state == CX88_I2C_RESTART)
   {
      /* If we're purging the last write before a repeated start,
       * we don't want to send STOP but do want to send START on the
       * next transfer */
      ctrl = CX23885_I2C_CONTROL_NOSTOP;
   }
   return cx23885_i2c_commit(bus, bus->last_write.byte, ctrl, bus->last_write.timeout);
}

static int cx23885_i2c_read(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   int error = IIC_NOERR;
   uint32_t ctrl = CX23885_I2C_CONTROL_READ;
   if (!(transaction->last))
   {
      ctrl |= (CX23885_I2C_CONTROL_NOSTOP | CX23885_I2C_CONTROL_EXTEND);
   }
   if (bus->transaction_state == CX88_I2C_ACTIVE)
   {
      /* If this read is the first transaction after a start, move
       * the state up to START since reads don't go through an initial
       * caching step */
      bus->transaction_state = CX88_I2C_START;
   }
   error = cx23885_i2c_commit(bus, 0, ctrl, transaction->timeout);
   transaction->byte = cx88_read_reg(bus->host_sc, cx23885_i2c_regs[bus->bus_number].reg_rdata);
   return error;
}

/* Conducts an I2C transaction. We cache writes and issue them with the next transaction
 * in order to properly handle the STOP condition (reads do not require this because
 * of the 'last' flag), so this function first commits any cached writes and then,
 * if transaction is non-NULL, commits transaction (if it is a read) or caches it. */
static int cx88_i2c_transfer(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   int error = IIC_NOERR;
   /* Flush any pending write */
   if ((bus->transaction_state > CX88_I2C_ACTIVE) && bus->pending_write)
   {
      if (bus->host_sc->board_data->bridge == CX88_BRIDGE_CX23880)
      {
         error = cx23880_i2c_write(bus, transaction);
      }
      else
      {
         error = cx23885_i2c_write(bus, transaction);
      }
   }
   bus->pending_write = FALSE;
   /* Commit or cache the present transaction */
   if (!error && (transaction != NULL))
   {
      if (transaction->type == CX88_I2C_READ)
      {
         if (bus->host_sc->board_data->bridge == CX88_BRIDGE_CX23880)
         {
            error = cx23880_i2c_read(bus, transaction);
         }
         else
         {
            error = cx23885_i2c_read(bus, transaction);
         }
      }
      else
      {
         /* If this is a write, just cache the transaction */
         bus->last_write = *transaction;
         bus->pending_write = TRUE;
      }
   }
   /* Now transition the state */
   switch (bus->transaction_state)
   {
      case CX88_I2C_ACTIVE:
         bus->transaction_state = CX88_I2C_START;
         break;
      case CX88_I2C_START:
         bus->transaction_state = CX88_I2C_STARTED;
         break;
      case CX88_I2C_RESTART:
         bus->transaction_state = CX88_I2C_ACTIVE;
         break;
      default:
         break;
   }
   return error;
}

/* iicbus start callback */
int cx88_i2c_start(device_t dev, u_char slave, int timeout)
{
   DIAGNOSTIC(device_printf(dev, "Starting I2C transaction for device 0x%x\n", (slave >> 1)))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   cx88_i2c_dev *i2c_dev = CX88_I2C_GET_DEV(dev, (slave >> 1));
   /* Ensure we own the bus & device */
   if ((bus == NULL) || (i2c_dev == NULL) || (cx88_i2c_lock_bus(bus, i2c_dev->iotype, timeout) != 0))
   {
      return IIC_ESTATUS;
   }
   if (bus->transaction_state != CX88_I2C_INACTIVE)
   {
      cx88_i2c_unlock_bus(bus);
      return IIC_ESTATUS;
   }
   bus->transaction_state = CX88_I2C_ACTIVE;
   /* This is annoying--iicbus takes slave addresses as 8-bit values,
    * including the R/W bit--why? */
   bus->slave_addr = slave >> 1;
   bus->bus_speed = i2c_dev->speed;
   bus->iotype = i2c_dev->iotype;
   cx88_i2c_set_intr(bus, TRUE); /* Enable i2c interrupt */
   return IIC_NOERR;
}

/* iicbus stop callback */
int cx88_i2c_stop(device_t dev)
{
   DIAGNOSTIC(device_printf(dev, "Stopping I2C transaction\n"))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int error = IIC_NOERR;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   /* Flush any pending write */
   error = cx88_i2c_transfer(bus, NULL);
   cx88_i2c_terminate(bus);
   return error;
}

/* iicbus restart callback */
int cx88_i2c_restart(device_t dev, u_char slave, int timeout)
{
   DIAGNOSTIC(device_printf(dev, "Restarting I2C transaction\n"))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   cx88_i2c_dev *i2c_dev = CX88_I2C_GET_DEV(dev, (slave >> 1));   
   int error = IIC_NOERR;
   if ((bus == NULL) || (i2c_dev == NULL) ||
       (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   bus->transaction_state = CX88_I2C_RESTART;
   /* Flush any pending write */
   error = cx88_i2c_transfer(bus, NULL);
   bus->slave_addr = slave >> 1;
   bus->bus_speed = i2c_dev->speed;
   bus->iotype = i2c_dev->iotype;
   return error;
}

/* iicbus reset callback */
int cx88_i2c_reset(device_t dev, u_char speed, u_char addr, u_char * oldaddr)
{
   cx88_i2c_stop(dev);
   if (oldaddr != NULL)
   {
      *oldaddr = 0;
   }
   return IIC_NOERR;
}

/* iicbus write callback */
int cx88_i2c_write (
   device_t dev,
   const char *buf,
   int len,
   int *bytes,
   int timeout)
{
   DIAGNOSTIC(device_printf(dev, "I2C write of %d bytes\n", len))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int i, error = IIC_NOERR;
   cx88_i2c_transaction transaction;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   transaction.type = CX88_I2C_WRITE;
   transaction.timeout = timeout;
   for (i = 0; i < len; ++i)
   {
      transaction.byte = buf[i];
      error = cx88_i2c_transfer(bus, &transaction);
      if (error)
      {
         break;
      }
      ++(*bytes); /* Increment byte counter */
   }
   return error;
}

/* iicbus read callback */
int cx88_i2c_read (
   device_t dev,
   char *buf,
   int len,
   int *bytes,
   int last,
   int delay)
{
   DIAGNOSTIC(device_printf(dev, "I2C read of %d bytes\n", len))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int i, error = IIC_NOERR;
   cx88_i2c_transaction transaction;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   transaction.type = CX88_I2C_READ;
   transaction.timeout = delay;
   for (i = 0; i < len; ++i)
   {
      transaction.last = ((last == IIC_LAST_READ) && (i == (len - 1)));
      error = cx88_i2c_transfer(bus, &transaction);
      if (error)
      {
         break;
      }
      buf[i] = transaction.byte;
      ++(*bytes); /* Increment byte counter */
   }
   return error;
}

void cx88_i2c_init_device(cx88_i2c_dev *dev)
{
   dev->dev = NULL;
   dev->speed = CX88_I2C_SPEED_100KHZ;
   dev->addr = CX88_I2C_INVALID_ADDR;
   dev->iotype = CX88_I2C_IO_NORMAL;
}

/* Creates an i2c bus instance and initializes devices in the i2c softc */
int cx88_i2c_attach(cx88_i2c_softc *sc, uint8_t bus_number, cx88_common_softc *host, device_t parent)
{
   uint8_t i;
   sc->num_i2c_devices = 0;
   for (i = 0; i < CX88_COMMON_MAX_I2C_DEVICES; ++i)
   {
      /* Initialize I2C device array to default (unconfigured) settings */
      cx88_i2c_init_device(&(sc->i2c_devices[i]));
   }
   return cx88_i2c_create(&(sc->bus), bus_number, host, parent);
}

int cx88_i2c_detach(cx88_i2c_softc *sc)
{
   int error = 0;
   uint8_t i;
   if (sc->bus.board_data != NULL)
   {
      /* Cancel any outstanding I/O */
      mtx_lock_spin(&(sc->bus.board_data->isr_lock));
      sc->bus.io_canceled = TRUE;
      wakeup(&(sc->bus));
      mtx_unlock_spin(&(sc->bus.board_data->isr_lock));
   }
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      if ((error = cx88_i2c_remove_device(&(sc->i2c_devices[i]))))
      {
         break;
      }
   }
   if (!error)
   {
      error = cx88_i2c_destroy(&(sc->bus));
   }
   return error;
}

/* Adds/removes/modifies i2c device entries as specified by the addresses in conf */
int cx88_i2c_update_devices(cx88_i2c_softc *sc, cx88_common_i2c_config *conf)
{
   DIAGNOSTIC(device_printf(sc->bus.host_sc->dev, "Configuring i2c devices\n")); 
   int error = cx88_i2c_lock_bus(&(sc->bus), CX88_I2C_IO_NORMAL, 0);
   if (error)
   {
      return error;
   }
   uint8_t i;
   if (conf->in_num_i2c_devices >= CX88_COMMON_MAX_I2C_DEVICES)
   {
      error = EINVAL;
   }
   /* If the new list of devices is larger than the previous one, 
    * remove the 'extra' devices */
   cx88_lock_device_tree(sc->bus.host_sc);
   for (i = conf->in_num_i2c_devices; !error && (i < sc->num_i2c_devices); ++i)
   {
      error = cx88_i2c_remove_device(&(sc->i2c_devices[i]));
   }
   cx88_unlock_device_tree(sc->bus.host_sc);
   for (i = 0; !error && (i < conf->in_num_i2c_devices); ++i)
   {  
      const char *name = ((sc->i2c_devices[i].dev == NULL) ? NULL : device_get_name(sc->i2c_devices[i].dev));
      if ((conf->devices[i].in_addr != sc->i2c_devices[i].addr) || (sc->i2c_devices[i].iotype != CX88_I2C_IO_NORMAL) ||
          (name == NULL) || (strncmp(name, conf->devices[i].in_driver, CX88_COMMON_MAX_DRIVER_NAME - 1) != 0))
      {
         conf->devices[i].in_driver[CX88_COMMON_MAX_DRIVER_NAME - 1] = '\0';
         /* Only reconfigure the device if the address is different */
         cx88_lock_device_tree(sc->bus.host_sc);
         error = cx88_i2c_remove_device(&(sc->i2c_devices[i]));
         if (!error)
         {
            error = cx88_i2c_add_device(
               &(sc->bus),
               conf->devices[i].in_addr,
               conf->devices[i].in_speed,
               CX88_I2C_IO_NORMAL,
               conf->devices[i].in_driver,
               &(sc->i2c_devices[i]));
         }
         cx88_unlock_device_tree(sc->bus.host_sc);
      }
      else
      {
         error = cx88_i2c_set_speed(&(sc->i2c_devices[i]), conf->devices[i].in_speed);
      }
      if (sc->i2c_devices[i].dev != NULL)
      {
         conf->devices[i].out_instance = device_get_unit(sc->i2c_devices[i].dev);
      }
      else
      {
         conf->devices[i].out_instance = -1;
      }
   }
   sc->num_i2c_devices = conf->in_num_i2c_devices;
   cx88_i2c_unlock_bus(&(sc->bus));
   return error;
}

cx88_i2c_dev *cx88_i2c_get_dev(cx88_i2c_softc *sc, uint8_t addr)
{
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      /* Determine whether we have a device with the given address */
      if (sc->i2c_devices[i].addr == addr)
      {
         return (&(sc->i2c_devices[i]));
      }  
   }
   return NULL;
}

/* Module declarations */
static int cx88_i2c_mod_handler(module_t mod, int what, void *arg)
{
   return 0;
}

static moduledata_t cx88_i2c_mod_data =
{
   "cx88i2c",
   cx88_i2c_mod_handler,
   NULL
};

DECLARE_MODULE(cx88_i2c, cx88_i2c_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_DEPEND(cx88_i2c, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);
MODULE_DEPEND(cx88_i2c, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_DEPEND(cx88_i2c, iicdev, 0, INT_MAX, INT_MAX);
MODULE_VERSION(cx88_i2c, CX88_I2C_MODVER);
