/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/types.h>
#include <sys/bus.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/uio.h>
#include <dev/iicbus/iiconf.h>
#include <dev/iicbus/iicbus.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include "iicdev.h"

#define BUFFER_SIZE 128

#ifdef _DIAGNOSTIC
#define DIAGNOSTIC(stmt) stmt;
#else
#define DIAGNOSTIC(stmt)
#endif

/* Softc structure */
typedef struct
{
   device_t dev; /* device pointer */
   uint8_t addr; /* 7-bit address */
   struct cdev *devnode; /* device node */
   boolean_t is_valid; /* device valid state indicator */
   char buffer[BUFFER_SIZE]; /* transfer buffer */
   struct sx lock;
} iicdev_softc;

static devclass_t iicdev_devclass;

static int iicdev_probe(device_t dev);
static int iicdev_attach(device_t dev);
static int iicdev_detach(device_t dev);
static int iicdev_suspend(device_t dev);
static int iicdev_resume(device_t dev);

static d_open_t iicdev_open;
static d_close_t iicdev_close;
static d_write_t iicdev_io;

/* Device node descriptor */
static struct cdevsw iicdev_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "iicdev",
   .d_open = iicdev_open,  /* open */
   .d_close = iicdev_close, /* close */
   .d_read = iicdev_io,  /* read */
   .d_write = iicdev_io, /* write */
};

/* Open callback */
static int iicdev_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   DIAGNOSTIC(printf("Opening iicdev\n"))
   return 0;
}

/* Close callback */
static int iicdev_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   DIAGNOSTIC(printf("Closing iicdev\n"))
   return 0;
}

/* Perform I/O on the device */
static int iicdev_io(struct cdev *dev, struct uio *uio, int ioflag)
{
   DIAGNOSTIC(printf("Performing I/O on iicdev\n"))
   int error;
   iicdev_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   device_t parent;
   uint8_t addr = (sc->addr << 1) | (uio->uio_rw == UIO_READ ? 1 : 0);
   parent = device_get_parent(sc->dev);
   sx_xlock(&(sc->lock));
   if (!sc->is_valid)
   {
      sx_xunlock(&(sc->lock));
      return ENXIO;
   }
   /* First make sure we can own the bus without interference from other drivers */
   if ((error = iicbus_request_bus(parent, sc->dev, IIC_DONTWAIT)))
   {
      sx_xunlock(&(sc->lock));
      return error;
   }
   if ((error = iicbus_start(parent, addr, 0)))
   {
      DIAGNOSTIC(device_printf(sc->dev, "Failed to start iic transaction\n"))
      iicbus_release_bus(parent, sc->dev);
      sx_xunlock(&(sc->lock));
      return error;
   }
   /* We can only transfer up to BUFFER_SIZE bytes in 1 shot, so loop until
    * everything has been transferred */
   while (!error && (uio->uio_resid > 0))
   {
      int num_bytes = MIN(uio->uio_resid, BUFFER_SIZE);
      if (uio->uio_rw == UIO_WRITE)
      {
         DIAGNOSTIC(device_printf(sc->dev, "IICDEV write of %d bytes\n", num_bytes))
         int sent_bytes = 0;
         error = uiomove(sc->buffer, num_bytes, uio);
         if (!error)
         {
            error = iicbus_write(parent, sc->buffer, num_bytes, &sent_bytes, 0);
         }
      }
      else if (uio->uio_rw == UIO_READ)
      {
         DIAGNOSTIC(device_printf(sc->dev, "IICDEV read of %d bytes\n", num_bytes))
         int received_bytes = 0;
         error = iicbus_read(parent, sc->buffer, num_bytes, &received_bytes,
            ((uio->uio_resid <= BUFFER_SIZE) ? IIC_LAST_READ : 0), 0);
         if (!error)
         {
            error = uiomove(sc->buffer, received_bytes, uio);
         }
      }
      else
      {
         DIAGNOSTIC(device_printf(sc->dev, "Invalid uio direction\n"))
         error = EINVAL;
      }
   }
   iicbus_stop(parent);
   iicbus_release_bus(parent, sc->dev);
   sx_xunlock(&(sc->lock));
   return error;
}

/* Probe DEVMETHOD */
static int iicdev_probe(device_t dev)
{
   device_set_desc(dev, "I2C device");
   return 0;
}

/* Attach DEVMETHOD */
static int iicdev_attach(device_t dev)
{
   iicdev_softc *sc = device_get_softc(dev);
   sc->dev = dev;
   sc->is_valid = TRUE;
   struct iicbus_ivar *devaddr = IICBUS_IVAR(dev);
   if (devaddr == NULL)
   {
      return ENXIO;  
   }
   sc->addr = (uint8_t) (devaddr->addr);
   DIAGNOSTIC(device_printf(dev, "Address %d\n", sc->addr))
   /* Basic initialization is done in iicdev_attach(), but we don't actually create the
    * devnode until the address is set, as it doesn't make sense to access the device
    * earlier than that. */
   sc->devnode = make_dev(&iicdev_cdevsw, device_get_unit(dev), UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(dev));
   if (sc->devnode == NULL)
   {
      return ENOMEM;
   }
   sc->devnode->si_drv1 = sc;
   sx_init(&(sc->lock), "iicdev lock");
   return 0;
}

/* Detach DEVMETHOD */
static int iicdev_detach(device_t dev)
{
   iicdev_softc *sc = device_get_softc(dev);
   sc->is_valid = FALSE;
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sx_destroy(&(sc->lock));
   }
   return bus_generic_detach(dev);
}

/* Suspend DEVMETHOD */
static int iicdev_suspend(device_t dev)
{
   iicdev_softc *sc = device_get_softc(dev);
   sx_xlock(&(sc->lock));
   sc->is_valid = FALSE;
   sx_xunlock(&(sc->lock));
   return 0;
}

/* Resume DEVMETHOD */
static int iicdev_resume(device_t dev)
{
   iicdev_softc *sc = device_get_softc(dev);
   sc->is_valid = TRUE;
   return 0;
}

static device_method_t iicdev_methods[] =
{
   /* Device interface */
   DEVMETHOD(device_probe,             iicdev_probe),
   DEVMETHOD(device_attach,            iicdev_attach),
   DEVMETHOD(device_detach,            iicdev_detach),
   DEVMETHOD(device_shutdown,          iicdev_detach),
   DEVMETHOD(device_suspend,           iicdev_suspend),
   DEVMETHOD(device_resume,            iicdev_resume),

   {NULL, NULL}
};

static driver_t iicdev_driver =
{
   IICDEV_NAME,
   iicdev_methods,
   sizeof(iicdev_softc)
};

DRIVER_MODULE(iicdev, iicbus, iicdev_driver, iicdev_devclass, NULL, NULL);
MODULE_DEPEND(iicdev, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_VERSION(iicdev, IICDEV_MODVER);

