/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_mpeg.h"

/* Minimum and maximum data buffer sizes (in bytes) */
#define MAX_MPEG_BUFFER_SIZE (1 << 19)
#define MIN_MPEG_BUFFER_SIZE (1 << 15)

/* Maximum line size, in bytes (dictated by capacity of line length register) */
#define MAX_MPEG_LINE_SIZE 0xFFF

static int cx88_mpeg_dmaprog(cx88_buffer *buffer, uint8_t bufindex);
static int cx88_mpeg_configure(cx88_mpeg_softc *sc, cx88_mpeg_config *conf, struct thread *context);
static int cx88_mpeg_transfer(cx88_mpeg_softc *sc, cx88_mpeg_wait *wait, boolean_t nonblocking);
static int cx88_mpeg_start_capture(cx88_mpeg_softc *sc, cx88_mpeg_capture *capture);
static void cx88_mpeg_reconfigure(cx88_mpeg_softc *sc);

static d_open_t cx88_mpeg_open;
static d_close_t cx88_mpeg_close;
static d_mmap_t cx88_mpeg_mmap;
static d_ioctl_t cx88_mpeg_ioctl;
static d_poll_t cx88_mpeg_poll;

/* Devnode descriptor */
static struct cdevsw cx88_mpeg_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88mpeg",
   .d_open = cx88_mpeg_open,   /* open */
   .d_close = cx88_mpeg_close, /* close */
   .d_mmap = cx88_mpeg_mmap,   /* mmap */
   .d_ioctl = cx88_mpeg_ioctl, /* ioctl */
   .d_poll = cx88_mpeg_poll,   /* poll */
};

static struct unrhdr *cx88_mpeg_units = NULL;

static const cx88_mpeg_stream_desc cx88_mpeg_streams[] = 
{
   {
      .stream = CX88_MPEG_CX23880_TS,
      .dma_channel = CX88_CH_MPEG_TS,
      .reg_gp_count = CX23880_MPEG_REG_GP_COUNT,
      .reg_dma_control = CX23880_MPEG_REG_DMA_CONTROL,
      .reg_line_length = CX23880_MPEG_REG_LINE_LENGTH,
      .reg_sop_control = CX23880_MPEG_REG_SOP_CONTROL,
      .reg_gen_control = CX23880_MPEG_REG_GEN_CONTROL,
      .reg_bad_packet_stat = CX23880_MPEG_REG_BAD_PACKET_STAT,
      .reg_sop_stat = CX23880_MPEG_REG_SOP_STATUS,
      .reg_fifo_stat = CX23880_MPEG_REG_FIFO_STATUS,
      .reg_misc = CX23880_MPEG_REG_MISCELLANEOUS,
      .reg_pci_intr_mask = CX23880_REG_PCI_INTMASK,
      .pci_intr_mask = CX23880_PCI_INT_MPEGTS,
      .reg_intr_mask = CX23880_MPEG_REG_INT_MASK,
      .reg_intr_stat = CX23880_MPEG_REG_INT_STAT,
      .reg_intr_mstat = CX23880_MPEG_REG_INT_MSTAT,
      .intr_mask = MPEG_INT_RISCI1 | MPEG_INT_SYNC | MPEG_INT_OPCODE_ERR |
         MPEG_INT_PAR_ERR | MPEG_INT_RIP_ERR | MPEG_INT_PCI_ABORT,
      .reg_devcntl2 = CX23880_REG_DEVCNTL2
   },
   {
      .stream = CX88_MPEG_CX23885_TS1,
      .dma_channel = CX88_CH_MPEG_TS,
      .reg_gp_count = CX23885_VIDEO_B_REG_GP_COUNT,
      .reg_dma_control = CX23885_VIDEO_B_REG_DMA_CONTROL,
      .reg_line_length = CX23885_VIDEO_B_REG_LINE_LENGTH,
      .reg_sop_control = CX23885_VIDEO_B_REG_SOP_CONTROL,
      .reg_gen_control = CX23885_VIDEO_B_REG_GEN_CONTROL,
      .reg_bad_packet_stat = CX23885_VIDEO_B_REG_BAD_PACKET_STAT,
      .reg_sop_stat = CX23885_VIDEO_B_REG_SOP_STATUS,
      .reg_fifo_stat = CX23885_VIDEO_B_REG_FIFO_STATUS,
      .reg_misc = CX23885_VIDEO_B_REG_MISCELLANEOUS,
      .reg_ts_clock = CX23885_VIDEO_B_REG_TS_CLOCK,
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_VIDEO_B,
      .reg_intr_mask = CX23885_VIDEO_B_REG_INT_MASK,
      .reg_intr_stat = CX23885_VIDEO_B_REG_INT_STAT,
      .reg_intr_mstat = CX23885_VIDEO_B_REG_INT_MSTAT,
      .intr_mask = MPEG_INT_RISCI1 | MPEG_INT_SYNC | MPEG_INT_OPCODE_ERR,
      .reg_devcntl2 = CX23885_REG_DEVCNTL2
   },
   {
      .stream = CX88_MPEG_CX23885_TS2,
      .dma_channel = CX88_CH_MPEG_TS_2,
      .reg_gp_count = CX23885_VIDEO_C_REG_GP_COUNT,
      .reg_dma_control = CX23885_VIDEO_C_REG_DMA_CONTROL,
      .reg_line_length = CX23885_VIDEO_C_REG_LINE_LENGTH,
      .reg_sop_control = CX23885_VIDEO_C_REG_SOP_CONTROL,
      .reg_gen_control = CX23885_VIDEO_C_REG_GEN_CONTROL,
      .reg_bad_packet_stat = CX23885_VIDEO_C_REG_BAD_PACKET_STAT,
      .reg_sop_stat = CX23885_VIDEO_C_REG_SOP_STATUS,
      .reg_fifo_stat = CX23885_VIDEO_C_REG_FIFO_STATUS,
      .reg_misc = CX23885_VIDEO_C_REG_MISCELLANEOUS,
      .reg_ts_clock = CX23885_VIDEO_C_REG_TS_CLOCK,
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_VIDEO_C,
      .reg_intr_mask = CX23885_VIDEO_C_REG_INT_MASK,
      .reg_intr_stat = CX23885_VIDEO_C_REG_INT_STAT,
      .reg_intr_mstat = CX23885_VIDEO_C_REG_INT_MSTAT,
      .intr_mask = MPEG_INT_RISCI1 | MPEG_INT_SYNC | MPEG_INT_OPCODE_ERR,
      .reg_devcntl2 = CX23885_REG_DEVCNTL2
   }
};

int cx88_mpeg_setup(
   cx88_mpeg_softc *sc,
   cx88_mpeg_stream_t stream,
   cx88_common_softc *common,
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t))
{
   int error = 0;
   if (stream >= CX88_MPEG_NUM_STREAMS)
   {
      return EINVAL;
   }
   sc->stream = &(cx88_mpeg_streams[stream]);
   sc->common = common;
   sc->i2c_accessor = i2c_accessor;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   sc->devnode = NULL;
   mtx_init(&(sc->isrlock), "CX88 MPEG ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 MPEG SX lock");
   if ((error = cx88_init_channel(&(sc->channel), sc->common, sc->stream->dma_channel, MAX_MPEG_BUFFER_SIZE)))
   {
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      return error;
   }
   if (((sc->unit = alloc_unr(cx88_mpeg_units)) == -1) ||
      ((sc->devnode = make_dev(&cx88_mpeg_cdevsw, sc->unit, UID_ROOT, GID_WHEEL, 0644, 
         "cx88mpeg%d", sc->unit)) == NULL))
   {
      if (sc->unit >= 0)
      {
         free_unr(cx88_mpeg_units, sc->unit);
      }
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      return ENOMEM;
   }
   sc->devnode->si_drv1 = sc;
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED);
   return 0;
}

void cx88_mpeg_teardown(cx88_mpeg_softc *sc)
{
   /* If the device node is NULL, we haven't been initialized, so don't 
    * do anything stupid (like trying to acquire uninitialized locks) */
   if (sc->devnode == NULL)
   {
      return;
   }
   sx_slock(&(sc->sxlock));
   cx88_mpeg_reset(sc, CX88_MPEG_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel (&(sc->channel)); /* Release all channel and buffer resources */
   sx_xunlock(&(sc->sxlock));
   /* Device file should be destroyed as the second step in cleanup*/
   destroy_dev(sc->devnode);
   sc->devnode = NULL;
   free_unr(cx88_mpeg_units, sc->unit);
   /* Now remove static parts of the softc */
   cx88_destroy_channel (&(sc->channel));
   mtx_destroy(&(sc->isrlock));
   sx_destroy(&(sc->sxlock));
}

/* Open file operation */
static int cx88_mpeg_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_mpeg_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common->dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }
   sc->in_use = TRUE;
   sx_xunlock(&sc->sxlock);
   return 0;
}

/* Close file operation */
static int cx88_mpeg_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_mpeg_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common->dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   cx88_mpeg_reset(sc, sc->state);
   /* If we're using user-allocated buffers, release them as
   * they can't be persisted across open() and close() in
   * different processes */
   cx88_free_userbufs(&(sc->channel), NULL);
   if (sc->reconfig_requested)
   {
      cx88_mpeg_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   sc->in_use = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Configures channel, buffers, and I2C devices */
static int cx88_mpeg_configure(cx88_mpeg_softc *sc, cx88_mpeg_config *config, struct thread *context)
{
   int error = 0;
   if (sc->state < CX88_MPEG_UNCONFIGURED)
   {
      return ENXIO;
   }
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED); /* Ensure there is no outstanding capture */
   size_t buffer_size, last_buffer_size;
   uint8_t num_buffers = 0;
   /* Validate ioctl parameters */
   if ((config->inout_num_packets == 0) || (config->in_packet_size == 0) || 
      ((config->in_buftype == CX88_MPEG_BUF_USER) && (config->bufconfig.in_userbuf == NULL)) ||
      ((config->in_buftype != CX88_MPEG_BUF_USER) && (config->in_buftype != CX88_MPEG_BUF_KERNEL)))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common->dev, "Configuring MPEG TS channel\n"))
   /* Allocate the channel */
   error = cx88_configure_channel(
      &(sc->channel),
      0,
      config->in_packet_size,
      MAX_MPEG_LINE_SIZE,
      ((config->in_buftype == CX88_MPEG_BUF_KERNEL) ? 0 : config->inout_num_packets));
   if (!error)
   {
      uint32_t total_lines = (config->in_packet_size * config->inout_num_packets) / sc->channel.line_size;
      uint32_t max_lines_per_buffer = MAX_MPEG_BUFFER_SIZE / sc->channel.line_size;
      if (max_lines_per_buffer == 0)
      {
         return EINVAL;
      }
      num_buffers = (total_lines + max_lines_per_buffer - 1) / max_lines_per_buffer;
      if (num_buffers < 2)
      {
         /* We need at least two buffers so the application can process one while the
          * other is being filled by the driver */
         num_buffers = 2;
      }
      uint32_t lines_per_buffer = (total_lines + num_buffers - 1) / num_buffers;
      uint32_t total_size = total_lines * sc->channel.line_size;
      buffer_size = lines_per_buffer * sc->channel.line_size;
      last_buffer_size = total_size - (buffer_size * (num_buffers - 1));
      if (buffer_size < MIN_MPEG_BUFFER_SIZE)
      {
         return EINVAL;
      }
      config->inout_num_packets = total_size / config->in_packet_size;
      config->out_num_buffers = num_buffers;
      config->out_packets_per_buffer = buffer_size / config->in_packet_size;
      if (config->in_buftype == CX88_MPEG_BUF_KERNEL)
      {
         /* Calculate the total amount of memory the application will need to mmap(), along
          * with the buffer size the application will need to use in calculating the offsets
          * of individual buffers within that mmap'ed region */
         config->bufconfig.out_kernelbuf.out_buffer_size = round_page(buffer_size);
         config->bufconfig.out_kernelbuf.out_mmap_size = 
            ((num_buffers - 1) * config->bufconfig.out_kernelbuf.out_buffer_size) + last_buffer_size;
      }
      DIAGNOSTIC(printf("Packets per buffer = %ld\n", buffer_size / config->in_packet_size))
      /* Request the buffer structures */
      error = cx88_request_buffers(&(sc->channel), (uint8_t)num_buffers, 0);
   }
   if (!error)
   {
      uint32_t i;
      for (i = 0; i < num_buffers; ++i)
      {
         void *bufptr = NULL;
         struct thread *bufcontext = NULL;
         if (config->in_buftype == CX88_MPEG_BUF_USER)
         {
            /* Each cx88 buffer is a segment of the user-allocated buffer */
            bufptr = (uint8_t *)(config->bufconfig.in_userbuf) + (i * buffer_size);
            bufcontext = context;
         }
         /* Configure each buffer */
         error = cx88_configure_buffer(
            &(sc->channel.buffers[i]),
            bufptr, /* This will be NULL for kernel-allocated buffers */
            bufcontext,
            ((i == (num_buffers - 1)) ? last_buffer_size : buffer_size),
            ((i == (num_buffers - 1)) ? 0 : (i + 1)),
            cx88_mpeg_dmaprog,
            CX88_RISC_REPROGRAM_IF_BUFFER_CHANGED);
         if (error)
         {
            break;
         }
      }
   }
   if (!error)
   {
      /* Now commit all changes we've made to the channel configuration */
      error = cx88_commit_channel(&(sc->channel));
   }
   if (!error)
   {
      sc->state = CX88_MPEG_CONFIGURED;
   }
   else
   {
      cx88_release_channel(&(sc->channel));
   }
   return error;
}

void cx88_mpeg_reset(cx88_mpeg_softc *sc, cx88_mpeg_state_t state)
{
   mtx_lock(&(sc->isrlock));
   /* Stop any data transfer */
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, 0);
   /* Disable interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 0);
   /* Reset capture status values */
   sc->buffer_available = 0;
   sc->isr_status = CX88_MPEG_WAIT_ABORTED;
   sc->overflows = 0;
   sc->bad_packets = 0;
   /* Our state value should never be higher than CX88_MPEG_CONFIGURED, as we won't be capturing once reset is done */
   sc->state = CX88_MPEG_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   wakeup(sc);
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
   
   /* Protect access to the shared PCI interrupt and PINMUX I/O registers. */
   cx88_lock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_pci_intr_mask,
      (cx88_read_reg(sc->common, sc->stream->reg_pci_intr_mask) & ~(sc->stream->pci_intr_mask)));
   if ((sc->stream->stream == CX88_MPEG_CX23880_TS) && !(sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT))
   {
      cx88_write_reg(sc->common, CX23880_REG_PINMUX_IO,
         (cx88_read_reg(sc->common, CX23880_REG_PINMUX_IO) & ~CX23880_PINMUX_MPEG_PAR));
   }
   cx88_unlock_board(sc->common);

   sc->gen_ctrl = 0;
   cx88_write_reg(sc->common, sc->stream->reg_gen_control, (MPEG_GEN_STAT_CLEAR | MPEG_GEN_SW_RESET));
   pause("cx88p", 1); /* Wait for reset state to clear */
   cx88_write_reg(sc->common, sc->stream->reg_gen_control, 0);
}

/* Initiate data capture */
static int cx88_mpeg_start_capture(cx88_mpeg_softc *sc, cx88_mpeg_capture *capture)
{
   int error = 0;
   cx88_mpeg_reset(sc, sc->state); /* Make sure we're in a clean, valid state first */
   if (sc->state != CX88_MPEG_CONFIGURED)
   {
      return ENXIO;
   }
   if ((sc->channel.packet_size > 0xFFF) || (capture->in_start_byte_count > 0xF))
   {
      device_printf(sc->common->dev, "Invalid packet size or start byte count\n");
      return EINVAL;  
   }
   if ((error = cx88_start_channel(&(sc->channel))))
   {
      return error;
   }
   sc->isr_status = CX88_MPEG_CAP_NOERR;
	sc->last_sync_buffer = 0;
   sc->buffer_available = 0;
   sc->packet_count = 0;
   sc->state = CX88_MPEG_CAPTURING;
   cx88_write_reg(sc->common, sc->stream->reg_line_length, sc->channel.line_size);
   if (sc->stream->stream == CX88_MPEG_CX23885_TS1)
   {
      cx88_write_reg(sc->common, CX23885_VIDEO_B_REG_SOURCE_SELECT, CX23885_VIDEO_B_SRC_SEL_TS);
   }
   if (sc->stream->stream != CX88_MPEG_CX23880_TS)
   {
      cx88_write_reg(sc->common, sc->stream->reg_ts_clock, CX23885_VIDEO_TS_CLOCK_ENABLE);
   }
   /* Configure start byte and packet length--we must see 1 start byte before 
    * data transfer can begin */
   cx88_write_reg(sc->common, sc->stream->reg_sop_control,
      (cx88_write_field(MPEG_SOP_START_BYTE, capture->in_start_byte) |
      cx88_write_field(MPEG_SOP_PACKET_LENGTH, sc->channel.packet_size) |
      cx88_write_field(MPEG_SOP_START_COUNT, capture->in_start_byte_count)));

   /* Configure general control register based on application input */
   sc->gen_ctrl = ((capture->in_options & CX88_MPEG_SERIAL_INPUT) ? MPEG_GEN_SERIAL : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_PUNCTURED_CLOCK) ? MPEG_GEN_PUNC_CLK : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_POS_POLARITY) ? MPEG_GEN_POSEDGE : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_LITTLE_ENDIAN) ? MPEG_GEN_BIT_REVERSE : 0);

   if ((sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT) && (capture->in_options & CX88_MPEG_BYTEWIDE_SOP))
   {
      cx88_write_reg(sc->common, sc->stream->reg_sop_stat, MPEG_SOP_BYTEWIDE);
   }
   else
   {
      cx88_write_reg(sc->common, sc->stream->reg_sop_stat, 0);
   }
   
   /* Enable FIFO overflow and bad packet checks */
   if (sc->stream->stream == CX88_MPEG_CX23880_TS)
   {
      /* For CX23885/7, FIFO overflow reporting seems to be broken--this register always
       * reports 4095 (max) overflows--it behaves the same way under Windows and Linux too. */
      cx88_write_reg(sc->common, sc->stream->reg_fifo_stat, MPEG_FIFO_OVERFLOW_INTR);
      /* Enabling bad packet reporting also has some undesirable effects w/ the 885/7.  
       * The bad packet counters work, but the detection of a bad packet seems to 
       * cause the TS to hang ~90% of the time. */
      cx88_write_reg(sc->common, sc->stream->reg_bad_packet_stat, MPEG_BAD_PACKET_ENABLE);
   }

   /* Protect access to the shared PCI interrupt and PINMUX I/O registers. */
   cx88_lock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_pci_intr_mask,
      (cx88_read_reg(sc->common, sc->stream->reg_pci_intr_mask) | sc->stream->pci_intr_mask));
   if (sc->stream->stream == CX88_MPEG_CX23880_TS)
   {
      if (!(sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT))
      {
         /* Configure parallel input */
         cx88_write_reg(sc->common, CX23880_REG_PINMUX_IO,
            (cx88_read_reg(sc->common, CX23880_REG_PINMUX_IO) | CX23880_PINMUX_MPEG_PAR));
      }
      else
      {
         cx88_write_reg(sc->common, CX23880_REG_PINMUX_IO,
            (cx88_read_reg(sc->common, CX23880_REG_PINMUX_IO) & (~CX23880_PINMUX_MPEG_PAR)));
      }
   }
   cx88_unlock_board(sc->common);
   cx88_write_reg(sc->common, sc->stream->reg_misc, 0);
   cx88_write_reg(sc->common, sc->stream->reg_gen_control, sc->gen_ctrl);
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_stat,
      cx88_read_reg(sc->common, sc->stream->reg_intr_stat));      
   /* Enable interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, sc->stream->intr_mask);
   cx88_write_reg(sc->common, sc->stream->reg_devcntl2, CX88_RUN_RISC);
   /* Start the data transfer */
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, (MPEG_RISC_ENABLE | MPEG_FIFO_ENABLE));
   return error;
}

/* Either wait for data to become available (if !nonblocking), or check if data is available.
 * If data is available, then return buffer count and transfer status to application */
static int cx88_mpeg_transfer(cx88_mpeg_softc *sc, cx88_mpeg_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   uint16_t prev_overflows = 0, overflows = 0, prev_bad_packets, bad_packets;
   wait->out_buffer_available = 0;
   wait->out_status = CX88_MPEG_CAP_NOERR;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_MPEG_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   while (!error && !(sc->isr_status & CX88_MPEG_CAP_FATAL) && (sc->buffer_available == 0))
   {
      if (!nonblocking)
      {
         error = msleep(sc, &(sc->isrlock), PZERO | PCATCH, "cx88mp", (wait->in_timeout_s * hz));
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   if (sc->buffer_available > 0)
   {
      /* Perform any necessary cache/bounce buffer synchronization.
       * Do this in a loop to account for the unlikely case where
       * we've missed an interrupt and multiple buffers need to be 
       * synchronized */
      uint8_t sync_bufindex = sc->last_sync_buffer;
      do
      {
         if (sync_bufindex >= sc->channel.num_buffers)
         {
            sync_bufindex = 0;
         }
         cx88_buffer *buffer = &(sc->channel.buffers[sync_bufindex]);
         /*Synchronize the buffer */
         cx88_mark_transfer_complete(buffer);
         cx88_prepare_for_transfer(buffer);
      }
      while ((++sync_bufindex) != sc->buffer_available);
      sc->last_sync_buffer = sc->buffer_available;
   }
   /* Copy buffer and transfer status */
   wait->out_buffer_available = sc->buffer_available;
   wait->out_status = sc->isr_status;
   sc->isr_status = CX88_MPEG_CAP_NOERR;
   wait->out_packet_count = sc->packet_count;
   /* Read FIFO overflow and bad packet counters */
   prev_bad_packets = sc->bad_packets;
   bad_packets = cx88_read_field(cx88_read_reg(sc->common, sc->stream->reg_bad_packet_stat), MPEG_BAD_PACKET_COUNT);
   sc->bad_packets = bad_packets;
   if (sc->stream->stream == CX88_MPEG_CX23880_TS)
   {
      prev_overflows = sc->overflows;
      overflows = cx88_read_field(cx88_read_reg(sc->common, sc->stream->reg_fifo_stat), MPEG_FIFO_OVERFLOW_COUNT);
      sc->overflows = overflows;
   }
   sc->buffer_available = 0;
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_s)
   {
      wait->out_status |= CX88_MPEG_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_MPEG_WAIT_ABORTED;
   }
   /* Compute the number of FIFO overflows since the last call to cx88_mpeg_transfer() */
   if (overflows != prev_overflows)
   {
      wait->out_status |= CX88_MPEG_HW_OVERFLOW;
      if (overflows < prev_overflows)
      {
         /* The overflow counter has wrapped around */
         wait->out_num_overflows = overflows + 0x1000 - prev_overflows;
      }
      else
      {
         wait->out_num_overflows = overflows - prev_overflows;
      }
   }
   /* Compute the number of bad packets since the last call to cx88_mpeg_transfer() */
   if (bad_packets != prev_bad_packets)
   {
      wait->out_status |= CX88_MPEG_BAD_PACKET;
      if (bad_packets < prev_bad_packets)
      {
         /* The bad packet counter has wrapped around */
         wait->out_num_bad_packets = bad_packets + 0x1000 - prev_bad_packets;
      }
      else
      {
         wait->out_num_bad_packets = bad_packets - prev_bad_packets;
      }
   }
   return error;
}

/* Ioctl file operation */
static int cx88_mpeg_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_mpeg_softc *sc = dev->si_drv1;
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common->dev);
         break;
      case CX88_MPEG_CONFIG:
         sx_xlock(&(sc->sxlock));
         error = cx88_mpeg_configure(sc, (cx88_mpeg_config*)data, td);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG:
         {
            cx88_common_i2c_config *conf = (cx88_common_i2c_config*)data;
            cx88_i2c_softc *i2c_sc = sc->i2c_accessor(sc->common, conf->in_bus_number);
            if (i2c_sc == NULL)
            {
               error = EINVAL;
            }
            else
            {
               error = cx88_i2c_update_devices(i2c_sc, conf);
            }
         }
         break;
      case CX88_MPEG_CAPTURE:
         sx_xlock(&(sc->sxlock));
         error = cx88_mpeg_start_capture(sc, (cx88_mpeg_capture*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_MPEG_STOP:
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         cx88_mpeg_reset(sc, sc->state);
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_MPEG_WAIT:
         sx_slock(&(sc->sxlock));
         error = cx88_mpeg_transfer(sc, (cx88_mpeg_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO:
         cx88_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX88_COMMON_PERIPH_RESET:
         cx88_peripheral_reset(sc->common, *((uint32_t*)data));
         break;
      case CX23885_COMMON_MC_GPIO:
         error = cx23885_mc_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX23885_COMMON_BRIDGE:
         snprintf((char*)data, CX88_COMMON_MAX_DRIVER_NAME, "%s", device_get_nameunit(sc->common->dev));
         break;
      case CX23885_COMMON_GET_STREAMS:
      {
         uint32_t *stream = (uint32_t*)data;
         switch(sc->stream->stream)
         {
            case CX88_MPEG_CX23885_TS1:
               *stream = CX23885_STREAM_MPEG1;
               break;
            case CX88_MPEG_CX23885_TS2:
               *stream = CX23885_STREAM_MPEG2;
               break;
            default:
               *stream = CX23885_STREAM_NONE;
               break;
         }
         break;
      }
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Select/poll file operation */
static int cx88_mpeg_poll(struct cdev *dev, int events, struct thread *td)
{
   cx88_mpeg_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_MPEG_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->buffer_available > 0)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

int cx88_mpeg_filter(cx88_mpeg_softc *sc)
{
   if (cx88_read_reg(sc->common, sc->stream->reg_intr_mstat) != 0)
   {
      /* Mask out the hardware interrupt until the ISR can handle it */
      cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 0);
      return FILTER_SCHEDULE_THREAD;
   }
   return FILTER_STRAY;
}

/* Interrupt service routine */
void cx88_mpeg_intr(cx88_mpeg_softc *sc)
{
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt */
   uint32_t intrs = cx88_read_reg(sc->common, sc->stream->reg_intr_stat);
   if (intrs & sc->stream->intr_mask)
   {
      if (sc->state != CX88_MPEG_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      /* Check for fatal errors */
      if (intrs & (MPEG_INT_OPCODE_ERR | MPEG_INT_SYNC))
      {
         sc->isr_status |= CX88_MPEG_DMA_PROGRAM;
      }
      if (intrs & (MPEG_INT_PCI_ABORT | MPEG_INT_PAR_ERR | MPEG_INT_RIP_ERR))
      {
         sc->isr_status |= CX88_MPEG_BUS_ERROR;
      }
      if (sc->isr_status & CX88_MPEG_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_write_reg(sc->common, sc->stream->reg_dma_control, 0);
      }
      if (intrs & MPEG_INT_RISCI1)
      {
         /* The counter indicates the most recently completed buffer */
         uint32_t bufindex = cx88_read_reg(sc->common, sc->stream->reg_gp_count);
         if (bufindex < sc->channel.num_buffers)
         {
            /* Available index is 1-based */
            sc->buffer_available = bufindex + 1;
            /* Adjust packet count */
            sc->packet_count += (sc->channel.buffers[bufindex].data_buffer_size / sc->channel.packet_size);
         }
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(sc->common, sc->stream->reg_intr_stat, intrs);
      cx88_write_reg(sc->common, sc->stream->reg_intr_mask, sc->stream->intr_mask);
      wakeup(sc);
      selwakeup(&(sc->selectors));
   }
   mtx_unlock(&(sc->isrlock));
}

/* RISC program generator */
static int cx88_mpeg_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (bufindex == 0)
   {
      /* Initial SYNC at beginning of capture */
      cx88_risc_sync(buffer, 0, CX88_RISC_NO_SYNC_INTR, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   }
   /* Transfer the entire buffer */
   error = cx88_risc_transfer_dmabuf (buffer, 0, (uint32_t)(buffer->data_buffer_size), CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   if (!error)
   {
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
       * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         (bufindex == (buffer->channel->num_buffers - 1) ? CX88_RISC_SYNC_SIZE : 0),
         CX88_TRIGGER_RISCI1,
         ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT));
   }
   return error;
}

/* Mmap file operation */
static int cx88_mpeg_mmap(
   struct cdev *dev,
#if __FreeBSD_version >= 900006
   vm_ooffset_t offset,
   vm_paddr_t *paddr,
   int nprot,
   vm_memattr_t *memattr)
#else
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
#endif
{
   cx88_mpeg_softc *sc = dev->si_drv1;
   sx_slock(&(sc->sxlock));
   DIAGNOSTIC(device_printf(sc->common->dev, "Attempting to mmap device memory\n"))
   int error = cx88_common_mmap(&(sc->channel), (vm_ooffset_t)offset, paddr, nprot);
   sx_sunlock(&(sc->sxlock));
   return error;
}

/* Suspend DEVMETHOD */
int cx88_mpeg_suspend(cx88_mpeg_softc *sc)
{
   sx_slock(&(sc->sxlock));
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel (&(sc->channel)); /* Release all channel and buffer resources */
   sx_xunlock(&(sc->sxlock));
   return 0;
}

static void cx88_mpeg_reconfigure(cx88_mpeg_softc *sc)
{
   cx88_free_sram(&(sc->channel));
   if (sc->state > CX88_MPEG_UNCONFIGURED)
   {
      sc->state = CX88_MPEG_UNCONFIGURED;
   }
}

/* Resource reallocation DEVMETHOD */
int cx88_mpeg_request_reconfig(cx88_mpeg_softc *sc)
{
   int error = 0;
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if (sc->in_use)
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_mpeg_reconfigure(sc);
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error;
}

static int cx88mpeg_mod_handler(module_t mod, int what, void *arg)
{
   switch(what)
   {
      case MOD_LOAD:
         cx88_mpeg_units = new_unrhdr(0, INT_MAX, NULL);
         if (cx88_mpeg_units == NULL)
         {
            return ENOMEM;
         }
         return 0;
      case MOD_UNLOAD:
         if (cx88_mpeg_units != NULL)
         {
            delete_unrhdr(cx88_mpeg_units);
            cx88_mpeg_units = NULL;
         }
         return 0;
      default:
         return 0;
   }
   return 0;
}

/* Module declarations */
static moduledata_t cx88mpeg_mod_data =
{
   "cx88mpegcore",
   cx88mpeg_mod_handler,
   NULL
};

DECLARE_MODULE(cx88mpegcore, cx88mpeg_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_DEPEND(cx88mpegcore, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);
MODULE_DEPEND(cx88mpegcore, cx88_i2c, CX88_I2C_MINVER, CX88_I2C_MODVER, CX88_I2C_MAXVER);
MODULE_VERSION(cx88mpegcore, CX88_MPEG_MODVER);
