/*-
 * Copyright 2009 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#ifndef __CX88_MPEG_H__
#define __CX88_MPEG_H__

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include <sys/limits.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_mpeg_regs.h"
#include "cx88_mpeg_ioctl.h"
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"

#define CX88_MPEG_MINVER 5
#define CX88_MPEG_MODVER 5
#define CX88_MPEG_MAXVER 5

typedef enum
{
   CX88_MPEG_CX23880_TS = 0,
   CX88_MPEG_CX23885_TS1,
   CX88_MPEG_CX23885_TS2,
   CX88_MPEG_NUM_STREAMS
} cx88_mpeg_stream_t;

typedef struct
{
   cx88_mpeg_stream_t stream;
   cx88_channel_t dma_channel;
   uint32_t reg_gp_count;
   uint32_t reg_dma_control;
   uint32_t reg_line_length;
   uint32_t reg_sop_control;
   uint32_t reg_gen_control;
   uint32_t reg_bad_packet_stat;
   uint32_t reg_sop_stat;
   uint32_t reg_fifo_stat;
   uint32_t reg_misc;
   uint32_t reg_ts_clock;
   uint32_t reg_pci_intr_mask;
   uint32_t pci_intr_mask;
   uint32_t reg_intr_mask;
   uint32_t reg_intr_stat;
   uint32_t reg_intr_mstat;
   uint32_t intr_mask;
   uint32_t reg_devcntl2;
} cx88_mpeg_stream_desc;

/* Driver states */
typedef enum
{
   CX88_MPEG_INVALID = 0, /* Uninitialized */
   CX88_MPEG_UNCONFIGURED, /* Channel/buffers/I2C devices not yet configured */
   CX88_MPEG_CONFIGURED, /* Configured but not capturing */
   CX88_MPEG_CAPTURING /* Configured and capturing */
} cx88_mpeg_state_t;

/* MPEG softc structure */
typedef struct
{
   const cx88_mpeg_stream_desc *stream;
   cx88_common_softc *common;
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t);
   cx88_channel channel; /* MPEG TS DMA channel */
   struct cdev *devnode; /* Device file structure */
   int unit; /* Unit number for devnode */
   struct mtx isrlock; /* Synchronize with interrupt service routine */
   struct sx sxlock; /* Driver entry point lock */
   struct selinfo selectors; /* Selectors polling for an interrupt */
   uint32_t gen_ctrl; /* TS_GEN_CNTRL register value */
   volatile uint8_t isr_status; /* Error status from interrupt service routine */
   uint8_t last_sync_buffer; /* 1-based index of most recently synchronized buffer */
   volatile uint8_t buffer_available; /* 1-based index of most recent completed buffer */
   uint64_t packet_count; /* Total packets transferred since current capture began */
   boolean_t in_use; /* Indicates whether device is active */
   uint16_t overflows; /* FIFO overflow counter */
   uint16_t bad_packets; /* Bad (!188-byte) packet counter */
   boolean_t reconfig_requested; /* Indicates a sibling driver has requested resource reallocation */
   cx88_mpeg_state_t state; /* Current driver state */
} cx88_mpeg_softc;

int cx88_mpeg_setup(
   cx88_mpeg_softc *sc,
   cx88_mpeg_stream_t stream,
   cx88_common_softc *common,
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t));
void cx88_mpeg_teardown(cx88_mpeg_softc *sc);
int cx88_mpeg_filter(cx88_mpeg_softc *sc);
void cx88_mpeg_intr(cx88_mpeg_softc *sc);
int cx88_mpeg_suspend(cx88_mpeg_softc *sc);
void cx88_mpeg_reset(cx88_mpeg_softc *sc, cx88_mpeg_state_t state);
int cx88_mpeg_request_reconfig(cx88_mpeg_softc *sc);

#endif
